/*
 * File: $Id: Plane3D.java,v 1.1.1.1 2006/04/19 09:38:47 hans Exp $
 * History
 * Date: 19.01.2005  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2003 - 2004 by
 *
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/**
 * Plane3D.java
 *
 * Abstraction of a 3D plane with two direction vectors and one point which is
 * is an element of the plane.
 * Created on January 19, 2005, 4:44 PM
 */

package org.spectrim.imagegraphicobjects;

import java.awt.Point;

/**
 * Constructor of a 3D-plane.
 * @author  hans
 */
public class Plane3D
{

    /**
     * Creates a 3D plane through point rho with direction vector r and s.
     * @param rho
     * @param r
     * @param s
     */
    public Plane3D( Point3D rho, Point3D r, Point3D s )
    {
        m_rho = rho;
        m_r = r;
        m_s = s;

        // Compute the normal vector
        m_n = m_r.computeNormalVector( m_s );

        m_k = m_n.m_x1 * rho.m_x1 + m_n.m_x2 * rho.m_x2 + m_n.m_x3 * rho.m_x3;

    }

    /**
     * Creates a 3D plane through point p with direction vector r and s.
     * @param rho
     * @param r
     * @param s
     * @param initialImageWidth
     * @param initialImageHeight
     * @param currentImageWidth
     * @param currentImageHeight
     * @param FOV_DIR_1
     * @param FOV_DIR_2
     * @param sliceThickness
     */
    public Plane3D( Point3D rho, Point3D r, Point3D s,
                    int initialImageWidth, int initialImageHeight,
                    int currentImageWidth, int currentImageHeight,
                    double FOV_DIR_1, double FOV_DIR_2,
                    double sliceThickness )
    {
        m_rho   = rho;
        m_r     = r;
        m_s     = s;
        m_sliceThickness = sliceThickness;

        m_currentImageWidth     = currentImageWidth;
        m_currentImageHeight    = currentImageHeight;

        m_initialImageWidth     = initialImageWidth;
        m_initialImageHeight    = initialImageHeight;

//        System.out.println( "Initial image width  : " + m_initialImageWidth );
//        System.out.println( "Initial image height : " + m_initialImageHeight );
//
//        System.out.println( "Current image width  : " + m_currentImageWidth );
//        System.out.println( "Current image height : " + m_currentImageWidth );


        m_FOV_DIR_1 = FOV_DIR_1;
        m_FOV_DIR_2 = FOV_DIR_2;

        // Compute the normal vector
        m_n = m_r.computeNormalVector( m_s );

        m_k = m_n.m_x1 * rho.m_x1 + m_n.m_x2 * rho.m_x2 + m_n.m_x3 * rho.m_x3;

    }


    /**
     * Returns the normal vector of the plane.
     * @return 
     */
    public Point3D getNormalVector( )
    {
        Point3D p = new Point3D( );
        return m_r.computeNormalVector( m_s );
    }

    /**
     * Returns a single point which is an element of the plane.
     * @return 
     */
    public Point3D getPointOnPlane( )
    {
        return m_rho;
    }

    /**
     * Substitutes a given point into the normal equation and returns the
     * result.
     * @param p
     * @return 
     */
    public double computePoint( Point3D p )
    {
        return  ( m_n.m_x1 * p.m_x1 + m_n.m_x2 * p.m_x2 + m_n.m_x3 * p.m_x3 );
    }

    /**
     * Returns the constant k of the normal equation representation of a
     * plane:
     *                 a.x + b.y + c.z = k.
     * @return 
     */
    public double getNormalPlaneEquationConstant( )
    {
        return m_k;
    }

    /**
     * Returns the distance from a given Point3D p to the plane.
     * @param p
     * @return 
     */
    public double getDistanceToPlane( Point3D p )
    {
        return ( Math.abs( m_n.m_x1 * p.m_x1 + m_n.m_x2 * p.m_x2 + m_n.m_x3 * p.m_x3 - m_k ) );
    }

    /**
     * Returns the projected point p.
     * @param q
     * @return 
     */
    public Point3D getProjectedPoint( Point3D q )
    {
        Point3D pq = new Point3D( );

        double alfa = ( m_k - q.m_x1 * m_n.m_x1 - q.m_x2 * m_n.m_x2 - q.m_x3 * m_n.m_x3 ) / ( ( m_n.m_x1 * m_n.m_x1 ) + ( m_n.m_x2 * m_n.m_x2 ) + ( m_n.m_x3 * m_n.m_x3 ) );

        pq.m_x1 = alfa * m_n.m_x1 + q.m_x1;
        pq.m_x2 = alfa * m_n.m_x2 + q.m_x2;
        pq.m_x3 = alfa * m_n.m_x3 + q.m_x3;

        return pq;
    }

    /**
     *
     * For some applications we need the closest Point in the discrete slice
     *
     * @param p
     * @return
     */
    public Point computeClosestPoint2D( Point3D p )
    {
        Point point2D = null;
        double distance = getDistanceToPlane( p );

        if ( distance < m_sliceThickness )
        {
            Point3D projPoint = getProjectedPoint( p );

            double detXY = ( m_s.m_x1 * m_r.m_x2 - m_s.m_x2 * m_r.m_x1 );
            double detXZ = ( m_s.m_x1 * m_r.m_x3 - m_s.m_x3 * m_r.m_x1 );
            double detYZ = ( m_s.m_x2 * m_r.m_x3 - m_s.m_x3 * m_r.m_x2 );

            double mhu = -1;

//            if ( ( detXY >= detYZ ) && ( detXY >= detYZ ) )
//            {
                mhu = getMhuFrom_X_Y( projPoint );
//            }
//
//            if ( ( detXZ >= detXY ) && ( detXZ >= detYZ ) )
//            {

              if ( mhu == 0.0 )
              {
                  mhu = getMhuFrom_X_Z( projPoint );
              }

//            }
//
//            if ( ( detYZ >= detXY ) && ( detYZ >= detXZ ) )
//            {

              if ( mhu == 0.0 )
              {
                 mhu = getMhuFrom_Y_Z( projPoint );
              }

//            }

            double lambda = getLambda( mhu, projPoint );

            int x = ( ( m_currentImageWidth - m_initialImageWidth ) / 2 ) + (int)( ( lambda * m_initialImageWidth / m_FOV_DIR_1 ) + 0.5 );
            int y = ( ( m_currentImageHeight - m_initialImageHeight ) / 2 ) + (int)( (   mhu * m_initialImageHeight / m_FOV_DIR_2 ) + 0.5 );

            return new Point( x, y );
        }
        else
        {
            return null;
        }
    }

    /**
     *
     * Computes mhu from the first two equations
     *
     * @param p
     * @return
     */
    public double getMhuFrom_X_Y( Point3D p )
    {
        double mhu = 0.0;

        if ( m_s.m_x1 * m_r.m_x2 - m_s.m_x2 * m_r.m_x1 != 0.0 )
        {
            mhu =  ( ( p.m_x1 * m_r.m_x2 - p.m_x2 * m_r.m_x1 ) - ( m_rho.m_x1 * m_r.m_x2-m_rho.m_x2 * m_r.m_x1 ) ) / ( m_s.m_x1 * m_r.m_x2 - m_s.m_x2 * m_r.m_x1 );
        }

        return mhu;
    }

    /**
     * Returns the lambda multiplier.
     * @param mhu
     * @param p
     * @return
     */
    public double getLambda( double mhu, Point3D p )
    {
        double lambda = -1;

        if ( ( m_r.m_x1 >= m_r.m_x2 ) && ( m_r.m_x1 >= m_r.m_x3 ) )
        {
            lambda = ( p.m_x1 - m_rho.m_x1 - mhu * m_s.m_x1 ) / m_r.m_x1;
        }

        if ( ( m_r.m_x2 >= m_r.m_x1 ) && ( m_r.m_x2 >= m_r.m_x3 ) )
        {
            lambda = ( p.m_x2 - m_rho.m_x2 - mhu * m_s.m_x2 ) / m_r.m_x2;
        }

        if ( ( m_r.m_x3 >= m_r.m_x1 ) && ( m_r.m_x3 >= m_r.m_x2 ) )
        {
            lambda = ( p.m_x3 - m_rho.m_x3 - mhu * m_s.m_x3 ) / m_r.m_x3;
        }

        return lambda;
    }


    /**
     *
     * Computes mhu from the first and the last equation.
     *
     * @param p
     * @return
     */
    public double getMhuFrom_X_Z( Point3D p )
    {
        double mhu = 0.0;

        if ( m_s.m_x1 * m_r.m_x3 - m_s.m_x3 * m_r.m_x1 != 0.0 )
        {
            mhu =  ( ( p.m_x1 * m_r.m_x3 - p.m_x3 * m_r.m_x1 ) - ( m_rho.m_x1 * m_r.m_x3 - m_rho.m_x3 * m_r.m_x1 ) ) / ( m_s.m_x1 * m_r.m_x3 - m_s.m_x3 * m_r.m_x1 );
        }

        return mhu;
    }

    /**
     *
     * Computes mhu from the first and the last equation
     *
     * @param p
     * @return
     */
    public double getMhuFrom_Y_Z( Point3D p )
    {

        double mhu =  0.0;

        if ( m_s.m_x2 * m_r.m_x3 - m_s.m_x3 * m_r.m_x2  != 0.0 )
        {
            mhu =  ( ( p.m_x2 * m_r.m_x3 - p.m_x3 * m_r.m_x2 ) - ( m_rho.m_x2 * m_r.m_x3 - m_rho.m_x3 * m_r.m_x2 ) ) / ( m_s.m_x2 * m_r.m_x3 - m_s.m_x3 * m_r.m_x2 );
        }

        return mhu;
    }

    /**
     *
     * Computes the intersection point between the plane and the line segment 
     * constructed from the two given points:
     * 
     * Returns null if plane not in-between line, line is located 
     * in-plane or parallel.
     *
     * @param A
     * @param B
     * @return
     */
    public Point3D getLineIntersection( Point3D A, Point3D B )
    {
        Point3D u = B.subtractPoint(A);
        Point3D w = A.subtractPoint(this.m_rho);
        double D = Point3D.dotProduct(this.m_n, u);
        double N = - Point3D.dotProduct(this.m_n, w);
        //System.out.println("Distance: "+Math.abs(D));
        if ( Math.abs(D) < 0.00000001)
        {
            return null;
        }
        double SI = N / D;
        //System.out.println("DistanceSI: "+SI);
        if(SI < 0.0 || SI > 1.0)
            return null;

        return A.addPoint(u.scalarMultiplication(SI));
    }

    // Member attributes
    private Point3D m_rho;
    private Point3D m_r;
    private Point3D m_s;

    private double  m_sliceThickness;

    private Point3D m_n;
    private double  m_k;

    private int     m_currentImageWidth;
    private int     m_currentImageHeight;
    private double  m_FOV_DIR_1;
    private double  m_FOV_DIR_2;
    private int     m_initialImageWidth;
    private int     m_initialImageHeight;
}
