/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.spectrim.imageprocessing;

import org.spectrim.imagegraphicobjects.Distance;
import org.spectrim.imagegraphicobjects.Point3D;

import java.awt.Point;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Date;

/**
 * Utility class to handle MeasuredDistancesFiles.
 * @author slotboom
 */
public class MeasuredDistancesFile
{
    private final DstHeaderStruct hdr;
    private RandomAccessFile f;
    private String fileMode = "";
    private final String m_imafile;
    

    /**
     * Constructor of a MeasuredDistancesFile-object.
     * @param fname
     * @param w
     * @param h
     * @param imafile file reference to which this file belongs ...
     */
    public MeasuredDistancesFile( String fname, int w, int h, String imafile )
    {
        hdr = new DstHeaderStruct( fname, w, h );
        m_imafile = imafile;
    }
    
    /**
     * Opens the file in mode 'mode'. 
     * @param mode
     * @return 
     */
    public boolean open( String mode )
    {
        try
        {
            //close( );
            f = new RandomAccessFile( hdr.fileName, mode );
        }
        catch ( IOException e )
        {
            System.out.println( "File " + hdr.fileName + " not found" );
            return false;
        }
        catch ( IllegalArgumentException e )
        {
            System.out.println( "Illegal file mode: " + mode );
            return false;
        }
        fileMode = mode;
        return true;
    }

    
    /**
     * Closes the file in the file system.
     */
    public void close( )
    {
        if ( f == null ) return;
        try
        {
            f.close( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
        f = null;
        fileMode = "";
    }

    /**
     * Writes the header of the measured distances file of the 
     * MeasuredDistancesFile object.
     */
    public void writeHeader( )
    {
        if ( f == null ) return;
        try
        {
            f.seek(0);
            f.writeBytes( "#$distances$\n");
            f.writeBytes( "# " + hdr.fileName + "\n" );
            Date date = new Date( );
            f.writeBytes( "# automatically created at " + date.toString( ) + "\n" );

            String machine = "<machine>";
            f.writeBytes( "# on " + machine + " " +
                          System.getProperty( "os.name" ) + " " +
                          System.getProperty( "os.arch" ) +
                          " from " +
                          System.getProperty("user.name") + "\n" );
            f.writeBytes( "# \n");
            f.writeBytes( "imagefile=" + m_imafile + "\n" );
            f.writeBytes( "width=" + hdr.imageWidth + "\n" );
            f.writeBytes( "height=" + hdr.imageHeight + "\n" );
            f.writeBytes( "normal={-0|0|1}\n" );
            f.writeBytes( "pos={0|0|<slice pos>}\n" );
            f.writeBytes( "slice=<slice id>\n" );
            f.writeBytes( "# \n" );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }

    /**
     * Writes the distance data attributes to the file.
     * @param p
     * @param zoom
     */
    public void writeDistanceData( Distance p, float zoom )
    {
        if ( f == null ) return;
        try
        {
            f.writeBytes( "# \n" );

            f.writeBytes( "distance=" + p.getDistanceName( ) + "\n" );

            f.writeBytes( p.getDistanceValue( ) + "\n" );

            f.writeBytes( p.getDistanceNumber( ) + "\n" );

            // begin Point
            f.writeBytes( p.getBeginPoint( ).x + "\n" );
            f.writeBytes( p.getBeginPoint( ).y + "\n" );

            //end Point
            f.writeBytes( p.getEndPoint( ).x + "\n" );
            f.writeBytes( p.getEndPoint( ).y + "\n" );

            // begin Point3D
            f.writeBytes( p.getBeginPoint3D( ).m_x1 + "\n" );
            f.writeBytes( p.getBeginPoint3D( ).m_x2 + "\n" );
            f.writeBytes( p.getBeginPoint3D( ).m_x3 + "\n" );

            //end Point3D
            f.writeBytes( p.getEndPoint3D( ).m_x1 + "\n" );
            f.writeBytes( p.getEndPoint3D( ).m_x2 + "\n" );
            f.writeBytes( p.getEndPoint3D( ).m_x3 + "\n" );

            f.writeBytes( p.getImageNumber( ) + "\n" );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }

    
    /**
     * Writes the distance attributes to the file in one line.
     * @param p
     * @param zoom
     */
    public void writeDistanceDataOnOneLine( Distance p, float zoom )
    {
        if ( f == null ) return;
        try
        {
            f.writeBytes( "# \n" );

            f.writeBytes( p.getImageNumber( ) + "\t" );
            
            f.writeBytes( "distance=" + p.getDistanceName( ) + "\t" );

            f.writeBytes( p.getDistanceValue( ) + "\t" );

            f.writeBytes( p.getDistanceNumber( ) + "\t" );

            // begin Point
            f.writeBytes( p.getBeginPoint( ).x + "\t" );
            f.writeBytes( p.getBeginPoint( ).y + "\t" );

            //end Point
            f.writeBytes( p.getEndPoint( ).x + "\t" );
            f.writeBytes( p.getEndPoint( ).y + "\t" );

            // begin Point3D
            f.writeBytes( p.getBeginPoint3D( ).m_x1 + "\t" );
            f.writeBytes( p.getBeginPoint3D( ).m_x2 + "\t" );
            f.writeBytes( p.getBeginPoint3D( ).m_x3 + "\t" );

            //end Point3D
            f.writeBytes( p.getEndPoint3D( ).m_x1 + "\t" );
            f.writeBytes( p.getEndPoint3D( ).m_x2 + "\t" );
            f.writeBytes( p.getEndPoint3D( ).m_x3 + "\t" );

            f.writeBytes( p.getImageNumber( ) + "\n" );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }
    
    

    /**
     * Finishes writing the file ..
     * @return
     */
    public boolean finish( )
    {
        if ( f == null ) return false;
        try
        {
            f.writeBytes( "# end\n" );
            f.close( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
            return false;
        }
        return true;
    }


    /**
     * Reads the header information from file..
     */
    public void readHeader( )
    {
        if ( f == null ) return;
        String line;
        try
        {
            f.seek( 0 );
            skipComments( );
        }
        catch ( IOException e )
        {
            System.out.println( e );
        }
    }
    
    /**
     * Reads distance data object from file..
     * @param zoom
     * @return
     */
    public Distance readDistanceData( float zoom )
    {
        Distance p;
        String parsLine;
        try
        {
            String distanceName = gotoNextDistanceEntry( );

            //String distanceName = readDistanceName( );

            //if ( indexString == null ) return null;

            //int index = Integer.valueOf( indexString.substring( 7 ) ).intValue( );
            if ( distanceName == null )
            {
                return null;
            }

            p = new Distance( );

            p.setDistanceName( distanceName );

            parsLine = f.readLine( );
            p.setDistanceValue( new Float( parsLine ) );

            parsLine = f.readLine( );
            p.setDistanceNumber( new Integer( parsLine ) );

            parsLine = f.readLine( );
            int bpx = new Integer( parsLine );

            parsLine = f.readLine( );
            int bpy = new Integer( parsLine );

            p.setBeginPoint( new Point( bpx, bpy ) );

            parsLine = f.readLine( );
            int epx = new Integer( parsLine );

            parsLine = f.readLine( );
            int epy = new Integer( parsLine );

            p.setEndPoint( new Point( epx, epy ) );

            parsLine = f.readLine( );
            float bp1_3d = new Float( parsLine );

            parsLine = f.readLine( );
            float bp2_3d = new Float( parsLine );

            parsLine = f.readLine( );
            float bp3_3d = new Float( parsLine );


            p.setBeginPoint3D( new Point3D( bp1_3d, bp2_3d, bp3_3d ) );

            parsLine = f.readLine( );
            float ep1_3d = new Float( parsLine );

            parsLine = f.readLine( );
            float ep2_3d = new Float( parsLine );

            parsLine = f.readLine( );
            float ep3_3d = new Float( parsLine );

            p.setEndPoint3D( new Point3D( ep1_3d, ep2_3d, ep3_3d ) );
            
            // set image number it belongs to
            parsLine = f.readLine( );

            int imageNumber = Integer.parseInt(parsLine );
            p.setImageNumber( imageNumber );

            return p;

        }
        catch ( IOException e )
        {
            System.out.println( e );
            return ( Distance )null;
        }
        
    }

    /**
     * Skips all lines starting with a '#';
     * These lines are comments.
     */
    private void skipComments( ) throws IOException
    {
        String line;
        do
        {
            line = f.readLine( );
        }
        while ( line.startsWith( "#" ) );
    }

    /**
     * Returns the line starting with 'regio='.
     */
    private String gotoNextDistanceEntry( ) throws IOException
    {
        String line;
        do
        {
            line = f.readLine( );
        } while ( line != null && !line.startsWith( "distance=" ) );

        if ( line == null )
        {
            return line;
        }
        else
        {
            return line.substring( line.indexOf( "=" ) + 1, line.length( ) );
        }
    }

    /**
     * Reads the line with contains regio m_label information.
     */
    private String readDistanceName( ) throws IOException
    {
        String line;
        do
        {
            line = f.readLine( );
        }
        while ( line != null && !line.startsWith( "label=" ) );
        if ( line == null )
        {
            return line;
        }
        else
        {
            return line.substring( line.indexOf( "=" ) + 1, line.indexOf( ":::" ) );
        }
    }
}

class DstHeaderStruct
{
    String fileName;
    int imageWidth;
    int imageHeight;

    public DstHeaderStruct( String name, int w, int h )
    {
        fileName = name;
        imageWidth = w;
        imageHeight = h;
    }
}
