/*
 * File: $Id: MorphologicFilters.java,v 1.1.1.1 2006/04/19 09:38:48 hans Exp $
 * History
 * Date: 20.03.2007 Hans Slotboom  Changes:
 *
 * Copyright (C) 2007 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


package org.spectrim.imageprocessing;

/**
 * Utility class that enables morphologic filtering being applied to the 
 * images.
 * @author hans
 */
public class MorphologicFilters 
{
    
    /**
     * Creates a new instance of MorphologicFilters with a default N4 kernel.
     */
    public MorphologicFilters( ) 
    {
        m_standardStructuringElement_N4 = true;
        
        m_filterSizeX = 1;
        m_filterSizeY = 1;
        m_filterSize = 1;
    }
    
    /**
     * Constructor: creates a new instance of MorphologicFilters. 
     * @param filterSize
     */
    public MorphologicFilters( int filterSize )
    {
        m_standardStructuringElement_N4 = false;
        m_filterSize = filterSize;
    }
    
     /**
     * Creates a new instance of MorphologicFilters. 
     * @param filterSizeX
     * @param filterSizeY
     */
    public MorphologicFilters( int filterSizeX, int filterSizeY )
    {
        m_standardStructuringElement_N4 = false;
        m_filterSizeX = filterSizeX;
        m_filterSizeY = filterSizeY;
        
    }
     
    /**
     * 
     * Perform erosion of the boolean-type input image x[][]
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @return the filtered boolean image
     */
    public boolean[][] erode( boolean[][] x, int imageWidth, int imageHeight )
    {
        // First, create an appropriate morphologic operator
        MorphologicOperator imageOperator;
        if ( m_standardStructuringElement_N4 == false )
        {
            imageOperator = new MorphologicOperator( m_filterSizeX, m_filterSizeY );
        }
        else
        {
            imageOperator = new MorphologicOperator( );
        }
        
        x = imageOperator.erode( imageWidth, imageHeight, x );
        
        return x;
    }
    
    /** 
     * Perform erosion of the boolean-type input image x[][].
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @return the filtered boolean image
     */
    public boolean[][] dilate( boolean[][] x, int imageWidth, int imageHeight )
    {
        // First, create an appropriate morphologic operator
        MorphologicOperator imageOperator;
        if ( m_standardStructuringElement_N4 == false )
        {
            imageOperator = new MorphologicOperator( m_filterSizeX, m_filterSizeY );
        }
        else
        {
            imageOperator = new MorphologicOperator( );
        }
        
        x = imageOperator.dilate( imageWidth, imageHeight, x );
        
        return x;
    }
    
    /**
     * 
     * Applies a simple Morphologic Filter to the image dataset x.
     * 
     * @param x
     * @param imageWidth
     * @param imageHeight
     * @param minimumThresshold
     * @param maximumThresshold
     * @param filterSize
     * @return 
     */
    public int[] applySimpleMorphologicFilter( int[] x, int imageWidth, int imageHeight, 
                                               int minimumThresshold, int maximumThresshold, 
                                               int filterSize )
    {
        int[] mask = new int[ x.length ];
        for ( int i = 0 ; i < mask.length ; i++ )
        {
            mask[i] = 1;
        }
        
        MaximumMask maxMask = new MaximumMask( maximumThresshold, 0 );
        mask = maxMask.applyMaximumMask( mask, x );
        
        MinimumMask minMask = new MinimumMask( minimumThresshold, 0 );
        mask = minMask.applyMinimumMask( mask, x );
        
        MorphologicOperator imageOperator;
        
        if ( m_standardStructuringElement_N4 == false )
        {
            imageOperator = new MorphologicOperator( filterSize );
        }
        else
        {
            imageOperator = new MorphologicOperator( );
        }
        
        // first erode the region
        mask = imageOperator.erode( imageWidth, imageHeight, mask );
       
        // first dilate the region
        mask = imageOperator.dilate( imageWidth, imageHeight, mask );
        
        for ( int i = 0; i < x.length; i++ )
        {
            //x[ i ] = x[ i ] * mask[ i ];
            x[ i ] = mask[ i ];
        }
        
        return x;
    }
    
     /**
     * Applies a simple Morphologic Filter an images stack the image dataset x.
     * @param x
     * @param numberOfImages
     * @param imageWidth
     * @param imageHeight
     * @param minimumThresshold
     * @param maximumThresshold
     * @param filterSize
     * @return 
     */
    public int[] applySimpleMorphologicFilter( int[] x, int numberOfImages, 
                                               int imageWidth, int imageHeight, 
                                               int minimumThresshold, 
                                               int maximumThresshold, 
                                               int filterSize )
    {
        // First mask the data
        int[] mask = new int[ x.length ];
        for ( int i = 0 ; i < mask.length ; i++ )
        {
            mask[i] = 1;
        }
        
        MaximumMask maxMask = new MaximumMask( maximumThresshold, 0 );
        mask = maxMask.applyMaximumMask( mask, x );
        
        MinimumMask minMask = new MinimumMask( minimumThresshold, 0 );
        mask = minMask.applyMinimumMask( mask, x );
        
        MorphologicOperator imageOperator;
        if ( m_standardStructuringElement_N4 == false )
        {
            imageOperator = new MorphologicOperator( filterSize );
        }
        else
        {
            imageOperator = new MorphologicOperator( );
        }
        
        // The morphologic filter is applied on the imagestack image by image
        int[] subMask = new int [imageWidth*imageHeight];
        
        for ( int i = 0; i < numberOfImages; i++ )
        {
            // System.out.println( "Filtering image : " + i );
            
            int offset = i * imageWidth * imageHeight;
            
            for ( int j = 0; j < imageWidth * imageHeight; j++ )
            {
                subMask[j] = mask[ offset + j ];
            }
            
            // first erode the image
            subMask = imageOperator.erode( imageWidth, imageHeight, subMask );
       
            // than dilate the image
            subMask = imageOperator.dilate( imageWidth, imageHeight, subMask );
        
            for ( int k = 0; k < subMask.length; k++ )
            {
                x[ offset + k ] = x[ offset + k ] * subMask[ k ];
            }
        }
        
        return x;
    }
    
    // Attributes
    
    private boolean m_standardStructuringElement_N4;
    
    private int m_filterSize  = 0;
    
    private int m_filterSizeX = 0;
    private int m_filterSizeY = 0;
}
