/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package org.spectrim.spectra;

import org.spectrim.lib.DoubleRange;
import java.text.DecimalFormat;
import org.spectrim.model.GlobalModel;
import org.spectrim.utilities.MatlabIO;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.ArrayRealVector;
import org.apache.commons.math3.linear.DecompositionSolver;
import org.apache.commons.math3.linear.QRDecomposition;
import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.linear.RealVector;
import org.spectrim.batchprocessing.BatchProcessing;
import org.spectrim.curvefitting.Exponential;
import org.spectrim.curvefitting.ExponentialFit;

/**
 *
 * @author nunopbarros
 */
public class Spectra {
    
    protected GlobalModel model;
    protected HashMap<String, Spectrum> data = new HashMap<>();
    protected int vectorsize = 0;
    protected int zerofillingsize;
    protected double spectralWidth;
    protected double resonanceFrequency;//Hz
    protected double ppmReference;
    protected int gridsize1;
    protected int gridsize2;
    protected int gridsize3;
    protected double[] xAxisTD;//ms
    protected double[] xAxisFDppm;//ppm
    protected double[] xAxisFDHz;//MHz
    protected double specScaleFD;
    protected HashMap<String, Double> qscores = new HashMap<>();
    protected double globalFreqShift;
    protected int nrClusters = 0;
    protected ArrayList<String> coordsLocalShiftsExecuted;
    protected  int totalLocalShiftsFinished = 0;
    protected boolean isFftDone;

    public void setVectorSize( int vectorSize ) 
    {
        vectorsize = vectorSize;
    }

    public void complexConjugateAllSpectra() 
    {
        String key;
        Iterator iter = data.keySet( ).iterator( );
        
        while ( iter.hasNext( ) ) 
        {
            key = (String) iter.next( );
            data.get(key).complexConjugate( );
        }
    }
    
    public enum Mode {
        
        ABS, DISP, MAG
    };
    
    public enum Domain {
        
        FD_PPM("FD - ppm"), FD_HZ("FD - MHz"), TD("TD - ms");
        public final String txt;
        
        private Domain(String txt) {
            this.txt = txt;
        }
    };
    
    public enum CombinationFunction {
        
        MEAN("mean"), MEAN_STDDEV("mean +- std dev"), STD_DEV("std deviation"),
        BOXPLOT_VIEW("\"boxplot\" (min,25th,median,75th,max)"), MIN_MAX("min max"),
        MIN_MEDIAN_MAX("min median max"), MEDIAN("median"), SEPARATE("separate");
        
        public final String txt;
        
        private CombinationFunction(String txt) {
            this.txt = txt;
        }
    };
    
    protected Domain defaultDomain = Domain.FD_PPM;
    protected CombinationFunction defaultCombinationFunction = CombinationFunction.SEPARATE;
    
    public Spectra( float[] abs, float[] disp, int dim1, int dim2, int dim3, 
                    double spectwidth, double resonfreq, double ppmref, 
                    boolean isFftDone ) 
    {
        gridsize1 = dim1;
        gridsize2 = dim2;
        gridsize3 = dim3;
        this.isFftDone = isFftDone;
        spectralWidth = spectwidth;
        resonanceFrequency = resonfreq;//Hz
        vectorsize = abs.length / (dim1 * dim2 * dim3);
        ppmReference = ppmref;
        calculateAxis();
        splitRawData(abs, disp);
    }
    
    Spectra( GlobalModel model, HashMap<String, Spectrum> data, 
             int vectorsize, int zerofillingsize, double spectralWidth,
             double resonanceFrequency, double ppmReference, 
             int gridsize1, int gridsize2, int gridsize3, double[] xAxisTD,
             double[] xAxisFDppm, double[] xAxisFDHz, double specScaleFD, HashMap<String, Double> qscores, Domain domain,
             CombinationFunction combfunc, boolean isFftDone ) 
    {
        this.model = model;
        this.data = (HashMap<String, Spectrum>) data.clone();
        this.vectorsize = vectorsize;
        this.zerofillingsize = zerofillingsize;
        this.spectralWidth = spectralWidth;
        this.resonanceFrequency = resonanceFrequency;//Hz
        this.ppmReference = ppmReference;
        this.gridsize1 = gridsize1;
        this.gridsize2 = gridsize2;
        this.gridsize3 = gridsize3;
        this.xAxisTD = xAxisTD;
        this.xAxisFDppm = xAxisFDppm;
        this.xAxisFDHz = xAxisFDHz;
        this.specScaleFD = specScaleFD;
        this.qscores = (HashMap<String, Double>) qscores.clone();
        this.defaultCombinationFunction = combfunc;
        this.defaultDomain = domain;
        this.isFftDone = isFftDone;
    }
    
    public Domain getDefaultDomain() {
        return defaultDomain;
    }
    
    public void setDefaultDomain(Domain defaultDomain) {
        this.defaultDomain = defaultDomain;
    }
    
    public CombinationFunction getDefaultCombinationFunction() {
        return defaultCombinationFunction;
    }
    
    public void setDefaultCombinationFunction(CombinationFunction defaultCombinationFunction) {
        this.defaultCombinationFunction = defaultCombinationFunction;
    }
    
    public int[] getGridSize( ) 
    {
        return new int[]{gridsize1, gridsize2, gridsize3};
    }
    
    public FittedSpectra cloneAsFittedSpectra( ) 
    {
        return new FittedSpectra( model, data, vectorsize, zerofillingsize, spectralWidth,
                                  resonanceFrequency, ppmReference, 
                                  gridsize1, gridsize2, gridsize3, xAxisTD,
                                  xAxisFDppm, xAxisFDHz, specScaleFD, qscores, 
                                  defaultDomain, defaultCombinationFunction );
    }
    
    public void clearSpectraValues( ) 
    {
        data.clear( );
    }
    
    public void setModel( GlobalModel model ) 
    {
        this.model = model;
    }
    
    public void setSpectrum( float[] absFID, float[] dispFID, String ID ) 
    {
        //This method should be used only to update spectra that was already created from the DICOM.
        //Created to allow import from jMRUI.
        if ( data.containsKey( ID ) ) 
        {
            this.vectorsize = absFID.length;
            this.zerofillingsize = 0;
            //Not foreseeen situations were std dev of Noise should not be retained.
            //maintaining this value with the one calculated from the initial raw data
            //makes it robust to pre-processing steps such as Apodization, which would
            //change drastically values of SNR or the normalization with noise std dev.
            Spectrum spect = new Spectrum(absFID, dispFID, this);
            float FDNoise = data.get(ID).getNoiseStdDevFD();
            float TDNoise = data.get(ID).getNoiseStdDevTD();
            spect.setNoiseStdFD(FDNoise);
            spect.setNoiseStdTD(TDNoise);
            data.put( ID, spect );
        }
    }
    
    public int getVectorSize( ) 
    {
        return vectorsize;
    }
    
    public int getZeroFillingSize( ) 
    {
        return zerofillingsize;
    }
    
    public HashMap<String, Spectrum> getData( ) 
    {
        return data;
    }
    
    public void normalizeSpectra( boolean bool ) 
    {
        this.model.normalizeSpectra( bool, false );
    }
    
    protected void splitRawData( float[] abs, float[] disp ) 
    {
        float[] absaux = new float[vectorsize];
        float[] dispaux = new float[vectorsize];
        Spectrum spectaux;
        for ( int k = 0; k < gridsize3; k++ ) 
        {
            for ( int j = 0; j < gridsize2; j++ ) 
            {
                for ( int i = 0; i < gridsize1; i++ ) 
                {
                    for ( int t = 0; t < (vectorsize); t++ ) 
                    {
                        int aLoc = k * gridsize1 * gridsize2 * vectorsize
                                   + j * gridsize1 * vectorsize
                                   + i * vectorsize + t;
                        absaux[t]  = abs[aLoc];
                        dispaux[t] = disp[aLoc];
                    }
                    spectaux = new Spectrum( absaux, dispaux, this );
                    data.put( i + "_" + j + "_" + k, spectaux );
                }
            }
        }
    }
    
    protected void calculateAxis() {
        calculateTDXaxis();
        calculateFDXaxisHertz();
        calculateFDXaxisPPM();
    }

    /**
     * Calculates the time domain xaxis in ms
     */
    protected void calculateTDXaxis() {
        xAxisTD = new double[vectorsize + zerofillingsize];
        
        for (int i = 0; i < (vectorsize + zerofillingsize); i++) {
            xAxisTD[i] = (double) i * (1000d / spectralWidth); //ms
        }
    }

    /**
     * Calculates the frequency domain xaxis in MHz
     */
    protected void calculateFDXaxisHertz() {
        int l = (vectorsize + zerofillingsize);
        xAxisFDHz = new double[l];
        for (int i = 0; i < l; i++) {
            xAxisFDHz[i] = resonanceFrequency / 1000000 + ( (double) i / (double) (l - 1) - 0.5d ) * spectralWidth / 1000000; //MHz
        }
    }

    /**
     * Calculates the frequency domain xaxis in ppm
     */
    protected void calculateFDXaxisPPM() {
        int l = (vectorsize + zerofillingsize);
        xAxisFDppm = new double[l];
        for (int i = 0; i < l; i++) {
            xAxisFDppm[i] = ppmReference + ((double) i / (double) (l - 1) - 0.5d) * (1000000 * spectralWidth / resonanceFrequency); //ppm
        }
    }
    
    public double getResonanceFrequency() {//Hz
        return resonanceFrequency;
    }
    
    public double getPPMReference() {
        return ppmReference;
    }
    
    public double getDwellTime() {
        return 1 / spectralWidth;
    }
    
    public double getFDstep() {
        return specScaleFD;
    }
    
    public double[] getXAxis() {
        return getXAxis(defaultDomain);
    }
    
    public double[] getXAxis(Domain domain) {
        switch (domain) {
            case TD:
                return xAxisTD;
            case FD_HZ:
                return xAxisFDHz;
            case FD_PPM:
                return xAxisFDppm;
        }
        return null;
    }
    
    public int getIndexScalePPM(double ppm) {
        return (int) Math.round(((ppm - ppmReference) * resonanceFrequency / (1000000 * spectralWidth) + 0.5d) * (vectorsize + zerofillingsize - 1));
    }
    
    public int getIndexScaleTD(double time) { //ms
        return (int) Math.round(time * spectralWidth / 1000);
    }
    
    public double getScaleTD() {//ms
        return (this.xAxisTD[1] - this.xAxisTD[0]);
    }
    
    public double getScaleHz() {//MHz
        return (this.xAxisFDHz[1] - this.xAxisFDHz[0]);
    }
    
    public double getScalePPM() {//ppm
        return (this.xAxisFDppm[1] - this.xAxisFDppm[0]);
    }
    
    public void zerofilling( int nzeros ) 
    {
        if ( zerofillingsize != nzeros ) 
        {
            zerofillingsize = nzeros;
            Set<String> keys = data.keySet( );
            Iterator iter = keys.iterator( );
            String aux;
            calculateAxis( );
            while ( iter.hasNext( ) ) 
            {
                aux = (String) iter.next( );
                data.get(aux).zerofilling( nzeros );
            }
        }
    }
    
    /**
     * This method truncates the data in time domain. 
     * @param truncation_point 
     */
    public void truncate_td_signal( int truncation_point ) 
    {
        
        if ( this.vectorsize != truncation_point ) 
        {
            Set<String> keys = data.keySet( );
            Iterator iter = keys.iterator( );
            String aux;
            calculateAxis( );
            while ( iter.hasNext( ) ) 
            {
                aux = (String) iter.next( );
                data.get( aux ).truncate_data( truncation_point );
            }
        }
        
        vectorsize = truncation_point;
        calculateAxis( );
       
    }
    
    public boolean isNormalized( ) 
    {
        if ( model != null ) 
        {
            return model.isSpectraNormalized( );
        } 
        else 
        {
            return false;
        }
    }
    
    public void autoPhase( float minppm, float maxppm ) 
    {
        autoPhase( data.keySet( ).iterator( ), minppm, maxppm );
    }
    
    public void autoPhase(ArrayList<String> ids, float minppm, float maxppm) {
        autoPhase(ids.iterator(), minppm, maxppm);
    }
    
    
    public void manualPhase(ArrayList<String> ids, float zeroOrder, float firstOrder) {
        manualPhase(ids.iterator(), zeroOrder, firstOrder);
    }
    
    public void autoPhase(Iterator iter, float minppm, float maxppm) {
        String aux;
        while (iter.hasNext()) {
            aux = (String) iter.next();
            data.get(aux).autoPhase(minppm, maxppm);
        }
    }
    
    public void manualPhase(Iterator iter, float zeroOrder, float firstOrder) {
        String aux;
        while (iter.hasNext()) 
        {
            aux = (String) iter.next( );
            data.get(aux).manualPhase(zeroOrder, firstOrder);
        }
    }
    
    public void apodizeExponential(double a) {
        apodizeExponential(data.keySet().iterator(), a);
    }
    
    public void apodizeExponential(ArrayList<String> ids, double a) {
        apodizeExponential(ids.iterator(), a);
    }
    
    public void apodizeExponential(Iterator iter, double a) {
        String aux;
        while (iter.hasNext()) {
            aux = (String) iter.next();
            data.get(aux).apodizeExponential(a);
        }
    }
    
    public void apodizeGaussian(double a) {
        apodizeGaussian(data.keySet().iterator(), a);
    }
    
    public void apodizeGaussian(ArrayList<String> ids, double a) {
        apodizeGaussian(ids.iterator(), a);
    }
    
    public void apodizeGaussian(Iterator iter, double a) {
        String aux;
        while (iter.hasNext()) {
            aux = (String) iter.next();
            data.get(aux).apodizeGaussian(a);
        }
    }
    
    public float[] getSpectrumData(int i, int j, int k, Mode mode) {
        return getSpectrumData(i + "_" + j + "_" + k, defaultDomain, mode);
    }
    
    public float[] getSpectrumData(String id, Mode mode) {
        return getSpectrumData(id, defaultDomain, mode);
    }
    
    public ArrayList<float[]> getSpectrumData(ArrayList<String> ids, Mode mode) {
        return getSpectrumData(ids, defaultCombinationFunction, defaultDomain, mode);
    }
    
    public float[] getSpectrumData(int i, int j, int k, Domain domain, Mode mode) {
        return getSpectrumData(i + "_" + j + "_" + k, domain, mode);
    }
    
    public float[] getSpectrumData(String id, Domain domain, Mode mode) {
        if (data.containsKey(id)) {
            switch (domain) {
                case TD:
                    switch (mode) {
                        case ABS:
                            return data.get(id).getAbsorptionTD();
                        case DISP:
                            return data.get(id).getDispersionTD();
                        case MAG:
                            return data.get(id).getMagnitudeTD();
                    }
                case FD_PPM:
                case FD_HZ:
                    switch (mode) {
                        case ABS:
                            return data.get(id).getAbsorptionFD();
                        case DISP:
                            return data.get(id).getDispersionFD();
                        case MAG:
                            return data.get(id).getMagnitudeFD();
                    }
            }
        }
        return null;
    }
    
    public ArrayList<float[]> getSpectrumData(ArrayList<String> ids, CombinationFunction combination, Domain domain, Mode mode) {
        if ( !containsID( ids ) ) 
        {
            return null;
        }
        Iterator iter = ids.iterator( );
        ArrayList<float[]> result = new ArrayList<>( );
        float[] res = new float[ vectorsize + zerofillingsize ];
        
        switch (combination) {
            case SEPARATE:
                if (containsID(ids)) {
                    while (iter.hasNext()) {
                        String key = (String) iter.next();
                        if (data.containsKey(key)) {
                            switch (domain) {
                                case TD:
                                    switch (mode) {
                                        case ABS:
                                            result.add(data.get(key).getAbsorptionTD());
                                            break;
                                        case DISP:
                                            result.add(data.get(key).getDispersionTD());
                                            break;
                                        case MAG:
                                            result.add(data.get(key).getMagnitudeTD());
                                            break;
                                    }
                                    break;
                                case FD_PPM:
                                case FD_HZ:
                                    switch (mode) {
                                        case ABS:
                                            result.add(data.get(key).getAbsorptionFD());
                                            break;
                                        case DISP:
                                            result.add(data.get(key).getDispersionFD());
                                            break;
                                        case MAG:
                                            result.add(data.get(key).getMagnitudeFD());
                                            break;
                                    }
                            }
                        }
                    }
                }
                break;
            case MEAN:
                result.add(getMeanSpectrum(ids, domain, mode));
                break;
            case STD_DEV:
                res = getStdDevSpectrum(ids, domain, mode);
                result.add(res);
                break;
            case MEAN_STDDEV:
                float resplus[] = new float[res.length];
                float resminus[] = new float[res.length];
                res = getMeanSpectrum(ids, domain, mode);
                float[] stdDev = getStdDevSpectrum(ids, domain, mode, res);
                for (int i = 0; i < res.length; i++) {
                    resplus[i] = res[i] + stdDev[i];
                    resminus[i] = res[i] - stdDev[i];
                }
                result.add(res);
                result.add(resplus);
                result.add(resminus);
                break;
            case BOXPLOT_VIEW:
                result.addAll(getPercentilesSpectra(ids, domain, mode, new float[]{0, 25, 50, 75, 100}));
                break;
            case MIN_MAX:
                result.addAll(getPercentilesSpectra(ids, domain, mode, new float[]{0, 100}));
                break;
            case MIN_MEDIAN_MAX:
                result.addAll(getPercentilesSpectra(ids, domain, mode, new float[]{0, 50, 100}));
                break;
            case MEDIAN:
                result.addAll(getPercentilesSpectra(ids, domain, mode, new float[]{50}));
                break;
        }
        return result;
    }
    
    public float[] getMeanSpectrum(ArrayList<String> ids, Domain domain, Mode mode) {
        float[] res = new float[vectorsize + zerofillingsize];
        if (!containsID(ids)) {
            return res;
        }
        
        if (mode == Mode.MAG) {
            float[] mean_abs = getMeanSpectrum(ids, domain, mode.ABS);
            float[] mean_disp = getMeanSpectrum(ids, domain, mode.DISP);
            for (int i = 0; i < res.length; i++) {
                res[i] = (float) Math.sqrt(Math.pow(mean_abs[i], 2) + Math.pow(mean_disp[i], 2));
            }
        } else {
            float[] aux = new float[res.length];
            for (String key : ids) {
                switch (domain) {
                    case TD:
                        switch (mode) {
                            case ABS:
                                aux = data.get(key).getAbsorptionTD();
                                break;
                            case DISP:
                                aux = data.get(key).getDispersionTD();
                        }
                        break;
                    case FD_PPM:
                    case FD_HZ:
                        switch (mode) {
                            case ABS:
                                aux = data.get(key).getAbsorptionFD();
                                break;
                            case DISP:
                                aux = data.get(key).getDispersionFD();
                        }
                }
                for (int i = 0; i < res.length; i++) {
                    res[i] = res[i] + aux[i] / ids.size();
                }
            }
        }
        return res;
    }
    
    private ArrayList<float[]> getPercentilesSpectra(ArrayList<String> ids, Domain domain, Mode mode, float[] percentiles) {
        assert (percentiles.length > 0) : "Percentiles array empty!";
        for (float val : percentiles) {
            assert (val >= 0 && val <= 100) : "Percentile out of range [0 100]. Value:" + val;
        }
        
        ArrayList<float[]> results = new ArrayList<>();
        if (!containsID(ids)) {
            return null;
        }
        float[] aux = new float[vectorsize + zerofillingsize];
        ArrayList<float[]> values2sort = new ArrayList<>();
        for (int i = 0; i < aux.length; i++) {
            values2sort.add(new float[ids.size()]);
        }
        
        for (int j = 0; j < ids.size(); j++) {
            String key = ids.get(j);
            switch (domain) {
                case TD:
                    switch (mode) {
                        case ABS:
                            aux = data.get(key).getAbsorptionTD();
                            break;
                        case DISP:
                            aux = data.get(key).getDispersionTD();
                            break;
                        case MAG:
                            aux = data.get(key).getMagnitudeTD();
                    }
                    break;
                case FD_PPM:
                case FD_HZ:
                    switch (mode) {
                        case ABS:
                            aux = data.get(key).getAbsorptionFD();
                            break;
                        case DISP:
                            aux = data.get(key).getDispersionFD();
                            break;
                        case MAG:
                            aux = data.get(key).getMagnitudeFD();
                            break;
                    }
            }
            
            for (int i = 0; i < aux.length; i++) {
                values2sort.get(i)[j] = aux[i];
            }
        }
        
        for (int i = 0; i < percentiles.length; i++) {
            results.add(new float[aux.length]);
        }
        
        for (int j = 0; j < aux.length; j++) {
            Arrays.sort(values2sort.get(j));
            for (int i = 0; i < percentiles.length; i++) {
                results.get(i)[j] = values2sort.get(j)[Math.round((ids.size() - 1) * percentiles[i] / 100)];
            }
        }
        return results;
    }
    
    private float[] getStdDevSpectrum(ArrayList<String> ids, Domain domain, Mode mode) {
        float[] meanSpectrum = getMeanSpectrum(ids, domain, mode);
        return getStdDevSpectrum(ids, domain, mode, meanSpectrum);
    }
    
    private float[] getStdDevSpectrum(ArrayList<String> ids, Domain domain, Mode mode, float[] meanSpectrum) {
        float[] res = new float[vectorsize + zerofillingsize];
        if (!containsID(ids)) {
            return res;
        }
        float[] aux = new float[res.length];
        for (String key : ids) {
            if (data.containsKey(key)) {
                switch (domain) {
                    case TD:
                        switch (mode) {
                            case ABS:
                                aux = data.get(key).getAbsorptionTD();
                                break;
                            case DISP:
                                aux = data.get(key).getDispersionTD();
                                break;
                            case MAG:
                                aux = data.get(key).getMagnitudeTD();
                        }
                        break;
                    case FD_PPM:
                    case FD_HZ:
                        switch (mode) {
                            case ABS:
                                aux = data.get(key).getAbsorptionFD();
                                break;
                            case DISP:
                                aux = data.get(key).getDispersionFD();
                                break;
                            case MAG:
                                aux = data.get(key).getMagnitudeFD();
                        }
                }
                for (int i = 0; i < res.length; i++) {
                    res[i] = res[i] + (float) Math.pow(meanSpectrum[i] - aux[i], 2) / ids.size();
                }
            }
        }
        for (int i = 0; i < res.length; i++) {
            res[i] = (float) Math.sqrt(res[i]);
        }
        return res;
    }
    
    public boolean containsID(String ID) {
        return data.containsKey(ID);
    }
    
    public boolean containsID(ArrayList<String> ids) {
        if (ids.isEmpty()) {
            return false;
        }
        boolean res = true;
        for (String id : ids) {
            if (!containsID(id)) {
                res = false;
                break;
            }
        }
        return res;
    }
    
    public void setPhase(double[] phase, int i, int j, int k) {
        if (data.containsKey(i + "_" + j + "_" + k)) {
            data.get(i + "_" + j + "_" + k).setPhase(phase[0], phase[1]);
        }
    }
    
    public void setPhase(String id, double[] phase) {
        if (data.containsKey(id)) {
            data.get(id).setPhase(phase[0], phase[1]);
        }
    }
    
    public void setPhase(ArrayList<String> ids, double[] phase) {
        Iterator iter = ids.iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            if (data.containsKey(id)) {
                data.get(id).setPhase(phase[0], phase[1]);
            }
        }
    }
    
    public void setPhase(double[] phase) {
        Iterator iter = data.keySet().iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            if (data.containsKey(id)) {
                data.get(id).setPhase(phase[0], phase[1]);
            }
        }
    }
    
    public void setPhase(HashMap<String, double[]> phases) {
        Iterator iter = phases.keySet().iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            if (data.containsKey(id)) {
                double[] phase = phases.get(id);
                data.get(id).setPhase(phase[0], phase[1]);
            }
        }
    }
    
    public void addPhase(ArrayList<String> ids, double[] phase) {
        Iterator iter = ids.iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            if (data.containsKey(id)) {
                data.get(id).addPhase(phase[0], phase[1]);
            }
        }
    }
    
    public double[] getPhase(int i, int j, int k) {
        return data.get(i + "_" + j + "_" + k).getPhaseParam();
    }
    
    public double[] getPhase(String id) {
        return data.get(id).getPhaseParam();
    }
    
    public double[] getMeanPhase(ArrayList<String> ids) {
        double[] res = {0, 0};
        int nr = 0;
        for (String key : ids) {
            res[0] += data.get(key).getPhaseParam()[0];
            res[1] += data.get(key).getPhaseParam()[1];
            nr++;
        }
        res[0] = res[0] / (double) nr;
        res[1] = res[1] / (double) nr;
        return res;
    }
    
    public double[] getStdDevPhase(ArrayList<String> ids) {
        double[] res = {0, 0};
        int nr = 0;
        double[] meanPhase = getMeanPhase(ids);
        for (String key : ids) {
            res[0] += Math.pow(data.get(key).getPhaseParam()[0] - meanPhase[0], 2);
            res[1] += Math.pow(data.get(key).getPhaseParam()[1] - meanPhase[1], 2);
            nr++;
        }
        res[0] = Math.sqrt(res[0] / (double) nr);
        res[1] = Math.sqrt(res[1] / (double) nr);
        return res;
    }
    
    public HashMap<String, double[]> getPhase(ArrayList<String> ids) {
        HashMap<String, double[]> result = new HashMap<>();
        Iterator iter = ids.iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            double[] res = data.get(id).getPhaseParam();
            result.put(id, res);
        }
        return result;
    }
    
    public HashMap<String, double[]> getPhase() {
        HashMap<String, double[]> result = new HashMap<>();
        Iterator iter = data.keySet().iterator();
        while (iter.hasNext()) {
            String id = (String) iter.next();
            double[] res = data.get(id).getPhaseParam();
            result.put(id, res);
        }
        return result;
    }
    
    public double getFrequencyShift(String id) {
        return data.get(id).getFrequencyShift();
    }
    
    public void saveSpectraToMAT(String folder) {
        saveSpectraToMAT(data.keySet().iterator(), folder, data.size());
    }
    
    public void saveSpectraToMAT(ArrayList<String> ids, String folder) {
        saveSpectraToMAT(ids.iterator(), folder, ids.size());
    }
    
    public void saveSpectraToMAT(Iterator iter, String folder, int nr) {
        MatlabIO matio = new MatlabIO();
        float[] aux1;
        float[] aux2;
        float[] aux3;
        float[] aux4;
        float[][] absFD = new float[this.vectorsize + this.zerofillingsize][nr];
        float[][] dispFD = new float[this.vectorsize + this.zerofillingsize][nr];
        float[][] absTD = new float[this.vectorsize + this.zerofillingsize][nr];
        float[][] dispTD = new float[this.vectorsize + this.zerofillingsize][nr];
        String[] ids = new String[nr];
        String idstring;
        int i = 0;
        String name = this.model.getPatientName();
        name.replace("^", "");
        name.replace("ä", "a");
        name.replace("Ä", "A");
        name.replace("ö", "o");
        name.replace("Ö", "O");
        name.replace("ü", "u");
        name.replace("Ü", "U");
        name.replace("^", "");
        name.replace("ä", "a");
        name.replace("Ä", "A");
        name.replace("ö", "o");
        name.replace("Ö", "O");
        name.replace("ü", "u");
        name.replace("Ü", "U");
        while (iter.hasNext()) {
            idstring = (String) iter.next();
            aux1 = data.get(idstring).getAbsorptionFD();
//            matio.addArrayToList("AbsFD" + idstring, aux);
            aux2 = data.get(idstring).getDispersionFD();
//            matio.addArrayToList("DispFD" + idstring, aux);
            aux3 = data.get(idstring).getAbsorptionTD();
//            matio.addArrayToList("AbsTD" + idstring, aux);
            aux4 = data.get(idstring).getDispersionTD();
//            matio.addArrayToList("DispTD" + idstring, aux);
            for (int j = 0; j < aux1.length; j++) {
                absFD[j][i] = aux1[j];
                dispFD[j][i] = aux2[j];
                absTD[j][i] = aux3[j];
                dispTD[j][i] = aux4[j];
            }
            
            ids[i] = name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber() + "_" + idstring;
            i++;
        }
        matio.addArrayToList("absFD_" + name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber(), absFD);
        matio.addArrayToList("dispFD_" + name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber(), dispFD);
        matio.addArrayToList("absTD_" + name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber(), absTD);
        matio.addArrayToList("dispTD_" + name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber(), dispTD);
        matio.addArrayToList("ids_" + name + "_" + this.model.getStudyDate() + "_" + this.model.getSeriesNumber(), ids);
        matio.addArrayToList("xAxisTD", this.xAxisTD);
        matio.addArrayToList("xAxisFDppm", this.xAxisFDppm);
        matio.addArrayToList("xAxisFDHz", this.xAxisFDHz);
        matio.save(folder);
    }
    
    public void savePeakAnalysis(String folder) {
        savePeakAnalysis(folder, data.keySet().iterator());
    }
    
    public void savePeakAnalysis(String folder, ArrayList<String> ids) {
        savePeakAnalysis(folder, ids.iterator());
    }
    
    public void savePeakAnalysis(String folder, Iterator iter) {
        
        MatlabIO matio = new MatlabIO();
        double[] snrh2o = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] snrcho = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] snrcr = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] snrnaa = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] maxh2o = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] maxcho = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] maxcr = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        double[] maxnaa = new double[this.gridsize1 * this.gridsize2 * this.gridsize3];
        
        String idstring;
        while (iter.hasNext()) {
            idstring = (String) iter.next();
            String[] parts = idstring.split("_");
            int pos = Integer.valueOf(parts[0]) + Integer.valueOf(parts[1]) * gridsize1
                    + Integer.valueOf(parts[2]) * (gridsize1 * gridsize2);
            snrh2o[pos] = data.get(idstring).getPeakSNR(4.6f, 4.8f);
            snrcho[pos] = data.get(idstring).getPeakSNR(3.1f, 3.285f);
            snrcr[pos] = data.get(idstring).getPeakSNR(2.927f, 3.1f);
            snrnaa[pos] = data.get(idstring).getPeakSNR(1.905f, 2.105f);
            maxh2o[pos] = data.get(idstring).getMaxPos(4.6f, 4.8f);
            maxcho[pos] = data.get(idstring).getMaxPos(3.1f, 3.285f);
            maxcr[pos] = data.get(idstring).getMaxPos(2.927f, 3.1f);
            maxnaa[pos] = data.get(idstring).getMaxPos(1.905f, 2.105f);
        }
        matio.addArrayToList("SNRH20", snrh2o);
        matio.addArrayToList("SNRCho", snrcho);
        matio.addArrayToList("SNRCr", snrcr);
        matio.addArrayToList("SNRNAA", snrnaa);
        matio.addArrayToList("MaxH20", maxh2o);
        matio.addArrayToList("MaxCho", maxcho);
        matio.addArrayToList("MaxCr", maxcr);
        matio.addArrayToList("MaxNAA", maxnaa);
        
        matio.addArrayToList("VectorSize", new float[]{this.vectorsize + this.zerofillingsize});
        
        matio.save(folder);
    }
    
    public void printPeakAnalysis() {
        printPeakAnalysis(data.keySet().iterator());
    }
    
    public void printPeakAnalysis(ArrayList<String> ids) {
        printPeakAnalysis(ids.iterator());
    }
    
    public void printPeakAnalysis(Iterator iter) {
        double noise;
        double snrh2o;
        double snrcho;
        double snrcr;
        double snrnaa;
        if (data.size() == 1) {
            this.model.printTxt("\n>>Noise\tSNR H20\tSNR Cho\tSNR Cr\tSNR NAA");
        } else {
            this.model.printTxt("\n>>Position\tNoise\tSNR H20\tSNR Cho\tSNR Cr\tSNR NAA");
        }
        String outputtxt = "";
        String idstring;
        while (iter.hasNext()) {
            idstring = (String) iter.next();
            
            noise = data.get(idstring).getNoiseStdDevFD();
            snrh2o = data.get(idstring).getPeakSNR(4.6f, 4.8f);
            snrcho = data.get(idstring).getPeakSNR(3.1f, 3.285f);
            snrcr = data.get(idstring).getPeakSNR(2.927f, 3.1f);
            snrnaa = data.get(idstring).getPeakSNR(1.905f, 2.105f);
            if (data.size() == 1) {
                outputtxt = noise + "\t" + snrh2o + "\t" + snrcho + "\t" + snrcr + "\t" + snrnaa;
            } else {
                if (!outputtxt.matches("")) {
                    outputtxt = outputtxt + "\n" + idstring + "\t" + noise + "\t" + snrh2o + "\t" + snrcho + "\t" + snrcr + "\t" + snrnaa;
                } else {
                    outputtxt = idstring + "\t" + noise + "\t" + snrh2o + "\t" + snrcho + "\t" + snrcr + "\t" + snrnaa;
                }
            }
        }
        if (data.size() == 1) {
            model.printNumeric(outputtxt, true);
        } else {
            this.model.printTxt(outputtxt);
        }
    }
    
    public void resetShift() {
        globalFreqShift = 0;
        resetShift(data.keySet().iterator());
    }
    
    public void resetShift(ArrayList<String> ids) {
        globalFreqShift = 0;
        resetShift(ids.iterator());
    }
    
    public void resetShift(Iterator iter) {
        globalFreqShift = 0;
        String aux;
        while (iter.hasNext()) {
            aux = (String) iter.next();
            data.get(aux).resetFrequencyShift();
        }
    }
    
    public double validatePPM(double ppmVal) {
        ppmVal = ppmVal < xAxisFDppm[0] ? xAxisFDppm[0] : ppmVal;
        ppmVal = ppmVal > xAxisFDppm[xAxisFDppm.length - 1] ? xAxisFDppm[xAxisFDppm.length - 1] : ppmVal;
        return ppmVal;
    }
    
    public double[] validatePPM(double[] ppmVals) {
        for (int i = 0; i < ppmVals.length; i++) {
            ppmVals[i] = ppmVals[i] < xAxisFDppm[0] ? xAxisFDppm[0] : ppmVals[i];
            ppmVals[i] = ppmVals[i] > xAxisFDppm[xAxisFDppm.length - 1] ? xAxisFDppm[xAxisFDppm.length - 1] : ppmVals[i];
        }
        return ppmVals;
    }
    
    public void calcPPMShift(double[] minvals, double[] maxvals, double[] targets, int[] multiplicity, double SNRLimit, double minPPM, double maxPPM, double maxLocalShift) {
        minvals = validatePPM(minvals);
        maxvals = validatePPM(maxvals);
        targets = validatePPM(targets);
        minPPM = validatePPM(minPPM);
        maxPPM = validatePPM(maxPPM);
        calcGlobalFrequencyShift(data.keySet(), minvals, maxvals, targets, multiplicity, SNRLimit, minPPM, maxPPM);
        calcLocalFrequencyShifts(data.keySet(), minPPM, maxPPM, maxLocalShift);
    }
    
    public void calcPPMShift(Set<String> ids, double[] minvals, double[] maxvals, double[] targets, int[] multiplicity, double SNRLimit, double minPPM, double maxPPM, double maxLocalShift) {
        minvals = validatePPM(minvals);
        maxvals = validatePPM(maxvals);
        targets = validatePPM(targets);
        minPPM = validatePPM(minPPM);
        maxPPM = validatePPM(maxPPM);
        calcGlobalFrequencyShift(ids, minvals, maxvals, targets, multiplicity, SNRLimit, minPPM, maxPPM);
        calcLocalFrequencyShifts(ids, minPPM, maxPPM, maxLocalShift);
    }
    
    private void calcGlobalFrequencyShift(Set<String> coords, double[] minvals, double[] maxvals, double[] targets, int[] multiplicity, double SNRLimit, double minPPM, double maxPPM) {
        assert (targets.length == minvals.length && targets.length == maxvals.length && targets.length == multiplicity.length) : "Arrays with different sizes - Not possible to determine shift for scale correction";
        assert (targets.length == 0) : "Empty Arrays - Not possible to determine shift for scale correction";
        DecimalFormat decformat = new DecimalFormat("#.####");
        int firstIndex = getIndexScalePPM(minPPM);
        int length = getIndexScalePPM(maxPPM) - firstIndex + 1;
        float[] meanMagSignal = new float[length];
        float[] meanAbsSignal = new float[length];
        float[] meanDispSignal = new float[length];

        //using the mean value of the std of the noise of the several signals will allow calculation the mean SNR and NOT the SNR of the
        //mean of the signals. Otherwise we would get higher values of SNR that would increase with the nr of signals
        //considered. In this way we get something independent of the nr of signals and the SNR threshold does not need to be adapted.
        float meanStdDevNoise = 0f;
        int count = 0;
        Iterator<String> iterExcitedVoxels = model.getExcitedVoxels().iterator();
        while (iterExcitedVoxels.hasNext()) {
            String id = iterExcitedVoxels.next();
            if (isQScoresEmpty() || (qscores.containsKey(id) && qscores.get(id) >= 0.3)) {
                float[] auxAbs = data.get(id).getAbsorptionFD(minPPM, maxPPM);
                float[] auxDisp = data.get(id).getDispersionFD(minPPM, maxPPM);
                meanStdDevNoise += data.get(id).getNoiseStdDevFD();
                count++;
                for (int i = 0; i < length; i++) {
                    meanAbsSignal[i] += auxAbs[i];
                    meanDispSignal[i] += auxDisp[i];
                }
            }
        }
        for (int i = 0; i < length; i++) {
            meanMagSignal[i] = (float) Math.sqrt(Math.pow(meanAbsSignal[i], 2) + Math.pow(meanDispSignal[i], 2)) / count;
        }
        meanStdDevNoise /= count;
        double globalFreqShiftIncrement = 0;
        double snrsum = 0;

        //Calculate shifts for each metabolite
        for (int i = 0; i < targets.length; i++) {
            //The intervals are shifted with the previous detected global shift
            int lowerB = getIndexScalePPM(minvals[i] + globalFreqShift) - firstIndex; //
            int upperB = getIndexScalePPM(maxvals[i] + globalFreqShift) - firstIndex;
            model.printTxt("Interval " + i + "(includes previous shift): [" + decformat.format(minvals[i] + globalFreqShift) + ","
                    + decformat.format(maxvals[i] + globalFreqShift) + "], Previous Global Shift applied:" + decformat.format(globalFreqShift));
            float max = Float.NEGATIVE_INFINITY;
            int maxIndex = 0;
            double metabShift = 0;
            //If multiplicity is even we need to find the axis of symmetry instead of the maximum
            if (multiplicity[i] % 2 == 0) {
                float max2 = Float.NEGATIVE_INFINITY;
                int maxIndex2 = 0;
                int rad = getIndexScalePPM(0.1) - getIndexScalePPM(0);//radius of 0.1ppm
                int midIndex = 0;
                double maxCorr = 0;
                
                for (int j = lowerB; j <= upperB; j++) {
                    double corr = 0;
                    for (int k = 1; k <= rad; k++) {
                        corr = corr + meanMagSignal[j - k] * meanMagSignal[j + k];
                    }
                    if (corr > maxCorr) {
                        maxCorr = corr;
                        midIndex = j;
                    }
                }
                model.printTxt("symmetry point at:" + decformat.format(getXAxis(Domain.FD_PPM)[midIndex + firstIndex]));
                
                for (int j = lowerB; j < midIndex; j++) {
                    if (meanMagSignal[j] > max) {
                        max = meanMagSignal[j];
                        maxIndex = j;
                    }
                }
                for (int j = midIndex + 1; j <= upperB; j++) {
                    if (meanMagSignal[j] > max2) {
                        max2 = meanMagSignal[j];
                        maxIndex2 = j;
                    }
                }
                max = (max + max2) / 2;
                maxIndex = Math.round(((float) maxIndex + (float) maxIndex2) / 2);
                metabShift = targets[i] - getXAxis(Domain.FD_PPM)[maxIndex + firstIndex];
                model.printTxt("Symmetry axis found at: " + decformat.format(getXAxis(Domain.FD_PPM)[maxIndex + firstIndex]));
            } else {
                for (int j = lowerB; j < upperB; j++) {
                    if (meanMagSignal[j] > max) {
                        max = meanMagSignal[j];
                        maxIndex = j;
                    }
                }
                double[][] matrixData = {
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex - 3 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex - 3 + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex - 2 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex - 2 + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex - 1 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex - 1 + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex + 1 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex + 1 + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex + 2 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex + 2 + firstIndex], 1},
                    {Math.pow(getXAxis(Domain.FD_PPM)[maxIndex + 3 + firstIndex], 2), getXAxis(Domain.FD_PPM)[maxIndex + 3 + firstIndex], 1}
                };
                RealMatrix X = new Array2DRowRealMatrix(matrixData);
                RealVector y = new ArrayRealVector(new double[]{meanMagSignal[maxIndex - 3], meanMagSignal[maxIndex - 2], meanMagSignal[maxIndex - 1],
                    meanMagSignal[maxIndex], meanMagSignal[maxIndex + 1], meanMagSignal[maxIndex + 2], meanMagSignal[maxIndex + 3]}, false);
                DecompositionSolver solver = new QRDecomposition(X).getSolver();
                RealVector w = solver.solve(y);
                double ppmMax = -w.getEntry(1) / (2 * w.getEntry(0));
                model.printTxt("Max found at: " + decformat.format(ppmMax));
                metabShift = targets[i] - ppmMax;
                max = (float) (w.getEntry(0) * (ppmMax * ppmMax) + w.getEntry(1) * ppmMax + w.getEntry(2));
            }
            
            float snr = max / meanStdDevNoise;
            if (snr > SNRLimit) {
                globalFreqShiftIncrement += snr * metabShift;
                snrsum += snr;
                model.printTxt("Peak from interval [" + decformat.format(minvals[i]) + ", " + decformat.format(maxvals[i]) + "] included! (SNR:" + decformat.format(snr) + ">" + decformat.format(SNRLimit) + ")");
            } else {
                model.printTxt("Peak from interval [" + decformat.format(minvals[i]) + ", " + decformat.format(maxvals[i]) + "] below SNR threshold (SNR:" + decformat.format(snr) + "<" + decformat.format(SNRLimit) + ")");
            }
        }
        
        if (snrsum > 0) {
            globalFreqShiftIncrement = globalFreqShiftIncrement / snrsum;
        }
        model.printTxt("New detected Global Shift: " + decformat.format(globalFreqShift + globalFreqShiftIncrement) + "|| increment:" + decformat.format(globalFreqShiftIncrement));
        this.globalFreqShift += globalFreqShiftIncrement;
        for (String coord : coords) {
            data.get(coord).frequencyShift(globalFreqShift);
        }
    }
    
    private void calcLocalFrequencyShifts(Set<String> coords, double minPPM, double maxPPM, double maxLocalShift) {
        int firstIndex = getIndexScalePPM(minPPM);
        int length = getIndexScalePPM(maxPPM) - firstIndex + 1;
        float[] meanMagSignal = new float[length];
        float[] meanAbsSignal = new float[length];
        float[] meanDispSignal = new float[length];

        //using the mean value of the std of the noise of the several signals will allow calculation the mean SNR and NOT the SNR of the
        //mean of the signals. Otherwise we would get higher values of SNR that would increase with the nr of signals
        //considered. In this way we get something independent of the nr of signals and the SNR threshold does not need to be adapted.
        int count = 0;
        Iterator<String> iterExcitedVoxels = model.getExcitedVoxels().iterator();
        while (iterExcitedVoxels.hasNext()) {
            String id = iterExcitedVoxels.next();
            if (isQScoresEmpty() || (qscores.containsKey(id) && qscores.get(id) >= 0.5)) {
                float[] auxAbs = data.get(id).getAbsorptionFD(minPPM, maxPPM);
                float[] auxDisp = data.get(id).getDispersionFD(minPPM, maxPPM);
                count++;
                for (int i = 0; i < length; i++) {
                    meanAbsSignal[i] += auxAbs[i];
                    meanDispSignal[i] += auxDisp[i];
                }
            }
        }
        for (int i = 0; i < length; i++) {
            meanMagSignal[i] = (float) Math.sqrt(Math.pow(meanAbsSignal[i], 2) + Math.pow(meanDispSignal[i], 2)) / count;
        }
        totalLocalShiftsFinished = 0;
        coordsLocalShiftsExecuted = new ArrayList(coords);
        //NOW discover individual shifts
        for (String coord : coords) {
            data.get(coord).calcLocalPPMShift(meanMagSignal, minPPM, maxPPM, maxLocalShift);//starts new thread
        }
    }
    
    public synchronized void localShiftThreadFinished() {
        DecimalFormat decformat = new DecimalFormat("#.####");
        this.totalLocalShiftsFinished++;
        if (totalLocalShiftsFinished == coordsLocalShiftsExecuted.size()) {
            double meanCorrelation = 0;
            double meanShift = 0;
            int count = 0;
            for (String coord : coordsLocalShiftsExecuted) {
                meanShift += data.get(coord).getFrequencyShift();
                meanCorrelation += data.get(coord).getCorrelationAlignment();
                count++;
            }
            model.printTxt("Calculation of local shifts FINISHED!");
            meanShift /= count;
            meanCorrelation /= count;
            model.printTxt("Mean Shift: " + decformat.format(meanShift));
            model.printTxt("Mean Correlation: " + decformat.format(meanCorrelation));
            int outliers = hampelFilterFreqShift();
//            model.printTxt("Outliers detected and substituted with median (1st pass: " + outliers);
            outliers = hampelFilterFreqShift();
//            model.printTxt("Outliers detected and substituted with median (2nd pass: " + outliers);
            BatchProcessing.resume();
        } else {
            if (totalLocalShiftsFinished % (coordsLocalShiftsExecuted.size() / 10) == 0) {
                model.printTxt("Calculating local shifts: " + totalLocalShiftsFinished + " out of " + coordsLocalShiftsExecuted.size());
            }
        }
    }
    
    private int hampelFilterFreqShift() {
        Iterator<String> iter = data.keySet().iterator();
        HashMap<String, Double> newShifts = new HashMap<>();
        while (iter.hasNext()) {
            String coord = iter.next();
            if (data.get(coord).isFrequencyShiftCorrected()) {
                List<Double> values = new ArrayList<>();
                String[] coords = coord.split("_");
                for (int i = -1; i <= 1; i++) {
                    for (int j = -1; j <= 1; j++) {
                        String coordi = (Integer.valueOf(coords[0]) + i) + "_"
                                + (Integer.valueOf(coords[1]) + j) + "_" + coords[2];
                        if (data.containsKey(coordi) && data.get(coordi).isFrequencyShiftCorrected()) {
                            values.add(data.get(coordi).getFrequencyShift());
                        }
                    }
                }
                Collections.sort(values);
                double median = values.get(values.size() / 2);
                double mean = 0;
                double stdDev = 0;
                int count = 0;
                for (int i = -2; i <= 2; i++) {
                    for (int j = -2; j <= 2; j++) {
                        String coordi = (Integer.valueOf(coords[0]) + i) + "_"
                                + (Integer.valueOf(coords[1]) + j) + "_" + coords[2];
                        if (data.containsKey(coordi) && data.get(coordi).isFrequencyShiftCorrected()) {
                            mean += data.get(coordi).getFrequencyShift();
                            count++;
                        }
                    }
                }
                mean /= count;
                for (int i = -2; i <= 2; i++) {
                    for (int j = -2; j <= 2; j++) {
                        String coordi = (Integer.valueOf(coords[0]) + i) + "_"
                                + (Integer.valueOf(coords[1]) + j) + "_" + coords[2];
                        if (data.containsKey(coordi) && data.get(coordi).isFrequencyShiftCorrected()) {
                            stdDev += Math.pow(data.get(coordi).getFrequencyShift() - mean, 2);
                        }
                    }
                }
                stdDev = Math.sqrt(stdDev / count);
                if (Math.abs(data.get(coord).getFrequencyShift() - median) / stdDev >= 1.5) {
                    newShifts.put(coord, median);
                }
            }
        }
        iter = newShifts.keySet().iterator();
        while (iter.hasNext()) {
            String coord = iter.next();
            data.get(coord).frequencyShift(newShifts.get(coord));
        }
        return newShifts.size();
    }
    
    public double getCorrelationAlignment(String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getCorrelationAlignment();
        }
        return res;
    }
    
    public double getPeakSNR(double minppm, double maxppm, int i, int j, int k) {
        double res = 0;
        if (data.containsKey(i + "_" + j + "_" + k)) {
            res = data.get(i + "_" + j + "_" + k).getPeakSNR(minppm, maxppm);
        }
        return res;
    }
    
    public double getPeakSNR(double minppm, double maxppm, String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getPeakSNR(minppm, maxppm);
        }
        return res;
    }
    
    public double getMeanFDSNR(double minppm, double maxppm, int i, int j, int k) {
        double res = 0;
        if (data.containsKey(i + "_" + j + "_" + k)) {
            res = data.get(i + "_" + j + "_" + k).getMeanFDSNR(minppm, maxppm);
        }
        return res;
    }
    
    public double getMeanFDSNR(double minppm, double maxppm, String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getMeanFDSNR(minppm, maxppm);
        }
        return res;
    }
    
    public double getMeanTDSNR(double minTime, double maxTime, int i, int j, int k) {
        double res = 0;
        if (data.containsKey(i + "_" + j + "_" + k)) {
            res = data.get(i + "_" + j + "_" + k).getMeanTDSNR(minTime, maxTime);
        }
        return res;
    }
    
    public double getMeanTDSNR(double minTime, double maxTime, String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getMeanTDSNR(minTime, maxTime);
        }
        return res;
    }
    
    public double getTDRelativeChange(double minTime, double maxTime, int i, int j, int k) {
        double res = 0;
        if (data.containsKey(i + "_" + j + "_" + k)) {
            res = data.get(i + "_" + j + "_" + k).getTDRelativeChange(minTime, maxTime);
        }
        return res;
    }
    
    public double getTDRelativeChange(double minTime, double maxTime, String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getTDRelativeChange(minTime, maxTime);
        }
        return res;
    }
    
    public double getAveragePower(String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getAveragePower();
        }
        return res;
    }
    
    public double getNoiseFD(String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getNoiseStdDevFD();
        }
        return res;
    }
    
    public double getNoiseTD(String ID) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getNoiseStdDevTD();
        }
        return res;
    }
    
    public double getMagnitudeSumTD(String ID, double minppm, double maxppm) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getMagnitudeSumTD(minppm, maxppm);
        }
        return res;
    }
    
    public double getMagnitudeSumFD(String ID, double minppm, double maxppm) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getMagnitudeSumFD(minppm, maxppm);
        }
        return res;
    }
    
    public double getAbsorptionSumFD(String ID, double minppm, double maxppm) {
        double res = 0;
        if (data.containsKey(ID)) {
            res = data.get(ID).getAbsorptionSumFD(minppm, maxppm);
        }
        return res;
    }
    
    public double[] getTDStatistics(String ID) {
        double[] res = new double[6];
        if (data.containsKey(ID)) {
            res = data.get(ID).getTDStatistics();
        }
        return res;
    }
    
    public double[] getFDStatistics(String ID) {
        double[] res = new double[6];
        if (data.containsKey(ID)) {
            res = data.get(ID).getFDStatistics();
        }
        return res;
    }
    
    public ArrayList<Double> getSpectralQualityFeatures(String coord) {
        ArrayList<Double> vals = new ArrayList<>();
        if (data.containsKey(coord)) {
            
            vals.add(getPeakSNR(0.5, 1.1, coord));
            vals.add(getPeakSNR(1.1, 1.6, coord));
            vals.add(getPeakSNR(1.6, 2.2, coord));
            vals.add(getPeakSNR(2.2, 2.6, coord));
            vals.add(getPeakSNR(2.6, 3.1, coord));
            vals.add(getPeakSNR(3.1, 3.5, coord));
            vals.add(getPeakSNR(3.5, 4.1, coord));
            vals.add(getPeakSNR(4.1, 5.5, coord));
            
            vals.add(getMeanFDSNR(0.5, 1.1, coord));
            vals.add(getMeanFDSNR(1.1, 1.6, coord));
            vals.add(getMeanFDSNR(1.6, 2.2, coord));
            vals.add(getMeanFDSNR(2.2, 2.6, coord));
            vals.add(getMeanFDSNR(2.6, 3.1, coord));
            vals.add(getMeanFDSNR(3.1, 3.5, coord));
            vals.add(getMeanFDSNR(3.5, 4.1, coord));
            vals.add(getMeanFDSNR(4.1, 5.5, coord));
            
            double meanGlobalSNR = getMeanFDSNR(0, 9, coord);
            double meanLeftSNR = getMeanFDSNR(6, 9, coord);
            vals.add(meanLeftSNR);
            vals.add(meanGlobalSNR);
            vals.add(meanLeftSNR / meanGlobalSNR);
            
            double[] tdstats = getTDStatistics(coord);
            vals.add(tdstats[0]);
            vals.add(tdstats[1]);
            vals.add(tdstats[2]);
            vals.add(tdstats[3]);
            vals.add(tdstats[4]);
            vals.add(tdstats[5]);
            
            double[] fdstats = getFDStatistics(coord);
            vals.add(fdstats[0]);
            vals.add(fdstats[1]);
            vals.add(fdstats[2]);
            vals.add(fdstats[3]);
            vals.add(fdstats[4]);
            vals.add(fdstats[5]);
            
            for (double time = 0; time < 200; time += 25) {
                vals.add(getMeanTDSNR(time, time + 25, coord));
            }
            for (double time = 0; time < 200; time += 25) {
                vals.add(getTDRelativeChange(time, time + 25, coord));
            }
        }
        return vals;
    }
    
    public ArrayList<String> getSpectralQualityFeaturesNames() {
        ArrayList<String> vals = new ArrayList<>();
        vals.add("SNR [0.5; 1.1]");
        vals.add("SNR [1.1; 1.6]");
        vals.add("SNR [1.6; 2.2]");
        vals.add("SNR [2.2; 2.6]");
        vals.add("SNR [2.6; 3.1]");
        vals.add("SNR [3.1; 3.5]");
        vals.add("SNR [3.5; 4.1]");
        vals.add("SNR [4.1; 5.5]");
        vals.add("mean SNR [0.5; 1.1]");
        vals.add("mean SNR [1.1; 1.6]");
        vals.add("mean SNR [1.6; 2.2]");
        vals.add("mean SNR [2.2; 2.6]");
        vals.add("mean SNR [2.6; 3.1]");
        vals.add("mean SNR [3.1; 3.5]");
        vals.add("mean SNR [3.5; 4.1]");
        vals.add("mean SNR [4.1; 5.5]");
        vals.add("mean SNR [6; 9]");
        vals.add("mean SNR [0; 9]");
        vals.add("mean [6; 9]/mean [0; 9]");
        vals.add("TD max");
        vals.add("TD time max");
        vals.add("TD mean");
        vals.add("TD std dev");
        vals.add("TD skew");
        vals.add("TD kurtosis");
        vals.add("FD max");
        vals.add("FD ppm max");
        vals.add("FD mean");
        vals.add("FD std dev");
        vals.add("FD skew");
        vals.add("FD kurtosis");
        vals.add("TD mean SNR [0; 25]");
        vals.add("TD mean SNR [25; 50]");
        vals.add("TD mean SNR [50; 75]");
        vals.add("TD mean SNR [75; 100]");
        vals.add("TD mean SNR [100; 125]");
        vals.add("TD mean SNR [125; 150]");
        vals.add("TD mean SNR [150; 175]");
        vals.add("TD mean SNR [175; 200]");
        vals.add("TD Relative Change [0; 25]");
        vals.add("TD Relative Change [25; 50]");
        vals.add("TD Relative Change [50; 75]");
        vals.add("TD Relative Change [75; 100]");
        vals.add("TD Relative Change [100; 125]");
        vals.add("TD Relative Change [125; 150]");
        vals.add("TD Relative Change [150; 175]");
        vals.add("TD Relative Change [175; 200]");
        return vals;
    }
    
    public ArrayList<Double> getMetabolicFeatures(String coord) {
        ArrayList<Double> vals = new ArrayList<>();
        double[] stats;
        if (data.containsKey(coord)) {
            double cho = getPeakSNR(3.1, 3.35, coord);
            double cr = getPeakSNR(2.8, 3.1, coord);
            double naa = getPeakSNR(1.8, 2.2, coord);
            double glx = getPeakSNR(2.2, 2.4, coord);
            double lip1_3 = getPeakSNR(1.05, 1.6, coord);
            double lip0_9 = getPeakSNR(0.5, 1.05, coord);
            double mIn = getPeakSNR(3.35, 3.70, coord);
            
            double mcho = getMeanFDSNR(3.1, 3.35, coord);
            double mcr = getMeanFDSNR(2.8, 3.1, coord);
            double mnaa = getMeanFDSNR(1.8, 2.2, coord);
            double mglx = getMeanFDSNR(2.2, 2.4, coord);
            double mlip1_3 = getMeanFDSNR(1.05, 1.6, coord);
            double mlip0_9 = getMeanFDSNR(0.5, 1.05, coord);
            double mmIn = getMeanFDSNR(3.35, 3.70, coord);

            //Ratios
            vals.add(mIn / cr);
            
            vals.add(mIn / lip0_9);
            vals.add(lip0_9 / cr);
            
            vals.add(glx / cr);
            vals.add(mIn / glx);
            vals.add(lip0_9 / glx);
            
            vals.add(cho / cr);
            vals.add(cho / glx);
            vals.add(cho / lip0_9);
            vals.add(cho / mIn);
            
            vals.add(cho / naa);
            vals.add(cr / naa);
            vals.add(lip0_9 / naa);
            vals.add(mIn / naa);
            vals.add(glx / naa);
            
            vals.add(lip1_3 / lip0_9);
            vals.add(lip1_3 / cr);
            vals.add(lip1_3 / naa);
            vals.add(lip1_3 / cho);
            vals.add(lip1_3 / glx);
            vals.add(lip1_3 / mIn);
            
            vals.add(mmIn / mcr);
            
            vals.add(mmIn / mlip0_9);
            vals.add(mlip0_9 / mcr);
            
            vals.add(mglx / mcr);
            vals.add(mmIn / mglx);
            vals.add(mlip0_9 / mglx);
            
            vals.add(mcho / mcr);
            vals.add(mcho / mglx);
            vals.add(mcho / mlip0_9);
            vals.add(mcho / mmIn);
            
            vals.add(mcho / mnaa);
            vals.add(mcr / mnaa);
            vals.add(mlip0_9 / mnaa);
            vals.add(mmIn / mnaa);
            vals.add(mglx / mnaa);
            
            vals.add(mlip1_3 / mlip0_9);
            vals.add(mlip1_3 / mcr);
            vals.add(mlip1_3 / mnaa);
            vals.add(mlip1_3 / mcho);
            vals.add(mlip1_3 / mglx);
            vals.add(mlip1_3 / mmIn);

            //Lip 0.9
            vals.add(lip0_9);
            vals.add(mlip0_9);
            stats = data.get(coord).getFDMagnitudeStatistics(0.5, 1.05);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //Lip 1.3 + Lactate
            vals.add(lip1_3);
            vals.add(mlip1_3);
            stats = data.get(coord).getFDMagnitudeStatistics(1.05, 1.6);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);
            stats = data.get(coord).getFDMagnitudeGradientStatistics(1.05, 1.6);
            vals.add(stats[0]);
            vals.add(stats[1]);
            vals.add(stats[2]);
            vals.add(stats[3]);
            vals.add(stats[4]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //NAA
            vals.add(naa);
            vals.add(mnaa);
            stats = data.get(coord).getFDMagnitudeStatistics(1.8, 2.2);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //Glx
            vals.add(glx);
            vals.add(mglx);
            stats = data.get(coord).getFDMagnitudeStatistics(2.2, 2.4);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);
            stats = data.get(coord).getFDMagnitudeGradientStatistics(2.2, 2.4);
            vals.add(stats[0]);
            vals.add(stats[1]);
            vals.add(stats[2]);
            vals.add(stats[3]);
            vals.add(stats[4]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //Creatine
            vals.add(cr);
            vals.add(mcr);
            stats = data.get(coord).getFDMagnitudeStatistics(2.8, 3.1);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //Choline
            vals.add(cho);
            vals.add(mcho);
            stats = data.get(coord).getFDMagnitudeStatistics(3.1, 3.35);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);

            //Myo-inositol
            vals.add(mIn);
            vals.add(mmIn);
            stats = data.get(coord).getFDMagnitudeStatistics(3.35, 3.7);
            vals.add(stats[3]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);
            stats = data.get(coord).getFDMagnitudeGradientStatistics(3.35, 3.7);
            vals.add(stats[0]);
            vals.add(stats[1]);
            vals.add(stats[2]);
            vals.add(stats[3]);
            vals.add(stats[4]);
            vals.add(stats[5]);
            vals.add(stats[6]);
            vals.add(stats[7]);
            
        }
        return vals;
    }
    
    public ArrayList<String> getMetabolicFeaturesNames() {//UPDATE THIS
        ArrayList<String> vals = new ArrayList<>();
        vals.add("Cho/Cr");
        vals.add("Cho/NAA");
        vals.add("Lip1.3/Lip0.9");
        
        String range = "[0.5; 1.05]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        
        range = "[1.05; 1.6]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        vals.add("grad min " + range);
        vals.add("grad ppm min " + range);
        vals.add("grad max " + range);
        vals.add("grad ppm max " + range);
        vals.add("grad mean " + range);
        vals.add("grad std dev " + range);
        vals.add("grad skewness " + range);
        vals.add("grad kurtosis " + range);
        
        range = "[1.8; 2.2]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        
        range = "[2.8; 3.1]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        
        range = "[3.1; 3.35]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        
        range = "[3.35; 3.7]";
        vals.add("SNR " + range);
        vals.add("mean SNR " + range);
        vals.add("ppm max " + range);
        vals.add("std dev " + range);
        vals.add("skewness " + range);
        vals.add("kurtosis " + range);
        vals.add("grad min " + range);
        vals.add("grad ppm min " + range);
        vals.add("grad max " + range);
        vals.add("grad ppm max " + range);
        vals.add("grad mean " + range);
        vals.add("grad std dev " + range);
        vals.add("grad skewness " + range);
        vals.add("grad kurtosis " + range);
        
        return vals;
    }
    
    public double[] expDecayParam(String ID) {
        Exponential curve = new Exponential(0, 0, 0);
        int lengthfit = getVectorSize() / 2;
        float[] subspectra = new float[lengthfit];
        double[] subaxis = new double[lengthfit];
        float[] abstd = data.get(ID).getMagnitudeTD();
        double[] axis = getXAxis(Domain.TD);
        for (int i = 0; i < lengthfit; i++) {
            subspectra[i] = abstd[i];
            subaxis[i] = axis[i];
        }
        ExponentialFit fit = new ExponentialFit(curve, subspectra, subaxis);
        float valA = (abstd[0] + abstd[1] + abstd[2] + abstd[3] + abstd[4] + abstd[5] + abstd[6] + abstd[7] + abstd[8] + abstd[9]) / 10;
        fit.setStartParameters(new double[]{0, 0.01, valA});
        DoubleRange[] ranges = new DoubleRange[3];
//        ranges[0] = new DoubleRange(-spectra.getAxisTD()[spectra.getVectorSize()/10], spectra.getAxisTD()[spectra.getVectorSize()/10]);
        ranges[0] = new DoubleRange(0, 0);
        ranges[1] = new DoubleRange(0.001, 10);
        ranges[2] = new DoubleRange(valA / 10, valA * 10);
        fit.setParameterRanges(ranges);
        fit.performRandomMinimization(32, 25, false);
        double par[] = fit.getMinimizingParameters();
        curve.setParameters(par[0], par[1], par[2]);
        return new double[]{par[1], par[2], fit.getMinimumFunctionValue()};
    }
    
    public void setReference(Spectra ref) {
        Iterator iter = data.keySet().iterator();
        String aux;
        String aux2;
        while (iter.hasNext()) {
            aux = (String) iter.next();
            String[] parts = aux.split("_");
//            aux2 = parts[0] + "_" + String.valueOf(Integer.valueOf(parts[1]) - 1) + "_" + parts[2];
//            if (ref.hasID(aux2)) {
//                data.get(aux).setReferenceVal(((Spectrum) ref.getData().get(aux2)).getMagnitudeSum(2f, 7.4f));
//            } else {
            data.get(aux).setReferenceVal(((Spectrum) ref.getData().get(aux)).getMagnitudeSumFD(2f, 7.4f));
//            }
        }
    }
    
    public void setQualityScores(HashMap<String, Double> qscores) {
        this.qscores = qscores;
    }
    
    public HashMap<String, Double> getQualityScores() {
        return qscores;
    }
    
    public double getQualityScore(String voxel) {
        if (qscores.containsKey(voxel)) {
            return qscores.get(voxel);
        }
        return -1;
    }
    
    public ArrayList<String> getQualitySelection(double threshold) {
        ArrayList<String> selection = new ArrayList<>();
        if (qscores.isEmpty()) {
            return null;
        } else {
            Iterator<String> iter = qscores.keySet().iterator();
            while (iter.hasNext()) {
                String id = iter.next();
                if (qscores.get(id) >= threshold) {
                    selection.add(id);
                }
            }
            return selection;
        }
    }
    
    public HashMap<String, float[]> getDiff(Spectra s, ArrayList<String> vs) {
        HashMap<String, float[]> d = new HashMap<>();
        for (String v : vs) {
            if (this.data.containsKey(v) && s.data.containsKey(v)) {
                d.put(v, this.data.get(v).getDiff(s.data.get(v)));
            }
        }
        return d;
    }
    
    public HashMap<String, float[]> getDiff(Spectra s, ArrayList<String> vs, double minPPM, double maxPPM) {
        HashMap<String, float[]> d = new HashMap<>();
        for (String v : vs) {
            if (this.data.containsKey(v) && s.data.containsKey(v)) {
                d.put(v, this.data.get(v).getDiff(s.data.get(v), minPPM, maxPPM));
            }
        }
        return d;
    }
    
    public HashMap<String, Float> getNoiseStdMap(ArrayList<String> vs) {
        HashMap<String, Float> d = new HashMap<>();
        for (String v : vs) {
            if (this.data.containsKey(v)) {
                d.put(v, this.data.get(v).getNoiseStdDevFD());
            }
        }
        return d;
    }
    
    public boolean isQScoresEmpty() {
        return qscores.isEmpty();
    }
    
}
