/*
 * File: $Id: MRLineModelSpectralResponse.java,v 1.1.1.1 2006/04/19 09:38:49 hans Exp $
 * History
 * Date: 22. Juli 2003, 14:51  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


package org.spectrim.tdfdfit;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.util.Random;
import javax.swing.JOptionPane;
import org.spectrim.signalprocessing.FFT;


/**
 * This class models the spectral response. 
 * @author  hans
 */
public class MRLineModelSpectralResponse extends SingleVoxelMeasuredSpectralData 
{
    
    /** 
     * Creates a new instance of MRLineModelSpectralResponse. 
     */
    public MRLineModelSpectralResponse( ) 
    {
        m_lineModelParameters = new MRFittingParameters( );
        m_lineModelParameters.setArea( 0.0f );
        m_lineModelParameters.setOffset( 0.0f );
        m_lineModelParameters.setGaussianWidth( 0.0f );
        m_lineModelParameters.setLorentzianWidth( 0.0f );
        m_lineModelParameters.setPhase( 0.0f );
        m_MRSignalProperties = MRSignalProperties.getInstance( );
        calculateResponse( );
    }
    
    /** 
     * Creates a new instance of MRLineModelSpectralResponse.
     * @param samplePoints
     */
    public MRLineModelSpectralResponse( int samplePoints ) 
    {
        super( samplePoints );
        m_lineModelParameters = new MRFittingParameters( );
    }
    
    /**
     * Creates a new instance of MRLineModelSpectralResponse.
     * @param fitPar object containing the parameters of the time/frequency response
     * has to be calculated
     * @param mrSignalProperties
     */
    public MRLineModelSpectralResponse( MRFittingParameters fitPar, 
                                        MRSignalProperties mrSignalProperties ) 
    {
        super( mrSignalProperties.getNumberOfSamplePoints( ) );
        m_lineModelParameters   = fitPar;
        m_MRSignalProperties    = mrSignalProperties;
        //m_lineModelParameters = new MRFittingParameters( );
        //autoSetXaxis( m_MRSignalProperties.getNumberOfSamplePoints( ) );        
        calculateResponse( );
    }

    /**
     * Creates a new instance of MRLineModelSpectralResponse.
     * @param fitPar object containing the parameters of the time/frequency response
     * has to be calculated
     * @param numberOfPoints
     */
    public MRLineModelSpectralResponse( MRFittingParameters fitPar, 
                                        int numberOfPoints ) 
    {
        m_lineModelParameters   = fitPar;
        m_MRSignalProperties    = MRSignalProperties.getInstance( );
        m_MRSignalProperties.setNumberOfSamplePoints( numberOfPoints );
        m_MRSignalProperties.setTruncationPoint( numberOfPoints );
        // m_lineModelParameters = new MRFittingParameters( );
        // autoSetXaxis( m_MRSignalProperties.getNumberOfSamplePoints( ) );        
        calculateResponse( );
    }
    
    
    
    
    /**
     * Returns the normalization factor.
     * @return 
     */
    public float getNormalizationFactor( )
    {
       float normalizationFactor = (float)Math.sqrt( m_absorptionChannelTD[0] * m_absorptionChannelTD[0] + 
                                                     m_dispersionChannelTD[0] * m_dispersionChannelTD[0] );
       return normalizationFactor;
    }
    
    
    /**
     * Creates a new instance of MRLineModelSpectralResponse.
     * @param mRLineModelSpectralResponse
     */
    public MRLineModelSpectralResponse( MRLineModelSpectralResponse mRLineModelSpectralResponse ) 
    {
        
        m_lineModelParameters = mRLineModelSpectralResponse.getFittingParameters( );
        m_MRSignalProperties = mRLineModelSpectralResponse.getMRSignalProperties( );
        
        m_absorptionChannelTD = new float[m_MRSignalProperties.getNumberOfSamplePoints( )];
        m_dispersionChannelTD = new float[m_MRSignalProperties.getNumberOfSamplePoints( )];
        
        System.arraycopy( mRLineModelSpectralResponse.getAbsorptionChannelTD( ), 0, m_absorptionChannelTD, 
                          0, mRLineModelSpectralResponse.getAbsorptionChannelTD( ).length );
        System.arraycopy( mRLineModelSpectralResponse.getDispersionChannelTD( ), 0, m_dispersionChannelTD, 
                          0, mRLineModelSpectralResponse.getDispersionChannelTD( ).length );
        
        setSpectralDataPointsTD( m_absorptionChannelTD, m_dispersionChannelTD );
        
    }
    
    /**
     * Sets the fitting parameters.
     * @param fitpar
     */
    public void setFittingParameters( MRFittingParameters fitpar )
    {
        m_lineModelParameters = fitpar;
    }
    
    /**
     * Sets the fitting parameters.
     * @return 
     */
    public MRFittingParameters getFittingParameters( )
    {
        return m_lineModelParameters;
    }
    
    /**
     * Sets the MR signal properties.
     * @param mrSignalProperties
     */
    public void setMRSignalProperties( MRSignalProperties mrSignalProperties )
    {
        m_MRSignalProperties = mrSignalProperties;
        // autoSetXaxis( m_MRSignalProperties.getNumberOfSamplePoints() );   
    }
    
    /**
     * Returns the signal properties parameters.
     * @return 
     */
    public MRSignalProperties getMRSignalProperties( )
    {
        return m_MRSignalProperties;
    }
    
    
    /**
     * This adds noise to a given response. 
     * @param noiseRange
     */
    public void addNoiseInTimeDomain( float noiseRange )
    {
        
        Random rnd = new Random( );
        
        for ( int i = 0 ; i < m_absorptionChannelTD.length; i++ )
        {
            m_absorptionChannelTD[i] = m_absorptionChannelTD[i] + 2f * ( rnd.nextFloat( ) - 0.5f ) * noiseRange;
            m_dispersionChannelTD[i] = m_dispersionChannelTD[i] + 2f * ( rnd.nextFloat( ) - 0.5f ) * noiseRange;
        }
        float[][] fftArray = new float[ m_absorptionChannelTD.length][2];
        for ( int i = 0; i < m_absorptionChannelTD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelTD[ i ];
            fftArray[i][1] = m_dispersionChannelTD[ i ];
        }

        // bring the FD in sync with the TD
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );

        for ( int i = 0 ; i <= m_absorptionChannelTD.length - 1 ; i++ )
        {
            int dd = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelTD.length;
            m_absorptionChannelFD[dd] = fftArray[i][0];
            m_dispersionChannelFD[dd] = fftArray[i][1];
        }
        
    }
    
    
    /**
     * Calculates the response of the belonging to the fit parameters.
     **/
    private void calculateResponse( )
    {
        int itrue, i, hap;
        float  f_o, n_shift, g_width2, l_arg, flap2,
	dt, dead_phase, dt_new, d_pha, dummy1, d_re,
	a, b ,c ,d, d_dummy, b_dummy, argu, cs, sn;
	
        if ( ( m_MRSignalProperties != null ) && ( m_lineModelParameters != null ) &&
             ( m_MRSignalProperties.getNumberOfSamplePoints( ) > 0 ) )
        {
            m_absorptionChannelTD = new float[ m_MRSignalProperties.getNumberOfSamplePoints( ) ];
            m_dispersionChannelTD = new float[ m_MRSignalProperties.getNumberOfSamplePoints( ) ] ;
	
            m_absorptionChannelFD = new float[ m_MRSignalProperties.getNumberOfSamplePoints( ) ];
            m_dispersionChannelFD = new float[ m_MRSignalProperties.getNumberOfSamplePoints( ) ];
            
            itrue = 1;
  
            f_o = twopidiv360 * ( (m_MRSignalProperties.getFirstOrderPhase( ) ) / 2.0f );
            n_shift = m_MRSignalProperties.getFirstOrderPhase( ) / 360.0f;
            float Hz_pp = m_MRSignalProperties.getFrequencyScale( );
            
            g_width2 = (1.0f / Hz_pp) * (float)( Math.PI * Math.PI ) * m_lineModelParameters.getGaussianWidth( ) * m_lineModelParameters.getGaussianWidth( )/ ( 1.67f * 1.67f );
            l_arg = (float)(Math.PI) *   m_lineModelParameters.getLorentzianLineWidth( );
            flap2 = (float)( m_MRSignalProperties.getNumberOfSamplePoints( ) / 2.0 );
            dt = n_shift/( (float)m_MRSignalProperties.getNumberOfSamplePoints( ) );
            dead_phase = (float)( 2.0 * Math.PI ) * ( m_lineModelParameters.getOffset( ) + flap2 ) * dt;
            dt_new = 1.0f / ( (float)( m_MRSignalProperties.getNumberOfSamplePoints( ) ) );
            d_pha = (float)(2.0 * Math.PI) * ( m_lineModelParameters.getOffset( ) ) * dt_new;
            dummy1 = (float)Math.exp( -g_width2 * dt * dt - dt * l_arg );

            m_absorptionChannelTD[0] = ( m_lineModelParameters.getArea( ) * dummy1 );
            m_dispersionChannelTD[0] = ( 0.0f );
 
            a = (float)Math.sin( d_pha );
            b = (float)Math.cos( d_pha );
            c = (float)Math.cos( d_pha );
            d = (float)Math.sin( d_pha );
 
            dt = ( n_shift + ( 1.0f ) ) / ( (float)( m_MRSignalProperties.getNumberOfSamplePoints( ) ) );

            dummy1 = (float)Math.exp( -g_width2 * dt * dt - dt * l_arg );

            m_absorptionChannelTD[1] = ( m_lineModelParameters.getArea( ) * dummy1 * b );
            m_dispersionChannelTD[1] = ( m_lineModelParameters.getArea( ) * dummy1 * d );


            for ( i = 2 ; i <= m_MRSignalProperties.getNumberOfSamplePoints( ) - 1; i++ )
            {
                if ( i <  m_MRSignalProperties.getTruncationPoint( ) )
            	{
                    d_dummy = a * b + c * d;
                    b_dummy = c * b - a * d;
                    dt = ( n_shift + (float)( i ) ) / ((float)( m_MRSignalProperties.getNumberOfSamplePoints( ) ) );
                    dummy1 = (float)Math.exp( -g_width2 * dt * dt - dt * l_arg );
                    m_absorptionChannelTD[i] = ( m_lineModelParameters.getArea( ) * dummy1 * b_dummy );
                    m_dispersionChannelTD[i] = ( m_lineModelParameters.getArea( ) * dummy1 * d_dummy );
                    b = b_dummy;
                    d = d_dummy;
		}
		else
		{
                    m_absorptionChannelTD[i] = 0.0f;
                    m_dispersionChannelTD[i] = 0.0f;
		}    
            }
            
            if ( m_lineModelParameters.m_fitWithReferenceLineShapeFunction == true )
            {
                complexMultiply( m_absorptionChannelTD , m_dispersionChannelTD, 
                                 m_MRSignalProperties.getReferenceLineReal( ),
                                 m_MRSignalProperties.getReferenceLineImaginary( ) );
            }
	
            if ( ( m_lineModelParameters.m_fitWithArbitraryLineShapeFunction == true ) && 
                 ( m_MRSignalProperties.getNumberOfLineShapeFunctionCoefficients( ) != 0 ) )
            {
                complexMultiply( m_absorptionChannelTD , m_dispersionChannelTD, 
                                 m_MRSignalProperties.getLineShapeFunctionReal( ),
                                 m_MRSignalProperties.getLineShapeFunctionImaginary( ) );
            }

            m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 0.5f;
            m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 0.5f;

            /* added to handle first points in FID */

            for ( i = 0; i < m_MRSignalProperties.getIgnoreNumberOfPoints( ); i++ )
            {
                m_absorptionChannelTD[i] = 0.0f;
                m_dispersionChannelTD[i] = 0.0f;
            }
            m_fftArray = new float[ m_absorptionChannelTD.length][2];
            for ( i = 0; i < m_absorptionChannelTD.length ; i++ )
            {
                m_fftArray[i][0] = m_absorptionChannelTD[ i ];
                m_fftArray[i][1] = m_dispersionChannelTD[ i ];
            }
        
            FFT fft = new FFT( );
            m_fftArray = fft.fft_1d( m_fftArray );

            for ( i = 0 ; i <= m_absorptionChannelTD.length - 1 ; i++ )
            {
                int dd = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelTD.length;
                m_absorptionChannelFD[dd] = m_fftArray[i][0];
                m_dispersionChannelFD[dd] = m_fftArray[i][1];
            }
	
            argu = m_lineModelParameters.getPhase( ) * (float)( Math.PI / 180.0 );
            cs = (float)Math.cos( argu - f_o + dead_phase );
            sn = (float)Math.sin( argu - f_o + dead_phase );

            for ( i = 0; i < this.m_absorptionChannelFD.length; i++ )
            {
		d_re = m_absorptionChannelFD[ i ] * cs - m_dispersionChannelFD[ i ] * sn;
		m_dispersionChannelFD[i] = m_absorptionChannelFD[ i ]  * sn + m_dispersionChannelFD[ i ] * cs;
		m_absorptionChannelFD[i] = d_re;

		d_re = m_absorptionChannelTD[ i ] * cs - m_dispersionChannelTD[ i ] * sn;
		m_dispersionChannelTD[i] = m_absorptionChannelTD[ i ]*sn + m_dispersionChannelTD[ i ]*cs;
		m_absorptionChannelTD[i] = d_re;
            }
        }      
    }

    /**
     * Performs a complex multiplication of two complex arrays.
     */
    private void complexMultiply( float[] yre,    float[] yim, 
                                  float[] ref_re, float[] ref_im )
    {
        float real;
	for ( int i = 0; i < yre.length; i++ )
	{
            real = yre[i] * ref_re[i] - yim[i] * ref_im[i];
            yim[i] = yre[i] * ref_im[i] + yim[i] * ref_re[i];
            yre[i] = real;
	}
    }
    
    
    /**
     * This method is meant to be able to additionally amplify, shift, 
     * broaden, and zero order phase the response.
     * @param fittingPars
     */
    public void applyMRLineModelSpectralResponse( MRFittingParameters fittingPars )
    {
        
        MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( fittingPars, m_MRSignalProperties );
        
//        XYPlotFrame xypf0 = new XYPlotFrame( "***MRLineModelResponse::applyMRLineModelSpectralResponse->FD", null );
//        float[] xaxis0 = new float[m_absorptionChannelTD.length];
//        
//        for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
//        {
//            xaxis0[i] = (float)i;
//            if ( i < 256 )
//            {
//                System.out.println( i + " abs  : " + resp.m_absorptionChannelFD[i] );
//                System.out.println( i + " disp : " + resp.m_dispersionChannelFD[i] );
//            }
//        }
//        xypf0.addCurve( new Curve( xaxis0, resp.m_absorptionChannelTD ), true );
//        xypf0.addCurve( new Curve( xaxis0, resp.m_dispersionChannelTD ), true );
        
        complexMultiply( m_absorptionChannelTD, 
                         m_dispersionChannelTD, 
                         resp.getAbsorptionChannelTD( ), 
                         resp.getDispersionChannelTD( ) );
        
        // Call the super class method to sync the time and frequency domain
        // data
        this.setSpectralDataPointsTD( m_absorptionChannelTD, m_dispersionChannelTD );
        
//        XYPlotFrame xypf = new XYPlotFrame( "MRLineModelResponse::applyMRLineModelSpectralResponse->FD", null );
//        float[] xaxis = new float[m_absorptionChannelTD.length];
//        for ( int i = 0; i < m_absorptionChannelTD.length/64; i++ )
//        {
//            xaxis[i] = (float)i;
//        }
//        
//        xypf.addCurve( new Curve( xaxis, resp.getAbsorptionChannelTD( ) ), true );
//        xypf.addCurve( new Curve( xaxis, resp.getDispersionChannelTD( ) ), true );
    }
    
    /**
     * Returns the number of sample points in the spectral response.
     * @return 
     */
    public int getNumberOfSampledPoints( )
    {
        return this.m_absorptionChannelTD.length;
    }
    
    // Attributes
    private MRFittingParameters m_lineModelParameters;
    private MRSignalProperties m_MRSignalProperties;

    float[][] m_fftArray;
    float f_0;
    private final float twopidiv360 = (float)( 2.0f * Math.PI / 360.0f );

    /**
     * Normalization using the first complex point of the response.
     */
    void normalizeOnTDzeroPart( )
    {
        float normalizationFactor = (float)Math.sqrt( m_absorptionChannelTD[0] * m_absorptionChannelTD[0] + 
                                                      m_dispersionChannelTD[0] * m_dispersionChannelTD[0] );
        if ( normalizationFactor!= 0f )
        {                                              
            for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
            {
                m_absorptionChannelTD[i] = m_absorptionChannelTD[i] / normalizationFactor;
                m_dispersionChannelTD[i] = m_dispersionChannelTD[i] / normalizationFactor;
                m_absorptionChannelFD[i] = m_absorptionChannelFD[i] / normalizationFactor;
                m_dispersionChannelFD[i] = m_dispersionChannelFD[i] / normalizationFactor;
            }
        }
        else
        {
            JOptionPane.showMessageDialog( null, "Cannot normalize numeric pattern \n Cause: Devision by zero." );
        }
    }
    
    /**
     * Writes the current data file to a TDFDFit formatted binary file.
     * @param filePathAndName
     * @param reverseFDOrder
     * @return 
     */
    public String writeAsTdfdfitFileForFitting( String filePathAndName, 
                                                boolean reverseFDOrder )
    {
        
        // TDFDFittingDefaults fitDef = TDFDFittingDefaults.getInstance( );
        // String path = fitDef.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH );
        // String path = fitDef.getProperty(TDFDFittingDefaults.LAST_USED_TDFDFIT_MODEL_PATH );
         // Remove the file extension
        while ( filePathAndName.contains( ".bin" ) == true )
        {
            int index = filePathAndName.lastIndexOf( "." );
            filePathAndName = filePathAndName.substring( 0, index );
        }
        
        // String dataFileName = path + "Numeric_Patterns" + System.getProperty( "file.separator" ) + filePathAndName;
        
        // System.out.println( "Data filename : " + filePathAndName + ".bin" );
        
        try
        {
            
            // m_absorptionChannelTD[0] = m_absorptionChannelTD[0] / 2f;
            // m_dispersionChannelTD[0] = m_dispersionChannelTD[0] / 2f;
            
            this.calculateFDRepresentation( );
            
            float[] reverseAbsorbtion = new float[m_absorptionChannelFD.length];
            float[] reverseDispersion = new float[m_dispersionChannelFD.length];
            
            if ( reverseFDOrder == true )
            {
                // TDFDFIT expects the spectrum in reversed order!
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    reverseAbsorbtion[i] = m_absorptionChannelFD[m_absorptionChannelFD.length - i - 1];
                    reverseDispersion[i] = -m_dispersionChannelFD[m_dispersionChannelFD.length - i - 1];
                }
            }
            else
            {
                // TDFDFIT expects the spectrum in reversed order!
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    reverseAbsorbtion[i] = m_absorptionChannelFD[i];
                    reverseDispersion[i] = m_dispersionChannelFD[i];
                }
            }
            try ( RandomAccessFile writer = new RandomAccessFile( filePathAndName + ".bin", "rw" ) ) 
            {
                ByteBuffer buf, buf1;
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    buf = ByteBuffer.allocate( 4 );

                    buf.putFloat( reverseAbsorbtion[i] );


                    writer.write( buf.get( 3 ) );
                    writer.write( buf.get( 2 ) );
                    writer.write( buf.get( 1 ) );
                    writer.write( buf.get( 0 ) );

                    buf1 = ByteBuffer.allocate( 4 );

                    buf1.putFloat( reverseDispersion[i] );
                    writer.write( buf1.get( 3 ) );
                    writer.write( buf1.get( 2 ) );
                    writer.write( buf1.get( 1 ) );
                    writer.write( buf1.get( 0 ) );

                }   //m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2f;
                    //m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2f;

                // calculateFDRepresentation( );
            }
        }
        
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted measured datafile" );
        }
        return filePathAndName + ".bin";
    }
    
    void doComplexConjugationTD( )
    {
        for ( int i = 0; i < this.m_absorptionChannelTD.length; i++ )
        {
            m_dispersionChannelTD[i] = -m_dispersionChannelTD[i];
        }
        calculateFDRepresentation( );
    }
}