/*
 * File: $Id: MRSignalProperties.java,v 1.1.1.1 2006/04/19 09:38:49 hans Exp $
 * History
 * Date: 03.06.2003 2003, 12:19    Authors: Hans Slotboom
 * Changes: 30-09-2003 - making a singleton class of it
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */
 
 
package org.spectrim.tdfdfit;

import java.io.RandomAccessFile;
import java.io.IOException;

/**
 * 
 * Container class carrying information about the basic MR signal properties.
 *
 * @author  hans
 */
public class MRSignalProperties 
{
    /** 
     * Creates a new default instance of MRSignalProperties. 
     */
    private MRSignalProperties( )
    {
        m_fitModelName  = "New TDFDFit model";
        m_lsfCoefs      = 0;
        m_ignore        = 0;
        m_phi_1         = 0.0f;
        m_truncate      = 1024;
        m_frscl         = 1.0f;
        m_nucleusType   = PROTON;
        m_minimumNoiseIntervalPointNumber = 50;
        m_maximumNoiseIntervalPointNumber = 250;
    }
    
    /**
     * Returns the one and only instance of MRSignalProperties.
     * @return 
     */
    public static MRSignalProperties getInstance( )
    {
        if ( theInstance == null )
        {
            theInstance = new MRSignalProperties( );
        }
        return theInstance;
    }
    
    /** 
     * Resets the model to default.
     */
    public void resetMRSignalProperties( )
    {
        m_fitModelName  = "New TDFDFit model";
        m_lsfCoefs      = 0;
        m_ignore        = 0;
        m_phi_1         = 0.0f;
        m_truncate      = 1024;
        m_frscl         = 1.0f;
        m_nucleusType   = PROTON;
        m_minimumNoiseIntervalPointNumber = 50;
        m_maximumNoiseIntervalPointNumber = 250;
    }
    
    /**
     *  Sets the instance members of the MRSignalProperties object.
     * @param frscl
     * @param ignore
     * @param truncate
     * @param phi_1
     * @param lsf_re
     * @param lsf_im
     * @param ref_re
     * @param ref_im
     */
    public void setMRSignalProperties( float frscl, int ignore, int truncate, float phi_1, 
                                       float[] lsf_re, float[] lsf_im, 
                                       float[] ref_re, float[] ref_im )
    {
	m_frscl = frscl;
	m_phi_1 = phi_1;
	m_ignore = ignore;
	m_truncate = truncate;
        
        m_lsf_re = new float[ lsf_re.length ];
        System.arraycopy( lsf_re, 0, m_lsf_re, 0, lsf_re.length );
        
        m_lsf_im = new float[ lsf_im.length ]; 
        System.arraycopy( lsf_im, 0, m_lsf_im, 0, lsf_im.length );
        
        m_ref_re = new float[ ref_re.length ];
        System.arraycopy( ref_re, 0, m_ref_re, 0, ref_re.length );
        
        m_ref_im = new float[ ref_im.length ];
        System.arraycopy( ref_im, 0, m_ref_im, 0, ref_im.length );
    }
    
    /**
     * Sets the number of sample points.
     * @param samples
     */
    public void setNumberOfSamplePoints( int samples )
    {
        m_samplePoints = samples;
    }
    
    public void setHz_pp( float frscl )
    {
        m_frscl = frscl;
    }
    
    
    /**
     * Sets the number of line shape coefficients to be used in the fit.
     */
    void setNumberOfLineShapeFunctionCoefs( int lsfcoefs )
    {
	m_lsfCoefs = lsfcoefs;
    }

    /**
     * Sets the frequency scale in Hz/point of the acquired signal.
     */
    void setFrequencyScale( float frscl )
    {
	m_frscl = frscl;
    }
  
    /**
     * Sets first order phase of the signal, this corresponds to a time delay
     * in the time domain acquisition.
     */
    void setFirstOrderPhase( float phi_1 )
    {
	m_phi_1 = phi_1;
    }
    
    /**
     * Sets the number of points to ignore at the beginning of the time domain 
     * response.
     */
    void setIgnoreNumberOfPoints( int ignore )
    {
	m_ignore = ignore;
    }
    
    /**
     * Set the point after which the FID should be truncated; i.e. the values 
     * set to zero
     * @param truncate
     */
    public void setTruncationPoint( int truncate )
    {
	m_truncate = truncate;
    }

    /**
     * Sets the line shape function object.
     * @param lsf_re
     * @param lsf_im
     */
    public void setLineShapeFunction( float[] lsf_re, float[] lsf_im )
    {
        m_lsf_re = new float[lsf_re.length];
        m_lsf_im = new float[lsf_im.length];
        
        System.arraycopy( lsf_re, 0, m_lsf_re, 0, lsf_re.length );
        System.arraycopy( lsf_im, 0, m_lsf_im, 0, lsf_im.length );
    }
    
    /**
     * Sets the reference line objects .
     * @param ref_re
     * @param ref_im
     */   
    public void setReferenceLine( float[] ref_re, float[] ref_im )
    {
        m_ref_re = new float[ref_re.length];
        m_ref_im = new float[ref_im.length];
        
        System.arraycopy( ref_re, 0, m_ref_re, 0, ref_re.length );
        System.arraycopy( ref_im, 0, m_ref_im, 0, ref_im.length );       
    }

    /**
     * Sets the name of the fitting project.
     * @param modelName
     */
    public void setFitModelName( String modelName )
    {
      m_fitModelName = modelName;   
    }
    
    
    /**
     * Returns the point after which the response is truncated.
     * @return 
     */
    public int getTruncationPoint( )
    {
	return m_truncate;
    }
    
    /**
     * Returns the number of sample points.
     * @return 
     */
    public int getNumberOfSamplePoints( )
    {
        return m_samplePoints;
    }
    
    /**
     * Returns the number of line shape coefficients.
     * @return 
     */
    public int getNumberOfLineShapeFunctionCoefficients( )
    {
	return m_lsfCoefs;
    }
    
    /**
     * Returns the frequency scale.
     * @return 
     */
    public float getFrequencyScale( )
    {
	return m_frscl;
    }
    
    /**
     * Returns the first order phase.
     * @return 
     */
    public float getFirstOrderPhase( )
    {
	return m_phi_1;
    }
    
    /**
     * Returns the number of points to ignore at the beginning of the response.
     * @return 
     */
    public int getIgnoreNumberOfPoints( )
    {
	return m_ignore;
    }
    
    public float[] getLineShapeFunctionReal( )
    {
	return m_lsf_re;
    }

    public float[] getLineShapeFunctionImaginary( )
    {
	return m_lsf_im;
    }

    public float[] getReferenceLineReal( )
    {
	return m_ref_re;
    }

    public float[] getReferenceLineImaginary( )
    {
	return m_ref_im;
    }

    /**
     * Returns the name of the fitting project.
     * @return 
     */
    public String getFitModelName( )
    {
      return m_fitModelName;   
    }
    
    /**
     * Sets the attribute determining whether the non analytic functions
     * has to be determined during the fit.
     * @param determineNALSF
     */ 
    public void setDetermineNonAnalyticLSF( boolean determineNALSF )
    {
        m_determineNonAnalyticLSF = determineNALSF;
    }

    /**
     * Returns the attribute determining whether the non analytic functions
     * has to be determined during the fit.
     * @return 
     */ 
    public boolean getDetermineNonAnalyticLSF(  )
    {
        return m_determineNonAnalyticLSF;
    }
    
    /**
     * Sets the attribute determining whether the fit uses a reference line.
     * @param setUseReferenceLine
     */ 
    public void setUseReferenceLine( boolean setUseReferenceLine )
    {
        m_useReferenceLine = setUseReferenceLine;
    }

    /**
     * Returns the attribute determining whether the fit uses a reference line.
     * @return 
     */ 
    public boolean getUseReferenceLine( )
    {
        return m_useReferenceLine;
    }
       
    /**
     * Reads the attributes from a RandomAccesFile.
     * @param in
     */
    public void read( RandomAccessFile in )
    {
        boolean oldVersion = true;
        
        try
        {
            m_fitModelName = in.readUTF( );
            
            if ( m_fitModelName.compareTo( "VERSION_1" ) == 0 )
            {
                m_fitModelName = in.readUTF( );
                oldVersion = false;
            }
            
            m_samplePoints = in.readInt( );
            m_lsfCoefs = in.readInt( );
            m_phi_1 = in.readFloat( );
            m_frscl = in.readFloat( );
            m_ignore = in.readInt( );
            m_truncate = in.readInt( );
            
            m_resonanceFrequency = in.readLong( );
            m_ppmValueInCenterOfSpectrum = in.readFloat( );
            m_nucleusType = in.readUTF( );
            
            m_determineNonAnalyticLSF = in.readBoolean( );
            m_useReferenceLine = in.readBoolean( );
            
            if ( oldVersion == false )
            {
                m_minimumNoiseIntervalPointNumber = in.readInt( );
                m_maximumNoiseIntervalPointNumber = in.readInt( );
            }
            
            // input of lsf & ref functions has also to be realized
        }
        catch ( IOException e )
        {
            System.out.println( "IOExeption in MRSignalProperties::read( )" );
        }
    }
    
    /**
     * Writes attributes to RandomAccessFile.
     */
    void write( RandomAccessFile out )
    {
        // try to write all essential attributes to file
        try
        {
            out.writeUTF( "VERSION_1" );
            out.writeUTF( m_fitModelName );
            out.writeInt( m_samplePoints );
            out.writeInt( m_lsfCoefs );
            out.writeFloat( m_phi_1 ); 
            out.writeFloat( m_frscl );
            out.writeInt( m_ignore );
            out.writeInt( m_truncate );
            
            out.writeLong( m_resonanceFrequency );
            out.writeFloat( m_ppmValueInCenterOfSpectrum );
            out.writeUTF( m_nucleusType );
            
            out.writeBoolean( m_determineNonAnalyticLSF );
            out.writeBoolean( m_useReferenceLine );                 
        
            out.writeInt( m_minimumNoiseIntervalPointNumber );
            out.writeInt( m_maximumNoiseIntervalPointNumber );
            // output of lsf functions has also to be realized
        }
        catch ( IOException e )
        {
            
        }
    }
    
    /**
     * Computes the acquisition bandwidth.
     * @return 
     */
    public float getAcquistitionBandwidth( ) 
    {
        return ( m_samplePoints * m_frscl );
    }
    
    
    /**
     * Returns the resonance frequency the model is valid for.
     * @return 
     */
    public long getResonanceFrequency( )
    {
        return m_resonanceFrequency;
    }
    
    /**
     * Returns the resonance frequency the model is valid for.
     * @param resonanceFrequency 
     */
    public void setResonanceFrequency( long resonanceFrequency )
    {
        m_resonanceFrequency = resonanceFrequency;
    }
    
    /**
     * The ppm value in the center point of the spectrum.
     * @return ppm value in the center of the spectrum
     */
    public float getPPMValueInCenterOfSpectrum( ) 
    {
        return m_ppmValueInCenterOfSpectrum;
    }
    
    /**
     * Sets the ppm value in the center point of the spectrum.
     * @param ppmValueInCenterOfSpectrum the center frequency of the spectrum 
     * in ppm.
     */
    public void setPPMValueInCenterOfSpectrum( float ppmValueInCenterOfSpectrum ) 
    {
        m_ppmValueInCenterOfSpectrum = ppmValueInCenterOfSpectrum;
    }
    
     /**
     * Returns the maximum ppm value in the center point of the spectrum.
     * in ppm.
     */
    public float getMaximalPPMValueOfSpectrum( ) 
    {
        
        float bandwithInHz = m_samplePoints * m_frscl;
        float bandwithInPPM = (bandwithInHz / this.m_resonanceFrequency)*1000000f;
        float maxPPMValue = m_ppmValueInCenterOfSpectrum + bandwithInPPM/2f;
        return maxPPMValue;
    }
    
     /**
     * Returns the minimum ppm value in the center point of the spectrum.
     * in ppm.
     */
    public float getMinimalPPMValueOfSpectrum( ) 
    {
        
        float bandwithInHz = m_samplePoints * m_frscl;
        float bandwithInPPM = (bandwithInHz / this.m_resonanceFrequency)*1000000f;
        float minPPMValue = m_ppmValueInCenterOfSpectrum - bandwithInPPM/2f;
        return minPPMValue;
    }
    
     /**
     * Returns the ppm range value in the center point of the spectrum.
     * in ppm.
     * @return 
     */
    public float getPPMRangeOfSpectrum( ) 
    {
        float bandwithInHz = m_samplePoints * m_frscl;
        float bandwithInPPM = (bandwithInHz / this.m_resonanceFrequency)*1000000f;
        return bandwithInPPM;
    }
    
    
    /**
     * Returns the nucleus type as a string (eg 1H, 31P, 13C etc.).
     * @return m_nucleusType nucleus type
     */
    public String getNucleusType( ) 
    {
        return m_nucleusType;
    }
    
    /**
     * Sets the nucleus type as a string (like 1H, 31P, 13C etc).
     * @param nucleusType
     */
    public void setNucleusType( String nucleusType ) 
    {
        m_nucleusType = nucleusType;
        
        switch ( m_nucleusType ) 
        {
            case PROTON:   
                            m_ppmValueInCenterOfSpectrum = 4.7f; break;
            case CARBON13: 
                            m_ppmValueInCenterOfSpectrum = 0.0f; break;
            case SODIUM23: 
                            m_ppmValueInCenterOfSpectrum = 0.0f; break;
            case PHOSPHORUS31: 
                            m_ppmValueInCenterOfSpectrum = 0.0f; break;
            default: 
                            m_ppmValueInCenterOfSpectrum = 0.0f; break;
        }   
    }
    
    
    public int getMinimumNoiseIntervalPointNumber( )
    {
        return m_minimumNoiseIntervalPointNumber;
    }
    
    public int getMaximumNoiseIntervalPointNumber( )
    {
        return m_maximumNoiseIntervalPointNumber;
    }
    
    public void setMinimumNoiseIntervalPointNumber( int minNoisePointNumber )
    {
        m_minimumNoiseIntervalPointNumber = minNoisePointNumber ;
    }
    
    public void setMaximumNoiseIntervalPointNumber( int maxNoisePointNumber  )
    {
        m_maximumNoiseIntervalPointNumber = maxNoisePointNumber;
    }
    
    
    
    // Attributes
    private float[] m_lsf_re, m_lsf_im;
    private float[] m_ref_re, m_ref_im;
    
    private String  m_fitModelName;
    private int     m_samplePoints;
    
    private long    m_resonanceFrequency;
    private float   m_ppmValueInCenterOfSpectrum; 
    private String  m_nucleusType = "1H";
   
    private float   m_frscl;
    private int     m_ignore;
    private int     m_lsfCoefs;
    private float   m_phi_1;
    private int     m_truncate;
    
    private int     m_minimumNoiseIntervalPointNumber;
    private int     m_maximumNoiseIntervalPointNumber;
    
    private boolean m_determineNonAnalyticLSF = false;
    private boolean m_useReferenceLine = false;
    
    // static private attribute is the instance itself (Singleton!)
    private static  MRSignalProperties theInstance = null;
    
    public final static String PROTON       = "1H";
    public final static String CARBON13     = "13C";
    public final static String SODIUM23     = "23Na";
    public final static String PHOSPHORUS31 = "31P";

    public void setAcquisitionBandwidth( double bandwidth ) 
    {
       m_frscl =  (float)bandwidth / m_samplePoints; 
    }

}
