/*
 * File: $Id: SingleVoxelSpectralDataSeries.java,v 1.3 2006/11/08 12:18:40 hans Exp $
 * History
 * Date: 24.03.2005, 8:57  Authors: Hans Slotboom  
 *
 * Changes:
 *
 * Copyright (C) 2005 -  by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */


/*
 * SingleVoxelSpectralDataSeries.java
 *
 * Created on March 24, 2005, 8:57 AM
 */

package org.spectrim.tdfdfit;


import java.awt.Color;

import java.io.RandomAccessFile;
import java.io.IOException;

import java.nio.ByteBuffer;

import java.util.LinkedHashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.LinkedList;

import org.spectrim.dicom.FileInfo;
import org.spectrim.dicom.DicomImageData;
import org.spectrim.dicom.DicomImageSeriesData;
import org.spectrim.signalprocessing.FFT;
import org.spectrim.imagegraphicobjects.Spectroscopic3DVoxel;
import org.spectrim.imagegraphicobjects.Point3D;
import org.spectrim.lib.Convert;
import org.spectrim.xyplot.Curve;

/**
 *
 * Class that handles the the data series of a single voxel spectrum 
 * repetitively measured and different conditions. This data type 
 * appears measuring a SVS dataset and storing the data separately
 *
 * @author  hans
 */
public class SingleVoxelSpectralDataSeries 
{
    
    /** 
     * Creates a new instance of SingleVoxelSpectralDataSeries 
     */
    public SingleVoxelSpectralDataSeries( )
    {
        
    }
      
    /** 
     * Creates a new instance of SingleVoxelSpectralDataSeries 
     */
    public SingleVoxelSpectralDataSeries( Color color )
    {
        m_curveColor = color;
    }
    
   /**
     * Creates a new instance of SingleVoxelSpectralDataSeries and reserves N sample
     * points for the response
     */
    public SingleVoxelSpectralDataSeries( int samplePoints, int instances )
    {
        m_instances = instances;
        m_numberOfSamplePoints = samplePoints; 
        
        allocateArrays( samplePoints, instances );
    }
    
   /**
     * Creates a new instance of SingleVoxelSpectralDataSeries and reserves N sample
     * points for the response
     */
    public SingleVoxelSpectralDataSeries( int samplePoints, int instances, Color color )
    {
       m_instances = instances;
       m_numberOfSamplePoints = samplePoints; 
       m_curveColor = color;
       
       allocateArrays( samplePoints, instances );
    }
    
    /**
     * Creates a new instance of SingleVoxelSpectralDataSeries and reserves N sample
     * points for the response
     */
    public SingleVoxelSpectralDataSeries( LinkedList<String> fileReferences, Color color )
    {
       
       m_fileReferences = fileReferences;
       
       m_instances = m_fileReferences.size( );
       
       int counter = 0;
       Iterator it = fileReferences.iterator( );
       while ( it.hasNext( ) )
       {
           String fileName = (String)it.next( );
           DicomImageData data = new DicomImageData( fileName );
           
           if ( counter == 0 )
           {
               setSiemensSpecificDicomHeaderInfo( data.getSiemensSpecificDataFromFile( ) );
               setDicomHeaderInfo( data.getHeader( ) );
               m_resonanceFrequency = ( new Long( (String)( data.getSiemensSpecificDataFromFile( ).get( "sTXSPEC.asNucleusInfo[0].lFrequency" ) ) ) ).longValue( );
               
               m_sampleDwellTime    = ( new Integer( (String)( data.getSiemensSpecificDataFromFile( ).get( "sRXSPEC.alDwellTime[0]" ) ) ) ).intValue( );
               
               m_acqFrequency =  1000f * 1000000f / ( 2f * m_sampleDwellTime );
               
               System.out.println( "m_acqFreq = " + m_acqFrequency );
               
           }
           
           addSVSSeries( new Integer( data.getHeader( ).getInstanceNumber( ).trim( ) ).intValue( ) - 1, data.getHeader( ).getNonImageData( ) );
           
           m_imageNumberToFilenameMap.put( new Integer( data.getHeader( ).getInstanceNumber( ).trim( ) ) - 1, fileName );
           counter++;
       }
       
       m_curveColor = color;
       // define the first specrum to be the current one
       m_currentInstance = 0;
       
    }

    /**
     * Creates a new instance of SingleVoxelSpectralDataSeries and reserves N sample
     * points for the response
     */
    public SingleVoxelSpectralDataSeries( DicomImageSeriesData dicomDataSeries, Color color )
    {
       
       m_instances = dicomDataSeries.getNumberOfImages( );
       
       int counter = 0;
       Iterator it = dicomDataSeries.getImageSeriesIterator( );
       
       while ( it.hasNext( ) )
       {
           DicomImageData data = dicomDataSeries.getDicomImageData( (String)it.next( ) );
           
           if ( counter == 0 )
           {
               setSiemensSpecificDicomHeaderInfo( data.getSiemensSpecificDataFromFile( ) );
               setDicomHeaderInfo( data.getHeader( ) );
               m_resonanceFrequency = ( new Long( (String)( data.getSiemensSpecificDataFromFile( ).get( "sTXSPEC.asNucleusInfo[0].lFrequency" ) ) ) ).longValue( );
               
               m_sampleDwellTime    = ( new Integer( (String)( data.getSiemensSpecificDataFromFile( ).get( "sRXSPEC.alDwellTime[0]" ) ) ) ).intValue( );
               
               m_acqFrequency =  1000f * 1000000f / ( 2f * m_sampleDwellTime );
               
               System.out.println( "m_acqFreq = " + m_acqFrequency );
               
           }
           
           addSVSSeries( new Integer( data.getHeader( ).getInstanceNumber( ).trim( ) ).intValue( ) - 1, data.getHeader( ).getNonImageData( ) );

           m_imageNumberToFilenameMap.put( new Integer( data.getHeader( ).getInstanceNumber( ).trim( ) ) - 1, data.getHeader( ).getSOPInstanceUID( ) );
           counter++;
       }
       
       m_curveColor = color;
       // define the first specrum to be the current one
       m_currentInstance = 0;
       
    }

    
    
    /**
     * Adds a single voxel data array to the series of single voxel spectra 
     */
    private void addSVSSeries( int instanceNumber, byte[] siemensDicomRawData )
    {
        
        m_currentInstance = instanceNumber;
        
        byte[] dummy = new byte[4];
        
        if ( m_bothChannelsTD == null )
        {
            m_numberOfSamplePoints = siemensDicomRawData.length / 8;
            allocateArrays( m_numberOfSamplePoints, m_instances );
        }
        
        for ( int i = 0; i < siemensDicomRawData.length / 4; i++ )
        {
            dummy[ 0 ] = siemensDicomRawData[ 4 * i + 3 ];
            dummy[ 1 ] = siemensDicomRawData[ 4 * i + 2 ];
            dummy[ 2 ] = siemensDicomRawData[ 4 * i + 1 ];
            dummy[ 3 ] = siemensDicomRawData[ 4 * i + 0 ];
                
            m_bothChannelsTD[ instanceNumber * m_numberOfSamplePoints + i ] = Convert.byte2float( dummy, 0 );
        }
        
        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {
            m_absorptionChannelTD[ instanceNumber * m_numberOfSamplePoints + i ] = m_bothChannelsTD[ instanceNumber * m_numberOfSamplePoints + 2 * i ];
            m_dispersionChannelTD[ instanceNumber * m_numberOfSamplePoints + i ] = m_bothChannelsTD[ instanceNumber * m_numberOfSamplePoints + 2 * i + 1 ];
        }
        
        m_absorptionChannelTD[ instanceNumber * m_numberOfSamplePoints] = m_absorptionChannelTD[ instanceNumber * m_numberOfSamplePoints]/2.0f;
        m_dispersionChannelTD[ instanceNumber * m_numberOfSamplePoints] = m_dispersionChannelTD[ instanceNumber * m_numberOfSamplePoints]/2.0f;
        
        calculateFDRepresentation( instanceNumber );
        
        autoSetXaxis( m_numberOfSamplePoints );
        
    }
    
    /**
     * Returns the persistent (from file) SVS instance as a 
     * SingleVoxelSpectralData object.
     * @param instanceNumber
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getPersistentInstance( int instanceNumber )
    {   
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );

        String fileName = m_imageNumberToFilenameMap.get( new Integer( instanceNumber ) );
        
        DicomImageData d_spec_data = new DicomImageData( fileName );
        m_currentDicomHeaderInfo = d_spec_data.getHeader( );
        
        svsData =  new SingleVoxelMeasuredSpectralData( m_currentDicomHeaderInfo.getNonImageData( ) );
        svsData.setDicomHeaderInfo( m_currentDicomHeaderInfo );
        m_currentSiemensHeaderInfo = d_spec_data.getSiemensSpecificDataFromFile( );
        svsData.setSiemensSpecificDicomHeaderInfo( m_currentSiemensHeaderInfo );        
        
        return svsData;
    }
    
    /**
     * Returns the current in memory SVS instance as a 
     * SingleVoxelSpectralData object.
     */
    public SingleVoxelMeasuredSpectralData getCurrentInstance( int instanceNumber )
    {   
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );

        String fileName = m_imageNumberToFilenameMap.get( new Integer( instanceNumber ) );
        
        DicomImageData d_spec_data = new DicomImageData( fileName );
        m_currentDicomHeaderInfo = d_spec_data.getHeader( );
        
        
        svsData =  new SingleVoxelMeasuredSpectralData( getAbsorptionChannelTD( instanceNumber ), getDispersionChannelTD( instanceNumber ) );
        
        svsData.setDicomHeaderInfo( m_currentDicomHeaderInfo );
        m_currentSiemensHeaderInfo = d_spec_data.getSiemensSpecificDataFromFile( );
        svsData.setSiemensSpecificDicomHeaderInfo( m_currentSiemensHeaderInfo );        
        
        return svsData;
    }
    
    /**
     * Allocates the working data arrays.
     * @param samplePoints
     * @param instances
     */
    protected void allocateArrays( int samplePoints, int instances )
    {
        m_bothChannelsTD = new float[2 * samplePoints * instances];
        m_absorptionChannelTD = new float[samplePoints * instances];
        m_dispersionChannelTD = new float[samplePoints * instances];
        m_xaxisTD = new float[ samplePoints ];

        m_bothChannelsFD = new float[2 * samplePoints * instances];
        m_absorptionChannelFD = new float[samplePoints * instances];
        m_dispersionChannelFD = new float[samplePoints * instances];
        m_xaxisFD = new float[ samplePoints ];
    }
    
    
    /**
     * Allocates the working data arrays again!
     * @param samplePoints
     * @param instances
     */
    public void reallocateArrays( int samplePoints, int instances )
    {
        m_instances = instances;
        m_numberOfSamplePoints = samplePoints; 
        
        m_bothChannelsTD = new float[2 * samplePoints * instances];
        m_absorptionChannelTD = new float[samplePoints * instances];
        m_dispersionChannelTD = new float[samplePoints * instances];
        m_xaxisTD = new float[samplePoints];

        m_bothChannelsFD = new float[2 * samplePoints * instances];
        m_absorptionChannelFD = new float[samplePoints * instances];
        m_dispersionChannelFD = new float[samplePoints * instances];
        m_xaxisFD = new float[samplePoints];
    }
    
    /**
     * Adds a time response to instance number "instance".
     * @param instance
     * @param absorption
     * @param dispersion
     */
    public void addTimeDomainResponse( int instance, float[] absorption, float[] dispersion )
    {
        m_currentInstance = instance;
        
        if ( ( absorption.length == m_numberOfSamplePoints ) &&
             ( dispersion.length == m_numberOfSamplePoints ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {
                m_absorptionChannelTD[instance * m_numberOfSamplePoints + i] += absorption[i];
                m_dispersionChannelTD[instance * m_numberOfSamplePoints + i] += dispersion[i];
            }
        }
        // brings the frequency domain response into sync
        calculateFDRepresentation( m_currentInstance );
    }
    
    /**
     * Adds a frequency response to the current response.
     * @param instance
     * @param dispersion
     * @param absorption
     */
    public void addFrequencyDomainResponse( int instance, float[] absorption, 
                                            float[] dispersion )
    {   
        m_currentInstance = instance;
        
        if ( ( absorption.length == m_numberOfSamplePoints ) &&
             ( dispersion.length == m_numberOfSamplePoints ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {
                m_absorptionChannelFD[instance * m_numberOfSamplePoints + i] += absorption[i];
                m_dispersionChannelFD[instance * m_numberOfSamplePoints + i] += dispersion[i];
            }
        }
        // brings the time domain response into sync
        calculateTDRepresentation( m_currentInstance );
    }
    
    /**
     * Subtract a time response to the current response.
     * @param instance
     * @param absorption
     * @param dispersion
     */
    public void subtractTimeDomainResponse( int instance, float[] absorption, 
                                            float[] dispersion )
    {
        m_currentInstance = instance;
        
        if ( ( absorption.length == m_numberOfSamplePoints ) &&
             ( dispersion.length == m_numberOfSamplePoints ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {
                m_absorptionChannelTD[instance * m_numberOfSamplePoints + i] -= absorption[i];
                m_dispersionChannelTD[instance * m_numberOfSamplePoints + i] -= dispersion[i];
            }
        }
        // brings the frequency domain response into sync
        calculateFDRepresentation( m_currentInstance );
    }
    
    /**
     * Subtract a frequency response to the current response.
     * @param instance
     * @param absorption
     * @param dispersion
     */
    public void subtractFrequencyDomainResponse( int instance, 
                                                float[] absorption, 
                                                float[] dispersion )
    {
        if ( ( absorption.length == m_numberOfSamplePoints ) &&
             ( dispersion.length == m_numberOfSamplePoints ) &&
             ( absorption.length == dispersion.length ) )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {
                m_absorptionChannelFD[instance * m_numberOfSamplePoints + i] -= absorption[i];
                m_dispersionChannelFD[instance * m_numberOfSamplePoints + i] -= dispersion[i];
            }
        }
        // brings the time domain response into sync
        calculateTDRepresentation( m_currentInstance );
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     * @param seriesNumber
     */
    protected void calculateFDRepresentation( int seriesNumber )
    {
        float[][] fftArray = new float[m_numberOfSamplePoints][2];
                
        for ( int i = 0 ; i < m_numberOfSamplePoints ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelTD[seriesNumber * m_numberOfSamplePoints + i];
            fftArray[i][1] = m_dispersionChannelTD[seriesNumber * m_numberOfSamplePoints + i];
        }
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        for ( int i = 0 ; i < m_numberOfSamplePoints ; i++ )
        {
            int d = ( i + m_numberOfSamplePoints / 2 ) % m_numberOfSamplePoints;
            m_absorptionChannelFD[seriesNumber * m_numberOfSamplePoints + d] = fftArray[i][0];
            m_dispersionChannelFD[seriesNumber * m_numberOfSamplePoints + d] = fftArray[i][1];
        }
    
        fftArray = null;
    }
    
    /**
     * Calculates the frequency domain representation of 
     * the absorption and dispersion channels.
     */
    private void calculateTDRepresentation( int seriesNumber )
    {
        float[][] fftArray = new float[m_numberOfSamplePoints][2];
        float[] bufAbs = new float[m_numberOfSamplePoints];
        float[] bufDisp = new float[m_numberOfSamplePoints];
               
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            fftArray[i][0] = m_absorptionChannelFD[ seriesNumber * m_numberOfSamplePoints + m_numberOfSamplePoints - 1 - i ];
            fftArray[i][1] = m_dispersionChannelFD[ seriesNumber * m_numberOfSamplePoints + m_numberOfSamplePoints - 1 - i ];
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
             bufAbs[ i ] = fftArray[i][0];
             bufDisp[ i ] = fftArray[i][1];
        }
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            int d = ( i + 1 + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelFD.length;
            fftArray[d][0] = bufAbs[i];
            fftArray[d][1] = bufDisp[i];
        }
        
        bufAbs = null;
        bufDisp = null;
        
        FFT fft = new FFT( );
        fftArray = fft.fft_1d( fftArray );
        
        for ( int i = 0 ; i < m_absorptionChannelFD.length ; i++ )
        {
            //int d = ( i + m_absorptionChannelTD.length / 2 ) % m_absorptionChannelFD.length;
            m_absorptionChannelTD[seriesNumber * m_numberOfSamplePoints + i] = fftArray[i][0] / m_numberOfSamplePoints;
            m_dispersionChannelTD[seriesNumber * m_numberOfSamplePoints + i] = fftArray[i][1] / m_numberOfSamplePoints;
        }
    
        fftArray = null;
    }
    
    /**
     * Sets the default time and frequency domain x-axis.
     */
    protected void autoSetXaxis( )
    {
        autoSetXaxis( m_absorptionChannelFD.length );
    }
    
    /**
     * Sets the default time and frequency domain x-axis.
     * @param length
     */
    protected void autoSetXaxis( int length )
    {
        
        TDFDFittingDefaults fitdef = TDFDFittingDefaults.getInstance( );
        fitdef.load( );
        
        String displayDomain = fitdef.getProperty( TDFDFittingDefaults.DISPLAY_DOMAIN );
        String displayUnits = fitdef.getProperty( TDFDFittingDefaults.DISPLAY_UNIT );
        
        if ( displayDomain.compareTo( TDFDFittingDefaults.TIME_DOMAIN ) == 0 )
        {
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.TIME_DOMAIN_POINTS ) == 0 )
            {
                calculateTDXaxisPoints( length );
            }
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.TIME_DOMAIN_SECONDS ) == 0 )
            {
                calculateTDXaxisSeconds( length );
            }
        }

        if ( displayDomain.compareTo( TDFDFittingDefaults.FREQ_DOMAIN ) == 0 )
        {
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_POINTS ) == 0 )
            {
                calculateFDXaxisPoints( length );
            }
            
            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_HZ ) == 0 )
            {
                calculateFDXaxisHertz( length );
            }

            if ( displayUnits.compareTo( TDFDFittingDefaults.FREQUENCY_DOMAIN_PPM ) == 0 )
            {
                calculateFDXaxisPPM( length );
            }
        }
    }
    
    /**
     * Calculates the time domain x-axis in point units.
     * @param length
     */
    protected void calculateTDXaxisPoints( int length )
    {
        m_xaxisTD = new float[ length ];
        
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisTD[i] = (float)i;
        }
    }

    /**
     * Calculates the frequency domain x-axis in point units.
     */
    protected void calculateFDXaxisPoints( int length )
    {
        m_xaxisFD = new float[ length ];
        
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisFD[i] = (float)( i - ( length / 2 ) );
        }
    }
    
    /**
     * Calculates the time domain x-axis in second units.
     * @param length
     */
    protected void calculateTDXaxisSeconds( int length )
    {
        m_xaxisTD = new float[ length ];
        
        MRSignalProperties cond = MRSignalProperties.getInstance( );
        float specWidth = cond.getFrequencyScale( ) * length;
        for ( int i= 0; i < length; i++ )
        {
            m_xaxisTD[i] = (float)i / specWidth;
        }
    }

    /**
     * Calculates the time domain x-axis in second units.
     * @param length
     */
    protected void calculateFDXaxisHertz( int length )
    {
        m_xaxisFD = new float[ length ];
                
        for ( int i = 0; i < length; i++ )
        {
            m_xaxisFD[i] = ( (float)( i - length / 2 ) / length ) * m_acqFrequency;
        }
    }
    
    /**
     * Calculates the time domain x-axis in second units.
     * @param length
     */
    protected void calculateFDXaxisPPM( int length )
    {
        
        m_xaxisFD = new float[ length ];
        
        for ( int i = 0; i < length; i++ )
        {
            /* TO DO make 4.7 dependent on nucleus choice or setting of resonance frequency*/
            m_xaxisFD[i] = 4.7f + ( (float)( i - length / 2 ) / length ) * m_acqFrequency / (float)m_resonanceFrequency;
        }
    }
    
    /** 
     * Set spectral data points.
     * @param instance
     * @param bothChannelsTD
     */
    public void setSpectralDataPointsTD( int instance, float[] bothChannelsTD )
    {
        m_currentInstance = instance;
        
        // bothChannels can directly be copied into m_bothChannels
        System.arraycopy( bothChannelsTD, 0, m_bothChannelsTD, 0, bothChannelsTD.length );
        if ( this.m_numberOfSamplePoints == bothChannelsTD.length / 2 )
        {
            for ( int i = 0; i < bothChannelsTD.length / 2; i++ )
            {
                m_absorptionChannelTD[ instance * m_numberOfSamplePoints + i ] = bothChannelsTD[ 2 * i ];
                m_dispersionChannelTD[ instance * m_numberOfSamplePoints + i ] = bothChannelsTD[ 2 * i + 1 ];
            }
        
            calculateFDRepresentation( m_currentInstance );
        
            // Set a default xaxis (time domain)
            autoSetXaxis( );
        }
        else
        {
            System.out.println( "Fatal error: Array size do not match in 'SingleVoxelSpectralDataSeries::setSpectralDataPointsTD' " );
        }
    }
    
    /** 
     * Set spectral data points.
     * @param instance
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion channel
     */
    public void setSpectralDataPointsTD( int instance, 
                                         float[] absorption, float[] dispersion )
    {
        
        m_currentInstance = instance;
        
        if ( ( m_numberOfSamplePoints == absorption.length ) && ( m_numberOfSamplePoints == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length ; i++ )
            {
                m_absorptionChannelTD[ instance * m_numberOfSamplePoints + i ] = absorption[ i ];
                m_dispersionChannelTD[ instance * m_numberOfSamplePoints + i ] = dispersion[ i ];
            }
                
            calculateFDRepresentation( instance );
        
            // Set a default xaxis (time domain)
            autoSetXaxis( );
        }
        else
        {
            System.out.println( "Fatal error: Array size do not match in 'SingleVoxelSpectralDataSeries::setSpectralDataPointsTD( )' " );
        }
    }

    /** 
     * Set spectral data points in time domain with a new number of sample points
     * This method allows for zero filling.
     * @param zeroFillFactor
     */
    public void setSpectralDataPointsTD( float zeroFillFactor )
    {
        
        // Backup data
        float[] absTD = new float[m_absorptionChannelTD.length];
        float[] dispTD = new float[m_dispersionChannelTD.length];
        float[] bothTD = new float[m_bothChannelsTD.length];
        
        for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
        {
            absTD[i] = m_absorptionChannelTD[i];
            dispTD[i] = m_dispersionChannelTD[i];
            
            bothTD[2*i] = m_absorptionChannelTD[i];
            bothTD[2*i+1] = m_dispersionChannelTD[i];
        }
        
        int oldNumberOfSamplePoints = m_numberOfSamplePoints;
        int newNumberOfSamplePoints = (int)(( m_absorptionChannelTD.length * zeroFillFactor ) / m_instances );
        
        int samplePoints;
        
        if ( zeroFillFactor < 1.0f )
        {
            samplePoints = newNumberOfSamplePoints;
        }
        else
        {
            samplePoints = oldNumberOfSamplePoints;
        }
        
        // Allocate new arrays
        reallocateArrays( newNumberOfSamplePoints, m_instances );
        
        // restore the data in the appropriate members
         // bring TD and FD representation in sync
        for ( int instanceNumber = 0 ; instanceNumber < m_instances ; instanceNumber++ )
        {
            for ( int i = 0 ; i < samplePoints ;i++ )
            {
                m_absorptionChannelTD[instanceNumber* newNumberOfSamplePoints + i] = absTD[instanceNumber * oldNumberOfSamplePoints + i];
                m_dispersionChannelTD[instanceNumber* newNumberOfSamplePoints + i] = dispTD[instanceNumber * oldNumberOfSamplePoints + i];
            }
        }
        
        for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
        {
            m_bothChannelsTD[2*i] = m_absorptionChannelTD[i];
            m_bothChannelsTD[2*i+1] = m_dispersionChannelTD[i];
        }
        
        
        // bring TD and FD representation in sync
        for ( int instance = 0 ; instance < m_instances ; instance++ )
        {
            calculateFDRepresentation( instance );
        }
        
        // Set a default xaxis (time domain)
        autoSetXaxis( );
    }

    
    /** 
     * Set spectral data points.
     * @param instance
     * @param absorption is a floating point array that contains the absorption channel
     * @param dispersion is a floating point array that contains the dispersion channel
     */
    public void setSpectralDataPointsFD( int instance, float[] absorption, float[] dispersion )
    { 
        m_currentInstance = instance;
        
        if ( ( m_numberOfSamplePoints == absorption.length ) && ( m_numberOfSamplePoints == dispersion.length ) )
        {
            for ( int i = 0; i < absorption.length ; i++ )
            {
                m_absorptionChannelFD[ instance * m_numberOfSamplePoints + i ] = absorption[ i ];
                m_dispersionChannelFD[ instance * m_numberOfSamplePoints + i ] = dispersion[ i ];
            }
        
            calculateTDRepresentation( m_currentInstance );
        
            // Set a default xaxis (time domain)
            autoSetXaxis( );
        }
        else
        {
            System.out.println( "Fatal error: Array size do not match in 'SingleVoxelSpectralDataSeries::setSpectralDataPointsFD(2args)' " );
        }
    }
    
    
    /** 
     * Returns the time domain absorption channel of all instances as 
     * one floating point block.
     *
     * @return returns the absorption channel floating point array
     */
    public float[] getAbsorptionChannelTD( )
    {
        return m_absorptionChannelTD;
    }

    /** 
     * Returns the time domain absorption channel of the current instance.
     * @return returns the absorption channel floating point array
     */
    public float[] getCurrentAbsorptionChannelTD( )
    {
        float[] abs_td = new float[m_numberOfSamplePoints];
        
        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {
            abs_td[i] = m_absorptionChannelTD[m_currentInstance * m_numberOfSamplePoints + i];
        }
        return abs_td;
    }

    
    /** 
     * Returns the frequency domain absorption channel of all instances as one
     * floating point block.
     * @return returns the absorption channel floating point array
     */
    public float[] getAbsorptionChannelFD( )
    {
        return m_absorptionChannelFD;
    }

    
   /** 
     * Returns the time domain absorption channel of the current instance.
     * @return returns the absorption channel floating point array
     */
    public float[] getCurrentAbsorptionChannelFD( )
    {
        float[] abs_fd = new float[m_numberOfSamplePoints];
        
        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {
            abs_fd[i] = m_absorptionChannelFD[m_currentInstance * m_numberOfSamplePoints + i];
        }
        return abs_fd;
    }
    
    
    /** 
     * Returns the time domain absorption channel as Curve object.
     * @return the time domain absorption channel as Curve object
     */
    public Curve getCurrentAbsorptionCurveTDPoints( )
    {   
        return new Curve( m_xaxisTD, 
                          getCurrentAbsorptionChannelTD( ), m_curveColor );
    }

    /** 
     * Returns the time domain absorption channel as Curve object.
     * @return the time domain absorption channel as Curve object
     */
    public Curve getCurrentAbsorptionCurveTDSeconds( )
    {
        return new Curve( m_xaxisTD, 
                          getCurrentAbsorptionChannelTD( ), m_curveColor );
    }
    
    
    /** 
     * Returns the frequency domain absorption channel as Curve object
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getCurrentAbsorptionCurveFDPoints( )
    {
        return new Curve( m_xaxisFD, 
                          getCurrentAbsorptionChannelFD( ), m_curveColor );
    }
    
    /** 
     * Returns the frequency domain absorption channel as Curve object
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getCurrentAbsorptionCurveFDHertz( )
    {
        return new Curve( m_xaxisFD, 
                          getCurrentAbsorptionChannelFD( ), m_curveColor );
    }
    
    
    /** 
     * Returns the frequency domain absorption channel as Curve object.
     * @return the frequency domain absorption channel as Curve object
     */
    public Curve getCurrentAbsorptionCurveFDPPM( )
    {
        return new Curve( m_xaxisFD, 
                          getCurrentAbsorptionChannelFD( ), m_curveColor );
    }
    
    /** 
     * Returns the time domain dispersion channel
     * @return returns the dispersion channel floating point array
     */
    public float[] getCurrentDispersionChannelTD( )
    {
        
        float[] disp_td = new float[m_numberOfSamplePoints];
        
        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {
            disp_td[i] = m_dispersionChannelTD[m_currentInstance * m_numberOfSamplePoints + i];
        }
        
        return disp_td;
    }    
    
    /** 
     * Returns the frequency domain dispersion channel.
     * @return returns the dispersion channel floating point array
     */
    public float[] getCurrentDispersionChannelFD( )
    {
        float[] disp_fd = new float[m_numberOfSamplePoints];
        
        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {
            disp_fd[i] = m_dispersionChannelFD[m_currentInstance * m_numberOfSamplePoints + i];
        }
        
        return disp_fd; 
    }    
    
    /** 
     * Returns the time domain dispersion channel as Curve object.
     * @return the time domain dispersion channel as Curve object
     */
    public Curve getCurrentDispersionCurveTDPoints( )
    {
        return new Curve( m_xaxisTD, getCurrentDispersionChannelTD( ), 
                          m_curveColor );
    }

    /** 
     * Returns the time domain dispersion channel as Curve object.
     * @return the time domain dispersion channel as Curve object
     */
    public Curve getCurrentDispersionCurveTDSeconds( )
    {
        return new Curve( m_xaxisTD, getCurrentDispersionChannelTD( ), 
                          m_curveColor );
    }    
    
    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getCurrentDispersionCurveFDPoints( )
    {
        return new Curve( m_xaxisFD, getCurrentDispersionChannelFD( ), 
                          m_curveColor );
    } 

    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getCurrentDispersionCurveFDHertz( )
    {
        return new Curve( m_xaxisFD, getCurrentDispersionChannelFD( ), 
                          m_curveColor );
    } 
    
    /** 
     * Returns the frequency domain dispersion channel as Curve object.
     * @return the frequency domain dispersion channel as Curve object
     */
    public Curve getCurrentDispersionCurveFDPPM( )
    {
        return new Curve( m_xaxisFD, getCurrentDispersionChannelFD( ), 
                          m_curveColor );
    }    
    
    /** 
     * Returns the time domain magnitude of the (absorption,dispersion) channel.
     * @return the time domain magnitude floating point array
     */
    public Curve getCurrentMagnitudeCurveTDPoints( )
    {
        float[] magnitude = new float[m_numberOfSamplePoints];
        
        float[] absTD = this.getCurrentAbsorptionChannelTD( );
        float[] dispTD = this.getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            magnitude[i] = (float)Math.sqrt( absTD[i] * absTD[i] + dispTD[i] * dispTD[i] );
        }
        
        return new Curve( m_xaxisTD, magnitude, m_curveColor );
    }

    /** 
     * Returns the time domain magnitude of the (absorption,dispersion) channel.
     * @return the time domain magnitude floating point array
     */
    public Curve getCurrentMagnitudeCurveTDSeconds( )
    {
        float[] magnitude = new float[m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            magnitude[i] = (float)Math.sqrt( absTD[i] * absTD[i] + dispTD[i] * dispTD[i] );
        }
        
        return new Curve( m_xaxisTD, magnitude, m_curveColor );
    }
    
    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getCurrentMagnitudeCurveFDPoints( )
    {
        float[] magnitude = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            magnitude[i] = (float)Math.sqrt( absFD[i] * absFD[i] + dispFD[i] * dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, magnitude, m_curveColor );
    }
    
    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getCurrentMagnitudeCurveFDHertz( )
    {
        float[] magnitude = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            magnitude[i] = (float)Math.sqrt( absFD[i] * absFD[i] + dispFD[i] * dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, magnitude, m_curveColor );
    }

    /** 
     * Returns the frequency domain magnitude of the 
     * (absorption,dispersion) channel.
     * @return returns the frequency domain magnitude floating point array
     */
    public Curve getCurrentMagnitudeCurveFDPPM( )
    {
        float[] magnitude = new float[m_numberOfSamplePoints];
        

        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            magnitude[i] = (float)Math.sqrt( absFD[i] * absFD[i] + dispFD[i] * dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, magnitude, m_curveColor );
    }
    
    /** 
     * Returns the power of the (absorption,dispersion) channel.
     * @return returns the power floating point array
     */
    public Curve getCurrentPowerCurveTDPoints( )
    {
        float[] power = new float[m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {   
            power[i] =  absTD[i] * absTD[i] + dispTD[i] * dispTD[i];
        }
        
        return new Curve( m_xaxisTD, power, m_curveColor );
    }    

    /** 
     * Returns the power of the (absorption,dispersion) channel.
     * @return returns the power floating point array
     */
    public Curve getCurrentPowerCurveTDSeconds( )
    {
        float[] power = new float[m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {   
            power[i] =  absTD[i] * absTD[i] + dispTD[i] * dispTD[i];
        }
        
        return new Curve( m_xaxisTD, power, m_curveColor );
    }

    /** 
     * Returns the frequency domain power of the (absorption,dispersion) 
     * channel.
     * @return the frequency domain power floating point array
     */
    public Curve getCurrentPowerCurveFDPoints( )
    {
        float[] power = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {   
            power[i] =  absFD[i] * absFD[i] + dispFD[i] * dispFD[i];
        }
        
        return new Curve( m_xaxisFD, power, m_curveColor );
    }    

    /** 
     * Returns the frequency domain power of the (absorption,dispersion) 
     * channel.
     * @return the frequency domain power floating point array
     */
    public Curve getCurrentPowerCurveFDHertz( )
    {
        float[] power = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {   
            power[i] =  absFD[i] * absFD[i] + dispFD[i] * dispFD[i];
        }
        
        return new Curve( m_xaxisFD, power, m_curveColor );
    }    
    
    /** 
     * Returns the frequency domain power of the (absorption,dispersion) 
     * channel.
     * @return the frequency domain power floating point array
     */
    public Curve getCurrentPowerCurveFDPPM( )
    {
        float[] power = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {   
            power[i] =  absFD[i] * absFD[i] + dispFD[i] * dispFD[i];
        }
        
        return new Curve( m_xaxisFD, power, m_curveColor );
    }    
    
    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getCurrentPhaseCurveTDPoints( )
    {
        float[] phase = new float[m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            phase[i] = (float)Math.atan2( absTD[i] , dispTD[i] );
        }
        
        return new Curve( m_xaxisTD, phase, m_curveColor );
    }

    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getCurrentPhaseCurveTDSeconds( )
    {
        float[] phase = new float[m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            phase[i] = (float)Math.atan2( absTD[i] , dispTD[i] );
        }
        
        return new Curve( m_xaxisTD, phase, m_curveColor );
    }

    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getCurrentPhaseCurveFDPoints( )
    {
        float[] phase = new float[m_numberOfSamplePoints];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            phase[i] = (float)Math.atan2( absFD[i] , dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, phase, m_curveColor );
    }
    
    
    /** 
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getCurrentPhaseCurveFDHertz( )
    {
        float[] phase = new float[ m_absorptionChannelFD.length ];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            phase[i] = (float)Math.atan2( absFD[i] , dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, phase, m_curveColor );
    }
    
    /**
     * Returns the phase of the (absorption, dispersion) channels.
     * @return returns the phase floating point array
     */
    public Curve getCurrentPhaseCurveFDPPM( )
    {
        float[] phase = new float[ m_absorptionChannelFD.length ];
        
        float[] absFD = getCurrentAbsorptionChannelFD( );
        float[] dispFD = getCurrentDispersionChannelFD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            phase[i] = (float)Math.atan2( absFD[i] , dispFD[i] );
        }
        
        return new Curve( m_xaxisFD, phase, m_curveColor );
    }
    
    
    /** 
     * Returns the complex signal with alternating absorption point
     * dispersion point of the current instance.
     * @return returns float array containing the sampled spectroscopic data points
     */
    public float[] getCurrentBothChannelsTD( )
    {
        float[] both = new float[2*m_numberOfSamplePoints];
        
        float[] absTD = getCurrentAbsorptionChannelTD( );
        float[] dispTD = getCurrentDispersionChannelTD( );
        
        for ( int i= 0; i < m_numberOfSamplePoints; i++ )
        {
            both[2*i] = absTD[i];
            both[2*i + 1] = dispTD[i];
        }
        
        return both;
    }
    
    /** 
     * Sets the sample frequency attribute.
     * @param sampleFrequency Frequency with which the MR signal is acquired 
     */    
    public void setSampleFrequency( float sampleFrequency )
    {
        m_sampleFrequency = sampleFrequency;
    }
    
    /** 
     * Returns the sample frequency with which the MR signal was acquired.
     * @return Returns the sample frequency in Hz 
     */    
    public float getSampleFrequency( )
    {
        return m_sampleFrequency;
    }
    
    
    /** 
     * Sets the sample frequency attribute. 
     * @param samplePoints
     */    
    public void setNumberOfSamplePoints( int samplePoints )
    {
        m_numberOfSamplePoints = samplePoints;
    }
    
    /** 
     * Returns the sample frequency with which the MR signal was acquired
     * @return Returns the sample frequency in Hz 
     */    
    public float getNumberOfSamplePoints( )
    {
        return m_numberOfSamplePoints;
    }
    
    /**
     * Clears both time and frequency domain channels. 
     */
    public void clear( )
    {
        // loops over all instances!!
        for ( int i = 0; i < m_absorptionChannelTD.length; i++ )
        {
            m_absorptionChannelTD[i] = 0f;
            m_dispersionChannelTD[i] = 0f;
            m_absorptionChannelFD[i] = 0f;
            m_dispersionChannelFD[i] = 0f;
        }
        // loops over all instances!!
        for ( int i = 0; i < 2*m_absorptionChannelTD.length; i++ )
        {
            m_bothChannelsTD[i] = 0f;
            m_bothChannelsFD[i] = 0f;
        }
    }

    /**
     * Writes attributes to RandomAccessFile.
     * @param out
     */
    public void write( RandomAccessFile out )
    {
        try
        {
            out.writeUTF( "<SingleVoxelSpectralDataSeries>" );
            out.writeUTF( "<SingleVoxelSpectralDataSeriesHeader>" );
            out.writeInt( m_numberOfSamplePoints );
            out.writeInt( m_instances );
            out.writeFloat( m_sampleFrequency );
            out.writeUTF( "</SingleVoxelSpectralDataSeriesHeader>" );
            
            out.writeUTF( "<AbsorptionChannels>" );
            if ( m_absorptionChannelTD != null )
            {
                // loops over all instances!!
                for ( int i = 0; i < this.m_absorptionChannelTD.length; i++)
                {
                    out.writeFloat( m_absorptionChannelTD[i] );
                }
            }
            out.writeUTF( "</AbsorptionChannels>" );
            out.writeUTF( "<DispersionChannels>" );
            if ( m_dispersionChannelTD != null )
            {
                // loops over all instances!!
                for ( int i = 0; i < this.m_absorptionChannelTD.length;i++ )
                {
                    out.writeFloat( m_dispersionChannelTD[i] );
                }
            }
            out.writeUTF( "</DispersionChannels>" );
            out.writeUTF( "</SingleVoxelSpectralDataSeries>" );
        }
        catch ( IOException e )
        {
        }
    }
    
    /**
     * Reads and parses the attributes from a RandomAccesFile.
     * @param in
     */
    public void read( RandomAccessFile in )
    {
        try
        {
            String tag = in.readUTF( );            
            if ( tag.compareTo( "<SingleVoxelSpectralDataSeries>" ) == 0 )
            {
                tag = in.readUTF( );
                if ( tag.compareTo( "<SingleVoxelSpectralDataSeriesHeader>" ) == 0 )
                {
                    
                    m_numberOfSamplePoints = in.readInt( );
                    m_instances = in.readInt( );
                    
                    // allocate memory for the data arrays
                    allocateArrays( m_numberOfSamplePoints, m_instances );
                    
                    m_sampleFrequency = in.readFloat( );
                    
                }
                tag = in.readUTF( );
            }
            
            tag = in.readUTF( );
            if ( tag.compareTo( "<AbsorptionChannels>" ) == 0 )
            {
                
                m_absorptionChannelTD = new float[ m_instances * m_numberOfSamplePoints ];
                for ( int i = 0 ; i < m_instances * m_numberOfSamplePoints; i++ )
                {
                    m_absorptionChannelTD[i] = in.readFloat( );
                }
                // read closing tag absorption channel
                tag = in.readUTF( );
            }
            
            tag = in.readUTF( );
            if ( tag.compareTo( "<DispersionChannels>" ) == 0 )
            {
                
                m_dispersionChannelTD = new float[  m_instances * m_numberOfSamplePoints ];
                for ( int i = 0 ; i < m_instances * m_numberOfSamplePoints; i++ )
                {
                    m_dispersionChannelTD[i] = in.readFloat( );
                }
                // read closing tag dispersion channel
                tag = in.readUTF( );
            }
            
            for ( int k = 0 ; k < m_instances * m_numberOfSamplePoints; k++ )
            {
                m_absorptionChannelTD[k * m_numberOfSamplePoints] = m_absorptionChannelTD[k * m_numberOfSamplePoints]/2.0f;
                m_dispersionChannelTD[k * m_numberOfSamplePoints] = m_dispersionChannelTD[k * m_numberOfSamplePoints]/2.0f;
            }
            
            // read the closing single voxel spectral data tag
            tag = in.readUTF( );
        }
        catch ( IOException e )
        {
            
        }
        
        // Brings the frequency domain response into sync
        for ( int i = 0 ; i < m_instances; i++ )
        {
            calculateFDRepresentation( i );
        }
    }
    
    /**
     * Writes the current file to tdfdfit file format (to test whether
     * the now implemented way is correct with 2DJ fitting.
     * @param fileName
     */
    public void writeAsTdfdfitFile( String fileName )
    {
        
        TDFDFittingDefaults fitDef = TDFDFittingDefaults.getInstance( );
        fitDef.load( );
        
        String path = fitDef.getProperty( "Fit project path" );
        fileName = path + System.getProperty( "file.separator" ) + fileName;
        System.out.println( "fileName : " + fileName );
        try
        {
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            for ( int k = 0; k < m_instances; k++ )
            {
                for ( int i = 0; i < m_numberOfSamplePoints; i++ )
                {
                    buf = ByteBuffer.allocate( 4 );
                
                    buf.putFloat( m_absorptionChannelFD[k * m_numberOfSamplePoints + i] );
                
                    writer.write( buf.get( 3 ) );
                    writer.write( buf.get( 2 ) );
                    writer.write( buf.get( 1 ) );
                    writer.write( buf.get( 0 ) );
                
                    buf1 = ByteBuffer.allocate( 4 );
                
                    buf1.putFloat( m_dispersionChannelFD[k * m_numberOfSamplePoints + i] );                
                
                    writer.write( buf1.get( 3 ) );
                    writer.write( buf1.get( 2 ) );
                    writer.write( buf1.get( 1 ) );
                    writer.write( buf1.get( 0 ) );
                }            
            }

            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit-formatted (2DJ??) measured datafile" );
        }
    }

    /**
     * Writes the current file to tdfdfit fileformat.
     * @param path
     * @param fileName
     */
    public void writeAsTdfdfitFile( String path, String fileName )
    {
        
        if ( path != null )
        {
            fileName = path + System.getProperty( "file.separator" ) + fileName;
        }
        
        System.out.println( "fileName : " + fileName );
        try
        {
            
            RandomAccessFile writer = new RandomAccessFile( fileName, "rw" );
            
            ByteBuffer buf, buf1;
            for ( int k = 0; k < m_instances; k++ )
            {            
                for ( int i = 0; i < m_absorptionChannelFD.length; i++ )
                {
                    buf = ByteBuffer.allocate( 4 );
                
                    buf.putFloat( m_absorptionChannelFD[k * m_numberOfSamplePoints + i] );
                
                    writer.write( buf.get( 3 ) );
                    writer.write( buf.get( 2 ) );
                    writer.write( buf.get( 1 ) );
                    writer.write( buf.get( 0 ) );
                
                    buf1 = ByteBuffer.allocate( 4 );
                
                    buf1.putFloat( m_dispersionChannelFD[k * m_numberOfSamplePoints + i] );                
                
                    writer.write( buf1.get( 3 ) );
                    writer.write( buf1.get( 2 ) );
                    writer.write( buf1.get( 1 ) );
                    writer.write( buf1.get( 0 ) );    
                }
            }
            
            writer.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "Cannot write tdfdfit (2dJ??)-formatted measured datafile" );
        }
    }
    
    /**
     * Reads a tdfdfit file format binary file and stores its contents in the
     * appropriate member variables.
     * @param numberOfInstances
     * @param numberOfComplexPoints
     * @param fileRef
     */
    public void readTDFDFitFile( int numberOfInstances, 
                                 int numberOfComplexPoints, String fileRef )
    {
        
        m_instances = numberOfInstances;
        m_numberOfSamplePoints = numberOfComplexPoints;
        
        try
        {          
            RandomAccessFile reader = new RandomAccessFile( fileRef, "rw" );
            
            ByteBuffer buf;
            byte byte0;
            byte byte1;
            byte byte2;
            byte byte3;
            float readFloat;
            for ( int k = 0; k < m_instances; k++ )
            {
                for ( int i = 0; i < m_numberOfSamplePoints; i++ )
                {
                
                    buf = ByteBuffer.allocate( 4 );
                
                    byte0 = reader.readByte( );
                    byte1 = reader.readByte( );
                    byte2 = reader.readByte( );
                    byte3 = reader.readByte( );
                  
                    buf.put( byte3 );
                    buf.put( byte2 );
                    buf.put( byte1 );
                    buf.put( byte0 );
                
                    readFloat = buf.getFloat( 0 );
                    buf.clear( );
                
                    m_absorptionChannelFD[k * m_numberOfSamplePoints + i] = readFloat;
                
                    buf = ByteBuffer.allocate( 4 );
                
                    byte0 = reader.readByte( );
                    byte1 = reader.readByte( );
                    byte2 = reader.readByte( );
                    byte3 = reader.readByte( );
                  
                    buf.put( byte3 );
                    buf.put( byte2 );
                    buf.put( byte1 );
                    buf.put( byte0 );
                
                    readFloat = buf.getFloat( 0 );
                    buf.clear( );
                
                    m_dispersionChannelFD[k * m_numberOfSamplePoints + i] = readFloat;
                }
                
                // the frequency domain response is read so generate the time 
                // domain response to get full representation
                calculateTDRepresentation( k );
            }
            reader.close( );
            
        }
        catch ( IOException e )
        {
            System.out.println( "4. Cannot read tdfdfit-formatted measured datafile" );
        }
        
        autoSetXaxis( );
    }
    
    /**
     * Sets the color of the curve to be drawn.
     * @param color
     */
    public void setCurveColor( Color color )
    {
        m_curveColor = color;
    }
    
    /**
     * Sets the Dicom file header information.
     * @param fileInfo
     */
    public void setDicomHeaderInfo( FileInfo fileInfo )
    {
        m_currentDicomHeaderInfo = fileInfo;
    }
    
    /**
     * Returns the Dicom file header information.
     * @return 
     */
    public FileInfo getDicomHeaderInfo( )
    {
        return m_currentDicomHeaderInfo;
    }
    
    /**
     * Returns the Dicom file header information.
     * @param instanceNumber
     * @return 
     */
    public FileInfo getDicomHeaderInfo( int instanceNumber )
    {     
        String fileName = m_imageNumberToFilenameMap.get( new Integer( instanceNumber ) );
        
        DicomImageData d_spec_data = new DicomImageData( fileName );
        
        return d_spec_data.getHeader( );
    }
    
    
    /**
     * Sets the LinkedHashMap containing all Siemens specific header 
     * information.
     * @param ssd
     */
    public void setSiemensSpecificDicomHeaderInfo( LinkedHashMap ssd )
    {
        m_currentSiemensHeaderInfo = ssd;
        
        String name;
        
        if ( ssd != null )
        {
            Iterator it = ssd.keySet( ).iterator( );
        
            while ( it.hasNext( ) )
            {
                Object key = it.next( );
                
                // System.out.println( "" + (String)key );
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dThickness" ) == 0 )
                {
                    m_thickVoxelSize = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dReadoutFOV" ) == 0 )
                {
                    m_readOutVoxelSize = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.dPhaseFOV" ) == 0 )
                {
                    m_phaseVoxelSize = new Double( (String)ssd.get( key ) );
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dSag" ) == 0 )
                {
                    if ( ((String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dSagVoxelPosition =  d.doubleValue( );
                    }
                    else
                    {
                        m_dSagVoxelPosition = 0.00;
                    }
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dCor" ) == 0 )
                {
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dCorVoxelPosition = d.doubleValue( );
                    }
                    else
                    {
                        m_dCorVoxelPosition = 0.00;
                    }
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sPosition.dTra" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dTraVoxelPosition = d.doubleValue( );
                    }
                    else
                    {
                        m_dTraVoxelPosition = 0.00;
                    }
                }

                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dSag" ) == 0 )
                { 
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dSagVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dSagVoxelOrientation = 0.00;
                    }
                
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dCor" ) == 0 )
                {
                    if ( ((String)ssd.get( key )).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dCorVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dCorVoxelOrientation = 0.00;
                    }
                    
                }
                
                if ( ((String)key).compareTo( "sSpecPara.sVoI.sNormal.dTra" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dTraVoxelOrientation = d.doubleValue( );
                    }
                    else
                    {
                        m_dTraVoxelOrientation = 0.00;
                    }
                }
                
                if ( ( (String)key).compareTo( "sSpecPara.sVoI.dInPlaneRot" ) == 0 )
                {
                    if ( ( (String)ssd.get( key ) ).compareTo( "" ) != 0 )
                    {
                        Double d = new Double( (String)ssd.get( key ) );
                        m_dInPlaneRotation = d.doubleValue( );
                    }
                    else
                    {
                        m_dInPlaneRotation = 0.0;
                    }
                }           
            }
        }        
        
        
        if ( ( this.m_dTraVoxelOrientation > this.m_dCorVoxelOrientation ) &&
             ( this.m_dTraVoxelOrientation > this.m_dSagVoxelOrientation ) )
        {
            computeVoxelForTransversalPrescription( );
        }
        
        if ( ( this.m_dCorVoxelOrientation > this.m_dTraVoxelOrientation ) &&
             ( this.m_dCorVoxelOrientation > this.m_dSagVoxelOrientation ) )
        {
            computeVoxelForCoronalPrescription( );
        }
        
        if ( ( this.m_dSagVoxelOrientation > this.m_dTraVoxelOrientation ) &&
             ( this.m_dSagVoxelOrientation > this.m_dCorVoxelOrientation ) )
        {
            computeVoxelForSagitalPrescription( );
        }
        
    }

    
    /**
     * Computes a voxel with Coronal prescription.
     */
    private void computeVoxelForSagitalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dCorVoxelOrientation, -m_dTraVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );

        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dSagVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        
        Point3D point1 = new Point3D(  - m_thickVoxelSize / 2.,
                                         m_phaseVoxelSize / 2. ,
                                       - m_readOutVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        
        Point3D point2 = new Point3D(    m_thickVoxelSize / 2.,
                                         m_phaseVoxelSize / 2. ,
                                       - m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        

        Point3D point3 = new Point3D( m_thickVoxelSize / 2.,
                                      m_phaseVoxelSize / 2. ,
                                      m_readOutVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        
        Point3D point4 = new Point3D( - m_thickVoxelSize / 2.,
                                        m_phaseVoxelSize / 2. ,
                                        m_readOutVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );

        Point3D point5 = new Point3D( - m_thickVoxelSize / 2.,
                                      - m_phaseVoxelSize / 2. ,
                                      - m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        
        Point3D point6 = new Point3D(   m_thickVoxelSize / 2.,
                                      - m_phaseVoxelSize / 2. ,
                                      - m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        

        Point3D point7 = new Point3D(   m_thickVoxelSize / 2.,
                                      - m_phaseVoxelSize / 2. ,
                                        m_readOutVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D( - m_thickVoxelSize / 2.,
                                      - m_phaseVoxelSize / 2. ,
                                        m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
    
    /**
     * Computes a voxel with Transversal prescription.
     */
    private void computeVoxelForCoronalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dTraVoxelOrientation, -m_dSagVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );

        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dCorVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        
        Point3D point1 = new Point3D(  - m_phaseVoxelSize / 2., 
                                         m_thickVoxelSize / 2.,
                                       - m_readOutVoxelSize / 2. );
        
        
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        Point3D point2 = new Point3D(    m_phaseVoxelSize / 2., 
                                         m_thickVoxelSize / 2.,
                                       - m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        

        Point3D point3 = new Point3D(    m_phaseVoxelSize / 2., 
                                         m_thickVoxelSize / 2.,
                                         m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        Point3D point4 = new Point3D(  - m_phaseVoxelSize / 2., 
                                         m_thickVoxelSize / 2.,
                                         m_readOutVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );
        

        Point3D point5 = new Point3D(  - m_phaseVoxelSize / 2., 
                                       - m_thickVoxelSize / 2.,
                                       - m_readOutVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        Point3D point6 = new Point3D(    m_phaseVoxelSize / 2., 
                                      -  m_thickVoxelSize / 2.,
                                      -  m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        
        Point3D point7 = new Point3D(    m_phaseVoxelSize / 2., 
                                      -  m_thickVoxelSize / 2.,
                                         m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D( -  m_phaseVoxelSize / 2., 
                                      -  m_thickVoxelSize / 2.,
                                         m_readOutVoxelSize / 2. );

        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
 
    /**
     * Computes a voxel with Transversal prescription
     */
    private void computeVoxelForTransversalPrescription( )
    {

        Point3D rotAxis1 = new Point3D( m_dCorVoxelOrientation, -m_dSagVoxelOrientation, 0.0 );
        rotAxis1 = rotAxis1.normalize( rotAxis1 );
        
        Point3D rotAxis2 = new Point3D( m_dSagVoxelOrientation, m_dCorVoxelOrientation, m_dTraVoxelOrientation );
        rotAxis2 = rotAxis2.normalize( rotAxis2 );

        
        double rotAngle1 = -Math.acos( m_dTraVoxelOrientation );
        double rotAngle2 = -m_dInPlaneRotation;
        
        // double rotAngle1 = 0.0;
        // double rotAngle2 = 0.0;
        // should be Point3D translate = new Point3D( m_dCorVoxelPosition, m_dSagVoxelPosition, m_dTraVoxelPosition );
        
        Point3D translate = new Point3D( m_dSagVoxelPosition, m_dCorVoxelPosition, m_dTraVoxelPosition );
        
        Point3D point1 = new Point3D(  - m_readOutVoxelSize / 2., 
                                         m_phaseVoxelSize / 2.,
                                       - m_thickVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point1 = point1.rotate3D( point1, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length( ) == 1.0 )
            {
                point1 = point1.rotate3D( point1, rotAxis2, rotAngle2 );
            }
            else
            {
                ;
            }
        }
        point1 = point1.translate( translate );
        
        
        Point3D point2 = new Point3D(   m_readOutVoxelSize / 2., 
                                        m_phaseVoxelSize / 2.,
                                      - m_thickVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point2 = point2.rotate3D( point2, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point2 = point2.rotate3D( point2, rotAxis2, rotAngle2 );
            }
        }
        point2 = point2.translate( translate );
        
        
        Point3D point3 = new Point3D(   m_readOutVoxelSize / 2., 
                                        m_phaseVoxelSize / 2.,
                                        m_thickVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point3 = point3.rotate3D( point3, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point3 = point3.rotate3D( point3, rotAxis2, rotAngle2 );
            }
        }
      
        point3 = point3.translate( translate );
        
        
        Point3D point4 = new Point3D(  - m_readOutVoxelSize / 2., 
                                         m_phaseVoxelSize / 2.,
                                         m_thickVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point4 = point4.rotate3D( point4, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point4 = point4.rotate3D( point4, rotAxis2, rotAngle2 );
            }
        }
        
        point4 = point4.translate( translate );
        
        Point3D point5 = new Point3D(  - m_readOutVoxelSize / 2., 
                                       - m_phaseVoxelSize / 2.,
                                       - m_thickVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point5 = point5.rotate3D( point5, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point5 = point5.rotate3D( point5, rotAxis2, rotAngle2 );
            }
        }
        
        point5 = point5.translate( translate );
        
        Point3D point6 = new Point3D(   m_readOutVoxelSize / 2., 
                                      - m_phaseVoxelSize / 2.,
                                      - m_thickVoxelSize / 2. );
        
        if ( rotAngle1 != 0.0 )
        {
            point6 = point6.rotate3D( point6, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point6 = point6.rotate3D( point6, rotAxis2, rotAngle2 );
            }
        }
        
        point6 = point6.translate( translate );
        
        Point3D point7 = new Point3D(   m_readOutVoxelSize / 2., 
                                      - m_phaseVoxelSize / 2.,
                                        m_thickVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point7 = point7.rotate3D( point7, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point7 = point7.rotate3D( point7, rotAxis2, rotAngle2 );
            }
        }
        
        point7 = point7.translate( translate );
        
        Point3D point8 = new Point3D(  - m_readOutVoxelSize / 2., 
                                       - m_phaseVoxelSize / 2.,
                                         m_thickVoxelSize / 2. );
        if ( rotAngle1 != 0.0 )
        {
            point8 = point8.rotate3D( point8, rotAxis1, rotAngle1 );
        }
        if ( rotAngle2 != 0.0 )
        {
            if ( rotAxis2.length() == 1.0 )
            {
                point8 = point8.rotate3D( point8, rotAxis2, rotAngle2 );
            }
        }
        
        point8 = point8.translate( translate );
        
        m_voxelGeometry = new Spectroscopic3DVoxel( point1, point2, point3, point4,
                                                    point5, point6, point7, point8 );
    }
 
    
    
    /**
     * Get single voxel geometry data.
     * @return 
     */
    public Spectroscopic3DVoxel getSpectroscopic3DVoxel( )
    {
        return m_voxelGeometry;
    }
    
    /**
     * Returns the LinkedHashMap containing all Siemens specific header 
     * information.
     * @return 
     */
    public LinkedHashMap getSiemensSpecificDicomHeaderInfo( )
    {
        return m_currentSiemensHeaderInfo;
    }

    /**
     * Returns the LinkedHashMap containing all Siemens specific header 
     * information.
     * @param instanceNumber
     * @return 
     */
    public LinkedHashMap getSiemensSpecificDicomHeaderInfo( int instanceNumber )
    {
        
        SingleVoxelSpectralData svsData = new SingleVoxelSpectralData( );
        String fileName = m_imageNumberToFilenameMap.get( new Integer( instanceNumber ) );
        
        DicomImageData d_spec_data = new DicomImageData( fileName );
    
        return d_spec_data.getSiemensSpecificDataFromFile( );
    }
    
    /**
     * Returns the string containing the voxel dimensions.
     * @return 
     */
    public String getVoxelDimensions( )
    {
        String voxelDimensions = "[ " +  m_currentSiemensHeaderInfo.get( "sSpecPara.sVoI.dThickness" ) + ", ";
        voxelDimensions += m_currentSiemensHeaderInfo.get( "sSpecPara.sVoI.dPhaseFOV" ) + ", ";
        voxelDimensions += m_currentSiemensHeaderInfo.get( "sSpecPara.sVoI.dReadoutFOV" ) + " ]";
        
        return voxelDimensions;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 0.
     * @return 
     */
    public String getReferenceFile0_UID( )
    {
        String refFile0 = ( (String)m_currentSiemensHeaderInfo.get( "tReferenceImage0" ) ).replace( '"', ' ' ).trim( );
        return refFile0;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 1.
     * @return 
     */
    public String getReferenceFile1_UID( )
    {
        String refFile1 = ( (String)m_currentSiemensHeaderInfo.get( "tReferenceImage1" ) ).replace( '"', ' ' ).trim( );
        return refFile1;
    }

    /**
     * Returns the MEDIA_STORAGE_SOP_INSTANCE_UID of reference file 2.
     * @return 
     */
    public String getReferenceFile2_UID( )
    {
        String refFile2 = ( (String)m_currentSiemensHeaderInfo.get( "tReferenceImage2" ) ).replace( '"', ' ' ).trim( );
        return refFile2;
    }
    
    /**
     * Sets the file reference of image reference Dicom file 0,
     * (ie a file system path).
     * @param fileRef0
     */
    public void setRefFile0_FileName( String fileRef0 )
    {
        m_referenceDicomFile0Path = fileRef0;
    }
    
    /**
     * Sets the file reference of image reference Dicom file 1,
     * (ie a file system path).
     * @param fileRef1
     */
    public void setRefFile1_FileName( String fileRef1 )
    {
        m_referenceDicomFile1Path = fileRef1;
    }

    /**
     * Sets the file reference of image reference Dicom file 2,
     * (ie a file system path).
     * @param fileRef2
     */
    public void setRefFile2_FileName( String fileRef2 )
    {
        m_referenceDicomFile2Path = fileRef2;
    }

    
    /**
     * Applies the QUALITY algorithm to the measured spectral data series.
     * @param applyFullQuality
     * @param reference
     */
    public void applyQuality( boolean applyFullQuality, 
                              SingleVoxelMeasuredSpectralData reference )
    {
        // Determine the phase of the water refererence signal
        float[] re = reference.getAbsorptionChannelTD( );
        float[] im = reference.getDispersionChannelTD( );
        
        float[] norm = new float[ m_numberOfSamplePoints * m_instances ];
        
        // importing spectrum halves the first time point
        float ref_scale = 2.0f * ( float )Math.sqrt( re[0] * re[0] + im[0] * im[0] );
        float[] waternorm = new float[re.length];
        float[] referencePhase = new float[re.length];
        float[] correctedSignalPhase = new float[ m_numberOfSamplePoints * m_instances ];
        for ( int k = 0; k < m_instances; k++ )
        {
            for ( int i= 0; i < re.length; i++ )
            {
                norm[k * m_numberOfSamplePoints + i] = (float)Math.sqrt( m_absorptionChannelTD[k * m_numberOfSamplePoints + i] * m_absorptionChannelTD[k * m_numberOfSamplePoints + i] + 
                                                                         m_dispersionChannelTD[k * m_numberOfSamplePoints + i] * m_dispersionChannelTD[k * m_numberOfSamplePoints + i] );
                waternorm[i] =(float)Math.sqrt( re[i] * re[i] + im[i] * im[i] ) / ref_scale;
            
                referencePhase[i] = (float)Math.atan2( im[i], re[i] );
                correctedSignalPhase[k * m_numberOfSamplePoints + i] = (float)Math.atan2( m_dispersionChannelTD[k * m_numberOfSamplePoints + i], m_absorptionChannelTD[k * m_numberOfSamplePoints +i] ) - referencePhase[i];
            }
        }
        
        TDFDFittingDefaults def = TDFDFittingDefaults.getInstance( );
        def.load( );
        
        if ( def.getProperty( TDFDFittingDefaults.QUALITY_TYPE ).compareTo( TDFDFittingDefaults.QUALITY_TYPE_AMPLITUDE_AND_PHASE ) == 0 )        
        {
            /*
             * This quality type corrects phase and amplitude
             */
            for ( int k = 0 ;  k < m_instances; k++ )
            {
                for ( int i = 0; i < re.length; i++ )
                {
                    m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ] = (float)( norm[k * m_numberOfSamplePoints + i] * Math.cos( correctedSignalPhase[k * m_numberOfSamplePoints + i] ) ) / waternorm[i];
                    m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ] = (float)( norm[k * m_numberOfSamplePoints + i] * Math.sin( correctedSignalPhase[k * m_numberOfSamplePoints + i] ) ) / waternorm[i];
                }
            }
        }
        else
        {
            /*
             * This quality type only corrects the phase
             */
            for ( int k = 0 ;  k < m_instances; k++ )
            {
                for ( int i = 0; i < re.length; i++ )
                {
                    m_absorptionChannelTD[k * m_numberOfSamplePoints + i] = (float)( norm[k * m_numberOfSamplePoints + i] * Math.cos( correctedSignalPhase[k * m_numberOfSamplePoints + i] ) );
                    m_dispersionChannelTD[k * m_numberOfSamplePoints + i] = (float)( norm[k * m_numberOfSamplePoints + i] * Math.sin( correctedSignalPhase[k * m_numberOfSamplePoints + i] ) );
                }
            }
        }
        
        // If response is loaded and set halving the first data point is done automatically
        m_absorptionChannelTD[0] = m_absorptionChannelTD[0] * 2.0f;
        m_dispersionChannelTD[0] = m_dispersionChannelTD[0] * 2.0f;
        
        System.out.println( "SingleVoxelMeasuredSpectralData::ApplyQuality - First point is doubled to get it in sync!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" );
        setSpectralDataPointsTDOfAllChannels( );

    }
    
    
    /** 
     * Set spectral data points and computes the FD representation;
     * this function assumes that the time domain arrays have been set
     * properly.
     */
    public void setSpectralDataPointsTDOfAllChannels( )
    {
        for ( int k = 0; k < m_instances; k++ )
        {
            calculateFDRepresentation( k );
        }
    }
    
    /** 
     * Sets the spectral data points and computes the TD representation;
     * this function assumes that the frequency domain arrays have been set
     * properly.
     */
    public void setSpectralDataPointsFDOfAllChannels( )
    {
        for ( int k = 0; k < m_instances; k++ )
        {
            calculateTDRepresentation( k );
        }
    }
    
    /**
     * Returns the number of instances.
     * @return 
     */
    public int getNumberOfInstances( )
    {
        return m_instances;
    }
    
    /**
     * Returns the absorption channel of instance "instanceNumber".
     * @param instanceNumber
     * @return 
     */
    public float[] getAbsorptionChannelTD( int instanceNumber )
    {
        float[] abs = new float[m_numberOfSamplePoints];
        
        for (int i=0; i < m_numberOfSamplePoints; i++ )
        {
            abs[i] = m_absorptionChannelTD[instanceNumber * m_numberOfSamplePoints + i];
        }
        
        return abs;
    }
    
    /**
     * Returns the absorption channel of instance "instanceNumber".
     * @param instanceNumber
     * @return 
     */
    public float[] getDispersionChannelTD( int instanceNumber )
    {
        float[] disp = new float[m_numberOfSamplePoints];
        
        for ( int i=0; i < m_numberOfSamplePoints; i++ )
        {
            disp[i] = m_dispersionChannelTD[instanceNumber * m_numberOfSamplePoints + i];
        }
        
        return disp;
    }
    
    /**
     * Returns the absorption channel of instance "instanceNumber".
     * @param instanceNumber
     * @return 
     */
    public float[] getAbsorptionChannelFD( int instanceNumber )
    {
        float[] abs = new float[m_numberOfSamplePoints];
        
        for (int i=0; i < m_numberOfSamplePoints; i++ )
        {
            abs[i] = m_absorptionChannelFD[instanceNumber * m_numberOfSamplePoints + i];
        }
        
        return abs;
    }
    
    /**
     * Returns the absorption channel of instance "instanceNumber".
     * @param instanceNumber
     * @return 
     */
    public float[] getDispersionChannelFD( int instanceNumber )
    {
        float[] disp = new float[m_numberOfSamplePoints];
        
        for ( int i=0; i < m_numberOfSamplePoints; i++ )
        {
            disp[i] = m_dispersionChannelFD[instanceNumber * m_numberOfSamplePoints + i];
        }
        
        return disp;
    }


    
    /**
     * Returns the sum of all responses as a SingleVoxelMeasuredSpectralData-
     * object.
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getSumSpectrum( )
    {
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );
        
        float[] abs = new float[m_numberOfSamplePoints]; 
        float[] disp = new float[m_numberOfSamplePoints]; 
        
        for ( int k = 0; k < m_instances; k++ )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {    
                abs[i] += m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ];
                disp[i] += m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ];
            }
        }
        
        svsData.setSpectralDataPointsTD( abs, disp );
        
        svsData.setDicomHeaderInfo( getDicomHeaderInfo( 0 ) );
        svsData.setSiemensSpecificDicomHeaderInfo( getSiemensSpecificDicomHeaderInfo( 0 ) );
        
        return svsData;
    }


    /**
     * Returns the sum of all responses as a SingleVoxelSpectralData-object.
     * @param finfo
     * @param hmap
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getSumSpectrum( FileInfo finfo, 
                                                           LinkedHashMap hmap )
    {
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );

        float[] abs = new float[m_numberOfSamplePoints];
        float[] disp = new float[m_numberOfSamplePoints];

        for ( int k = 0; k < m_instances; k++ )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {
                abs[i] += m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ];
                disp[i] += m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ];
            }
        }

        svsData.setSpectralDataPointsTD( abs, disp );

        svsData.setDicomHeaderInfo( finfo );
        svsData.setSiemensSpecificDicomHeaderInfo( hmap );

        return svsData;
    }


    
    /**
     * Returns the sum of all responses as a SingleVoxelSpectralData-object.
     * @param minimum
     * @param maximum
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getSumSpectrum( int minimum, int maximum )
    {
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );
           
        float[] abs = new float[m_numberOfSamplePoints]; 
        float[] disp = new float[m_numberOfSamplePoints]; 
        
        for ( int k = 0; k < m_instances; k++ )
        {
            if ( ( k >= minimum ) && ( k <= maximum ) )
            {
                for ( int i = 0; i < m_numberOfSamplePoints; i++ )
                {    
                    abs[i] += m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ];
                    disp[i] += m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ];
                }
            }
        }
        
        svsData.setSpectralDataPointsTD( abs, disp );
        
        svsData.setDicomHeaderInfo( getDicomHeaderInfo( minimum ) );
        svsData.setSiemensSpecificDicomHeaderInfo( getSiemensSpecificDicomHeaderInfo( minimum ) );
        
        return svsData;
    }
    
    
    /**
     * Returns the average of all responses as a SingleVoxelSpectralData-object.
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getAverageSpectrum( )
    {
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );
        
        float[] abs = new float[m_numberOfSamplePoints]; 
        float[] disp = new float[m_numberOfSamplePoints]; 
        
        for ( int k = 0; k < m_instances; k++ )
        {
            for ( int i = 0; i < m_numberOfSamplePoints; i++ )
            {    
                abs[i] += m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ];
                disp[i] += m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ];
            }
        }

        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {    
            abs[i] = abs[i] / m_instances;
            disp[i]= disp[i] / m_instances;
        }
        
        svsData.setSpectralDataPointsTD( abs, disp );
        
        svsData.setDicomHeaderInfo( getDicomHeaderInfo( 1 ) );
        svsData.setSiemensSpecificDicomHeaderInfo( getSiemensSpecificDicomHeaderInfo( 1 ) );
        
        return svsData;
    }
    
    
    
    /**
     * Computes the moving median output for a certain output channel 'cnt'
     * and filter width 'filterSize'.
     * @param cnt
     * @param filterSize
     * @return 
     */
    public SingleVoxelMeasuredSpectralData applyMovingMedianFilter( int cnt, int filterSize )
    {
        SingleVoxelMeasuredSpectralData retObj= new SingleVoxelMeasuredSpectralData( );
        
        return retObj;
    }
    
    /**
     * Computes the moving median output for a certain output channel 'cnt'
     * and filter width 'filterSize'
     */
    public SingleVoxelMeasuredSpectralData applyMovingAverageFilter( int cnt, int filterSize )
    {
        
        SingleVoxelMeasuredSpectralData retObj = new SingleVoxelMeasuredSpectralData( );

        // filter not completely working
        if ( cnt <= filterSize ) 
        {
            retObj = getAverageOfInstanceRangeSpectrum( 0 , filterSize );
        }
        
        // filter in working area
        if ( ( cnt > filterSize ) && ( cnt < m_instances - filterSize - 1 ) )
        {
            retObj = getAverageOfInstanceRangeSpectrum( cnt - filterSize , cnt + filterSize );
        }
        
        if ( cnt >= m_instances - filterSize - 1 )
        {
            retObj = getAverageOfInstanceRangeSpectrum( m_instances - filterSize  , m_instances - 1 );
        }
        
        return retObj;
    }
    
    /**
     * Computes the moving median output for a certain output channel 'cnt'
     * and filter width 'filterSize'.
     * @param cnt
     * @param filterSize
     * @return 
     */
    public SingleVoxelMeasuredSpectralData applySimpleAverageFilter( 
                                               int cnt, int filterSize )
    {
        
        SingleVoxelMeasuredSpectralData retObj = new SingleVoxelMeasuredSpectralData( );

        // filter not completely working
        if ( ( cnt > -1 ) && ( cnt + filterSize < this.getNumberOfInstances( ) ) )
        {
            retObj = getAverageOfInstanceRangeSpectrum( cnt , cnt + filterSize );
        }
        else
        {
            retObj = getAverageOfInstanceRangeSpectrum( cnt , getNumberOfInstances( ) );
        }
        
        return retObj;
    }
    
    
    
    /**
     * Computes a filtered response. 
     * @param cnt
     * @param filterType
     * @param filterSize
     * @return 
     */
    public SingleVoxelMeasuredSpectralData computeFilteredResponse( 
                                       int cnt, int filterType, int filterSize )
    {
        SingleVoxelMeasuredSpectralData retData = null;
        
        if ( filterType == 0 )
        {
            retData =  applyMovingMedianFilter( cnt, filterSize );
        }
        else if ( filterType == 1 )    
        {
            retData = applyMovingAverageFilter( cnt, filterSize );
        }
        else if ( filterType == 2 )
        {
            retData = applySimpleAverageFilter( cnt, filterSize );
        }
        
        return retData;
    }
    
    
    /**
     * Returns the average of all responses as a SingleVoxelSpectralData -object of
     * the responses from minimum - maximum.
     * @param minimum
     * @param maximum
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getAverageOfInstanceRangeSpectrum( 
                                                     int minimum , int maximum )
    {
        SingleVoxelMeasuredSpectralData svsData = new SingleVoxelMeasuredSpectralData( );
        
        float[] abs = new float[m_numberOfSamplePoints]; 
        float[] disp = new float[m_numberOfSamplePoints]; 
        
        for ( int k = 0; k < m_instances; k++ )
        {
            if ( ( k >= minimum ) && ( k <= maximum ) )
            {
                for ( int i = 0; i < m_numberOfSamplePoints; i++ )
                {    
                    abs[i] += m_absorptionChannelTD[ k * m_numberOfSamplePoints + i ];
                    disp[i] += m_dispersionChannelTD[ k * m_numberOfSamplePoints + i ];
                }
            }
        }

        for ( int i = 0; i < m_numberOfSamplePoints; i++ )
        {    
            abs[i] = abs[i] / m_instances;
            disp[i]= disp[i] / m_instances;
        }
        
        svsData.setSpectralDataPointsTD( abs, disp );
        
        svsData.setDicomHeaderInfo( getDicomHeaderInfo( 1 ) );
        svsData.setSiemensSpecificDicomHeaderInfo( getSiemensSpecificDicomHeaderInfo( 1 ) );
        
        return svsData;
    }
    
    
    /**
     * Returns the LinkedList containing the file references.
     * @return 
     */ 
    public LinkedList<String> getFileReferences( )
    {
        return m_fileReferences;
    }
    
    // Attributes
    private float m_sampleFrequency = 1000.0f;
    
    private int m_numberOfSamplePoints = 0;
    private int m_instances = 0;
    private int m_currentInstance = 0;
    
    // Both channel ( alterating abs, disp ) dispersion and absorption channel
    // in time domain (TD)
    protected float[] m_bothChannelsTD;
    protected float[] m_absorptionChannelTD;
    protected float[] m_dispersionChannelTD;
    protected float[] m_xaxisTD;
    
    // Both channel ( alterating abs, disp ) dispersion and absorption channel
    // in frequency domain (FD)
    protected float[] m_bothChannelsFD;
    protected float[] m_absorptionChannelFD;
    protected float[] m_dispersionChannelFD;
    protected float[] m_xaxisFD;
    
    protected FileInfo m_currentDicomHeaderInfo;
    
    private Color m_curveColor = Color.black;
    
    // Contains Siemens specific header information
    private LinkedHashMap m_currentSiemensHeaderInfo;
    
    // File pointers to files on which the single voxel spectroscopy was
    // planned
    private String m_referenceDicomFile0Path;
    private String m_referenceDicomFile1Path;
    private String m_referenceDicomFile2Path;
    
    private Spectroscopic3DVoxel m_voxelGeometry;
    
    private double m_thickVoxelSize = 0.00;
    private double m_readOutVoxelSize = 0.00;
    private double m_phaseVoxelSize = 0.00;
    
    private double m_dSagVoxelPosition = 0.00;
    private double m_dCorVoxelPosition = 0.00;
    private double m_dTraVoxelPosition = 0.00;
    
    private double m_dSagVoxelOrientation = 0.00;
    private double m_dCorVoxelOrientation = 0.00;
    private double m_dTraVoxelOrientation = 0.00;
    
    private double m_dInPlaneRotation;
    
    // Member containing all fileReferences to files of the time series
    private LinkedList<String> m_fileReferences;
    
    private TreeMap<Integer, String> m_imageNumberToFilenameMap = new TreeMap<Integer, String>( );
    
    // Value necessary to compute exact ppm values
    private long m_resonanceFrequency = 0;
    
    // Value from Siemens specific header to compute sample frequency
    private long m_sampleDwellTime = 0;
    private float m_acqFrequency = 0;
    
}
