/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.spectrim.tdfdfit.quantitativeresults;

import org.spectrim.lib.TextFileWriter;
import org.spectrim.xyplot.Curve;
import java.awt.Color;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.spectrim.tdfdfit.SingleVoxelMeasuredSpectralData;
import org.spectrim.tdfdfit.TDFDFittingDefaults;

/**
 * A class that handles normal value data bases: 
 * It can read itself from a stream and write itself to a stream.
 * @author slotboom
 */
public class NormalMetaboliteValuesAndSpectrumDatabase
{

    /**
     * Default constructor of a NormalMetaboliteValuesAndSpectrumDatabase 
     * object.
     */
    public NormalMetaboliteValuesAndSpectrumDatabase( )
    {
        System.out.println( "Default Constructor NormalMetaboliteValuesAndSpectrumDatabase is called!" );
    }


    /**
     * Alternative constructor of a NormalMetaboliteValuesAndSpectrumDatabase 
     * object.
     * @param dBaseName    - name of the database (and filename)
     */
    public NormalMetaboliteValuesAndSpectrumDatabase( String dBaseName )
    {
        m_dataBaseName = dBaseName;
        
        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        String dBasePath = defs.getProperty( TDFDFittingDefaults.TDFDFIT_NORMAL_VALUES_CONCENTRATIONS_PATH );
        String dataBaseFileRef = dBasePath + System.getProperty( "file.separator" ) + m_dataBaseName + ".ndb";
        
        try
        {
            m_dbFile = new RandomAccessFile( dataBaseFileRef, "rw" );
            
            // empty database having only a name.
            if ( m_dbFile.length( ) > 0 )
            {
                readDatabase( );
            }
        }
        catch ( FileNotFoundException ex )
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        } catch ( IOException ex )
        {
            Logger.getLogger(NormalMetaboliteValuesAndSpectrumDatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Returns the number of instances in the database.
     * @return
     */
    public int getNumberOfInstancesInDatabase( )
    {
        return m_normalResponses.size( );
    }

    /**
     * Drops the old database (also removes it from the file system).
     * @throws java.lang.IllegalArgumentException
     */
    public void dropDatabase( ) throws IllegalArgumentException
    {
        // System.out.println( "Dropdatabase( ) in NormalMetaboliteValuesAndSpectrumDatabase is called!" );
        try
        {
            m_dbFile.close( );
            m_dbFile = null;
        }
        catch ( IOException ex )
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        }

        // Make sure the file or directory exists and isn't write protected
        if ( m_dataBaseFile != null )
        {
            if ( !m_dataBaseFile.exists( ) )
            {
                throw new IllegalArgumentException( "Delete: no such file or directory: " + m_dataBaseName + ".ndb" );
            }

            if ( !m_dataBaseFile.canWrite( ) )
            {
                throw new IllegalArgumentException( "Delete: write protected: " + m_dataBaseName + ".ndb" );
            }

            // If it is a directory, make sure it is empty
            if ( m_dataBaseFile.isDirectory( ) )
            {
                String[] files = m_dataBaseFile.list( );
                if ( files.length > 0 )
                {
                    throw new IllegalArgumentException( "Delete: directory not empty: " + m_dataBaseName + ".ndb");
                }
            }

            // Attempt to delete it
            boolean success = m_dataBaseFile.delete( );

            if ( !success )
            {
                throw new IllegalArgumentException( "Delete: deletion failed" );
            }
        }
    }

    /**
     * Sets the name of the database.
     * @param dataBaseName
     */
    public void setDataBaseName( String dataBaseName )
    {
        m_dataBaseName = dataBaseName;
    }

    /**
     *
     * Adds a new instance to the normal value database.
     *
     * @param response
     * @param pars
     */
    public void addDatabaseObject( NormalDatabaseSpectrumResponse response,
                                   NormalDatabaseMetaboliteParameters pars )
    {

        m_normalResponses.add( response );
        m_normalMetabolitePars.add( pars );
        m_numberOfInstances = m_normalMetabolitePars.size( );

        writeDatabase( );
    }

    /**
     * Returns the response of database element number dBElementNumber
     * @param dBElementNumber
     * @return
     */
    public NormalDatabaseSpectrumResponse getResponse( int dBElementNumber )
    {
        if ( ( dBElementNumber > -1 ) && (dBElementNumber < m_normalResponses.size( ) ) )
        {
            return m_normalResponses.get( dBElementNumber );
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns the metabolite parameters of a normal value spectroscopy 
     * database belonging to database number. 
     * @param dBElementNumber
     * @return 
     */
    public  NormalDatabaseMetaboliteParameters getNormalDatabaseMetaboliteParameters( int dBElementNumber )
    {
        if ( ( dBElementNumber > -1 ) && (dBElementNumber < m_normalMetabolitePars.size( ) ) )
        {
            return m_normalMetabolitePars.get( dBElementNumber );
        }
        else
        {
            return null;
        }
    }

    /**
     * Initializes or opens a new database with name dataBaseName
     * @param dataBaseName
     * @param fieldStrength
     * @param resonanceFrequency
     * @param quantificationMethod
     * @param scannerType
     * @param sequenceType
     * @param tissueType
     * @param echoTime
     * @param repetitionTime
     * @param manufacturer
     */
    public NormalMetaboliteValuesAndSpectrumDatabase( String dataBaseName,
                                                      String fieldStrength,
                                                      String resonanceFrequency,
                                                      String quantificationMethod,
                                                      String scannerType,
                                                      String sequenceType,
                                                      String tissueType,
                                                      String echoTime,
                                                      String repetitionTime,
                                                      String manufacturer )
    {

        //System.out.println( "Constructor NormalMetaboliteValuesAndSpectrumDatabase is called!" );

        m_dataBaseName = dataBaseName;

        m_fieldStrength         = fieldStrength;
        m_quantificationMethod  = quantificationMethod;
        m_resonanceFrequency    = resonanceFrequency;
        m_scannerType           = scannerType;
        m_sequenceType          = sequenceType;
        m_tissueType            = tissueType;
        m_echoTime              = echoTime;
        m_repetitionTime        = repetitionTime;
        m_manufacturer          = manufacturer;

        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        String dBasePath = defs.getProperty( TDFDFittingDefaults.TDFDFIT_NORMAL_VALUES_CONCENTRATIONS_PATH );


        String dataBaseFileRef = m_dataBaseName + ".ndb";

        File file = new File( dBasePath + System.getProperty( "file.separator" ) + dataBaseFileRef );

        if ( file.exists( ) == true )
        {
            // open it!
            m_dataBaseAlreadyExists = true;
            try
            {
                m_dbFile = new RandomAccessFile( dBasePath + System.getProperty( "file.separator" ) + dataBaseFileRef, "rw" );
                readDatabase( );
            }
            catch ( FileNotFoundException ex )
            {
                Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
            }
        }
        else
        {
            // System.out.println( "Database does not exist yet!\n Create it now!" );

            m_numberOfInstances = 0;
            try
            {
                m_dbFile = new RandomAccessFile( dBasePath + System.getProperty( "file.separator" ) + dataBaseFileRef, "rw" );
            }
            catch ( FileNotFoundException ex )
            {
                Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
            }

            write( m_dbFile );

        }
    }


    /**
     * Returns whether the database already exists
     * @return
     */
    public boolean dataBaseAlreadyExists( )
    {
        return m_dataBaseAlreadyExists;
    }

    /**
     * Returns the name of the database.
     * @return
     */
    public String getDatabaseName( )
    {
        return m_dataBaseName;
    }

    /**
     * Reads the database from the file system.
     */
    public void readDatabase( )
    {
        try 
        {
            if ( m_dbFile.length( ) > 0 )
            {
                read( m_dbFile );
            }
        } catch (IOException ex) 
        {
            Logger.getLogger(NormalMetaboliteValuesAndSpectrumDatabase.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    /**
     * Writes the database to the file system.
     */
    public void writeDatabase( )
    {

        // Test whether the database already exists
        if ( m_dbFile != null )
        {
            // if yes: drop it!
            // delete the old database file
            dropDatabase( );
        }

        // now create a new RandomAccessFile

        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        String dBasePath = defs.getProperty( TDFDFittingDefaults.TDFDFIT_NORMAL_VALUES_CONCENTRATIONS_PATH );


        String dataBaseFileRef = m_dataBaseName + ".ndb";

        //System.out.println( ">>>>>>> Opening for writing : " + curDir + System.getProperty( "file.separator" ) + dataBaseFileRef );

        File file = new File( dBasePath + System.getProperty( "file.separator" ) + dataBaseFileRef );
        try
        {
            m_dbFile = new RandomAccessFile( dBasePath + System.getProperty( "file.separator" ) + dataBaseFileRef, "rw" );
            // write to db to the new random access file!
            write( m_dbFile );
            try
            {
                m_dbFile.close( );
            }
            catch ( IOException ex )
            {
                Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
            }
        } 
        catch ( FileNotFoundException ex )
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
    }


    /**
     * Searches the working directory for available databases ..
     * @return
     */
    public Object[][] getAvailableNormalMetaboliteValuesdBaseInformation( )
    {

        // Let the application search the default directory set by the user
        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        String curDir = defs.getProperty( TDFDFittingDefaults.TDFDFIT_NORMAL_VALUES_CONCENTRATIONS_PATH );
        String[] files = new File( curDir ).list( );
        
        int cntr = 0;
        for (String file : files) 
        {
            if ( file.endsWith( ".ndb" ) )
            {
                cntr++;
            }
        }

        Object[][] dBaseInfo = new Object[cntr][m_numberOfColumns];
        cntr=0;
        for ( String file : files ) 
        {
            System.out.println( "Database file : " + file );

            if ( file.endsWith( ".ndb" ) )
            {

                //NormalMetaboliteValuesAndSpectrumDatabase db = new NormalMetaboliteValuesAndSpectrumDatabase( file );

                m_dataBaseName = curDir + System.getProperty( "file.separator" ) + file;
                //System.out.println( "Parsing existing databases .. " );
                //System.out.println( "Parsing existing normal value database : " + m_dataBaseName );
                try
                {
                    m_dbFile = new RandomAccessFile( m_dataBaseName, "rw" );
                }
                catch ( FileNotFoundException ex )
                {
                    Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log(Level.SEVERE, null, ex );
                }

                readDatabase( );
                
                // String name 
                dBaseInfo[cntr][0]    = file.substring( 0, file.lastIndexOf( "." ) );
                
                // String fieldStrength 
                dBaseInfo[cntr][1]    = m_quantificationMethod ;

                // String quantificationMethod
                dBaseInfo[cntr][2]    = m_manufacturer;

                // String scannerType    = m_scannerType;
                dBaseInfo[cntr][3]    = m_fieldStrength ;

                // String sequenceType   = m_sequenceType;
                dBaseInfo[cntr][4]    = m_scannerType;

                // String tissueType     = m_tissueType;
                dBaseInfo[cntr][5]    = m_sequenceType ;

                // String echoTime       = m_echoTime;
                dBaseInfo[cntr][6]    = m_tissueType ;

                // String repetitionTime = m_repetitionTime;
                dBaseInfo[cntr][7]    = m_echoTime ;

                // String manufacturer   = m_manufacturer;
                dBaseInfo[cntr][8]    = m_repetitionTime;
                
                try 
                {
                    m_dbFile.close( );
                }
                catch ( IOException ex )
                {
                    Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
                }

                cntr++;
            }
        }

        return dBaseInfo;
    }

    /**
     * Returns the number of columns.
     * @return
     */
    public int getNumberOfColumns( )
    {
        return m_numberOfColumns;
    }

    /**
     * Returns the minimum normal absorption frequency domain signal for 
     * displaying for compared to groups.
     * @return
     */
    public float[] getMinimaOfNormalRangeSpectrumAbsorption( )
    {

        float[] minimalSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < minimalSignal.length; i++ )
        {
            minimalSignal[i] = Float.MAX_VALUE;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredReal( );

            for ( int i = 0; i < resp.length; i++ )
            {
                if ( resp[i] < minimalSignal[i] )
                {
                    minimalSignal[i] = (float)resp[i];
                }
            }
        }

        return minimalSignal;
    }


    /**
     * Returns the maximum absorption frequency domain signal of the 
     * compared to groups.
     * @return
     */
    public float[] getMaximaOfNormalRangeSpectrumAbsorption( )
    {

        float[] maximalSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < maximalSignal.length; i++ )
        {
            maximalSignal[i] = -Float.MIN_VALUE;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredReal( );

            for ( int i = 0; i < resp.length; i++ )
            {
                if ( resp[i] > maximalSignal[i] )
                {
                    maximalSignal[i] = (float)resp[i];
                }
            }
        }

        return maximalSignal;
    }

    /**
     * Returns the mean absorption frequency domain signal for displaying 
     * the results compared to groups.
     * @return
     */
    public float[] getMeanOfNormalRangeSpectrumAbsorption( )
    {

        float[] meanSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < meanSignal.length; i++ )
        {
            meanSignal[i] = 0f;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredReal( );

            for ( int i = 0; i < resp.length; i++ )
            {
                meanSignal[i] += (float)resp[i];
            }
        }

        for ( int i = 0; i < meanSignal.length; i++ )
        {
            meanSignal[i] = meanSignal[i] / m_normalResponses.size( );
        }

        return meanSignal;
    }



    /**
     * Returns the minimum dispersion frequency domain signal for displaying 
     * of results compared to groups.
     * @return
     */
    public float[] getMinimaOfNormalRangeSpectrumDispersion( )
    {

        float[] minimalSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < minimalSignal.length; i++ )
        {
            minimalSignal[i] = Float.MAX_VALUE;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredImag( );

            for ( int i = 0; i < resp.length; i++ )
            {
                if ( resp[i] < minimalSignal[i] )
                {
                    minimalSignal[i] = (float)resp[i];
                }
            }
        }

        return minimalSignal;
    }


    /**
     * Returns the maximum dispersion frequency domain signal for displaying of 
     * results comparing the groups.
     * @return
     */
    public float[] getMaximaOfNormalRangeSpectrumDispersion( )
    {

        float[] maximalSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < maximalSignal.length; i++ )
        {
            maximalSignal[i] = -Float.MIN_VALUE;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredImag( );

            for ( int i = 0; i < resp.length; i++ )
            {
                if ( resp[i] > maximalSignal[i] )
                {
                    maximalSignal[i] = (float)resp[i];
                }
            }
        }

        return maximalSignal;
    }


    /**
     * Returns the mean dispersion frequency domain signal for displaying 
     * of results comparing the groups.
     * @return
     */
    public float[] getMeanOfNormalRangeSpectrumDispersion( )
    {

        float[] meanSignal = new float[ m_normalResponses.get( 0 ).getMeasuredReal( ).length ];

        for ( int i = 0; i < meanSignal.length; i++ )
        {
            meanSignal[i] = 0f;
        }

        Iterator it = m_normalResponses.iterator( );

        while ( it.hasNext( ) )
        {
            NormalDatabaseSpectrumResponse normResp = (NormalDatabaseSpectrumResponse )it.next( );
            double[] resp = normResp.getMeasuredImag( );

            for ( int i = 0; i < resp.length; i++ )
            {
                meanSignal[i] += (float)resp[i];
            }

        }

        for ( int i = 0; i < meanSignal.length; i++ )
        {
            meanSignal[i] = meanSignal[i] / m_normalResponses.size( );
        }

        return meanSignal;
    }



    /**
     * Set all average and standard deviation attributes.
     */
    public void setAllAverageAndStDevsAttributes( )
    {
        m_averagePars = new NormalDatabaseMetaboliteParameters( );
        m_stDevsOfPars = new NormalDatabaseMetaboliteParameters( );

        double[] rawAveragePeakAreas = null;
        double[] rawAverageFractionCorrectedPeakAreas = null;
        double[] averageJMRUITypeConcentrations = null;
        double[] averageTrueTwoCompartmentModelConcentrations = null;
        double[] averageDFreqs = null;
        double[] averageLorentzWidths = null;
        double[] averageLorentzRWidths = null;
        double[] averagePhases = null;

        double[] rawStDevPeakAreas = null;
        double[] rawStDevFractionCorrectedPeakAreas = null;
        double[] stDevJMRUITypeConcentrations = null;
        double[] stDevTrueTwoCompartmentModelConcentrations = null;
        double[] stDevDFreqs = null;
        double[] stDevLorentzWidths = null;
        double[] stDevLorentzRWidths = null;
        double[] stDevPhases = null;


        if ( m_normalMetabolitePars.size( ) > 0 )
        {
            m_metaboliteNames = m_normalMetabolitePars.get( 0 ).getMetaboliteNames( );

            rawAveragePeakAreas = new double[m_metaboliteNames.size( )];
            rawAverageFractionCorrectedPeakAreas = new double[m_metaboliteNames.size( )];
            averageJMRUITypeConcentrations = new double[m_metaboliteNames.size()];
            averageTrueTwoCompartmentModelConcentrations = new double[m_metaboliteNames.size()];
            averageDFreqs = new double[m_metaboliteNames.size()];
            averageLorentzWidths = new double[m_metaboliteNames.size()];
            averageLorentzRWidths = new double[m_metaboliteNames.size()];
            averagePhases = new double[m_metaboliteNames.size()];

            rawStDevPeakAreas = new double[m_metaboliteNames.size()];
            rawStDevFractionCorrectedPeakAreas = new double[m_metaboliteNames.size()];
            stDevJMRUITypeConcentrations = new double[m_metaboliteNames.size()];
            stDevTrueTwoCompartmentModelConcentrations = new double[m_metaboliteNames.size()];
            stDevDFreqs = new double[m_metaboliteNames.size()];
            stDevLorentzWidths = new double[m_metaboliteNames.size()];
            stDevLorentzRWidths = new double[m_metaboliteNames.size()];
            stDevPhases = new double[m_metaboliteNames.size()];

            for ( int i = 0; i < m_metaboliteNames.size(); i++ )
            {

                Iterator it = m_normalMetabolitePars.iterator( );
                while ( it.hasNext( ) )
                {

                    NormalDatabaseMetaboliteParameters pars = (NormalDatabaseMetaboliteParameters)it.next( );

                    rawAveragePeakAreas[i] += pars.getRawPeakAreas( )[i];
                    rawAverageFractionCorrectedPeakAreas[i] += pars.getRawFractionCorrectedPeakAreas( )[i];
                    averageJMRUITypeConcentrations[i] += pars.getJMRUITypeConcentrations( )[i];
                    averageTrueTwoCompartmentModelConcentrations[i] += pars.getTrueTwoCompartmentModelConcentrations( )[i];

                    averageDFreqs[i] += pars.getDFreqs( )[i];
                    averageLorentzWidths[i] += pars.getLorentzWidths( )[i];
                    averageLorentzRWidths[i] += pars.getLorentzRWidths( )[i];
                    averagePhases[i] += pars.getPhases( )[i];

                }

                rawAveragePeakAreas[i] = rawAveragePeakAreas[i] / m_normalMetabolitePars.size( );
                rawAverageFractionCorrectedPeakAreas[i] = rawAverageFractionCorrectedPeakAreas[i] / m_normalMetabolitePars.size( );
                averageJMRUITypeConcentrations[i] = averageJMRUITypeConcentrations[i] / m_normalMetabolitePars.size( );
                averageTrueTwoCompartmentModelConcentrations[i] =  averageTrueTwoCompartmentModelConcentrations[i] / m_normalMetabolitePars.size( );

                averageDFreqs[i] = averageDFreqs[i] / m_normalMetabolitePars.size( );
                averageLorentzWidths[i] = averageLorentzWidths[i] / m_normalMetabolitePars.size( );
                averageLorentzRWidths[i] = averageLorentzRWidths[i] / m_normalMetabolitePars.size( );
                averagePhases[i] = averagePhases[i] / m_normalMetabolitePars.size( );


                Iterator it2 = m_normalMetabolitePars.iterator( );
                while ( it2.hasNext( ) )
                {

                    NormalDatabaseMetaboliteParameters pars = (NormalDatabaseMetaboliteParameters)it2.next( );

                    rawStDevPeakAreas[i] +=  ( rawAveragePeakAreas[i] - pars.getRawPeakAreas( )[i] ) * ( rawAveragePeakAreas[i] - pars.getRawPeakAreas( )[i] );
                    rawStDevFractionCorrectedPeakAreas[i] += ( rawAverageFractionCorrectedPeakAreas[i] - pars.getRawFractionCorrectedPeakAreas( )[i] ) * ( rawAverageFractionCorrectedPeakAreas[i] - pars.getRawFractionCorrectedPeakAreas( )[i] );
                    stDevJMRUITypeConcentrations[i] += ( averageJMRUITypeConcentrations[i] - pars.getJMRUITypeConcentrations( )[i] ) * ( averageJMRUITypeConcentrations[i] - pars.getJMRUITypeConcentrations( )[i] );
                    stDevTrueTwoCompartmentModelConcentrations[i] += ( averageTrueTwoCompartmentModelConcentrations[i] - pars.getTrueTwoCompartmentModelConcentrations( )[i] ) * ( averageTrueTwoCompartmentModelConcentrations[i] - pars.getTrueTwoCompartmentModelConcentrations( )[i] );

                    stDevDFreqs[i] += ( averageDFreqs[i] - pars.getDFreqs( )[i] ) * ( averageDFreqs[i] - pars.getDFreqs( )[i] );
                    stDevLorentzWidths[i] += ( averageLorentzWidths[i] - pars.getLorentzWidths( )[i] ) * ( averageLorentzWidths[i] - pars.getLorentzWidths( )[i] );
                    stDevLorentzRWidths[i] += ( averageLorentzRWidths[i] - pars.getLorentzRWidths( )[i] ) * ( averageLorentzRWidths[i] - pars.getLorentzRWidths( )[i] );
                    stDevPhases[i] += ( averagePhases[i] - pars.getPhases( )[i] ) * ( averagePhases[i] - pars.getPhases( )[i] );

                }

                if ( ( m_normalMetabolitePars.size( ) - 1 ) > 0 )
                {
                    rawStDevPeakAreas[i] = Math.sqrt( rawStDevPeakAreas[i] )/ ( m_normalMetabolitePars.size( ) - 1 );
                    rawStDevFractionCorrectedPeakAreas[i] = Math.sqrt( rawStDevFractionCorrectedPeakAreas[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevJMRUITypeConcentrations[i] = Math.sqrt( stDevJMRUITypeConcentrations[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevTrueTwoCompartmentModelConcentrations[i] = Math.sqrt( stDevTrueTwoCompartmentModelConcentrations[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevDFreqs[i] = Math.sqrt( stDevDFreqs[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevLorentzWidths[i] = Math.sqrt( stDevLorentzWidths[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevLorentzRWidths[i] = Math.sqrt( stDevLorentzRWidths[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                    stDevPhases[i] = Math.sqrt( stDevPhases[i] ) / ( m_normalMetabolitePars.size( ) - 1 );
                }
                else
                {
                    rawStDevPeakAreas[i] = -1.0;
                    rawStDevFractionCorrectedPeakAreas[i] = -1.0;
                    stDevJMRUITypeConcentrations[i] = -1.0;
                    stDevTrueTwoCompartmentModelConcentrations[i] = -1.0;
                    stDevDFreqs[i] = -1.0;
                    stDevLorentzWidths[i] = -1.0;
                    stDevLorentzRWidths[i] = -1.0;
                    stDevPhases[i] = -1.0;
                }
            }
        }

        // update the object with average values!
        if ( m_metaboliteNames != null )
        {
            m_averagePars.setMetaboliteNames( m_metaboliteNames );
            m_averagePars.setRawPeakAreas( rawAveragePeakAreas );
            m_averagePars.setRawFractionCorrectedPeakAreas( rawAverageFractionCorrectedPeakAreas );
            m_averagePars.setJMRUITypeConcentrations( averageJMRUITypeConcentrations );
            m_averagePars.setTrueTwoCompartmentModelConcentrations(  averageTrueTwoCompartmentModelConcentrations );
            m_averagePars.setDFreqs( averageDFreqs );
            m_averagePars.setLorentzWidths( averageLorentzWidths );
            m_averagePars.setLorentzRWidths( averageLorentzRWidths );
            m_averagePars.setPhases( averagePhases );

            //
            m_stDevsOfPars.setMetaboliteNames( m_metaboliteNames );
            m_stDevsOfPars.setRawPeakAreas( rawStDevPeakAreas );
            m_stDevsOfPars.setRawFractionCorrectedPeakAreas( rawStDevFractionCorrectedPeakAreas );
            m_stDevsOfPars.setJMRUITypeConcentrations( stDevJMRUITypeConcentrations );
            m_stDevsOfPars.setTrueTwoCompartmentModelConcentrations( stDevTrueTwoCompartmentModelConcentrations );
            m_stDevsOfPars.setDFreqs( stDevDFreqs );
            m_stDevsOfPars.setLorentzWidths( stDevLorentzWidths );
            m_stDevsOfPars.setLorentzRWidths(  stDevLorentzRWidths );
            m_stDevsOfPars.setPhases( stDevPhases );
        }
    }


    /**
     * Deletes the entry with dbEntryNumber from the database!
     * @param dbEntryNumber
     */
    void deleteEntry( int dbEntryNumber )
    {
        if ( ( ( dbEntryNumber > -1 ) ) && ( dbEntryNumber < m_normalResponses.size( ) ) )
        {

            m_normalResponses.remove( dbEntryNumber );
            m_normalMetabolitePars.remove( dbEntryNumber );
            m_numberOfInstances = m_normalMetabolitePars.size( );
            // commit!
            writeDatabase( );
        }
    }

    /**
     * Returns the average parameters of the selected database.
     * @return
     */
    public NormalDatabaseMetaboliteParameters getAveragesDatabaseParameters( )
    {
        if ( this.m_numberOfInstances > 1 )
        {
            setAllAverageAndStDevsAttributes( );
            return m_averagePars;
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns the standard deviation parameters of the selected database.
     * @return
     */
    public NormalDatabaseMetaboliteParameters getStDevsDatabaseParameters( )
    {
        if ( this.m_numberOfInstances > 1 )
        {
            setAllAverageAndStDevsAttributes( );
            return m_stDevsOfPars;
        }
        else
        {
            return null;
        }
    }


    /**
     * Returns the average spectrum as a SingleVoxelSpectralData object.
     * @return
     */
    public SingleVoxelMeasuredSpectralData getAverageSpectrum( )
    {

        SingleVoxelMeasuredSpectralData avSpectrum = new SingleVoxelMeasuredSpectralData( );

        avSpectrum.setSpectralDataPointsFD( getMeanOfNormalRangeSpectrumAbsorption( ),getMeanOfNormalRangeSpectrumDispersion( ) );
        avSpectrum.setSiemensSpecificDicomHeaderInfo( null );
        avSpectrum.setCurveDrawingStyle( Curve.FAT_SOLID_LINE );
        avSpectrum.setCurveColor( Color.CYAN );
        avSpectrum.setCurveLabel( "average spectrum" );

        return avSpectrum;
    }

    
    /**
     * Reads and parses the attributes from a RandomAccesFile-object.
     */
    void read( RandomAccessFile in )
    {
        try
        {
            // <NormalMetaboliteDatabase>
            String dummy = in.readUTF( );

            m_fieldStrength = in.readUTF( );
            // System.out.println( "m_fieldStrength " + m_fieldStrength );

            m_resonanceFrequency = in.readUTF( );
            // System.out.println( "m_fieldStrength " + m_fieldStrength );
            
            m_quantificationMethod = in.readUTF( );
            // System.out.println( "m_quantificationMethod " + m_quantificationMethod );

            m_scannerType   = in.readUTF( );
            // System.out.println( "m_scannerType " + m_scannerType );

            m_sequenceType  = in.readUTF( );
            // System.out.println( "m_sequenceType " + m_sequenceType );

            m_tissueType    = in.readUTF( );
            // System.out.println( "m_tissueType " + m_tissueType );

            m_echoTime      = in.readUTF( );
            // System.out.println( "m_echoTime " + m_echoTime );

            m_repetitionTime= in.readUTF( );
            // System.out.println( "m_repetitionTime " + m_repetitionTime );

            m_manufacturer  = in.readUTF( );
            // System.out.println( "m_manufacturer " + m_manufacturer );

            m_numberOfInstances = in.readInt( );
            // System.out.println( "m_numberOfInstances " + m_numberOfInstances );

            
            // be sure that all structures are empty before loading ..
            m_normalResponses = new LinkedList<NormalDatabaseSpectrumResponse>( );
            m_normalMetabolitePars = new LinkedList<NormalDatabaseMetaboliteParameters>( );

            for ( int i = 0; i < m_numberOfInstances; i++ )
            {
                // System.out.println( "NormalMetaboliteValuesAndSpectrumDatabase -> Reading instance : " + i );

                NormalDatabaseSpectrumResponse response = new NormalDatabaseSpectrumResponse( );
                response.read( in );

                NormalDatabaseMetaboliteParameters pars = new NormalDatabaseMetaboliteParameters( );
                pars.read( in );

                // read the last string from the stream!
                dummy = in.readUTF( );

                // store them in the data containing objects
                m_normalResponses.add( response );
                m_normalMetabolitePars.add( pars );

            }

            dummy = in.readUTF( );
            //System.out.println( "Last string read from stream : " + dummy );

            in.close( );
        }
        catch ( IOException ex )
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
    }

    /**
     * Writes attributes to RandomAccessFile.
     * @param out
     */
    public void write( RandomAccessFile out )
    {
        try
        {

            out.writeUTF( "<NormalMetaboliteDatabase>" );

            out.writeUTF( m_fieldStrength );
            out.writeUTF( m_resonanceFrequency );
            out.writeUTF( m_quantificationMethod );
            out.writeUTF( m_scannerType );
            out.writeUTF( m_sequenceType );
            out.writeUTF( m_tissueType );
            out.writeUTF( m_echoTime );
            out.writeUTF( m_repetitionTime );
            out.writeUTF( m_manufacturer );

            out.writeInt( m_numberOfInstances );

            for ( int i = 0; i < m_numberOfInstances; i++ )
            {
                out.writeUTF( "<NormalDatabaseInstance>" );

                NormalDatabaseSpectrumResponse normalResponses = m_normalResponses.get( i );
                normalResponses.write( out );
                
                NormalDatabaseMetaboliteParameters pars = m_normalMetabolitePars.get( i );
                pars.write( out );

                out.writeUTF( "</NormalDatabaseInstance>" );
            }

            out.writeUTF( "</NormalMetaboliteDatabase>" );
            out.close( );

        }
        catch ( IOException ex )
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
    }

    /**
     * Returns the echo time of the database.
     * @return 
     */
    private String getEchoTime( )
    {
        return m_echoTime;
    }
    
    /**
     * Returns the magnetic field strength of the database.
     * @return 
     */
    private String getFieldStrength( )
    {
        return m_fieldStrength;
    }

    /**
     * Returns the quantification method used for obtaining the results.
     * @return 
     */
    private String getQuantificationMethod( )
    {
        return m_quantificationMethod;
    }

    /**
     * Returns the manufacturer of the scanner.
     * @return 
     */
    private String getManufacturer( )
    {
        return m_manufacturer;
    }

    /**
     * Returns the repetition time of the pulse sequence.
     * @return 
     */
    private String getRepetitionTime( )
    {
        return m_repetitionTime;
    }
    
    /**
     * Returns the scanner type (model).
     * @return 
     */
    private String getScannerType( )
    {
        return m_scannerType;
    }

    /**
     * Returns the sequence type with which the response was obtained.
     * @return 
     */
    private String getSequenceType( )
    {
        return m_sequenceType;
    }

    /**
     * Returns the examined tissue type (organ).
     * @return 
     */
    private String getTissueType( )
    {
        return m_tissueType;
    }

    /**
     * Returns the parameters from the database.
     * @return
     */
    public LinkedList<NormalDatabaseMetaboliteParameters> getDatabaseParameters( )
    {
        return m_normalMetabolitePars;
    }

    /**
     * Returns the names in the same order as which the normal 
     * values are returned.
     *
     * @return
     */
    public LinkedList<String> getMetaboliteNames( )
    {
        if ( m_normalMetabolitePars.size( ) > 0 )
        {
            return m_normalMetabolitePars.get( 0 ).getMetaboliteNames( );
        }
        else
        {
            return null;
        }
    }

    /**
     * Exports the database to an ASCII file.
     * @param fileAndPath
     */
    public void exportToTextFile( String fileAndPath )
    {
        try 
        {
            
            TextFileWriter writer = new TextFileWriter( new FileWriter( fileAndPath ) );
            
            writer.writeStringLn( "Databasename" + "\t" + m_dataBaseName );
            
            for ( int i = 0; i < m_numberOfInstances; i++ )
            {
                
                NormalDatabaseMetaboliteParameters pars = m_normalMetabolitePars.get( i );
                
                for ( int j = 0 ; j < m_normalMetabolitePars.size( ) ; j++ )
                {
                    writer.writeString( pars.getMetaboliteNames( ).get(j) + "\t" );
                    writer.writeString( pars.getRawPeakAreas( )[j] + "\t" );
                    writer.writeString( pars.getRawFractionCorrectedPeakAreas( )[j] + "\t" );
                    writer.writeString( pars.getJMRUITypeConcentrations( )[j] + "\t" );
                    writer.writeString( pars.getTrueTwoCompartmentModelConcentrations( )[j] + "\t" );
                    writer.writeString( pars.getDFreqs( )[j] + "\t" );
                    writer.writeString( pars.getPhases( )[j] + "\t" );
                    writer.writeString( pars.getLorentzWidths( )[j] + "\t" );
                    writer.writeString( pars.getLorentzRWidths( )[j] + "\t" );       
                }
            }
            
            writer.flush( );
            writer.close( );
                
        }
        catch ( IOException ex ) 
        {
            Logger.getLogger( NormalMetaboliteValuesAndSpectrumDatabase.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
        
    }

    // Attributes
    private int              m_numberOfInstances = -1;
    private LinkedList<NormalDatabaseSpectrumResponse> m_normalResponses = new LinkedList<>( );
    private LinkedList<NormalDatabaseMetaboliteParameters> m_normalMetabolitePars = new LinkedList<>( );

    private String           m_dataBaseName;

    private RandomAccessFile m_dbFile;
    private File             m_dataBaseFile;
    private boolean          m_dataBaseAlreadyExists = false;

    private String           m_fieldStrength;       // in T!
    private String           m_resonanceFrequency;  // in Hz
    private String           m_quantificationMethod;
    private String           m_scannerType;
    private String           m_sequenceType;
    private String           m_tissueType;
    private String           m_echoTime;
    private String           m_repetitionTime;
    private String           m_manufacturer;
    private String           m_acquisitionBandwidth; // needed for display also!!

    private int              m_numberOfColumns = 9;

    private NormalDatabaseMetaboliteParameters m_averagePars;
    private NormalDatabaseMetaboliteParameters m_stDevsOfPars;

//    private double[] m_rawAveragePeakAreas;
//    private double[] m_rawAverageFractionCorrectedPeakAreas;
//    private double[] m_averageJMRUITypeConcentrations;
//    private double[] m_averageTrueTwoCompartmentModelConcentrations;
//    private double[] m_averageDFreqs;
//    private double[] m_averageLorentzWidths;
//    private double[] m_averageLorentzRWidths;
//    private double[] m_averagePhases;
//
//    private double[] m_rawStDevPeakAreas;
//    private double[] m_rawStDevFractionCorrectedPeakAreas;
//    private double[] m_stDevJMRUITypeConcentrations;
//    private double[] m_stDevTrueTwoCompartmentModelConcentrations;
//    private double[] m_stDevDFreqs;
//    private double[] m_stDevLorentzWidths;
//    private double[] m_stDevLorentzRWidths;
//    private double[] m_stDevPhases;

    private LinkedList<String> m_metaboliteNames = new LinkedList<>();

//    public final static String METAB_DATA_RAW_UNCORRECTED_TYPE                = "METAB_DATA_RAW_UNCORRECTED_TYPE";
//    public final static String METAB_DATA_RAW_COMPONENTS_ONLY_CORRECTED_TYPE  = "METAB_DATA_RAW_COMPONENTS_ONLY_CORRECTED_TYPE";
//    public final static String METAB_DATA_JMRUI_CONC_STYLE                    = "METAB_DATA_JMRUI_CONC_STYLE";
//    public final static String METAB_DATA_TRUECONC_TWO_COMPONENTS_CORRECTED   = "METAB_DATA_TRUECONC_TWO_COMPONENTS_CORRECTED";

}
