/*
 * File: $Id: TDFDFitPrototypeModelSpectrum.java,v 1.1.1.1 2006/04/19 09:38:49 hans Exp $
 * History
 * Date: 27.05.2003, 12:07  Authors: Hans Slotboom
 *
 * Changes:
 *
 * Copyright (C) 2003 by
 * 
 * Inselspital - Department of Neuroradiology
 * CH 3010 Bern Switzerland
 */ 
 
package org.spectrim.tdfdfit;

import java.awt.Cursor;
import java.awt.Color;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.HashMap;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Date;

import java.net.InetAddress;
import java.net.UnknownHostException;

import java.io.RandomAccessFile;
import java.io.IOException;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.File;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;
import javax.swing.JFileChooser;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.Random;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JOptionPane;
import org.spectrim.main.MetaboliteFeatures;
import org.spectrim.main.VoxelMetabolites;
import org.spectrim.randomforestregression.DataGeneratorMain;
import org.spectrim.xyplot.Curve;
import org.spectrim.xyplot.XYPlotFrame;



/**
 * This Singleton class contains all information to calculate the numeric time
 * and frequency domain model from the parametric and numeric informations
 * This class is thought to contain the complete model information of a 
 * in vivo spectrum containing lines, numeric patterns, parametric patterns, 
 * and metabolites.
 * @author  hans
 */
public class TDFDFitPrototypeModelSpectrumParameters implements ActionListener
{

    /** 
     * Creates a new instance of TDFDFitPrototypeModelSpectrumParameters. 
     */
    private TDFDFitPrototypeModelSpectrumParameters( )
    {
        TDFDFittingDefaults.getInstance( ).addActionListener( this );
    }

    /**
     * Returns the one and only instance of 
     * TDFDFitPrototypeModelSpectrumParameters.
     * @return 
     */
    public static TDFDFitPrototypeModelSpectrumParameters getInstance( )
    {
        if ( theInstance == null )
        {
            theInstance = new TDFDFitPrototypeModelSpectrumParameters( );
        }
        return theInstance;
    }
    
    /**
     * Resets the model to the default values.
     */
    public void resetModel( )
    {
        m_mrModelSpectrumName = "No model defined";
        m_simpleLineList = new LinkedList<>( );
        m_parametricPatternList = new LinkedList<>( );
        m_metaboliteList = new LinkedList<>( );
        m_numericPatternList = new LinkedList<>( );
        
    }
    
    /**
     * Sets which MRSignalProperties apply for this spectrum.
     * @param mrSignalProperties
     */
    public void setMRSignalProperties( MRSignalProperties mrSignalProperties )
    {
        m_MRSignalProperties = mrSignalProperties;
    }
    
    /**
     * Returns the applying MRSignalProperties for this spectrum.
     * @return 
     */
    public MRSignalProperties getMRSignalProperties( )
    {
        return m_MRSignalProperties;
    }    
    
    /** 
     * Returns a list of MRMetaboliteModel objects from the 
     * list MRMetaboliteModel.
     * @return 
     */
    public LinkedList getMetaboliteModels( )
    {
        return m_metaboliteList;
    }
    
    /**
     * Returns the name of the model as string.
     * @return 
     */
    @Override
    public String toString( )
    { 
        return m_mrModelSpectrumName;
    }
    
    /**
     * Defines the name of the spectral model.
     * @param specModelName
     */
    public void setModelSpectrumName( String specModelName )
    {
        m_mrModelSpectrumName = specModelName;
        if  ( m_fitModelTree != null )
        {
            m_fitModelTree.setAllPopupItemsEnabled( );
        }
    }
    
    /**
     * Defines the name of the spectral model.
     * @return 
     */
    public String getModelSpectrumName( )
    {
        TDFDFittingDefaults fdef = TDFDFittingDefaults.getInstance( );
        fdef.load( );
        if ( this.m_thisFitDoesWaterRemoval == false )
        {
            m_mrModelSpectrumName = fdef.getProperty( TDFDFittingDefaults.DEFAULT_TDFDFIT_FIT_MODEL_NAME );
        }
        else
        {
            m_mrModelSpectrumName = fdef.getProperty( TDFDFittingDefaults.TDFDFIT_WATER_MODEL_NAME );
        }
        return m_mrModelSpectrumName;
    }
    
    
    public void setFitDoesWaterRemoval( boolean removeWater )
    {
        m_thisFitDoesWaterRemoval = removeWater;
    }
    

    
    /**
     * Reads and parses the attributes from a RandomAccesFile.
     * @param in 
     */
    public void read( RandomAccessFile in )
    {
        // The user wants to load a new prototype model.
        // First reset the model.
        
        this.resetModel( );
        
        try
        {
            String tag = in.readUTF( );
            
            while ( tag.compareTo( "</MRModelSpectrum>" ) != 0 )
            {
                
                if ( tag.compareTo( "<MRFittingParameters>" ) == 0 )
                {
                    MRLineModelParameters linePars = new MRLineModelParameters( );
                    linePars.read( in );
                    // add the just created object to the appropriate map
                    m_simpleLineList.add( linePars );
                    //notifyLineRead( linePars );
                }
                
                if ( tag.compareTo( "<MRParametricPattern>" ) == 0 )
                {
                    // System.out.println( "tag : " + tag );
                    MRParametricPattern parPat = new MRParametricPattern( );
                    parPat.read( in );
                    // add the just created object to the appropriate map
                    m_parametricPatternList.add( parPat );
                }
                
                if ( tag.compareTo( "<MRNumericPattern>" ) == 0 )
                {
                    MRNumericPattern numPat = new MRNumericPattern( );
                    numPat.read( in );
                    m_numericPatternList.add( numPat );
                }
                
                if ( tag.compareTo( "<MRMetaboliteModel>" ) == 0 )
                {
                    MRMetaboliteModel metMod = new MRMetaboliteModel( );
                    metMod.read( in );
                    m_metaboliteList.add( metMod );
                }
                
                tag = in.readUTF( );  
                // System.out.println( "% tag = " + tag );
            }
        }
        catch ( IOException e )
        {
            System.out.println( "Error reading model file." );
        }
    }
        
    /**
     * Raises an ActionEvent notifying registered objects that a line has
     * read successfully from file.
     * 
     * @param eventType
     * @param obj
     * @param actionCommand 
     */
    public void notifyActionListeners( int eventType, 
                                        Object obj, 
                                        String actionCommand )
    {
        Iterator it = m_actionListeners.iterator( );
        while( it.hasNext( ) )
        {
            if ( obj != null )
            {
                ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( obj, eventType, actionCommand ) );
            }
            else
            {
                ( ( ActionListener )it.next( ) ).actionPerformed( new ActionEvent( this, eventType, actionCommand ) );
            }
        }
    }
    
    /**
     * Writes attributes to RandomAccessFile.
     * @param out 
     */
    void write( RandomAccessFile out )
    {
        try
        {
            out.writeUTF( "<MRModelSpectrum>" );
            
            // write all simple resonance lines
            writeSimpleLines( out );
            // write parametric patterns
            writeParametricPatterns( out );
            // write numeric patterns
            writeNumericPatterns( out );
            // write metabolites
            writeMetabolites( out );
            
            out.writeUTF( "</MRModelSpectrum>" );
        }
        catch ( IOException e )
        {
        }
    }
    
    /**
     * Called when registered object raises an ActionEvent.
     * @param e
     */
    @Override
    public void actionPerformed( ActionEvent e ) 
    {
        
        Object sender = e.getSource( );
        
        if ( e.getSource( ) == TDFDFittingDefaults.getInstance( ) )
        {
            // do nothing!
        }
        
        if ( ( sender instanceof TDFDFitModelTreeWithPopupMenu ) && ( m_MRSignalProperties != null ) )
        {
            m_fitModelTree = (TDFDFitModelTreeWithPopupMenu)e.getSource( );

            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_TREE_NODE_ADD_EVENT )
            {
                // Clear the model
                clearModel( );
                
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                extractModelTreeInformation( true );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_CLEARED )
            {
                // Clear the model
                clearModel( );
                
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                
                extractModelTreeInformation( true );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_TREE_NODE_DELETE_EVENT )
            {
                
                // Clear the model
                clearModel( );
                
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                
                extractModelTreeInformation( true );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_TREE_CREATED_EVENT )
            {
                // Clear the model
                clearModel( );
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                extractModelTreeInformation( true );  
            }
            
            // The user selected a metabolite node-notify listeners 
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_TREENODE_NAME_CHANGE_EVENT )
            {
                
                TDFDFitModelTreeWithPopupMenu treeView = (TDFDFitModelTreeWithPopupMenu)e.getSource( );
                
                // Let the registered objects know that a line node has been selected
                notifyActionListeners( NODE_NAME_CHANGED, this, treeView.getNewNameAfterRenaming( ) );
                       
                // TO DO prior knowledge should be updated conform the name change
                treeView.updatePriorKnowledgeInTreeModel( treeView.getOldNodeNameBeforeRenaming( ), 
                                                          treeView.getNewNameAfterRenaming( ) );
                
                // Clear the model
                clearModel( );
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                extractModelTreeInformation( true );  
                
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.CREATE_STARTING_VALUE_FILES )
            {
                createTDFDFitStartingValueFiles( );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.EXPORT_AS_PARAMETRIC_PATTERN )
            {
                exportAsParametricPattern( );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.EXPORT_AS_NUMERIC_PATTERN )
            {
                // performs the writing to a numeric pattern file ..
                exportSelectionAsNumericPattern( );
            }
            
            if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.EXPORT_AS_METABOLITE_PATTERN )
            {
                // performs the writing to a numeric pattern file ..
                exportSelectionAsMetabolitePattern( );
            }
           
        }
        
        // The user selected a MRLineModelParameters node - notify listeners 
        if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.LINE_MODEL_SELECTED_EVENT )
        {
            // Let the registered objects know that a line node has been selected
            doPriorKnowledge( true );
            notifyActionListeners( NODE_SELECTED, e.getSource( ), null );
        }
        
        // The user selected a MRNumericPattern node - notify listeners 
        if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.NUMERIC_PATTERN_SELECTED_EVENT )
        {
            // Let the registered objects know that a line node has been selected
            doPriorKnowledge( true );
            notifyActionListeners( NODE_SELECTED, e.getSource( ), null );    
        }
        
        // The user selected a parametric pattern node - notify listeners 
        if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.PARAMETRIC_PATTERN_SELECTED_EVENT ) 
        {
            // Let the registered objects know that a line node has been selected
            doPriorKnowledge( true );
            notifyActionListeners( NODE_SELECTED, e.getSource( ), null );
        }
        
        // The user selected a metabolite node- notify listeners 
         if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.METABOLITE_SELECTED_EVENT )
         {
             // Let the registered objects know that a line node has been selected
             doPriorKnowledge( true );
             notifyActionListeners( NODE_SELECTED, e.getSource( ), null );   
         }
         
        // The user selected a metabolite node- notify listeners 
         if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.ROOT_SELECTED_EVENT )
         {
             // Let the registered objects know that a line node has been selected
             notifyActionListeners( ROOT_SELECTED, e.getSource( ), null );
         }
        
        // The user selected a metabolite node notify listeners 
        if ( e.getID( ) == TDFDFitModelTreeWithPopupMenu.MODEL_ROOT_NAME_CHANGE_EVENT )
        {
            // Let the registered objects know that a line node has been selected
            m_mrModelSpectrumName = (String)e.getSource( );
            notifyActionListeners( MODEL_ROOT_NAME_CHANGED, e.getSource( ), null );
        }
        
        // Handle events from the MRLineModelParametersPanel
        if ( e.getSource( ) instanceof MRLineModelParametersPanel )
        {
            if ( e.getActionCommand( ).compareTo( MRLineModelParametersPanel.FIT_MODEL_SPECTRUM_CMD ) == 0 )
            {                
                String parameterFile = createTDFDFitStartingValueFiles( );
                String binaryDataFile = writeSpectrumToFitToTdfdfitFormat( );
                ((MRLineModelParametersPanel)( e.getSource( ) ) ).setCursor( new Cursor( Cursor.WAIT_CURSOR ) );
                startFit( binaryDataFile, parameterFile );
                ((MRLineModelParametersPanel)( e.getSource( ) ) ).setCursor( new Cursor( Cursor.DEFAULT_CURSOR ) );
                
                // Clear the model
                clearModel( );
                
                // Visit all nodes of the tree and extract the objects from it
                // and put it in the local member objects
                extractModelTreeInformation( true );
                ( ( MRLineModelParametersPanel)e.getSource( ) ).restoreFileDeletionSettings( );
            }
        }
        
        if ( e.getActionCommand( ).compareTo( "measured spectrum changed" ) == 0 )
        {
            SingleVoxelSpectralData resp = (SingleVoxelSpectralData)e.getSource( );
            m_measuredSpectralResponse.reallocateArrays( resp.getAbsorptionChannelTD( ).length );
            m_measuredSpectralResponse.autoSetXaxis( );
            
            resp.autoSetXaxis( );
            
            m_measuredSpectralResponse.setSpectralDataPointsFD( resp.getAbsorptionChannelFD( ), resp.getDispersionChannelFD( ) );
            m_measuredSpectralResponse.setSiemensSpecificDicomHeaderInfo( resp.getSiemensSpecificDicomHeaderInfo( ) );
            
            MRSignalProperties mrCond = MRSignalProperties.getInstance( );
            mrCond.setNumberOfSamplePoints( resp.getAbsorptionChannelFD( ).length );
        }
    }
    
  
    /**
     * Exports the currently selected leave of the tree model to a metabolite (.mpf) 
     * pattern.     
     */
    private void exportSelectionAsMetabolitePattern( )
    {
        
        if ( m_fitModelTree.getSelectedObject( ).getClass( ).equals( MRMetaboliteModel.class ) == true )
        {
            MRMetaboliteModel metModel = (MRMetaboliteModel)m_fitModelTree.getSelectedObject( );
            metModel.exportMetaboliteModel( );
        }
    } 
  
    /**
     * Exports the currently selected of the tree model to a numerical pattern.     
     */
    private void exportSelectionAsNumericPattern( )
    {
        
        if ( m_fitModelTree.getSelectedObject( ).getClass( ).equals( MRParametricPattern.class ) == true )
        {
            MRParametricPattern parPat = (MRParametricPattern)m_fitModelTree.getSelectedObject( );            
            parPat.exportParametricPatternAsNumericalPattern( );
        }
        
        if ( m_fitModelTree.getSelectedObject( ).getClass( ).equals( MRLineModelParameters.class ) == true )
        {
            MRLineModelParameters lineModel = (MRLineModelParameters)m_fitModelTree.getSelectedObject( );
            lineModel.exportMRFittingParametersAsNumericalPattern( );
        }
        
        if ( m_fitModelTree.getSelectedObject( ).getClass( ).equals( MRMetaboliteModel.class ) == true )
        {
            MRMetaboliteModel metModel = (MRMetaboliteModel)m_fitModelTree.getSelectedObject( );
            metModel.exportMetaboliteModelAsNumericalPattern( );
        }
        
    }
    
    /**
     * Exports only the lines of current model as one parametric pattern; 
     * in the current implementation this is only possible 
     * if the model consists of MRLineModelParameter and MRParametricPattern
     * objects.
     */
    private void exportAsParametricPattern( )
    {
        TDFDFittingDefaults defaults = TDFDFittingDefaults.getInstance( );
        defaults.load( );
        
        String path = defaults.getProperty( "Fit project path" );
         
        String parametricPatternDir = path + System.getProperty( "file.separator" ) + 
                                      "Parametric_Patterns" + 
                                      System.getProperty( "file.separator" );
        
        JFileChooser fileChooser = new JFileChooser( parametricPatternDir );
        
        fileChooser.setName( "Select a parametric pattern file" );
        
        int result = fileChooser.showSaveDialog( null );
        
        File fitParametricPatternFileReference = new File( "" );
        
        switch ( result )
        {
            case JFileChooser.APPROVE_OPTION:   // Approve Open button was pressed
                                                fitParametricPatternFileReference = fileChooser.getSelectedFile( );
                                                break;
            case JFileChooser.CANCEL_OPTION:     // Cancel or close dialog was selected
                                                break;
            case JFileChooser.ERROR_OPTION:      // the selected process did not terminate properly
                                                break;
        }
        
        if ( fitParametricPatternFileReference.getName( ).compareTo( "" ) != 0 )
        {
            // instantiate a temporary MRParametricPattern object
            MRParametricPattern parPat = new MRParametricPattern( this.getModelSpectrumName( ) );
            
            // iterate over all simple lines in the model
            Iterator it = m_simpleLineList.iterator( );
            
            MRSignalProperties mrCond = MRSignalProperties.getInstance( );
            float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
        
            while ( it.hasNext( ) )
            {
                MRLineModelParameters linPar = new MRLineModelParameters( );
                MRLineModelParameters obj = (MRLineModelParameters)it.next( );
                linPar.setFittingParameters( (MRFittingParameters)obj );
                
                if ( defaults.getProperty( TDFDFittingDefaults.EXPORT_MODEL_AS_PATTERN_APPLY_FILTER ).compareTo( TDFDFittingDefaults.NO ) == 0 )
                {
                    parPat.addSimpleLine( linPar );
                }
                else
                {
                    if ( Math.abs( linPar.getOffset( ) ) > new Float( defaults.getProperty( TDFDFittingDefaults.PATTERN_EXPORT_CUTOFF_FREQ ) ).floatValue( ) )
                    {
                         parPat.addSimpleLine( linPar );
                    }
                }
            } 
            
            // here other patterns are not written
            
            // finally write them to a file
            parPat.writeModelFile( fitParametricPatternFileReference.getAbsolutePath( ) );
            
        }
    }
    
    /**
     * Replaces all "\" by "\\" in order to be interpreted 
     * correctly by the operating system.
     * @param cmd 
     * @return 
     */
    private String generateProperCommand( String cmd )
    {
        String out = "";
        for ( int i = 0; i < cmd.length( ); i++ )
        {
            if ( cmd.substring( i, i + 1 ).compareTo( "\\" ) == 0 )
            {
                out += "\\" + "\\";
            }
            else
            {
                out += cmd.substring( i, i + 1 );
            }
        }
        return out;
    }
    
    /**
     * Starts the fit that corresponds to the current status of the model
     * and the loaded measured spectrum.
     */
    private void startFit( String dataFile, String parameterFile )
    {
        
        m_fitModelTree.setMRModelSpectrum( this );
        
        int     returnValue;
        String  cmd = null;
        
        TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
        fittingDefaults.load( );
        
        String tdfdfitExecutablePath = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH );
        String currentModelBasePath = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_MODELS_BASE_PATH );
        String currentModelName = fittingDefaults.getProperty( TDFDFittingDefaults.DEFAULT_TDFDFIT_FIT_MODEL_NAME );
        
        if ( tdfdfitExecutablePath == null )
        {
            String userHomeAsString = System.getProperty( "user.home" );
            userHomeAsString = userHomeAsString + System.getProperty( "file.speparator" ) + 
                               ".SpectrIm" + System.getProperty( "file.speparator" ) + 
                               "TDFDFitExecutableDir";
            fittingDefaults.setProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH, userHomeAsString );
            fittingDefaults.save( );
        }
        
        String tdfdfitExecutableName = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_EXECUTABLE_NAME );
        if ( tdfdfitExecutableName == null )
        {
            fittingDefaults.setProperty( TDFDFittingDefaults.TDFDFIT_EXECUTABLE_NAME, "tdfdfit2017.exe" );
            fittingDefaults.save( );
        }
        
        String parFileName = parameterFile;
       
        String dataFileName = dataFile;
        
        String TDFDFitCall;
        
        
        TDFDFitCall = tdfdfitExecutablePath + System.getProperty( "file.separator" )+ tdfdfitExecutableName;
        
        cmd = TDFDFitCall + " " + tdfdfitExecutablePath + " " + dataFileName + " " + parFileName;
        
        cmd =  generateProperCommand( cmd );
        
        // System.out.println( "Trying following system command : " + cmd );
        
        try 
        {
            Process p = Runtime.getRuntime( ).exec( cmd );
            
            InputStream in = p.getInputStream( );
            int c;
            while ( ( c = in.read( ) ) != -1 ) 
            {
                System.out.print( (char)c ) ;
            }

            returnValue = p.waitFor( );
            in.close( );
	}
        catch ( IOException ioX ) 
        {
            // not handled yet
            System.out.println( "IOException occurred" );
        }
        catch ( InterruptedException intX ) 
        {
            // not handled yet
            System.out.println( "InterruptedException occurred" );
	}
        
        // move the data to the patient database directory.
        String fileNameCopy = dataFileName;
        String verboseFileName = dataFileName.substring( 0, dataFileName.lastIndexOf( "." ) );
        String resultFileToParse = tdfdfitExecutablePath + System.getProperty( "file.separator" ) + verboseFileName + "_verbose.par";
        String fileNameBase = dataFileName.substring( 0, fileNameCopy.lastIndexOf( '.' ) );
        String moveDataToPath = currentModelBasePath + System.getProperty( "file.separator" ) + fileNameBase + System.getProperty( "file.separator" ) + "MasterTDFDFitData";
        File moveDataToPathDirectory = new File( moveDataToPath );
        if ( !moveDataToPathDirectory.exists( ) )
        {
            moveDataToPathDirectory.mkdir( );
        }
        
        moveDeleteTDFDFitRelatedFiles( moveDataToPath, fileNameBase, resultFileToParse, parFileName );
        
        /*
            String title = m_waterReferenceSpectralResponse.getDicomHeaderInfo( ).getPatientName( ) + " (" +
                           m_waterReferenceSpectralResponse.getDicomHeaderInfo( ).getPatientBirthDate() + ") - " + 
                           m_waterReferenceSpectralResponse.getDicomHeaderInfo( ).getSeriesDescription( ) + " -  Series # : " + 
                           m_waterReferenceSpectralResponse.getDicomHeaderInfo( ).getSeriesNumber( ) + " - " +
                           m_waterReferenceSpectralResponse.getDicomHeaderInfo( ).getAcquisitionDate( );
        */
        
        FitResultViewerFrame fitViewer = new FitResultViewerFrame( m_MRSignalProperties.getNumberOfSamplePoints( ), 
                                                                   moveDataToPath + System.getProperty( "file.separator" ) + dataFileName, "", true  );
        
        // This method reads the fitresults from file and updates the 
        // MRFittingParameter objects
        
        resultFileToParse = moveDataToPath + System.getProperty( "file.separator" ) + verboseFileName + "_verbose_master.par";
        
        if ( new File( resultFileToParse ).exists( ) )
        {
             readFitresultsToImproveStartingValues( resultFileToParse );
        }
        else
        {
            resultFileToParse = moveDataToPath + System.getProperty( "file.separator" ) + verboseFileName + "_verbose.par";
        }
        
       
        
    }
    
    public void moveDeleteTDFDFitRelatedFiles( String moveDataToPath,
                                               String fileNameBase,  
                                               String verboseParameterFileName, 
                                               String parameterFileName ) 
    {
         // Clean up the bin directory!
        TDFDFittingDefaults fitDef = TDFDFittingDefaults.getInstance( );
        fitDef.load( );
        
        String path = fitDef.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH );
        
        File fileDotBin = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".bin" );
        File fileDotFit = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".fit" );
        File fileDotFit0 = new File( path + System.getProperty( "file.separator" ) + fileNameBase    + ".fit0" );
        File fileDotFitRes = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".fitres" );
        File fileDotOrg = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".org" );
        File fileDotRes = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".res" );
        File fileDotCorOrg = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".cor" );
        File fileDotExl = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".exl" );
        File fileDotNln0 = new File( path + System.getProperty( "file.separator" ) + fileNameBase    + ".nln0" );
        File fileDotMat = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".mat" );
        File fileDotErr = new File( path + System.getProperty( "file.separator" ) + fileNameBase     + ".err" );
        
        File parameterStartFile = new File( parameterFileName );
        File verboseParameterStartFile = new File( verboseParameterFileName );
        
        
        String outputFileNamePar = parameterStartFile.getName( );
        outputFileNamePar = outputFileNamePar.replace( ".par", "_master.par" );
        String outputVerboseParameterStartFileName = verboseParameterStartFile.getName( );
        outputVerboseParameterStartFileName = outputVerboseParameterStartFileName.replace( ".par", "_master.par" );
        
        File moveDataToPathFile = new File( moveDataToPath );
        
        if ( parameterStartFile.exists( ) && moveDataToPathFile.exists( ) )
        {
            File outputFile = new File( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + outputFileNamePar  ); 
            if ( outputFile.exists( ) )
            {
                outputFile.delete( );
            }
            
            if ( new File( outputFileNamePar ).exists( ) )
            {
                System.out.println( "ParameterStartFile exists." );
                parameterStartFile.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_PAR ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_PAR, "no" );
                fitDef.save( );
            }
            
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_PAR ).compareTo( "yes" ) == 0 )
            {
                parameterStartFile.delete( );
            }
            else
            {
                parameterStartFile.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + outputFileNamePar ) );
            }
            
            // ============================================================================================================================================
            
            File outputVerboseFile = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + outputVerboseParameterStartFileName );
            
            if ( outputVerboseFile.exists( ) )
            {
                outputVerboseFile.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_VERBOSE_PAR ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_VERBOSE_PAR, "no" );
                fitDef.save( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_DOT_VERBOSE_PAR ).compareTo( "yes" ) == 0 )
            {
                verboseParameterStartFile.delete( );
            }
            else
            {
                verboseParameterStartFile.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + outputVerboseParameterStartFileName ) );
            }
               
        }
        
        if ( fileDotBin.exists( ) && moveDataToPathFile.exists( ) )
        {
            
            File binfile = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".bin"  );
            if ( binfile.exists( ) )
            {
                binfile.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN, "no" );
                fitDef.save( );
            }        
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN ).compareTo( "yes" ) == 0 )
            {
                fileDotBin.delete( );
            }
            else
            {
                fileDotBin.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".bin"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File fitfile = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fit"  );
            if ( fitfile.exists( ) )
            {
                fitfile.delete( );
            }
            
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT, "no" );
                fitDef.save( );
            }        
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT ).compareTo( "yes" ) == 0 )
            {
                fileDotFit.delete( );
            }
            else
            {
                fileDotFit.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fit"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            
            File fitfile0 = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fit0"  );
            if ( fitfile.exists( ) )
            {
                fitfile0.delete( );
            }
                                    
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT0 ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT0, "no" );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FIT0 ).compareTo( "yes" ) == 0 )
            {
                fileDotFit0.delete( );
            }
            else
            {
                fileDotFit0.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fit0"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File filedotfitres = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fitres"  );
            if ( filedotfitres.exists( ) )
            {
                filedotfitres.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FITRES ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FITRES, "no" );
                fitDef.save( );
            }        
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_FITRES ).compareTo( "yes" ) == 0 )
            {
                fileDotFitRes.delete( );
            }
            else
            {
                fileDotFitRes.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".fitres"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File file_dot_org =  new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".org"  );
            if ( file_dot_org.exists( ) )
            {
                file_dot_org.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ORG ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ORG, "no" );
                fitDef.save( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ORG ).compareTo( "yes" ) == 0 )
            {
                fileDotOrg.delete( );
            }
            else
            {
                fileDotOrg.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".org"  ) );
            }
            
             //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
             
            File file_dot_res = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".res"  );
            if ( file_dot_res.exists( ) )
            {
                file_dot_res.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_RES ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_RES, "no" );
                fitDef.save( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_RES ).compareTo( "yes" ) == 0 )
            {
                fileDotRes.delete( );
            }
            else
            {
                fileDotRes.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".res"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File file_dot_cororg = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".cor"  );
            if ( file_dot_cororg.exists( ) )
            {
                file_dot_cororg.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_CORORG ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_CORORG, "no" );
                fitDef.save( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_CORORG ).compareTo( "yes" ) == 0 )
            {
                fileDotCorOrg.delete( );
            }
            else
            {
                fileDotCorOrg.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".cor"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File file_dot_exl = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".exl"  );
            if ( file_dot_exl.exists( ) )
            {
                file_dot_exl.delete( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_EXL ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_EXL, "no" );
                fitDef.save( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_EXL ).compareTo( "yes" ) == 0 )
            {
                fileDotExl.delete( );
            }
            else
            {
                 fileDotExl.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".exl"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            
            File file_dot_nln_0 = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".nln0"  );
            if ( file_dot_nln_0.exists( ) )
            {
                file_dot_nln_0.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_NLN0 ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_NLN0, "no" );
                fitDef.save( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_NLN0 ).compareTo( "yes" ) == 0 )
            {
                fileDotNln0.delete( );
            }
            else
            {
                fileDotNln0.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".nln0"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File file_dot_mat = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".mat"  );
            if ( file_dot_mat.exists( ) )
            {
                file_dot_mat.delete( );
            } 
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_MAT ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_MAT, "no" );
                fitDef.save( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_MAT ).compareTo( "yes" ) == 0 )
            {
                fileDotMat.delete( );
            }
            else
            {
                fileDotMat.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".mat"  ) );
            }
            
            //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            
            File file_dot_err = new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".err"  );
            if ( file_dot_err.exists( ) )
            {
                file_dot_err.delete( );
            }
            
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ERR ) == null )
            {
                fitDef.setProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ERR, "no" );
                fitDef.save( );
            }
            if ( fitDef.getProperty( TDFDFittingDefaults.DELETE_TDFDFIT_OUTPUTFILE_BIN_ERR ).compareTo( "yes" ) == 0 )
            {
                fileDotErr.delete( );
            }
            else
            {
                fileDotErr.renameTo( new File ( moveDataToPathFile.getPath( ) + System.getProperty( "file.separator" ) + fileNameBase  + ".err"  ) );
            }
           
        }
        
        // renameTo does not move, in fact it is a copy therefore we also need to delete the source files explicitly!
        fileDotBin = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".bin" );
        if ( fileDotBin.exists( ) )
        {
            fileDotBin.delete( );
        }
        
        fileDotFit = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".fit" );
        if ( fileDotFit.exists( ) )
        {
            fileDotFit.delete( );
        }
        
        fileDotFit0 = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".fit0" );
        if ( fileDotFit0.exists( ) )
        {
            fileDotFit0.delete( );
        }
        
        fileDotFitRes = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".fitres" );
        if ( fileDotFitRes.exists( ) )
        {
            fileDotFitRes.delete( );
        }
        
        fileDotOrg = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".org" );
        if ( fileDotOrg.exists( ) )
        {
            fileDotOrg.delete( );
        }
        
        fileDotRes = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".res" );
        if ( fileDotRes.exists( ) )
        {
            fileDotRes.delete( );
        }
        
        fileDotCorOrg = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".cor" );
        if ( fileDotCorOrg.exists( ) )
        {
            fileDotCorOrg.delete( );
        }
        
        fileDotExl = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".exl" );
        if ( fileDotExl.exists( ) )
        {
            fileDotExl.delete( );
        }
        
        fileDotNln0 = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".nln0" );
        if ( fileDotNln0.exists( ) )
        {
            fileDotNln0.delete( );
        }
        
        fileDotMat = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".mat" );
        if ( fileDotMat.exists( ) )
        {
            fileDotMat.delete( );
        }
        
        fileDotErr = new File( path + System.getProperty( "file.separator" ) + fileNameBase  + ".err" );
        if ( fileDotErr.exists( ) )
        {
            fileDotErr.delete( );
        }
    }
    
    
    
    
    
    /**
     * This method reads the fit results from file and updates the 
     * MRFittingParameter objects.
     */
    private void readFitresultsToImproveStartingValues( String fileToParse )
    {
//        TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
//        String path = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH ) +  System.getProperty( "file.separator" );
//        
//        String parFileName =  path + System.getProperty( "file.separator" ) + "verbose.par" ;
        
        // TDFDFit produces in the directory the verbose.par result file
        // this file has to be parsed and the result values should be read 
        // into the FittingParameter objects
        
        DefaultMutableTreeNode root = (DefaultMutableTreeNode)m_fitModelTree.getModel( ).getRoot( );
        int cntr = 0;
        MRFittingParameters fitPar = new MRLineModelParameters( );
        
        try
        {            
            RandomAccessFile reader = new RandomAccessFile( fileToParse, "rw" );
            
            String str = reader.readLine( );
            // System.out.println( "string read " + str );

            while ( str.trim( ).compareTo( "tdfdfit_starting_values" ) != 0 )
            {
                str = reader.readLine( );
            }
            
            // read the following blank line
            reader.readLine( );
            int totalComponents = m_metaboliteList.size( ) + 
                                  m_numericPatternList.size( ) + 
                                  m_parametricPatternList.size( ) + 
                                  m_simpleLineList.size( );
            
            for ( int i = 0; i < totalComponents; i++ )
            {
                
                // local variable needed to identify whether current spectral
                // object is an instance of MRLineModelParameters; if it
                // is an ofset shift of half the spectral width has to occur
                // if not, no offset shift is needed
                boolean isObjectMRLineModelParameters = false;
                boolean isObjectNumericPattern        = false;
                Object obj = ((DefaultMutableTreeNode)root.getChildAt( i ) ).getUserObject( );
            
                if ( obj instanceof MRLineModelParameters )
                {
                    fitPar = (MRFittingParameters)obj;
                    isObjectMRLineModelParameters = true;
                }
                
                if ( obj instanceof MRParametricPattern )
                {
                    fitPar = ((MRParametricPattern)obj).getFittingParameters( );
                }
                
                if ( obj instanceof MRNumericPattern )
                {
                    fitPar = ((MRNumericPattern)obj).getFittingParameters( );
                    isObjectNumericPattern = true;
                }
                
                if ( obj instanceof MRMetaboliteModel )
                {
                    fitPar = ((MRMetaboliteModel)obj).getFittingParameters( );
                }
                         
                str = reader.readLine( );
                // System.out.println( str );
                int ii = 0;
                while ( str.charAt( ii++ ) == ' ' )
                {
                    // do nothing
                }
                str = str.substring( ii - 1 );
                int length = str.indexOf( " " );
                String firstToken = str.substring( 0, length );
                str = str.substring(length);
                
                ////////////////////////////////////////////////////////////////
                ii = 0;
                while ( str.charAt( ii++ ) == ' ' )
                {
                    // do nothing
                }
                str = str.substring( ii - 1 );
                
                length = str.indexOf( " " );
                String secondToken = str.substring( 0, length );
                str = str.substring( length );
                Float area = new Float( secondToken );
                
                ////////////////////////////////////////////////////////////////
                ii = 0;
                while ( str.charAt( ii++ ) == ' ' )
                {
                    // do nothing
                }
                str = str.substring( ii - 1 );                
                length = str.indexOf( " " );           
                String thirdToken = str.substring( 0, length );
                str = str.substring( length );
                Float offset = new Float( thirdToken );
                
                ////////////////////////////////////////////////////////////////
                ii = 0;
                while ( str.charAt( ii++ ) == ' ' )
                {
                    // do nothing
                }
                str = str.substring( ii - 1 );
                
                length = str.indexOf( " " );
                String fourthToken = str.substring( 0, length );
                str = str.substring( length );
                Float gauss = new Float( fourthToken );
                
                ////////////////////////////////////////////////////////////////
                
                ii = 0;
                while ( str.charAt( ii++ ) == ' ' )
                {
                    // do nothing
                }
                str = str.substring( ii - 1 );
                length = str.indexOf( " " );
                String fifthToken = str.substring( 0, length );
                str = str.substring( length );
                Float phase = new Float( fifthToken );
                
                ////////////////////////////////////////////////////////////////
                
                str = str.trim( );                
                Float lorentz = new Float( str );

//                System.out.println( area.toString( ) + " " + 
//                                    offset.toString( ) + " " + 
//                                    gauss.toString( ) + " " + 
//                                    phase.toString( ) + " " + 
//                                    lorentz.toString( ) );
 
                fitPar.setArea( area.floatValue( ) / (float)MRSignalProperties.getInstance( ).getNumberOfSamplePoints( ) );

                if ( ( isObjectMRLineModelParameters == true ) )
                {
                    fitPar.setOffset( offset.floatValue( ) - ( (float)MRSignalProperties.getInstance( ).getNumberOfSamplePoints( ) ) / 2.0f );
                }
                else
                {
                    fitPar.setOffset( offset.floatValue( ) );
                }

                fitPar.setGaussianWidth( gauss.floatValue( ) );
                
                fitPar.setPhase( phase.floatValue( ) );
                
                fitPar.setLorentzianWidth( lorentz.floatValue( ) );
            }
            reader.close( );
        }
        catch ( IOException e )
        {
            System.out.println( "IOException in reading >>filename<<_verbose.par" );
        }
    }
    
    /**
     * Writes the currently loaded measured spectrum of which a spectral model 
     * has to be made to the same directory as the starting value file.
     */
    private String writeSpectrumToFitToTdfdfitFormat( )
    {
        return m_measuredSpectralResponse.writeAsTdfdfitFileForFitting( getModelSpectrumName( ) + ".bin" );
    }
    
    /**
     * Creates starting value files on disk; in order that the fitting 
     * application TDFDFit can do the fitting in a separate process; and 
     * further returns the filename of the parameter file.
     * @return 
     */
    public String createTDFDFitStartingValueFiles( )
    {
        
        TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
        fittingDefaults.load( );
        
        String path = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH ) + System.getProperty( "file.separator" );
        String modelName = fittingDefaults.getProperty( TDFDFittingDefaults.DEFAULT_TDFDFIT_FIT_MODEL_NAME );
        
        // Remove the file extension
        while ( modelName.contains( "." ) == true )
        {
            int index = modelName.lastIndexOf( "." );
            modelName = modelName.substring(0, index );
        }
        
        String modelPathAndFile = path + modelName + ".par";
        // System.out.println( "Output model file : " + modelPathAndFile );
        try
        {
            TextFileWriter writer = new TextFileWriter( new FileWriter( modelPathAndFile ) );
            writer.setNumberFormat( "0.00000000E00" );
            
            writeStartingValueFileHeader( writer );
            writeParametricPatternFileReferences( writer );
            writeNumericPatternFileReferences( writer );
            writeFittingStrategyInformation( writer );
            
            // first write the starting values of the simple lines
            writeSimpleLineInformation( writer ); 
            // secondly write the starting value of the parametric patterns
            writeParametricPatternInformation( writer );
            // thirdly write the starting values of the numeric patterns
            writeNumericPatternInformation( writer );
            // fourthly, write metabolite prior knowledge
            writeMetaboliteInformation( writer );
            // write the complete prior knowledge block into the file
            writePriorKnowledgePartStartingValueFile( writer );
            
            // write lineshape coefficients
            writeLineShapeCoefficients( writer );
            // write the footer of the file
            writeStartingValueFileFooter( writer );
            
            writer.flush( );
            writer.close( );
           
        }
        catch( IOException e )
        {
            // not handled yet
        }
        
        return modelPathAndFile;
    }
    
    
    /**
     * Creates starting value files on disk; in order that the fitting 
     * application TDFDFit can do the fitting in a separate process; and 
     * further returns the filename of the parameter file.
     * @param ID
     * @return 
     */
    public String createTDFDFitStartingValueFiles( int ID )
    {
        TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
        fittingDefaults.load( );
        
        String path = fittingDefaults.getProperty( TDFDFittingDefaults.TDFDFIT_INSTALLATION_PATH );
        String modelName = getModelSpectrumName( );
        
        // Remove the file extension
        while ( modelName.contains( "." ) == true )
        {
            int index = modelName.lastIndexOf( "." );
            modelName = modelName.substring(0, index );
        }
        
        String modelPathAndFile;
                
        if ( path.endsWith( System.getProperty( "file.separator" ) ) == true )
        {
            modelPathAndFile = path + modelName + "_" + ID + ".par";
        }
        else
        {
            modelPathAndFile = path + System.getProperty( "file.separator" ) + modelName + "_" + ID + ".par";
        }
        
        // System.out.println( "Output model file : " + modelPathAndFile );
        try
        {
            TextFileWriter writer = new TextFileWriter( new FileWriter( modelPathAndFile ) );
            writer.setNumberFormat( "0.00000000E00" );
            
            writeStartingValueFileHeader( writer );
            writeParametricPatternFileReferences( writer );
            writeNumericPatternFileReferences( writer );
            writeFittingStrategyInformation( writer );
            
            // first write the starting values of the simple lines
            writeSimpleLineInformation( writer ); 
            // secondly write the starting value of the parametric patterns
            writeParametricPatternInformation( writer );
            // thirdly write the starting values of the numeric patterns
            writeNumericPatternInformation( writer );
            // fourthly, write metabolite prior knowledge
            writeMetaboliteInformation( writer );
            // write the complete prior knowledge block into the file
            writePriorKnowledgePartStartingValueFile( writer );
            
            // write lineshape coefficients
            writeLineShapeCoefficients( writer );
            // write the footer of the file
            writeStartingValueFileFooter( writer );
            
            writer.flush( );
            writer.close( );
           
        }
        catch( IOException e )
        {
            // not handled yet
        }
        
        return modelPathAndFile;
    }
    
    
    
    /**
     * Creates starting value files on disk; in order that the fitting application
     * TDFDFit can do the fitting in a separate process based on this starting values.
     * @param spectrumName
     */
    public void createStartingValueFile( String spectrumName )
    {
        TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
        fittingDefaults.load( );
        
        String path = fittingDefaults.getProperty( "Fit project database path" );
        String modelName = spectrumName;
        String outStr = path + System.getProperty( "file.separator" ) + modelName + ".par";
        outStr = outStr.trim( );
        // System.out.println( "Output model file : " + outStr );
        try
        {
            TextFileWriter writer = new TextFileWriter( new FileWriter( outStr ) );
            writer.setNumberFormat( "0.00000000E00" );
            
            writeStartingValueFileHeader( writer );
            writeParametricPatternFileReferences( writer );
            writeNumericPatternFileReferences( writer );
            writeFittingStrategyInformation( writer );
            
            // first write the starting values of the simple lines
            writeSimpleLineInformation( writer ); 
            // secondly write the starting value of the parametric patterns
            writeParametricPatternInformation( writer );
            // thirdly write the starting values of the numeric patterns
            writeNumericPatternInformation( writer );
            // fourthly, write metabolite prior knowledge
            writeMetaboliteInformation( writer );
            // write the complete prior knowledge block into the file
            writePriorKnowledgePartStartingValueFile( writer );
            
            // write lineshape coefficients
            writeLineShapeCoefficients( writer );
            // write the footer of the file
            writeStartingValueFileFooter( writer );
            
            writer.flush( );
            writer.close( );
            
            // write the parametric pattern starting value files
            writeParametricStartingValueFiles( );
        }
        catch( IOException e )
        {
            // not handled yet
        }
    }
    
    
    /**
     * Creates starting value files on disk to a given path and filename; 
     * in order that the fitting application
     * TDFDFit can do the fitting in a separate process.
     * @param path
     */
    public void createStartingValueFileToPath( String path )
    {

        try
        {
            TextFileWriter writer = new TextFileWriter( new FileWriter( path ) );
            writer.setNumberFormat( "0.00000000E00" );
            
            writeStartingValueFileHeader( writer );
            writeParametricPatternFileReferences( writer );
            writeNumericPatternFileReferences( writer );
            writeFittingStrategyInformation( writer );
            
            // first write the starting values of the simple lines
            writeSimpleLineInformation( writer ); 
            // secondly write the starting value of the parametric patterns
            writeParametricPatternInformation( writer );
            // thirdly write the starting values of the numeric patterns
            writeNumericPatternInformation( writer );
            // fourthly, write metabolite prior knowledge
            writeMetaboliteInformation( writer );
            // write the complete prior knowledge block into the file
            writePriorKnowledgePartStartingValueFile( writer );
            
            // write lineshape coefficients
            writeLineShapeCoefficients( writer );
            // write the footer of the file
            writeStartingValueFileFooter( writer );
            
            writer.flush( );
            writer.close( );
            
            // write the parametric pattern starting value files
            writeParametricStartingValueFiles( );
        }
        catch( IOException e )
        {
            // not handled yet
        }
    }
    
    /**
     * The method writes the first info block of a TDFDFit readable
     * starting value file.
     */ 
    private void writeStartingValueFileHeader( TextFileWriter bw ) throws IOException
    {
        // Get the information that controls the runtime behaviour of
        TDFDFittingDefaults fDefs = TDFDFittingDefaults.getInstance( );
        fDefs.load( );
        
        if ( fDefs.getProperty( TDFDFittingDefaults.FittingTolerance ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.FittingTolerance, "" + 1.0e-7 ); 
             fDefs.save( );
        }
        if ( fDefs.getProperty( TDFDFittingDefaults.FittingMode ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.FittingMode, "" + 1 ); 
             fDefs.save( );
        }
        if ( fDefs.getProperty( TDFDFittingDefaults.TDFDFitVerbosity ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.TDFDFitVerbosity, "" + 0 ); 
             fDefs.save( );
        }
        if ( fDefs.getProperty( TDFDFittingDefaults.ShowInfoOfComponent ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.ShowInfoOfComponent, "" + 1 ); 
             fDefs.save( );
        }
        if ( fDefs.getProperty( TDFDFittingDefaults.CalculateErrorsAfterFit ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.CalculateErrorsAfterFit, "" + 0 ); 
             fDefs.save( );
        }
        
        if ( fDefs.getProperty( TDFDFittingDefaults.NumberOfLineShapeFourierCoefficientsToFit ) == null )
        {
             fDefs.setProperty( TDFDFittingDefaults.NumberOfLineShapeFourierCoefficientsToFit, "" + 10 ); 
             fDefs.save( );
        }
        
        double fTol = ( new Double( "" + fDefs.getProperty( TDFDFittingDefaults.FittingTolerance ) ) );
        int fitMode = ( new Integer( "" + fDefs.getProperty( TDFDFittingDefaults.FittingMode ) ) );
        int verbose = ( new Integer( "" + fDefs.getProperty( TDFDFittingDefaults.TDFDFitVerbosity ) ) );
        int showLine = ( new Integer( "" + fDefs.getProperty( TDFDFittingDefaults.ShowInfoOfComponent ) ) );
        
        int calcError = 0;
        if ( ( (String)fDefs.getProperty( TDFDFittingDefaults.CalculateErrorsAfterFit ) ).compareTo( "yes" ) == 0 )
        {
            calcError = 1;
        }
        int numberOfLineshapeCoefficients = ( new Integer( "" + fDefs.getProperty( TDFDFittingDefaults.NumberOfLineShapeFourierCoefficientsToFit ) ) );
        
        
        bw.writeStringLn( "version_3" );
        bw.writeString( "n_pnts " );
        
        // verify if zero filling is performed!!
        int points = MRSignalProperties.getInstance( ).getNumberOfSamplePoints( );
        
        bw.writeIntegerLn( points );
        
        int startPointNumberNoiseInterval = MRSignalProperties.getInstance( ).getMinimumNoiseIntervalPointNumber();
        
        bw.writeString( "noistr " );
        bw.writeIntegerLn( startPointNumberNoiseInterval );
        
        int stopPointNumberNoiseInterval = MRSignalProperties.getInstance( ).getMaximumNoiseIntervalPointNumber();
        // System.out.println( "stopPointNumberNoiseInterval = " + stopPointNumberNoiseInterval );
        bw.writeString( "noistp " );
        bw.writeIntegerLn( stopPointNumberNoiseInterval );
        
        bw.writeString( "nLines " );
        bw.writeIntegerLn( m_simpleLineList.size( ) );
        bw.writeString( "lsfcfs " );
        bw.writeIntegerLn( numberOfLineshapeCoefficients );
        bw.writeString( "phi__1 " );
        bw.writeFloatLn( m_MRSignalProperties.getFirstOrderPhase( ) );
        bw.writeString( "Hz_p_p " );
        
        // Note: TDFDFit internally the application works with points, not with
        // Hz or even ppm. Therefore Hz_pp is always set to 1.0f!
        //bw.writeFloatLn( m_MRSignalProperties.getFrequencyScale( ) );
        bw.writeFloatLn( 1.0f );
        
        bw.writeString( "f__tol " );
        bw.writeDoubleLn( fTol );
        bw.writeString( "fit_mo " );
        bw.writeIntegerLn( fitMode );
        bw.writeString( "verbos " );
        bw.writeIntegerLn( verbose );
        bw.writeString( "showli " );
        bw.writeIntegerLn( showLine );
        bw.writeString( "clcerr " );
        bw.writeIntegerLn( calcError );
        bw.writeString( "sg_idl " );
        bw.writeStringLn( "none" );
        
        // Ranges should be 2 in case that randomfit is used
        bw.writeString( "ranges " );
        bw.writeIntegerLn( 3 );
        bw.writeString( "ignore " );
        MRFitStrategy fitStrat = MRFitStrategy.getInstance( );
        bw.writeIntegerLn( fitStrat.getFitStrategyStep( 0 ).getIgnoreFirstSamplePointsNumber( ) );
        bw.writeString( "pattrn " );
        bw.writeIntegerLn( m_numericPatternList.size( ) + m_parametricPatternList.size( ) );
    }
    
    /**
     * Writes the prior knowledge information to the starting value file
     * For the time being these are hard coded values; in future it should be
     * written user defined values via special GUI
     * @param bw
     * @throws IOException 
     */
    private void writeFittingStrategyInformation( TextFileWriter bw ) throws IOException
    {
        bw.writeStringLn( "   algor trunc  ampli  frequ  width  phase  voigt  lsfft  wtype  convo  niter  cmplx   n_int   min   max " );

        // bw.writeStringLn( "@    2   1024      1      1      1      1      1      0      1      0    100      2      1    450    750  " );    
        
        MRFitStrategy fitStrat = MRFitStrategy.getInstance( );
        for ( int i = 0; i < fitStrat.getNumberOfFitStrategySteps( ); i++ )
        {
            MRFitStrategyStep fitStratStep = fitStrat.getFitStrategyStep( i );
            bw.writeString( "@    " );
            bw.writeInteger( fitStratStep.getFitAlgorithType( ) );
            bw.writeString( "   " );
            bw.writeInteger( fitStratStep.getTruncateAfterPointNumber( ) );
            bw.writeString( "      " );
            if ( fitStratStep.getFitArea( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }
            
            if ( fitStratStep.getFitOffset( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }
            
            if ( fitStratStep.getFitGauss( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }
            
            if ( fitStratStep.getFitPhase( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }

            if ( fitStratStep.getFitLorentz( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }
            
            if ( fitStratStep.getApplyLineShapeFit( ) == true )
            {
                bw.writeString( "1      " );
            }
            else
            {
                bw.writeString( "0      " );
            }
            
            bw.writeInteger( fitStratStep.getLeastSquaresWeightingType( ) );
            bw.writeString( "      " );
            
            if ( fitStratStep.getApplyConvolution( ) == true )
            {
                bw.writeString( "1" );
                bw.writeString( "    " );
            }
            else
            {
                bw.writeString( "0" );
                bw.writeString( "    " );
            }
            
            bw.writeInteger( fitStratStep.getNumberOfIterations( ) );
            bw.writeString( "      " );
            
            bw.writeInteger( fitStratStep.getFittingType( ) );
            bw.writeString( "      " );
            
            bw.writeInteger( fitStratStep.getNumberOfRanges( ) );
            bw.writeString( "     " );
            
            for ( int j = 0; j < fitStratStep.getNumberOfRanges( ); j++ )
            {
                bw.writeInteger( fitStratStep.getFittingRangeMinimum( j ) );
                bw.writeString( "   " );
                bw.writeInteger( fitStratStep.getFittingRangeMaximum( j ) );
                bw.writeString( "    " );
            }
            
            bw.writeStringLn( "" );
        }
    }
    
    /**
     * Writes the starting value information of simple lines to the 
     * starting value file.
     * @param bw
     * @throws IOException 
     */
    private void writeSimpleLineInformation( TextFileWriter bw ) throws IOException
    {
        bw.writeStringLn( "number       area          frequency      gausswidth        phase        lorentzwidth" );
        
        Iterator it = m_simpleLineList.iterator( );
        int cntr = 1;
        
        MRSignalProperties mrCond = MRSignalProperties.getInstance( );
        float offsetShift = mrCond.getNumberOfSamplePoints( ) / 2;
              
        // System.out.println( "Offset shift " +  offsetShift );
        
        while ( it.hasNext( ) )
        {
            bw.writeInteger( cntr++ );
            bw.writeTab( );
            bw.writeTab( );
            MRLineModelParameters linPar = (MRLineModelParameters)it.next( );
            bw.writeFloat( linPar.getArea( ) );
            bw.writeTab( );
            bw.writeFloat( linPar.getOffset( ) + offsetShift );
            bw.writeTab( );
            bw.writeFloat( linPar.getGaussianWidth( ) );
            bw.writeTab( );
            bw.writeFloat( linPar.getPhase( ) );
            bw.writeTab( );
            bw.writeFloatLn( linPar.getLorentzianLineWidth( ) );
        }
    }

   
    /**
     * Writes the starting value information of simple lines to the 
     * starting value file.
     * @param bw
     * @throws IOException 
     */
    private void writeParametricPatternInformation( TextFileWriter bw ) throws IOException
    {
        Iterator it = m_parametricPatternList.iterator( );
        int cntr = m_simpleLineList.size( ) + 1 ;
        
        MRSignalProperties mrCond = MRSignalProperties.getInstance( );
        float offsetShift = mrCond.getNumberOfSamplePoints( ) / 2;
        
        while ( it.hasNext( ) )
        {
            bw.writeInteger( cntr++ );
            bw.writeTab( );
            bw.writeTab( );
            MRParametricPattern parPat = (MRParametricPattern)it.next( );
            MRFittingParameters fitPar = parPat.getFittingParameters( );
            bw.writeFloat( fitPar.getArea( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getOffset( ) /*+ offsetShift*/ );
            bw.writeTab( );
            bw.writeFloat( fitPar.getGaussianWidth( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getPhase( ) );
            bw.writeTab( );
            bw.writeFloatLn( fitPar.getLorentzianLineWidth( ) );
        }
    }


    /**
     * Writes the list of parametric patterns to the global starting file. 
     * @param bw
     * @throws IOException 
     */
    private void writeParametricPatternFileReferences( TextFileWriter bw ) throws IOException
    {
        Iterator it = this.m_parametricPatternList.iterator( );
        
        // Iterate over all MRParametricPattern objects
        while ( it.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)it.next( );
                
            TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
            fittingDefaults.load( );
            
            String modelName = fittingDefaults.getProperty( TDFDFittingDefaults.DEFAULT_TDFDFIT_FIT_MODEL_NAME );
            String path = fittingDefaults.getProperty( TDFDFittingDefaults.DEFAULT_USED_TDFDFIT_MODEL_PATH );       
            String parPatFileName = path + System.getProperty( "file.separator" ) + "Parametric_Patterns" + System.getProperty( "file.separator" ) + parPat.getName( ) + ".pat";
            bw.writeStringLn( parPatFileName );
            // Now write the .pat file itself to disk
            parPat.writeParPatternAsTDFDFitDotPatFile( parPatFileName );
        }
    }
    
    
    /**
     * Writes the list of parametric patterns to the global starting file.
     */
    private void writeMetabolitePatternFileReferences( TextFileWriter bw ) throws IOException
    {
        Iterator it = m_metaboliteList.iterator( );
        
        // Iterate over all MRParametricPattern objects
        while ( it.hasNext( ) )
        {
            MRMetaboliteModel metab = (MRMetaboliteModel)it.next( );
                
            TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
            fittingDefaults.load( );
            
            String path = fittingDefaults.getProperty( "Last used TDFDFit model path" );       
            String metabFileName = path + System.getProperty( "file.separator" ) + "Metabolites" + System.getProperty( "file.separator" ) + metab.getName( ) + ".pat";
            bw.writeStringLn( metabFileName );
            // Now write the .pat file itself to disk
            metab.writeMetaboliteAsTDFDFitDotPatFile( metabFileName );
        }
    }
    
    /**
     * Writes the list of numeric pattern file references to the 
     * global starting file.
     */
    private void writeNumericPatternFileReferences( TextFileWriter bw ) throws IOException
    {
        Iterator it = m_numericPatternList.iterator( );
            
        // Iterate over all MRNumericPattern objects
        while ( it.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)it.next( );
            
            TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
            fittingDefaults.load( );
            
            String path = fittingDefaults.getProperty( TDFDFittingDefaults.DEFAULT_USED_TDFDFIT_MODEL_PATH );
            String numPatFileName = path + System.getProperty( "file.separator" ) + "Numeric_Patterns" + System.getProperty( "file.separator" ) + numPat.getName( );
            if ( numPatFileName.contains( ".bin" ) == false )
            {
                numPatFileName = numPatFileName + ".bin";
            }
            
            bw.writeStringLn( numPatFileName );
            
            // Now write the numeric pattern as a bin file to disk ..
            numPat.writeNumericPatternAsTDFDFitDotBinFile( numPatFileName, true );
             
        }
    }
    
    /**
     * Writes the starting value information of parametric patterns 
     * to the starting value.
     */     
    private void writeParametricStartingValueFiles( )
    {
        // For each parametric pattern a special starting valuefile should be 
        // written which is referenced to in the global starting value file
        try
        {
            Iterator it = m_parametricPatternList.iterator( );
            
            // Iterate over all MRParametricPattern objects
            while ( it.hasNext( ) )
            {
                MRParametricPattern parPat = (MRParametricPattern) it.next( );
                
                TDFDFittingDefaults fittingDefaults = TDFDFittingDefaults.getInstance( );
                fittingDefaults.load( );
                
                String path = fittingDefaults.getProperty( "Fit project path" );
                
                String parPatFileName = path + System.getProperty( "file.separator" ) + parPat.getName( ) + ".pat";
                
                TextFileWriter writer = new TextFileWriter( new FileWriter( parPatFileName ) );
                writer.setNumberFormat( "0.00000000E00" );
                writer.writeStringLn( "start_pattern" );
                writer.writeString( "number_of_lines " );

                LinkedList lineList = parPat.getSimpleLineList( );

                writer.writeIntegerLn( lineList.size( ) );
                
                writer.writeStringLn( "parameters" );
                
                writer.writeStringLn( "name   line_type     area         frequency     gauss_width       zero_phase    lorentz_width" );
                
                Iterator iter = lineList.iterator( );
                int cntr = 1;
                
                MRSignalProperties mrCond = MRSignalProperties.getInstance( );
                float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
                
                while ( iter.hasNext( ) )
                {
                    MRLineModelParameters linPar = (MRLineModelParameters)iter.next( );
                    writer.writeString( "Line_" + cntr++ );
                    writer.writeTab( );
                    writer.writeString( "voigt" );
                    writer.writeTab( );
                    writer.writeFloat( linPar.getArea( ) );
                    writer.writeTab( );
                    writer.writeFloat( linPar.getOffset( ) + offsetShift  );
                    writer.writeTab( );
                    writer.writeFloat( linPar.getGaussianWidth( ) );
                    writer.writeTab( );
                    writer.writeFloat( linPar.getPhase( ) );
                    writer.writeTab( );
                    writer.writeFloatLn( linPar.getLorentzianLineWidth( ) );
                }
                writer.writeStringLn( "end_of_pattern" );
                // flush the data
                writer.flush( );
                // close the file
                writer.close( );
            }
        }
        catch ( IOException e )
        {
            // not handled yet
        }
    }
    
    /**
     * Writes the starting value information of numeric patterns 
     * to the starting value.
     * @param bw
     * @throws IOException 
     */
    private void writeNumericPatternInformation( TextFileWriter bw ) throws IOException
    {
        Iterator it = m_numericPatternList.iterator( );
        int cntr = m_simpleLineList.size( ) + m_parametricPatternList.size( ) + 1;
        
        MRSignalProperties mrCond = MRSignalProperties.getInstance( );
        float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
        
        while ( it.hasNext( ) )
        {
            bw.writeInteger( cntr++ );
            bw.writeTab( );
            bw.writeTab( );
            
            MRNumericPattern numPat = (MRNumericPattern)it.next( );
            MRFittingParameters fitPar = numPat.getFittingParameters( );
            bw.writeFloat( fitPar.getArea( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getOffset( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getGaussianWidth( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getPhase( ) );
            bw.writeTab( );
            bw.writeFloatLn( fitPar.getLorentzianLineWidth( ) );
        }
    }
    
    /**
     * Writes the metabolite starting value information to the major 
     * starting value file.
     */
    private void writeMetaboliteInformation( TextFileWriter bw ) throws IOException
    {
        Iterator it = m_metaboliteList.iterator( );
        int cntr = m_simpleLineList.size( ) + m_parametricPatternList.size( ) + m_numericPatternList.size( ) + 1;
        
        MRSignalProperties mrCond = MRSignalProperties.getInstance( );
        float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
        
        while ( it.hasNext( ) )
        {
            bw.writeInteger( cntr++ );
            bw.writeTab( );
            bw.writeTab( );
            MRMetaboliteModel metMod = (MRMetaboliteModel) it.next( );
            MRFittingParameters fitPar = metMod.getFittingParameters( );

            bw.writeFloat( fitPar.getArea( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getOffset( ) + offsetShift );
            bw.writeTab( );
            bw.writeFloat( fitPar.getGaussianWidth( ) );
            bw.writeTab( );
            bw.writeFloat( fitPar.getPhase( ) );
            bw.writeTab( );
            bw.writeFloatLn( fitPar.getLorentzianLineWidth( ) );
        }
    }
    
    /**
     * Writes the line shape coefficients into the .par master 
     * starting value file.
     */ 
    private void writeLineShapeCoefficients( TextFileWriter bw ) throws IOException
    {
        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        int numberOfFourierCoeffs = new Integer( "" + defs.get( TDFDFittingDefaults.NumberOfLineShapeFourierCoefficientsToFit ) );
        bw.writeStringLn( "lsf_coefs" );
        
        int mode = new Integer( "" + defs.get( TDFDFittingDefaults.FittingMode ) );
        
        if ( mode == 3 )
        {
            for ( int i = 0; i < 2 * numberOfFourierCoeffs; i++ )
            {
                if ( i == numberOfFourierCoeffs )
                {
                    bw.writeDoubleLn( 1.0 );
                }
                else
                {
                    bw.writeDoubleLn( 0.0 );
                }
            }
        }
         bw.writeStringLn( "end_of_lsf" );
    }
    
    /**
     * Writes the line shape coefficients into the .par master 
     * starting value file.
     */ 
    private void writeStartingValueFileFooter( TextFileWriter bw ) throws IOException
    {
        bw.writeStringLn( "-----------------------------------------------------------------------" );
        bw.writeStringLn( " This is a generated starting value file: do not edit !!! " );
        
        Calendar cal = new GregorianCalendar( );
        Date curDate = cal.getTime( );
        
        
        bw.writeStringLn( " File created at : " + curDate.toString( ) );
        bw.writeStringLn( " User            : " + System.getProperty( "user.name" ) );
        
        String ipAddrStr = "";
        try 
        {
            InetAddress addr = InetAddress.getLocalHost( );
            
            // Get IP Address
            byte[] ipAddr = addr.getAddress( );
    
            // Get hostname
            ipAddrStr = addr.toString( );
         } 
         catch ( UnknownHostException e ) 
         {
             // not handled yet
         }

        bw.writeStringLn( " Host            : " + ipAddrStr   );
        bw.writeStringLn( "-----------------------------------------------------------------------" );
        
    }
    
    
    /**
     * Creates lists of lines, parametric patterns, 
     * numeric patterns and metabolites.
     */
    private void createSpectralComponentNumberToNameMap( )
    {
        
        // Clear the hashmap and determine it again
        if ( m_numberToComponentNameMap != null )
        {    
            m_numberToComponentNameMap.clear( );
            m_numberToComponentNameMap = null;
        }
        
        m_numberToComponentNameMap = new HashMap<String,Integer>( );
        
        Iterator it = m_simpleLineList.iterator( );
        int cntr = 1;
        while ( it.hasNext( ) )
        {
            MRLineModelParameters linPar = (MRLineModelParameters)it.next( );
            String lineName = linPar.getName( );
            m_numberToComponentNameMap.put( lineName, new Integer( cntr++ ) );
            // System.out.println( "added to map: " + lineName + " - " + (cntr-1) );
        }
        
        Iterator it2 = this.m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)it2.next( );
            m_numberToComponentNameMap.put( parPat.getFittingParameters( ).getName( ), new Integer( cntr++ ) );
            // System.out.println( "added to map: " + parPat.getFittingParameters( ).getName( ) + " - " +(cntr-1) );
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)it3.next( );
            m_numberToComponentNameMap.put( numPat.getFittingParameters( ).getName( ), new Integer( cntr++ ) );
            // System.out.println( "added to map: " + numPat.getFittingParameters( ).getName( ) + " - " + (cntr-1) );
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRMetaboliteModel metMod = (MRMetaboliteModel)it4.next( );
            m_numberToComponentNameMap.put( metMod.getFittingParameters( ).getName( ), new Integer( cntr++ ) );
            //System.out.println( "added to map: " + metMod.getFittingParameters( ).getName( ) + " - " + (cntr-1) );
        }   
    }
    
    /**
     * This method writes the prior knowledge information to file in "classic"
     * TDFDFit format (so generates the prior knowledge part of the 
     * starting value file.
     */
     private void writePriorKnowledgePartStartingValueFile( TextFileWriter bw )
     {
        
        // first update the hastable containing the number to spectral component name map
        createSpectralComponentNumberToNameMap( );
        
        // write the prior of the simple line part
        writeMRLineModelParametersPriorKnowledge( bw );
        
        // write the prior of the parametric pattern part
        writeMRParametricPatternPriorKnowledge( bw );
        
        // write the prior of the numeric pattern part
        writeMRNumericPatternPriorKnowledge( bw );
        
        // write the prior of the metabolite part
        writeMRMetabolitePriorKnowledge( bw );

     }
     
    /**
     * Writes the simple line part from the prior knowledge portion of the
     * classical TDFDFit starting value file.
     */
    private void writeMRLineModelParametersPriorKnowledge( TextFileWriter bw )
    {
 
        Iterator it = m_simpleLineList.iterator( );
        int cntr = 1;
        MRSignalProperties mrCond = MRSignalProperties.getInstance();
        // float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
        float offsetShift = mrCond.getNumberOfSamplePoints( ) / 2;
        try
        {
            while ( it.hasNext( ) )
            {
                bw.writeString( "#" );
                bw.writeIntegerLn( cntr++ );
                bw.writeStringLn( "lsffit 0" );
            
                MRLineModelParameters linPar = (MRLineModelParameters)it.next( );
                // default model is always no Voigt; with this type you can make also pure Gaussian and
                // also pure Lorentzian lines : therefore linetype is hard coded 3 
                bw.writeString( "3" );
                bw.writeTab( );
                if ( linPar.getAreaReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( linPar.getAreaReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( linPar.getArea( ) );
                bw.writeTab( );
                bw.writeFloat( linPar.getMinimumArea( ) );
                bw.writeTab( );
                bw.writeFloat( linPar.getMaximumArea( ) );
                bw.newLine( );
                
                // also for offset frequency
                bw.writeString( "3" );
                bw.writeTab( );
                if ( linPar.getOffsetReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( linPar.getOffsetReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( linPar.getOffset( ) + offsetShift );
                bw.writeTab( );
                bw.writeFloat( linPar.getMinimumOffset( ) + offsetShift );
                bw.writeTab( );
                bw.writeFloatLn( linPar.getMaximumOffset( ) + offsetShift );
                
                // also for the gaussian width
                bw.writeString( "3" );
                bw.writeTab( );
                if ( linPar.getGaussianWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( linPar.getGaussianWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( linPar.getGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloat( linPar.getMinimumGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( linPar.getMaximumGaussianWidth( ) );
                
                // also for the zero order phase
                bw.writeString( "3" );
                bw.writeTab( );
                if ( linPar.getPhaseReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( linPar.getPhaseReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( linPar.getPhase( ) );
                bw.writeTab( );
                bw.writeFloat( linPar.getMinimumPhase( ) );
                bw.writeTab( );
                bw.writeFloatLn( linPar.getMaximumPhase( ) );
                
                // also for the lorentzian width
                bw.writeString( "3" );
                bw.writeTab( );
                if ( linPar.getLorentzianLineWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( linPar.getLorentzianLineWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                
                bw.writeTab( );
                bw.writeFloat( linPar.getLorentzianLineWidth( ) );
                bw.writeTab( );
                bw.writeFloat( linPar.getMinimumLorentzianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( linPar.getMaximumLorentzianWidth( ) );
            }  
        }
        catch ( IOException e )
        {
            //  not caught yet
        }
    }
    
    /**
     * Writes the parametric pattern part from the prior knowledge 
     * portion of the classical TDFDFit starting value file.
     */ 
    private void writeMRParametricPatternPriorKnowledge( TextFileWriter bw )
    {
        
        Iterator it = m_parametricPatternList.iterator( );
        
        int cntr = m_simpleLineList.size( ) + 1;
                
        try
        {
            while ( it.hasNext( ) )
            {
                bw.writeString( "#" );
                bw.writeIntegerLn( cntr++ );
                bw.writeStringLn( "lsffit 0" );
            
                MRFittingParameters fitPar = ( (MRParametricPattern)it.next( ) ).getFittingParameters( );

                // default model is always no Voigt; with this type you can make also pure Gaussian and
                // also pure Lorentzian lines : therefore linetype is hard coded 3 
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getAreaReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getAreaReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMaximumArea( ) );
                bw.newLine( );
                
                // also for offset frequency
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getOffsetReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getOffsetReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getOffset( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumOffset( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumOffset( ) );
                
                // also for the gaussian width
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getGaussianWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getGaussianWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumGaussianWidth( ) );
                
                // also for the zero order phase
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getPhaseReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getPhaseReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getPhase( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumPhase( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumPhase( ) );
                
                // also for the lorentzian width
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getLorentzianLineWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getLorentzianLineWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                
                bw.writeTab( );
                bw.writeFloat( fitPar.getLorentzianLineWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumLorentzianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumLorentzianWidth( ) );
            }
        }
        catch ( IOException e )
        {
            // not caught yet
        }
    }
    
    
    /**
     * Writes the numeric pattern part from the prior knowledge portion of the
     * classical TDFDFit starting value file.
     */     
    private void writeMRNumericPatternPriorKnowledge( TextFileWriter bw  )
    {
        Iterator it = m_numericPatternList.iterator( );
        
        int cntr = m_simpleLineList.size( ) + m_parametricPatternList.size( ) + 1;
        
        try
        {
            while ( it.hasNext( ) )
            {
                bw.writeString( "#" );
                bw.writeIntegerLn( cntr++ );
                bw.writeStringLn( "lsffit 0" );
            
                MRFittingParameters fitPar = ( (MRNumericPattern)it.next( ) ).getFittingParameters( );

                // default model is always no Voigt; with this type you can make also pure Gaussian and
                // also pure Lorentzian lines : therefore linetype is hard coded 3 
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getAreaReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getAreaReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMaximumArea( ) );
                bw.newLine( );
                
                // also for offset frequency
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getOffsetReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getOffsetReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getOffset( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumOffset( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumOffset( ) );
                
                // also for the gaussian width
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getGaussianWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getGaussianWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumGaussianWidth( ) );
                
                // also for the zero order phase
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getPhaseReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getPhaseReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getPhase( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumPhase( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumPhase( ) );
                
                // also for the lorentzian width
                bw.writeString( "4" );
                bw.writeTab( );
                if ( fitPar.getLorentzianLineWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getLorentzianLineWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                
                bw.writeTab( );
                bw.writeFloat( fitPar.getLorentzianLineWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumLorentzianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumLorentzianWidth( ) );
            }
        }
        catch ( IOException e )
        {
            //  not caught yet
        }        
    }
    
    /**
     * Writes the numeric pattern part from the prior knowledge portion of the
     * classical TDFDFit starting value file.
     */     
    private void writeMRMetabolitePriorKnowledge( TextFileWriter bw  )
    {
        Iterator it = m_metaboliteList.iterator( );
        
        int cntr = m_simpleLineList.size( ) + m_parametricPatternList.size( ) + 
                   m_numericPatternList.size( ) + 1;
        
        MRSignalProperties mrCond = MRSignalProperties.getInstance( );
        float offsetShift = mrCond.getFrequencyScale( ) * mrCond.getNumberOfSamplePoints( ) / 2;
                
        // System.out.println( "Offset shift " +  offsetShift );
        
        try
        {
            while ( it.hasNext( ) )
            {
                bw.writeString( "#" );
                bw.writeIntegerLn( cntr++ );
                bw.writeStringLn( "lsffit 0" );
            
                MRFittingParameters fitPar = ( (MRMetaboliteModel)it.next( ) ).getFittingParameters( );

                // default model is always no Voigt; with this type you can make also pure Gaussian and
                // also pure Lorentzian lines : therefore linetype is hard coded 3 
                bw.writeString( "3" );
                bw.writeTab( );
                if ( fitPar.getAreaReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getAreaReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumArea( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMaximumArea( ) );
                bw.newLine( );
                
                // also for offset frequency
                bw.writeString( "3" );
                bw.writeTab( );
                if ( fitPar.getOffsetReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getOffsetReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getOffset( ) + offsetShift );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumOffset( ) + offsetShift );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumOffset( ) + offsetShift );
                
                // also for the gaussian width
                bw.writeString( "3" );
                bw.writeTab( );
                if ( fitPar.getGaussianWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getGaussianWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumGaussianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumGaussianWidth( ) );
                
                // also for the zero order phase
                bw.writeString( "3" );
                bw.writeTab( );
                if ( fitPar.getPhaseReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getPhaseReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                bw.writeTab( );
                bw.writeFloat( fitPar.getPhase( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumPhase( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumPhase( ) );
                
                // also for the lorentzian width
                bw.writeString( "3" );
                bw.writeTab( );
                if ( fitPar.getLorentzianLineWidthReference( ).compareTo( "Free" ) == 0 )
                {
                    bw.writeString( "0" );
                }
                else
                {
                    Integer refInt = m_numberToComponentNameMap.get( fitPar.getLorentzianLineWidthReference( ) );
                    bw.writeInteger( refInt.intValue( ) );
                }
                
                bw.writeTab( );
                bw.writeFloat( fitPar.getLorentzianLineWidth( ) );
                bw.writeTab( );
                bw.writeFloat( fitPar.getMinimumLorentzianWidth( ) );
                bw.writeTab( );
                bw.writeFloatLn( fitPar.getMaximumLorentzianWidth( ) );
                
            }
        }
        catch ( IOException e )
        {
            //  not caught yet
        }        
    }
    
    /**
     * After a node name is changed, possible other components referencing
     * to this node should be updated as well.
     */
    private void upDatePriorKnowledge( String oldNodeName, String newNodeName )
    {
        TreePath path = m_fitModelTree.getSelectionPath( );
        
        // This node has been selected
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)( path.getLastPathComponent( ) );
        
        // Get the parent level of the node
        DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent( );
        
        m_priorKnowledgePossibilities = new LinkedList<String>( );
        m_priorKnowledgePossibilities.add( "Free" );
        
        // Visit all children of the parent node
        if ( parent.getChildCount( ) > 0 )
        {
            for ( Enumeration e = parent.children( ); e.hasMoreElements( ); )
            {
                DefaultMutableTreeNode dmtn = (DefaultMutableTreeNode)e.nextElement( );
                Object obj = dmtn.getUserObject( );
                m_priorKnowledgePossibilities.add( obj.toString( ) );
            }
        } 
    }
    
    /**
     * This method extracts all possible prior knowledge possibilities from
     * the tree model, taking into account that prior knowledge can only exist
     * among nodes having the same parent.
     */
    private void doPriorKnowledge( boolean sentNotification )
    {
        TreePath path = m_fitModelTree.getSelectionPath( );
        
        // This node has been selected
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)( path.getLastPathComponent( ) );
        
        DefaultMutableTreeNode parent;
        // Get the parent level of the node
        if ( node.getParent( ) != null )
        {
            parent = (DefaultMutableTreeNode)node.getParent( );
        }
        else
        {
            parent = (DefaultMutableTreeNode)m_fitModelTree.getModel( ).getRoot( );
        }
        
        m_priorKnowledgePossibilities = new LinkedList<String>( );
        m_priorKnowledgePossibilities.add( "Free" );
        
        // Visit all children of the parent node
        if ( parent.getChildCount( ) > 0 )
        {
            for ( Enumeration e = parent.children( ); e.hasMoreElements( ); )
            {
                DefaultMutableTreeNode dmtn = (DefaultMutableTreeNode)e.nextElement( );
                Object obj = dmtn.getUserObject( );
                m_priorKnowledgePossibilities.add( obj.toString( ) );
                // System.out.println( "prior knowledge list item : " + obj.toString( ) );
            }
        }
        
        if ( sentNotification == true )
        {
            notifyActionListeners( PRIOR_KNOWLEDGE_UPDATE, this, null );
        }
    }
    
    /**
     * This is the method that allows the change of MRFittingParameters object
     * of the currently selected object.
     * @param fitPar
     */
    public void setMRFittingParametersOfCurrentlySelectedObject( MRFittingParameters fitPar  )
    {
        // This node has been selected
        TreePath path = m_fitModelTree.getSelectionPath( );
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)( path.getLastPathComponent( ) );
        
        // Get the associated object
        Object obj = node.getUserObject( );
        
        if ( obj instanceof MRLineModelParameters )
        {
            ((MRLineModelParameters)obj).setFittingParameters( fitPar );
        }
        if ( obj instanceof MRParametricPattern )
        { 
            ((MRParametricPattern)obj).setFittingParameters( fitPar );
        }
        if ( obj instanceof MRNumericPattern )
        {
            ((MRNumericPattern)obj).setFittingParameters( fitPar );
        }
        if ( obj instanceof MRMetaboliteModel )
        {
            ((MRMetaboliteModel)obj).setFittingParameters( fitPar );
        }
        
        // Clear the model
        clearModel( );
                
        // Visit all nodes of the tree and extract the objects from it
        // and put it in the local member objects
                
        extractModelTreeInformation( true );
        
        doPriorKnowledge( true );
        
    }
    
    
    /**
     * Returns the LinkedList containing the prior knowledge possibilities.
     * @return 
     */
    public LinkedList getPriorKnowlegdePossibilityLinkedList( )
    {
        return m_priorKnowledgePossibilities;
    }
    
    /**
     * Explicitly clears the data containing TreeMaps :
     * Note a simple TreeMap::clear does not clear the TreeMaps
     * that data structures contain - Therefore a clear is made explicitly
     * by iterating over the TreeMaps and the TreeMaps therein.
     */   
    public void clearModel( )
    {
        
        // Clear the content of the linkedlists
        Iterator it = m_parametricPatternList.iterator( );
        while ( it.hasNext( ) )
        {
             MRParametricPattern parPat = (MRParametricPattern)it.next( );
             parPat.clear( );
        }
        
        Iterator it2 = m_numericPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
             MRNumericPattern numPat = (MRNumericPattern)it2.next( );
             numPat.clear( );
        }
        
        Iterator it3 = m_simpleLineList.iterator( );
        while ( it3.hasNext( ) )
        {
             MRLineModelParameters linPar = (MRLineModelParameters)it3.next( );
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
             MRMetaboliteModel metMod = (MRMetaboliteModel)it4.next( );
             metMod.clear( );
        }
        
        // Clear the maps itself
        m_metaboliteList.clear( );
        m_numericPatternList.clear( );
        m_parametricPatternList.clear( );
        m_simpleLineList.clear( );

        m_metaboliteList        = null;
        m_numericPatternList    = null;
        m_parametricPatternList = null;
        m_simpleLineList        = null;
                
        m_metaboliteList        = new LinkedList<MRMetaboliteModel>( );
        m_numericPatternList    = new LinkedList<MRNumericPattern>( );
        m_parametricPatternList = new LinkedList<MRParametricPattern>( );
        m_simpleLineList        = new LinkedList<MRLineModelParameters>( );
    }
    
    /**
     * Extracts all information of the FitModelTreeWithPopupMenu object
     * and puts it in local member variables.
     * @param sentNotification 
     */
    private void extractModelTreeInformation( boolean sentNotification )
    {
        DefaultMutableTreeNode root = (DefaultMutableTreeNode)m_fitModelTree.getModel( ).getRoot( );
        
        // In this method the spectral response is recalculated:
        // Set the response to zero before calculating it and adding
        // it to the member variable m_modelSpectralResponse
        if ( m_completeModelSpectralResponse != null )
        {
            m_completeModelSpectralResponse.clear( );
        }
        
        m_simpleLineList.clear( );
        m_numericPatternList.clear( );
        m_parametricPatternList.clear( );
        m_metaboliteList.clear( );
        
        // Iterates over all nodes of the JTree
        processAllNodes( root );
        
        // Calculate the complete new spectral response
        calculateCompleteModelSpectralResponse( );
        if ( ( m_fitModelTree != null ) && ( m_fitModelTree.getSelectedObject( ) != null ) )
        {
            if ( !( m_fitModelTree.getSelectedObject( ) instanceof String ) )  
            // Calculate the spectral response of the selected node only
            {
                calculateSelectedPartModelSpectralResponse( );
            }
        }
        
        // Sent a notification to all listeners that the model has changed
        if ( sentNotification == true )
        {
            notifyActionListeners( MODEL_CHANGED, this, null );
        }
    }
    
    /**
     * Computes the partial response of the selected part of the JTree model.
     */
    private void calculateSelectedPartModelSpectralResponse( )
    {
        
        boolean rotateOutputSignalNecessary = true;
        
        TreePath path = m_fitModelTree.getSelectionPath( );
        
        // check whether the tree is not empty
        if ( path != null )
        {
            
            if ( m_completeModelSpectralResponse == null )
            {
                m_completeModelSpectralResponse = null;
                m_completeModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
                m_completeModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
            }
            
            DefaultMutableTreeNode dmtn = (DefaultMutableTreeNode)path.getLastPathComponent( );

            // Create a new MRLineModelSpectralResponse
            m_selectedModelSpectralResponse = null;
            m_selectedModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
            m_selectedModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
            m_selectedModelSpectralResponse.setCurveColor( Color.blue );

            //Object obj = ( (TreeModelCheckBoxNodeData)dmtn.getUserObject( ) ).getObject( );
            
            if ( dmtn.getUserObject( ) instanceof String )
            //if ( obj instanceof String )
            {
                m_selectedModelSpectralResponse.addTimeDomainResponse( 
                                                m_completeModelSpectralResponse.getAbsorptionChannelTD( ), 
                                                m_completeModelSpectralResponse.getDispersionChannelTD( ) );
                rotateOutputSignalNecessary = false;
                
            }
            else
            {
                //Object obj = ( (TreeModelCheckBoxNodeData)dmtn.getUserObject( ) ).getObject( );
                
                Object obj = dmtn.getUserObject( );
                
                if ( obj instanceof MRLineModelParameters )
                {
                    MRLineModelParameters linModel = (MRLineModelParameters)obj;
                    MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( (MRFittingParameters)linModel, m_MRSignalProperties );
                    m_selectedModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );  
                }
            
                if ( obj instanceof MRNumericPattern )
                {
                    MRNumericPattern numPat = (MRNumericPattern)obj;
                    MRLineModelSpectralResponse resp = numPat.getDynamicSpectralResponse( MRSignalProperties.getInstance( ) );
                    m_selectedModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                }
            
                if ( obj instanceof MRParametricPattern )
                {
                    MRParametricPattern parPat = (MRParametricPattern)obj;
                    MRLineModelSpectralResponse resp = parPat.getSpectralResponse( m_MRSignalProperties );
                    m_selectedModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                }
            
                if ( obj instanceof MRMetaboliteModel )
                {
                    // Now MRMetabolite can also be written and read, 
                    // but still scaling of an external imported  MRMetabolite 
                    // seems to be not ok.
                
                    MRMetaboliteModel metabolite = (MRMetaboliteModel)obj;
                    MRLineModelSpectralResponse resp = metabolite.getSpectralResponse( m_MRSignalProperties );
                    m_selectedModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                }
            }
            
            if ( rotateOutputSignalNecessary )
            {
                m_selectedModelSpectralResponse.rotateInFrequencyDomain( -1 );
            }
        }     
    }
    
    /**
     * Calculates the spectral representation of the MRModelSpectrum (this).
     */
    private void calculateCompleteModelSpectralResponse( )
    {

        // Create a new MRLineModelSpectralResponse
        m_completeModelSpectralResponse = null;
        m_completeModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
        m_completeModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
        m_completeModelSpectralResponse.setCurveColor( Color.red );
        // Calculate the sum of all simple lines MRLineModelSpectralResponse 
        // objects - if any
        Iterator lineIter = m_simpleLineList.iterator( );
        while ( lineIter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)lineIter.next( );
            
            // Calculate the corresponding resonance line
            MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( (MRFittingParameters)simpleLine, m_MRSignalProperties );
            m_completeModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
        }
        
        // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
        // objects - if any
        Iterator parPatIter = m_parametricPatternList.iterator( );
        while ( parPatIter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
            MRLineModelSpectralResponse resp = parPat.getSpectralResponse( m_MRSignalProperties );
            m_completeModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
        }
        
        // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
        // objects
        Iterator mrNumPatIter = m_numericPatternList.iterator( );
        while ( mrNumPatIter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
            MRLineModelSpectralResponse resp = numPat.getDynamicSpectralResponse( m_MRSignalProperties );
            m_completeModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
        }
        
        // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
        // objects - if any
        Iterator mrMetabIter = m_metaboliteList.iterator( );
        while ( mrMetabIter.hasNext( ) )
        {
            MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
            MRLineModelSpectralResponse resp = metabModel.getSpectralResponse( m_MRSignalProperties );
            m_completeModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
        }
        
        // Mysterious one point rotation ...
        m_completeModelSpectralResponse.rotateInFrequencyDomain( -1 );
        // m_completeModelSpectralResponse.autoSetXaxis( );
        
    }
    
    
    
    public VoxelMetabolites updateVoxelMetabolites(  float[] predictedParametersRFRegression, VoxelMetabolites voxelMetabolitesToUpdate ) 
    {
        
        int componentCounter = 0;
        
        // Create a new MRLineModelSpectralResponse
        MRLineModelSpectralResponse upDatedCompleteModelSpectralResponse = null;
        upDatedCompleteModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
        upDatedCompleteModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
        upDatedCompleteModelSpectralResponse.setCurveColor( Color.red );
        
        // Calculate the sum of all simple lines MRLineModelSpectralResponse 
        // objects - if any
        Iterator lineIter = m_simpleLineList.iterator( );
        while ( lineIter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)lineIter.next( );
            
            // Now update all model parameters
            simpleLine.setOffset( simpleLine.getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            simpleLine.setGaussianWidth( simpleLine.getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            simpleLine.setPhase( simpleLine.getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            simpleLine.setLorentzianWidth( simpleLine.getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = simpleLine.getArea( );
            simpleLine.setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
            
            // Calculate the corresponding resonance line
            MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( (MRFittingParameters)simpleLine, m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            ////
            MetaboliteFeatures metabFeatures = voxelMetabolitesToUpdate.getFeatures( simpleLine.getName( ) );
            
            metabFeatures.setAmplitude( predictedParametersRFRegression[ 4 + componentCounter ] );
            metabFeatures.setFrqShift( simpleLine.getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabFeatures.setDamping( simpleLine.getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabFeatures.setPhase( simpleLine.getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabFeatures.setGaussianDamping( simpleLine.getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            ////
            
            // Now restore old model parameters
            simpleLine.setOffset( simpleLine.getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            simpleLine.setGaussianWidth( simpleLine.getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            simpleLine.setPhase( simpleLine.getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            simpleLine.setLorentzianWidth( simpleLine.getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            simpleLine.setArea( originalArea );
            
            componentCounter++;
        }
        
        // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
        // objects - if any
        Iterator parPatIter = m_parametricPatternList.iterator( );
        while ( parPatIter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
            
            // Now update all model parameters
            parPat.getFittingParameters( ).setOffset( parPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            parPat.getFittingParameters( ).setGaussianWidth( parPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            parPat.getFittingParameters( ).setPhase( parPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            parPat.getFittingParameters( ).setLorentzianWidth( parPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            
            float originalArea = parPat.getFittingParameters( ).getArea( );
            parPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
    
            MRLineModelSpectralResponse resp = parPat.getSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            ////
            MetaboliteFeatures metabFeatures = voxelMetabolitesToUpdate.getFeatures( parPat.getName( ) );
            
            metabFeatures.setAmplitude( predictedParametersRFRegression[ 4 + componentCounter ] );
            metabFeatures.setFrqShift( parPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabFeatures.setDamping( parPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabFeatures.setPhase( parPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabFeatures.setGaussianDamping( parPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            ////
            
            // Now restore old model parameters again
            parPat.getFittingParameters( ).setOffset( parPat.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            parPat.getFittingParameters( ).setGaussianWidth( parPat.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            parPat.getFittingParameters( ).setPhase( parPat.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            parPat.getFittingParameters( ).setLorentzianWidth( parPat.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            parPat.getFittingParameters( ).setArea( originalArea );
    
            componentCounter++;
        }
        
        // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
        // objects
        Iterator mrNumPatIter = m_numericPatternList.iterator( );
        while ( mrNumPatIter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
            
            // Now update all model parameters
            numPat.getFittingParameters( ).setOffset( numPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            numPat.getFittingParameters( ).setGaussianWidth( numPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            numPat.getFittingParameters( ).setPhase( numPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            numPat.getFittingParameters( ).setLorentzianWidth( numPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            
            float originalArea = numPat.getFittingParameters( ).getArea( );
            numPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
 
            MRLineModelSpectralResponse resp = numPat.getDynamicSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            ////
            MetaboliteFeatures metabFeatures = voxelMetabolitesToUpdate.getFeatures( numPat.getName( ) );
            
            metabFeatures.setAmplitude( predictedParametersRFRegression[ 4 + componentCounter ] );
            metabFeatures.setFrqShift( numPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabFeatures.setDamping( numPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabFeatures.setPhase( numPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabFeatures.setGaussianDamping( numPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            ////
            
            
            // Restore parameters
            numPat.getFittingParameters( ).setOffset( numPat.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            numPat.getFittingParameters( ).setGaussianWidth( numPat.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            numPat.getFittingParameters( ).setPhase( numPat.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            numPat.getFittingParameters( ).setLorentzianWidth( numPat.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            numPat.getFittingParameters( ).setArea( originalArea );
            
            componentCounter++;
        }
        
        // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
        // objects - if any
        Iterator mrMetabIter = m_metaboliteList.iterator( );
        while ( mrMetabIter.hasNext( ) )
        {
            MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
            
            // Now update all model parameters
            metabModel.getFittingParameters( ).setOffset( metabModel.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabModel.getFittingParameters( ).setGaussianWidth( metabModel.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            metabModel.getFittingParameters( ).setPhase( metabModel.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabModel.getFittingParameters( ).setLorentzianWidth( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            
            float originalArea = metabModel.getFittingParameters( ).getArea( );
            metabModel.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
   
            MRLineModelSpectralResponse resp = metabModel.getSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            //// 
            MetaboliteFeatures metabFeatures = voxelMetabolitesToUpdate.getFeatures( metabModel.getName( ) );
            
            metabFeatures.setAmplitude( predictedParametersRFRegression[ 4 + componentCounter ] );
            metabFeatures.setFrqShift( metabModel.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabFeatures.setDamping( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabFeatures.setPhase( metabModel.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabFeatures.setGaussianDamping( metabModel.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            ////
            
            // Now restore old model parameters
            metabModel.getFittingParameters( ).setOffset( metabModel.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabModel.getFittingParameters( ).setGaussianWidth( metabModel.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            metabModel.getFittingParameters( ).setPhase( metabModel.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabModel.getFittingParameters( ).setLorentzianWidth( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabModel.getFittingParameters( ).setArea( originalArea );
        
            componentCounter++;
        }
        
        upDatedCompleteModelSpectralResponse.autoSetXaxis( );
        
        return voxelMetabolitesToUpdate;
    }
    
    /**
     * Calculates the spectral representation of the MRModelSpectrum (this).
     * @param predictedParametersRFRegression
     * @return 
     */
    public MRLineModelSpectralResponse calculateCompleteModelSpectralResponseBasedUponRandomForestRegressionPredictedParameters( float[] predictedParametersRFRegression )
    {
        int componentCounter = 0;
        
        // Create a new MRLineModelSpectralResponse
        MRLineModelSpectralResponse upDatedCompleteModelSpectralResponse = null;
        upDatedCompleteModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
        upDatedCompleteModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
        upDatedCompleteModelSpectralResponse.setCurveColor( Color.red );
        
        // Calculate the sum of all simple lines MRLineModelSpectralResponse 
        // objects - if any
        Iterator lineIter = m_simpleLineList.iterator( );
        while ( lineIter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)lineIter.next( );
            
            // Now update all model parameters
            simpleLine.setOffset( simpleLine.getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            simpleLine.setGaussianWidth( simpleLine.getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            simpleLine.setPhase( simpleLine.getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            simpleLine.setLorentzianWidth( simpleLine.getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = simpleLine.getArea( );
            simpleLine.setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
            
            // Calculate the corresponding resonance line
            MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( (MRFittingParameters)simpleLine, m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            // Now restore old model parameters
            simpleLine.setOffset( simpleLine.getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            simpleLine.setGaussianWidth( simpleLine.getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            simpleLine.setPhase( simpleLine.getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            simpleLine.setLorentzianWidth( simpleLine.getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            simpleLine.setArea( originalArea );
            
            componentCounter++;
        }
        
        // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
        // objects - if any
        Iterator parPatIter = m_parametricPatternList.iterator( );
        while ( parPatIter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
            
            // Now update all model parameters
            parPat.getFittingParameters( ).setOffset( parPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            parPat.getFittingParameters( ).setGaussianWidth( parPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            parPat.getFittingParameters( ).setPhase( parPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            parPat.getFittingParameters( ).setLorentzianWidth( parPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = parPat.getFittingParameters( ).getArea( );
            parPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
    
            MRLineModelSpectralResponse resp = parPat.getSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            // Now restore old model parameters again
            parPat.getFittingParameters( ).setOffset( parPat.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            parPat.getFittingParameters( ).setGaussianWidth( parPat.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            parPat.getFittingParameters( ).setPhase( parPat.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            parPat.getFittingParameters( ).setLorentzianWidth( parPat.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            parPat.getFittingParameters( ).setArea( originalArea );
    
            componentCounter++;
        }
        
        // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
        // objects
        Iterator mrNumPatIter = m_numericPatternList.iterator( );
        while ( mrNumPatIter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
            
            // Now update all model parameters
            numPat.getFittingParameters( ).setOffset( numPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            numPat.getFittingParameters( ).setGaussianWidth( numPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            numPat.getFittingParameters( ).setPhase( numPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            numPat.getFittingParameters( ).setLorentzianWidth( numPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = numPat.getFittingParameters( ).getArea( );
            numPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
 
            MRLineModelSpectralResponse resp = numPat.getDynamicSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            // Restore parameters
            numPat.getFittingParameters( ).setOffset( numPat.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            numPat.getFittingParameters( ).setGaussianWidth( numPat.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            numPat.getFittingParameters( ).setPhase( numPat.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            numPat.getFittingParameters( ).setLorentzianWidth( numPat.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            numPat.getFittingParameters( ).setArea( originalArea );
            
            componentCounter++;
        }
        
        // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
        // objects - if any
        Iterator mrMetabIter = m_metaboliteList.iterator( );
        while ( mrMetabIter.hasNext( ) )
        {
            MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
            
            // Now update all model parameters
            metabModel.getFittingParameters( ).setOffset( metabModel.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabModel.getFittingParameters( ).setGaussianWidth( metabModel.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            metabModel.getFittingParameters( ).setPhase( metabModel.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabModel.getFittingParameters( ).setLorentzianWidth( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = metabModel.getFittingParameters( ).getArea( );
            metabModel.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
   
            MRLineModelSpectralResponse resp = metabModel.getSpectralResponse( m_MRSignalProperties );
            upDatedCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
            
            // Now restore old model parameters
            metabModel.getFittingParameters( ).setOffset( metabModel.getFittingParameters( ).getOffset( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabModel.getFittingParameters( ).setGaussianWidth( metabModel.getFittingParameters( ).getGaussianWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            metabModel.getFittingParameters( ).setPhase( metabModel.getFittingParameters( ).getPhase( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabModel.getFittingParameters( ).setLorentzianWidth( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) - predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            metabModel.getFittingParameters( ).setArea( originalArea );
    
            componentCounter++;
        }
        
        upDatedCompleteModelSpectralResponse.doComplexConjugationTD( );      
        upDatedCompleteModelSpectralResponse.autoSetXaxis( );
    
        return upDatedCompleteModelSpectralResponse;
    }
    
    
    /**
     * Update the model parameters of the default model.
     */ 
    public void restoreModelParametersToDefault( )
    {
         
        // restore the simple lines to default
        m_simpleLineList.clear( );
        
        Iterator it = m_copyOfSimpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            m_simpleLineList.add( (MRLineModelParameters)it.next( ) );
        }
        
        // restore the parametric pattern list to default..
        m_parametricPatternList.clear( );
        
        Iterator it2 = m_copyOfParametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            m_parametricPatternList.add( (MRParametricPattern)it2.next( ) );
        }
        
        // restore the numeric pattern list to default..
        m_numericPatternList.clear( );
        Iterator it3 = m_copyOfNumericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            m_numericPatternList.add( (MRNumericPattern)it3.next( ) );
        }
        
        // restore the metabolite list to default..
        m_metaboliteList.clear( );
        Iterator it4 = m_copyOfMetaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            m_metaboliteList.add( (MRMetaboliteModel)it4.next( ) );
        }
        
        // clear the copies again.
        m_copyOfSimpleLineList.clear( );
        m_copyOfParametricPatternList.clear( );
        m_copyOfNumericPatternList.clear( );
        m_copyOfMetaboliteList.clear( );
    }        
            
    /**
     * Update the model parameters of the default model.
     * @param predictedParametersRFRegression
     */ 
    public void updatePrototypeModeParametersBasedUponRandomForestRegressionPredictedParameters( float[] predictedParametersRFRegression )
    {        
        int componentCounter = 0;
        
        // Create a new MRLineModelSpectralResponse
        MRLineModelSpectralResponse upDatedCompleteModelSpectralResponse = null;
        upDatedCompleteModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
        upDatedCompleteModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
        upDatedCompleteModelSpectralResponse.setCurveColor( Color.red );
        
        // Calculate the sum of all simple lines MRLineModelSpectralResponse 
        // objects - if any
        Iterator lineIter = m_simpleLineList.iterator( );
        while ( lineIter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)lineIter.next( );
            MRLineModelParameters simpleLineCopy = new MRLineModelParameters( simpleLine );
            
            // save a copy of the MRLineModelParameters object ..
            m_copyOfSimpleLineList.add( simpleLineCopy );
            
            // Now update all model parameters
            simpleLine.setOffset( simpleLine.getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            simpleLine.setGaussianWidth( simpleLine.getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            simpleLine.setPhase( simpleLine.getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            simpleLine.setLorentzianWidth( simpleLine.getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
            float originalArea = simpleLine.getArea( );
            simpleLine.setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
            
            // Sync the prior knowledge as well 
            simpleLine.setAreaReferenceVal( predictedParametersRFRegression[ 4 + componentCounter ] );
            simpleLine.setOffsetReferenceVal( simpleLine.getOffset( ) );
            simpleLine.setLorenzianWidthReferenceVal(simpleLine.getLorentzianLineWidth( ) );
            simpleLine.setGaussianWidthReferenceVal( simpleLine.getGaussianWidth( ) );
            simpleLine.setPhaseReferenceVal( simpleLine.getPhase( ) );
            
            componentCounter++;
        }
        
        // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
        // objects - if any
        Iterator parPatIter = m_parametricPatternList.iterator( );
        while ( parPatIter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
            MRParametricPattern parPatCopy = new MRParametricPattern( parPat );
            
            // save a copy of the MRParametricPattern object ..
            m_copyOfParametricPatternList.add( parPatCopy );
            
            // Now update all model parameters
            parPat.getFittingParameters( ).setOffset( parPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            parPat.getFittingParameters( ).setGaussianWidth( parPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            parPat.getFittingParameters( ).setPhase( parPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            parPat.getFittingParameters( ).setLorentzianWidth( parPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
//            float originalArea = parPat.getFittingParameters( ).getArea( );
            parPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
    
            //sync the prior knowledge as well
            parPat.getFittingParameters( ).setOffsetReferenceVal(  parPat.getFittingParameters( ).getOffset( ) );
            parPat.getFittingParameters( ).setGaussianWidthReferenceVal( parPat.getFittingParameters( ).getGaussianWidth( ) );
            parPat.getFittingParameters( ).setPhaseReferenceVal( parPat.getFittingParameters( ).getPhase( ) );
            parPat.getFittingParameters( ).setLorenzianWidthReferenceVal( parPat.getFittingParameters( ).getLorentzianLineWidth( ) );
            parPat.getFittingParameters( ).setAreaReferenceVal( parPat.getFittingParameters( ).getArea( ) );
                
            componentCounter++;
        }
        
        // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
        // objects
        Iterator mrNumPatIter = m_numericPatternList.iterator( );
        while ( mrNumPatIter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
            MRNumericPattern numPatCopy = new MRNumericPattern( numPat );
            
             // save a copy of the MRNumericPattern object ..
            m_copyOfNumericPatternList.add( numPatCopy );
            
            
            // Now update all model parameters
            numPat.getFittingParameters( ).setOffset( numPat.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            numPat.getFittingParameters( ).setGaussianWidth( numPat.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            numPat.getFittingParameters( ).setPhase( numPat.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            numPat.getFittingParameters( ).setLorentzianWidth( numPat.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
//            float originalArea = numPat.getFittingParameters( ).getArea( );
            numPat.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
 
            
            
            //Sync the prior knowledge as well
            numPat.getFittingParameters( ).setOffsetReferenceVal( numPat.getFittingParameters( ).getOffset( ) );
            numPat.getFittingParameters( ).setGaussianWidthReferenceVal( numPat.getFittingParameters( ).getGaussianWidth( ) );
            numPat.getFittingParameters( ).setPhaseReferenceVal( numPat.getFittingParameters( ).getPhase( ) );
            numPat.getFittingParameters( ).setLorenzianWidthReferenceVal( numPat.getFittingParameters( ).getLorentzianLineWidth( ) );
            numPat.getFittingParameters( ).setAreaReferenceVal( numPat.getFittingParameters( ).getArea( ) );
            
            componentCounter++;
        }
        
        // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
        // objects - if any
        Iterator mrMetabIter = m_metaboliteList.iterator( );
        while ( mrMetabIter.hasNext( ) )
        {
            MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
            MRMetaboliteModel metabModelCopy = new MRMetaboliteModel( metabModel );
            
             // save a copy of the MRNumericPattern object ..
            m_copyOfMetaboliteList.add( metabModelCopy );
            
            
            // Now update all model parameters
            metabModel.getFittingParameters( ).setOffset( metabModel.getFittingParameters( ).getOffset( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_OFFSET] );
            metabModel.getFittingParameters( ).setGaussianWidth( metabModel.getFittingParameters( ).getGaussianWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_GAUSSWITDTH] );
            metabModel.getFittingParameters( ).setPhase( metabModel.getFittingParameters( ).getPhase( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_PHASE] );
            metabModel.getFittingParameters( ).setLorentzianWidth( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) + predictedParametersRFRegression[TDFDFitRandomModelSignalGeneratorFrame.OVERALL_LORENTZWIDTH] );
//            float originalArea = metabModel.getFittingParameters( ).getArea( );
            metabModel.getFittingParameters( ).setArea( predictedParametersRFRegression[ 4 + componentCounter ] );
   
            //Sync the prior knowledge as well
            metabModel.getFittingParameters( ).setOffsetReferenceVal( metabModel.getFittingParameters( ).getOffset( ) );
            metabModel.getFittingParameters( ).setGaussianWidthReferenceVal( metabModel.getFittingParameters( ).getGaussianWidth( ) );
            metabModel.getFittingParameters( ).setPhaseReferenceVal( metabModel.getFittingParameters( ).getPhase( ) );
            metabModel.getFittingParameters( ).setLorenzianWidthReferenceVal( metabModel.getFittingParameters( ).getLorentzianLineWidth( ) );
            metabModel.getFittingParameters( ).setAreaReferenceVal( metabModel.getFittingParameters( ).getArea( ) ); 
                
            componentCounter++;
        }
    }
    
    
    private double getRandomOffsetVariationValue( double minimalOffsetShift, double maximalOffsetShift )
    {
        Random rnd = new Random( );
        double randomVal = minimalOffsetShift + rnd.nextDouble( )  * ( maximalOffsetShift - minimalOffsetShift );
        return randomVal;
    }
    
    private double getRandomLorentzianVariationValue( double minimalLorentzianVariation, double maximalLorentzianVariation )
    {
        Random rnd = new Random( );
        double randomVal = minimalLorentzianVariation + rnd.nextDouble( ) * ( maximalLorentzianVariation - minimalLorentzianVariation );
        return randomVal;
    }
    
    private double getRandomGaussianVariationValue( double minimalGaussianVariation, double maximalGaussianVariation )
    {
        Random rnd = new Random( );
        double randomVal = minimalGaussianVariation + ( rnd.nextDouble( ) ) * ( maximalGaussianVariation - minimalGaussianVariation );
        return randomVal;   
    }
    
    private double getRandomPhaseVariationValue( double minimalPhaseVariation, double maximalPhaseVariation )
    {
        Random rnd = new Random( );
        double randomVal = minimalPhaseVariation + rnd.nextDouble( ) * ( maximalPhaseVariation - minimalPhaseVariation );
        return randomVal;         
                
    }
    
    private double getRandomAreaVariationValue( double currentArea, double minimalAreaFactor, double maximalAreaFactor )
    {
//        double minimalAreaVariation = currentArea * minimalAreaFactor;
//        double maximalAreaVariation = currentArea * maximalAreaFactor;
                
        Random rnd = new Random( );
        double randomVal = currentArea * ( minimalAreaFactor + rnd.nextDouble( ) * ( maximalAreaFactor - minimalAreaFactor ) );
        return randomVal;       
                
    }
    
    
    
    /**
     * Defines the path in which the patient database is stored.
     */
    private String defineFileNameForMachineLearningTrainingDataSet( )
    {
        
        File defaultPatientDBRootPath;
        
        String defaultPathString = System.getProperty( "user.home" );
        
        JFileChooser fileChooser = new JFileChooser( defaultPathString );
        
        fileChooser.setFileSelectionMode( JFileChooser.FILES_ONLY );
        
        int result = fileChooser.showOpenDialog( null );
        
        File output_file = null;
        
        switch ( result )
        {
            case JFileChooser.APPROVE_OPTION:   // Approve Open button was pressed
                                                output_file = fileChooser.getSelectedFile( );
                                                                                                
                                                break;
           case JFileChooser.CANCEL_OPTION:     output_file = null;
                                                // Cancel or close dialog was selected
                                                break;
           case JFileChooser.ERROR_OPTION:      // the selected process did not terminate properly
                                                output_file = null;
                                                break;
        }
        if ( output_file.getAbsolutePath( ) == null )
        {
            output_file = new File( defaultPathString );
        }
        return output_file.getAbsolutePath( );
    }
    
    
    /**
     * Calculates the spectral representation of the MRModelSpectrum (this).
     * @param TDFDFitRandomModelSignalGeneratorFrm
     * @return 
     */
    public MRLineModelSpectralResponse[] createGroundtruthModelSpectralResponsesAndRandomParameterVariationsAsCSV( TDFDFitRandomModelSignalGeneratorFrame TDFDFitRandomModelSignalGeneratorFrm )
    {
        
         boolean once = false;
        
        String outputGroundTruthFileName = defineFileNameForMachineLearningTrainingDataSet( );
        TextFileWriter groundTruthWriterForRandomForestRegression = null;
        TextFileWriter groundTruthTDSumSignalsWriter = null;
        TextFileWriter groundTruthTDArtefactSignalsWriter = null;
        
        
        int numberOfResponsesToCalculate = TDFDFitRandomModelSignalGeneratorFrm.getNumberOfResponsesToCalculate( );
        m_randomModelResponses = new MRLineModelSpectralResponse[numberOfResponsesToCalculate];
         
        MRLineModelSpectralResponse localCompleteModelSpectralResponse;
        
        double minimalAmplitudeFactor = TDFDFitRandomModelSignalGeneratorFrm.getMinimalAmplitudeFactor( );
        double maximalAmplitudeFactor = TDFDFitRandomModelSignalGeneratorFrm.getMaximalAmplitudeFactor( );      
        
        double minimalOffsetShift = TDFDFitRandomModelSignalGeneratorFrm.getOffsetLeftShift( );
        double maximalOffsetShift = TDFDFitRandomModelSignalGeneratorFrm.getOffsetRightShift( );              
        
        double minimalPhaseVariation = TDFDFitRandomModelSignalGeneratorFrm.getMinimalPhaseVariation( );
        double maximalPhaseVariation = TDFDFitRandomModelSignalGeneratorFrm.getMaximalPhaseVariation( );   

        double minimalLorentzianVariation = TDFDFitRandomModelSignalGeneratorFrm.getMinimalLorentzianVariation( );
        double maximalLorentzianVariation = TDFDFitRandomModelSignalGeneratorFrm.getMaximalLorentzianVariation( );   
        
        double minimalGaussianVariation = TDFDFitRandomModelSignalGeneratorFrm.getMinimalGaussianVariation( );
        double maximalGaussianVariation = TDFDFitRandomModelSignalGeneratorFrm.getMaximalGaussianVariation( ); 
        
        int minFDIndexPoint = TDFDFitRandomModelSignalGeneratorFrm.getMinimalOffsetRangeFrequencyIndexPoint( );
        int maxFDIndexPoint = TDFDFitRandomModelSignalGeneratorFrm.getMaximalOffsetRangeFrequencyIndexPoint( );
        int maxTDIndexPoint = TDFDFitRandomModelSignalGeneratorFrm.getNumberOfTDSignalIndexPoint( );
        
        
        float simulatedNoiseTD = TDFDFitRandomModelSignalGeneratorFrm.getTDNoiseLevel( );
        
        // Open the csv-file
        try 
        {
            groundTruthWriterForRandomForestRegression = new TextFileWriter( new FileWriter( outputGroundTruthFileName + ".csv" ) );
            groundTruthTDSumSignalsWriter = new TextFileWriter( new FileWriter( outputGroundTruthFileName + "_all_sum_TD_responses" + ".csv" ) );
            
            if ( ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeSmileyArtifactIntoBaseline( ) == true ) || 
                 ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeResidualWaterBaseline( ) == true ) || 
                 ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeRandomlyVaryingBaseline( ) == true ) )
            {
                groundTruthTDArtefactSignalsWriter = new TextFileWriter( new FileWriter( outputGroundTruthFileName + "_all_artefacts_only_TD_responses" + ".csv" ) );
            }
            
            
            groundTruthWriterForRandomForestRegression.setNumberFormat( "0.000000000000E00" );
            groundTruthTDSumSignalsWriter.setNumberFormat( "0.000000000000E00" );
            if ( groundTruthTDArtefactSignalsWriter != null )
            {
                groundTruthTDArtefactSignalsWriter.setNumberFormat( "0.000000000000E00" );
            }

            Iterator lineIterForColumnNames = m_simpleLineList.iterator( );

            boolean overall_parameter_headers_set = false;
            
            while ( lineIterForColumnNames.hasNext( ) )
            {
                // call the copy constructor to keep the original model parameters ...
                MRLineModelParameters simpleLine = new MRLineModelParameters( (MRLineModelParameters)lineIterForColumnNames.next( ) );
                if  ( overall_parameter_headers_set == false )
                {
                    groundTruthWriterForRandomForestRegression.writeString( "overall_offset," +
                                                                            "overall_gaussw," +
                                                                            "overall_zero_phase," +
                                                                            "overall_lorentz_width,");
                    overall_parameter_headers_set = true;
                }
                groundTruthWriterForRandomForestRegression.writeString( "line_" + simpleLine.getName( ) + "_area," );                                                   
            }
            
            Iterator numericPatternIterForColumnNames = m_numericPatternList.iterator( );
                
            while ( numericPatternIterForColumnNames.hasNext( ) )
            {
                // call the copy constructor to keep the original model parameters ...
                MRNumericPattern numPat = (MRNumericPattern)numericPatternIterForColumnNames.next( );
                if  ( overall_parameter_headers_set == false )
                {
                    groundTruthWriterForRandomForestRegression.writeString( "overall_offset," +
                                                                            "overall_gaussw," +
                                                                            "overall_zero_phase," +
                                                                            "overall_lorentz_width," );

                    overall_parameter_headers_set = true;
                }
                
            }
            
            // Store the model values lwidth, gwidth, phase and offset of naa
            
            ImperfectSpectrumModelParameters imperfectModelPars = ImperfectSpectrumModelParameters.getInstance( );
            
            String referenceSpectralComponentName = imperfectModelPars.getSelectedReferenceComponent( );

            
            
            // Create the initital model parameters for the area parameters of the numeric patterns
            float reference_component_area   = 0f;
            float reference_component_offset = 0f;
            float reference_component_lwidth = 0f;
            float reference_component_phase  = 0f;
            float reference_component_gwidth = 0f;
            
            Iterator mrNumPatIterForColumnNames = m_numericPatternList.iterator( );
            while ( mrNumPatIterForColumnNames.hasNext( ) )
            {
                MRNumericPattern numPat = (MRNumericPattern)mrNumPatIterForColumnNames.next( );
                groundTruthWriterForRandomForestRegression.writeString( "num_pat_" + numPat.getName( ) + "_area," );
                if ( numPat.getName( ).contains( referenceSpectralComponentName ) )
                {
                    reference_component_offset = numPat.getFittingParameters( ).getOffset( );
                    reference_component_lwidth = numPat.getFittingParameters( ).getLorentzianLineWidth( );
                    reference_component_phase  = numPat.getFittingParameters( ).getPhase( );
                    reference_component_gwidth = numPat.getFittingParameters( ).getGaussianWidth( );
                    reference_component_area   = numPat.getFittingParameters( ).getArea( );
                }
            }
            
            if ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeResidualWaterBaseline( ) == true ){
                int num_water = imperfectModelPars.getNumberOfWaterCompoments();
                for ( int ii = 0; ii < num_water; ii++ ){
                    groundTruthWriterForRandomForestRegression.writeString( "Water_area_"+ii+"," + 
                                                                            "Water_shift_"+ii+"," +
                                                                            "Water_phase_"+ii+"," +
                                                                            "Water_width_"+ii+",");                

                }
            }
            
            for( int z = minFDIndexPoint; z <= maxFDIndexPoint; z++ )
            {  
                //  int z= 0; z < number_of_points_TD ; z++
                groundTruthWriterForRandomForestRegression.writeString( "absFD[" + z + "]," );
            }
            
            
            for( int z = minFDIndexPoint; z <= maxFDIndexPoint; z++ )
            {  
                groundTruthWriterForRandomForestRegression.writeString( "dispFD[" + z + "]," );
            }
           
            for( int z = 0; z <= maxTDIndexPoint; z++ )
            { 
                groundTruthWriterForRandomForestRegression.writeString( "absTD[" + z + "]," );
            }
                
            for( int z = 0; z <= maxTDIndexPoint; z++ )
            {  
                groundTruthWriterForRandomForestRegression.writeString( "dispTD[" + z + "]," );
            }    

            groundTruthWriterForRandomForestRegression.writeString( "\n" ); 
        
            for ( int i = 0; i < numberOfResponsesToCalculate; i++ )
            {

                System.out.println( "Signal " + i + " is computed .." );
                
                boolean overall_parameter_values_set  = false;
                
                // Create a new MRLineModelSpectralResponse
                localCompleteModelSpectralResponse = null;
                localCompleteModelSpectralResponse = new MRLineModelSpectralResponse( m_MRSignalProperties.getNumberOfSamplePoints( ) );
                localCompleteModelSpectralResponse.setMRSignalProperties( m_MRSignalProperties );
                localCompleteModelSpectralResponse.setCurveColor( Color.red );
                // Calculate the sum of all simple lines MRLineModelSpectralResponse 
                // objects - if any
                Iterator lineIter = m_simpleLineList.iterator( );

                
                double randomOffsetVariationValue     = getRandomOffsetVariationValue( minimalOffsetShift, maximalOffsetShift );
                double randomLorentzianVariationValue = getRandomLorentzianVariationValue( minimalLorentzianVariation, maximalLorentzianVariation );
                double randomGaussianVariationValue   = getRandomGaussianVariationValue( minimalGaussianVariation, maximalGaussianVariation );
                double randomPhaseVariationValue      = getRandomPhaseVariationValue( minimalPhaseVariation, maximalPhaseVariation );
                double randomAreaVariationValue;

                while ( lineIter.hasNext( ) )
                {
                    
                    // call the copy constructor to keep the original model parameters ...
                    MRLineModelParameters simpleLine = new MRLineModelParameters( (MRLineModelParameters)lineIter.next( ) );
                    
                    
                    float oldArea       = simpleLine.getArea( );
                    float oldFreq       = simpleLine.getOffset( );
                    float oldLorentz    = simpleLine.getLorentzianLineWidth( );
                    float oldGauss      = simpleLine.getGaussianWidth( );
                    float oldPhase      = simpleLine.getPhase( );
                    
                    
                    System.out.println( "Writing line : " + simpleLine.getName( ) );
                    
                    // the area has to be randomized for each line individually
                    randomAreaVariationValue = getRandomAreaVariationValue( simpleLine.getArea( ), minimalAmplitudeFactor, maximalAmplitudeFactor );

                    simpleLine.setArea( (float)randomAreaVariationValue );
                    
                    simpleLine.setOffset( (float)( simpleLine.getOffset( ) + randomOffsetVariationValue ) );
                    simpleLine.setLorentzianWidth( (float)( simpleLine.getLorentzianLineWidth( ) + randomLorentzianVariationValue ) );
                    simpleLine.setGaussianWidth( (float)( simpleLine.getGaussianWidth( ) + randomGaussianVariationValue ) );
                    simpleLine.setPhase( (float)( simpleLine.getPhase( )+ randomPhaseVariationValue ) );

                    if ( overall_parameter_values_set == false )
                    {
                        // write simple line parameters to csv file
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)( simpleLine.getOffset( ) ) );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)( simpleLine.getGaussianWidth( )  ) );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)( simpleLine.getPhase( ) ) );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)( simpleLine.getLorentzianLineWidth( ) ) );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
         
                        overall_parameter_values_set = true;
          
                    }
                    
                    groundTruthWriterForRandomForestRegression.writeDouble( randomAreaVariationValue );
                    groundTruthWriterForRandomForestRegression.writeString( "," );
                    
                    // Calculate the corresponding resonance line
                    MRLineModelSpectralResponse resp = new MRLineModelSpectralResponse( (MRFittingParameters)simpleLine, m_MRSignalProperties );
                    localCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                    
                    simpleLine.setArea( oldArea );       
                    simpleLine.setOffset( oldFreq );
                    simpleLine.setLorentzianWidth( oldLorentz );
                    simpleLine.setGaussianWidth( oldGauss );
                    simpleLine.setPhase( oldPhase );
                   
                }
                
                // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
                // objects - if any
                Iterator parPatIter = m_parametricPatternList.iterator( );
                while ( parPatIter.hasNext( ) )
                {
                    MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
                    System.out.println( "Warning: createGroundtruthModelSpectralResponsesAndRandomParameterVariationsAsCSV does not support parametric patterns!");
                    float oldArea  = parPat.getFittingParameters( ).getArea( );
                    float oldFreq  = parPat.getFittingParameters( ).getOffset( );
                    float oldLorentz = parPat.getFittingParameters( ).getLorentzianLineWidth( );
                    float oldGauss = parPat.getFittingParameters( ).getGaussianWidth( );
                    float oldPhase = parPat.getFittingParameters( ).getPhase( );
                    
                    // the area has to be randomized for each line individually
                    randomAreaVariationValue = getRandomAreaVariationValue( oldArea, minimalAmplitudeFactor, maximalAmplitudeFactor );
                    
                    System.out.println( "Writing parametric pattern area : " + randomAreaVariationValue );
                    
                    parPat.getFittingParameters( ).setArea( (float)randomAreaVariationValue );
                    parPat.getFittingParameters( ).setOffset( (float)( parPat.getFittingParameters( ).getOffset( ) + randomOffsetVariationValue ) );
                    parPat.getFittingParameters( ).setLorentzianWidth( (float)( parPat.getFittingParameters( ).getLorentzianLineWidth( ) + randomLorentzianVariationValue ) );
                    parPat.getFittingParameters( ).setGaussianWidth( (float)( parPat.getFittingParameters( ).getGaussianWidth( ) + randomGaussianVariationValue ) );
                    parPat.getFittingParameters( ).setPhase( (float)( parPat.getFittingParameters( ).getPhase( ) + randomPhaseVariationValue ) );
                    MRLineModelSpectralResponse resp = parPat.getSpectralResponse( m_MRSignalProperties );
                    localCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                    
                    // set the old values back to prevent recursion on the same object!!
                    parPat.getFittingParameters( ).setArea(oldArea);
                    parPat.getFittingParameters( ).setOffset(oldFreq);
                    parPat.getFittingParameters( ).setLorentzianWidth(oldLorentz);
                    parPat.getFittingParameters( ).setPhase(oldPhase);
                    parPat.getFittingParameters( ).setGaussianWidth(oldGauss);
                    
                }
                
                // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
                // objects
                Iterator mrNumPatIter = m_numericPatternList.iterator( );
                while ( mrNumPatIter.hasNext( ) )
                {
                    MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
                    
                    float oldArea  = numPat.getFittingParameters( ).getArea( );
                    float oldFreq  = numPat.getFittingParameters( ).getOffset( );
                    float oldLorentz = numPat.getFittingParameters( ).getLorentzianLineWidth( );
                    float oldGauss = numPat.getFittingParameters( ).getGaussianWidth( );
                    float oldPhase = numPat.getFittingParameters( ).getPhase( );
                    
                    System.out.println( "Writing numeric pattern : " + numPat.getName( ) );
                    
                    if ( overall_parameter_values_set == false )
                    {
                        // write numeric pattern parameters to csv file
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomOffsetVariationValue );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomGaussianVariationValue );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomPhaseVariationValue );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomLorentzianVariationValue );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
         
                        overall_parameter_values_set = true;
          
                    }
                    
                    // the area has to be randomized for each line individually
                    randomAreaVariationValue = getRandomAreaVariationValue( oldArea, minimalAmplitudeFactor, maximalAmplitudeFactor );
                    
                    groundTruthWriterForRandomForestRegression.writeDouble( (double)randomAreaVariationValue );
                    
                    groundTruthWriterForRandomForestRegression.writeString( "," );
                    
                    
                    System.out.println( "Writing numeric pattern area : " + randomAreaVariationValue );
                    
                    numPat.getFittingParameters( ).setArea( (float)randomAreaVariationValue );
                    numPat.getFittingParameters( ).setOffset( (float)( numPat.getFittingParameters( ).getOffset( ) + randomOffsetVariationValue ) );
                    numPat.getFittingParameters( ).setLorentzianWidth( (float)( numPat.getFittingParameters( ).getLorentzianLineWidth( ) + randomLorentzianVariationValue ) );
                    numPat.getFittingParameters( ).setGaussianWidth( (float)( numPat.getFittingParameters( ).getGaussianWidth( ) + randomGaussianVariationValue ) );
                    numPat.getFittingParameters( ).setPhase( (float)( numPat.getFittingParameters( ).getPhase( ) + randomPhaseVariationValue ) );
                    
                    MRLineModelSpectralResponse resp = numPat.getDynamicSpectralResponse( m_MRSignalProperties );
                    localCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                    
                    // Set the old values back to prevent recursion on the same object!!
                    numPat.getFittingParameters( ).setArea( oldArea );
                    numPat.getFittingParameters( ).setOffset( oldFreq );
                    numPat.getFittingParameters( ).setLorentzianWidth( oldLorentz );
                    numPat.getFittingParameters( ).setPhase( oldPhase );
                    numPat.getFittingParameters( ).setGaussianWidth( oldGauss );
                    
                }

                // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
                // objects - if any
                Iterator mrMetabIter = m_metaboliteList.iterator( );
                while ( mrMetabIter.hasNext( ) )
                {
                    MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
                    MRLineModelSpectralResponse resp = metabModel.getSpectralResponse( m_MRSignalProperties );
                    localCompleteModelSpectralResponse.addTimeDomainResponse( resp.getAbsorptionChannelTD( ), resp.getDispersionChannelTD( ) );
                }

                float   absValT0;
                float   dispValT0;
                float   normValT0;
                int     responseSize = localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length;
                
                float[] xaxis = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                
                // define some variables for the output
                float[] baseLineComponentAbs = null;
                float[] baseLineComponentDisp = null; 
                
                float[] smileyFDAbs = null;
                float[] smileyFDDisp = null;

                if ( imperfectModelPars.getIncludeSmileyArtefacts( ) == true )
                {
                    // get the values of the user defined parameters ..
                    float maxSmileyAmplitudeFactor = imperfectModelPars.getMaxSmileyAmplitudeFactor( );
                    float maxSmileyPhaseDifference = imperfectModelPars.getMaxSmileyPhaseDifference( );
                    float maxSmileyDecayConstant   = imperfectModelPars.getMaxSmileyDecayConstant( );
                    
                    smileyFDAbs       =  new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    smileyFDDisp      =  new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ];
                
                    float[] respFDMagnitude   = localCompleteModelSpectralResponse.getMagnitudeCurveFDPoints( ).getYvalues( );
                
                    float   signalPhaseT0     = (float)Math.atan2( localCompleteModelSpectralResponse.getAbsorptionChannelTD( )[0], 
                                                                   localCompleteModelSpectralResponse.getDispersionChannelTD( )[0] ); 
                    float   maxFDVal          = SignalTools.getMaximum( respFDMagnitude );
                
                    float smileyDecayConstant  = maxSmileyDecayConstant * ( float )( 1.0f + Math.random( ) * smileyFDAbs.length );
                    float randomPhaseSmileyArt = ( float )( 2.0f * ( Math.random( ) - 0.5 ) * maxSmileyPhaseDifference );
                    float randomFDAmplitudeFactorSmileyArt = ( float )( 2.0 * ( Math.random( ) - 0.5 ) ) * maxSmileyAmplitudeFactor;
                
                    for ( int kk = 0; kk < smileyFDAbs.length/2; kk++ )
                    {
                        smileyFDAbs[kk]  = randomFDAmplitudeFactorSmileyArt * maxFDVal * (float)Math.exp( -kk * smileyFDAbs.length  / smileyDecayConstant ) * (float)Math.cos( signalPhaseT0 + randomPhaseSmileyArt );
                        smileyFDDisp[kk] = randomFDAmplitudeFactorSmileyArt * maxFDVal * (float)Math.exp( -kk * smileyFDAbs.length  / smileyDecayConstant ) * (float)Math.sin( signalPhaseT0 + randomPhaseSmileyArt );
                    }
                
                    for ( int kk = 0; kk < smileyFDAbs.length/2; kk++ )
                    {
                        smileyFDAbs[ (smileyFDAbs.length) - kk - 1 ]  = smileyFDAbs[kk];
                        smileyFDDisp[ (smileyFDAbs.length) - kk - 1 ] = smileyFDDisp[kk];
                    }
                
                    localCompleteModelSpectralResponse.addFrequencyDomainResponse( smileyFDAbs, smileyFDDisp );
                }
                
                if ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeRandomlyVaryingBaseline( ) == true )
                
                {
                    
                    float generalBaseLineAmplitudeFactor = imperfectModelPars.getGeneralBaseLineAmplitudeFactor( );
                    float generalBaseLineMaxPhase = imperfectModelPars.getGeneralBaseLineMaxPhase( );
                    float generalBaseLineOffsetVar = imperfectModelPars.getGeneralBaseLineOffsetVar( );
                    float generalBaseLineDamping = imperfectModelPars.getGeneralBaseLineDamping( );
                    
                    
                    baseLineComponentAbs  = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    baseLineComponentDisp = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    

                    absValT0  = localCompleteModelSpectralResponse.getAbsorptionChannelTD( )[0];
                    dispValT0 = localCompleteModelSpectralResponse.getDispersionChannelTD( )[0];

                    normValT0 = (float)Math.sqrt( absValT0 * absValT0 + dispValT0 * dispValT0 );
                    
                    float randomPhase = ( float ) ( 2.0 * ( Math.random( ) - 0.5 ) * generalBaseLineMaxPhase ); 
                    float randomAmplification =  ( float ) (  generalBaseLineAmplitudeFactor * Math.random( ) * normValT0 );
                    
                    baseLineComponentAbs  = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    baseLineComponentDisp = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    
                    float offsetVar = 2f * generalBaseLineOffsetVar - (float)( generalBaseLineOffsetVar * ( 2.0 * ( 0.5 - Math.random( ) ) ) );
                    
                    for ( int ii = 0; ii < responseSize; ii++ )
                    {
                        xaxis[ii] = (float)ii;
                        if ( ii == 0 )
                        {
                           baseLineComponentAbs[ii] = (float)( randomAmplification * Math.cos( randomPhase ) ) * 0.0f;
                           baseLineComponentDisp[ii] = (float)( randomAmplification * Math.sin( randomPhase ) ) * 0.0f;
                        }
                        else
                        {
                            baseLineComponentAbs[ii]  = (float)( randomAmplification * Math.cos( randomPhase ) * Math.sin(( Math.PI * offsetVar * ii ) / (float)responseSize  ) / ( ( ( ( Math.PI * generalBaseLineDamping * ii ) / (float)responseSize ) ) * ( ( Math.PI * generalBaseLineDamping * ii ) / (float)responseSize ) ) );
                            baseLineComponentDisp[ii] = (float)( randomAmplification * Math.sin( randomPhase ) * Math.sin(( Math.PI * offsetVar * ii ) / (float)responseSize  ) / ( ( ( ( Math.PI * generalBaseLineDamping * ii ) / (float)responseSize ) ) * ( ( Math.PI * generalBaseLineDamping * ii ) / (float)responseSize ) ) );
                        }
                    }

                    if ( once == false )
                    {
                        XYPlotFrame pltfrm = new XYPlotFrame( "Examples signals of groundtruth" , null );
                        pltfrm.addCurve( new Curve( xaxis, baseLineComponentAbs, Color.BLUE, "Baseline" )  , true );
                        once = true;
                    }

                    localCompleteModelSpectralResponse.addTimeDomainResponse( baseLineComponentAbs, baseLineComponentDisp );
                    
                }
                
                float[] waterLineTestTDAbs = null; 
                float[] waterLineTestTDDisp = null;
                
                if ( TDFDFitRandomModelSignalGeneratorFrm.getIncludeResidualWaterBaseline( ) == true )
                {
                    float[] resWaterAmpls   = imperfectModelPars.getResidualWaterAmplFactors( );
                    float[] resWaterOffsets = imperfectModelPars.getResidualWaterOffsetValues( );
                    float[] resWaterPhases  = imperfectModelPars.getResidualWaterPhaseValues( );
                    float[] resWaterLwidths = imperfectModelPars.getResidualWaterLWidthValues( );
                    
                    float[] randomWaterAmpls   = new float[resWaterAmpls.length];
                    float[] randomWaterShifts = new float[resWaterAmpls.length];
                    float[] randomWaterPhases  = new float[resWaterAmpls.length];
                    float[] randomWaterLwidths = new float[resWaterAmpls.length];
                    
                    
                    double  widthLFactorNAA2Water3T = 3.44f;
                    
                    waterLineTestTDAbs      = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ]; 
                    waterLineTestTDDisp     = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ];
                    
                    for ( int ii = 0; ii < resWaterAmpls.length; ii++ )
                    {
                        randomWaterAmpls[ii]   = 2f * (float)( Math.random( ) - 0.5f ) * resWaterAmpls[ii];
                        randomWaterShifts[ii]  = 2f * (float)( Math.random( ) - 0.5f ) * resWaterOffsets[ii];
                        randomWaterPhases[ii]  = 2f * (float)( Math.random( ) - 0.5f ) * resWaterPhases[ii];
                        randomWaterLwidths[ii] = (float)Math.random( ) * resWaterLwidths[ii];
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomWaterAmpls[ii] * reference_component_area );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomWaterShifts[ii] );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomWaterPhases[ii] );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                        groundTruthWriterForRandomForestRegression.writeDouble( (double)randomWaterLwidths[ii]* widthLFactorNAA2Water3T * reference_component_lwidth );
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                    }
                                   
                    for ( int ii = 0; ii < responseSize; ii++ )
                    {
                        xaxis[ii] = (float)ii;
                        
                        for ( int kk = 0; kk < randomWaterAmpls.length; kk++ )
                        {
                            waterLineTestTDAbs[ii]  +=  (float)randomWaterAmpls[kk] * reference_component_area * (float)Math.sin( Math.PI * randomWaterShifts[kk] * ii / (float)responseSize ) * (float)Math.exp(-ii * widthLFactorNAA2Water3T * randomWaterLwidths[kk] * reference_component_lwidth / (float)responseSize ) * (float)Math.cos(  Math.PI * randomWaterPhases[kk] / 180f );
                            waterLineTestTDDisp[ii] +=  (float)randomWaterAmpls[kk] * reference_component_area * (float)Math.cos( Math.PI * randomWaterShifts[kk] * ii / (float)responseSize ) * (float)Math.exp(-ii * widthLFactorNAA2Water3T * randomWaterLwidths[kk] * reference_component_lwidth / (float)responseSize ) * (float)Math.sin(  Math.PI * randomWaterPhases[kk] / 180f );
                        }
                    }
                    
                    if ( once == false )
                    {
                        XYPlotFrame pltfrm = new XYPlotFrame( "Examples residual water signal groundtruth" , null );
                        pltfrm.addCurve( new Curve( xaxis, waterLineTestTDAbs, Color.BLUE, "Residual water" )  , true );
                        once = true;
                    }
                    
                    localCompleteModelSpectralResponse.addTimeDomainResponse( waterLineTestTDAbs, waterLineTestTDDisp );
                    
                }
                
                float[] sumArtefactsOutAbs  = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ];
                float[] sumArtefactsOutDisp = new float[ localCompleteModelSpectralResponse.getAbsorptionChannelTD( ).length ];
                
                for ( int zz = 0; zz < sumArtefactsOutAbs.length ; zz++ )
                {
                    if ( waterLineTestTDAbs != null )
                    {
                        sumArtefactsOutAbs[zz]  += waterLineTestTDAbs[zz];
                        sumArtefactsOutDisp[zz] += waterLineTestTDDisp[zz]; 
                    }
                    
                    if ( baseLineComponentAbs != null )
                    {
                        sumArtefactsOutAbs[zz]  += baseLineComponentAbs[zz];
                        sumArtefactsOutDisp[zz] += baseLineComponentDisp[zz]; 
                    }
                    
                    if ( smileyFDAbs != null )
                    {
                        sumArtefactsOutAbs[zz]  += smileyFDAbs[zz]; 
                        sumArtefactsOutDisp[zz] += smileyFDDisp[zz]; 
                    }
                }
                               
                localCompleteModelSpectralResponse.autoSetXaxis( );
                
                localCompleteModelSpectralResponse.addNoiseInTimeDomain( simulatedNoiseTD );
                // localCompleteModelSpectralResponse.doComplexConjuation( );
                
                this.m_randomModelResponses[i] = localCompleteModelSpectralResponse;
                
                // Response "i" is defined now: write feautures to file system
                float [] output = localCompleteModelSpectralResponse.getAbsorptionChannelFD( );
                
                for( int z = minFDIndexPoint; z <= maxFDIndexPoint; z++ )
                {  
                    groundTruthWriterForRandomForestRegression.writeDouble( (double)output[z] );
                    groundTruthWriterForRandomForestRegression.writeString( "," );
                }
            
                output = localCompleteModelSpectralResponse.getDispersionChannelFD( );
            
                for( int z = minFDIndexPoint; z <= maxFDIndexPoint; z++ )
                {  
                    groundTruthWriterForRandomForestRegression.writeDouble( (double)output[z] );
                    groundTruthWriterForRandomForestRegression.writeString( "," );
                }
           
                output = localCompleteModelSpectralResponse.getAbsorptionChannelTD( );
                
                for( int z = 0; z <= maxTDIndexPoint; z++ )
                { 
                    groundTruthWriterForRandomForestRegression.writeDouble( (double)output[z] );
                    groundTruthWriterForRandomForestRegression.writeString( "," );
                }
                
                output = localCompleteModelSpectralResponse.getDispersionChannelTD( );
                for( int z = 0; z <= maxTDIndexPoint; z++ )
                {  
                    groundTruthWriterForRandomForestRegression.writeDouble( (double)output[z] );
                    if ( z < maxTDIndexPoint )
                    {
                        groundTruthWriterForRandomForestRegression.writeString( "," );
                    }
                }    

                groundTruthWriterForRandomForestRegression.writeStringLn( "" );
                
                // Also write the complete responses to a file
                output = localCompleteModelSpectralResponse.getAbsorptionChannelTD( );
                for ( int z = 0; z < output.length; z++ )
                {
                    groundTruthTDSumSignalsWriter.writeDouble( (double)output[z] );
                    groundTruthTDSumSignalsWriter.writeString( "," );    
                    if ( groundTruthTDArtefactSignalsWriter != null )
                    {
                        groundTruthTDArtefactSignalsWriter.writeDouble( (double)sumArtefactsOutAbs[z] );
                        groundTruthTDArtefactSignalsWriter.writeString( "," );
                    }
                }
                
                output = localCompleteModelSpectralResponse.getDispersionChannelTD( );
                
                for ( int z = 0;z < output.length; z++ )
                {
                    groundTruthTDSumSignalsWriter.writeDouble( (double)output[z] );
                    if ( groundTruthTDArtefactSignalsWriter != null )
                    {
                        groundTruthTDArtefactSignalsWriter.writeDouble( (double)sumArtefactsOutDisp[z] );
                    }
                    if ( z < output.length - 1 )
                    {
                        groundTruthTDSumSignalsWriter.writeString( "," );
                        if ( groundTruthTDArtefactSignalsWriter != null )
                        {
                            groundTruthTDArtefactSignalsWriter.writeString( "," );
                        }
                    }
                }
                
                groundTruthTDSumSignalsWriter.writeStringLn( "" );
                
                if ( groundTruthTDArtefactSignalsWriter != null )
                {
                    groundTruthTDArtefactSignalsWriter.writeStringLn( "" );
                }
                
            }
            
            groundTruthWriterForRandomForestRegression.flush( );
            groundTruthWriterForRandomForestRegression.close( );
            
            groundTruthTDSumSignalsWriter.flush( );
            groundTruthTDSumSignalsWriter.close( );
            
            if ( groundTruthTDArtefactSignalsWriter != null )
            {
                groundTruthTDArtefactSignalsWriter.flush( );
                groundTruthTDArtefactSignalsWriter.close( );
            }
        }
        catch ( IOException ex ) 
        {
            Logger.getLogger( DataGeneratorMain.class.getName( ) ).log( Level.SEVERE, null, ex );
        }
        
        // return the responses enabling the caller to verify its shape...
        return m_randomModelResponses;
        
    }
    
    
    /**
     * Processes all JTree-nodes of the tree and puts its contents in several 
     * TreeMap objects.
     */
    private void processAllNodes( DefaultMutableTreeNode node )
    {
        if ( node.getUserObject( ) instanceof MRLineModelParameters )
        {
            MRLineModelParameters linPar = (MRLineModelParameters)node.getUserObject( );
            
            m_currentSimpleLineName = linPar.getName( );
            
            // Depending on what parent the line has, it has to be added to a 
            // different Treemap objects
            
            DefaultMutableTreeNode parent = ((DefaultMutableTreeNode)((TreeNode)node).getParent( ) );
            
            if ( parent.getUserObject( ) instanceof String )
            {
                // add the parameters to the m_simpleLineList
                m_simpleLineList.add( linPar );                
            }
            
            if ( parent.getUserObject( ) instanceof MRParametricPattern )
            {
                // Now there are two possibilities either the parametric pattern
                // is directly connected to rootlevel object or it is connected
                // to a metabolite object
                DefaultMutableTreeNode grandparent = ( (DefaultMutableTreeNode)( (TreeNode)parent ).getParent( ) );
                
                if ( grandparent.getUserObject( ) instanceof String )
                {
                    m_parametricPatternList.getLast( ).addSimpleLine( linPar );
                }
                if ( grandparent.getUserObject( ) instanceof MRMetaboliteModel )
                {
                    // get the last MRMetaboliteModel
                    
                    Object key = m_metaboliteList.getLast( );
                    MRMetaboliteModel last_metabModel = (MRMetaboliteModel)key;
                    
                    last_metabModel.addLineToLatestAddedParametricPattern( linPar.getName( ), linPar );
                }
            }
            
            // A line needs to be added directly to the MetaboliteModel node
            if ( parent.getUserObject( ) instanceof MRMetaboliteModel )
            {
                MRMetaboliteModel metMod = (MRMetaboliteModel )parent.getUserObject( );
                metMod.addMRLineModelParameters( linPar );
            }
            
        }
        
        // Depending on what parent the line has, it has to be added to different
        // Treemap objects
        if ( node.getUserObject( ) instanceof MRParametricPattern )
        {
            
            Object nodeObject = node.getUserObject( );
            MRParametricPattern parPat = (MRParametricPattern)nodeObject;
            
            // Depending on what parent the line has, it has to be added to a different
            // Treemap objects
            
            DefaultMutableTreeNode parent = ((DefaultMutableTreeNode)((TreeNode)node).getParent( ) );
            
            if ( parent.getUserObject( ) instanceof String )
            {
                m_parametricPatternList.add( parPat );
            } 
            
            if ( parent.getUserObject( ) instanceof MRMetaboliteModel )
            {
                // get the last MRMetaboliteModel
                // MRMetaboliteModel model = (MRMetaboliteModel)m_metaboliteList.get( m_currentMetaboliteName );
                MRMetaboliteModel model = (MRMetaboliteModel)parent.getUserObject( );
                model.addParametericPattern( parPat );
            }
        }
        
        // Depending on what parent the line has, it has to be added to a different
        // LinkedList objects
        if ( node.getUserObject( ) instanceof MRMetaboliteModel )
        {
             m_currentMetaboliteName = ( (MRMetaboliteModel)node.getUserObject( ) ).getName( );
             m_metaboliteList.add( (MRMetaboliteModel)node.getUserObject( ) );
             MRMetaboliteModel mm = (MRMetaboliteModel)node.getUserObject( );
        }
        
        // Depending on what parent the line has, it has to be added to a different
        // LinkedList objects
        if ( node.getUserObject( ) instanceof MRNumericPattern )
        {
            MRNumericPattern numPat = (MRNumericPattern)node.getUserObject( );
            
            m_currentNumericPatternName =  numPat.getName( );
            
            DefaultMutableTreeNode parent = ((DefaultMutableTreeNode)((TreeNode)node).getParent( ) );
            
            if ( parent.getUserObject( ) instanceof String )
            {
                m_numericPatternList.add( numPat );
            }
            
            if ( parent.getUserObject( ) instanceof MRMetaboliteModel )
            {
                MRMetaboliteModel metMod = (MRMetaboliteModel)parent.getUserObject( );
                metMod.addNumericPattern( m_currentNumericPatternName, numPat );
            }
        }
        // This method calls itself
        if ( node.getChildCount( ) >= 0 )
        {
            for ( Enumeration e = node.children( ); e.hasMoreElements( ); )
            {
                DefaultMutableTreeNode n = (DefaultMutableTreeNode)e.nextElement( );
                processAllNodes( n );
            }
        }
    }
    
    /**
     * Writes all metabolites to a random access file.
     */
    private void writeMetabolites( RandomAccessFile out )
    {
        // Iterate over m_metaboliteList
        // and call the member function write( ) for each element
        Iterator it = m_metaboliteList.iterator( );
        while ( it.hasNext( ) )
        {
            MRMetaboliteModel metModel = (MRMetaboliteModel)it.next( );
            metModel.write( out );
        }
    }
    
    /**
     * Writes all numeric patterns of the metabolite to a random access file.
     */
    private void writeNumericPatterns( RandomAccessFile out )
    {
        // get a reference to all the keys of the Treemap containing 
        // the numeric pattern information
        Iterator iter = m_numericPatternList.iterator( );
        // Iterate of all elements of the LinkedList
        while ( iter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)iter.next( );
            numPat.write( out );
        }
    }
    
    /**
     * Writes all parametric patterns of the metabolite to a random access file.
     */
    private void writeParametricPatterns( RandomAccessFile out )
    {        
        // get a reference to all the keys of the Treemap containing 
        // the numeric pattern information
        Iterator iter = m_parametricPatternList.iterator( );
        // Iterate of all elements of the TreeMap
        while ( iter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)iter.next( );
            parPat.write( out );
        } 
    }
    
    /**
     * Writes all metabolites to a random access file.
     */    
    private void writeSimpleLines( RandomAccessFile out )
    {   
        // get a reference to all the keys of the Treemap containing 
        // the simple line information
        Iterator iter = m_simpleLineList.iterator( );
        // Iterate of all elements of the TreeMap
        while ( iter.hasNext( ) )
        {
            MRLineModelParameters simpleLine = (MRLineModelParameters)iter.next( );
            simpleLine.write( out );
        }
    }
    
    /**
     * Returns the TreeMap containing the MRLineModelParameters objects.
     * @return 
     */ 
    public LinkedList getSimpleMRLines( )
    {
       return m_simpleLineList;
    }
    
    /**
     * This method was added to make macro molecular baseline definition 
     * possible in a easy and straightforward way.
     * @param par 
     */
    public void addSimpleLineFromObject( MRLineModelParameters par )
    {
        m_simpleLineList.add( par );
    }
    
    /**
     * Returns the LinkedList containing the MRParametricPattern objects.
     * @return 
     */ 
    public LinkedList getParametricPatterns( )
    {
       return m_parametricPatternList;
    }
    
    /**
     * Returns the LinkedList containing the MRParametricPattern objects.
     * @return 
     */ 
    public LinkedList getNumericPatterns( )
    {
       return m_numericPatternList;
    }    
    
    /**
     * Returns the LinkedList containing the MRMetaboliteModel objects.
     * @return 
     */ 
    public LinkedList getMetaboliteModelsList( )
    {
       return m_metaboliteList;
    }    
    
    
    /**
     * Adds an ActionListener to the list of action listeners who are notified
     * when something happens within this object.
     * @param e
     */
    public void addActionListener( ActionListener e )
    {
	m_actionListeners.add( e );
    }
      
    /**
     * Returns the overall response belonging to the set of model parameters.
     * @return 
     */
    public MRLineModelSpectralResponse getCompleteModelResponse( )
    {
        return m_completeModelSpectralResponse;
    }
    
    /**
     * Returns the overall response belonging to the set of model parameters.
     * @return 
     */
    public MRLineModelSpectralResponse getSelectedModelResponse( )
    {
        calculateSelectedPartModelSpectralResponse( );
        return m_selectedModelSpectralResponse;
    }
    
    /**
     * Returns the TDFDFitModelTreeWithPopupMenu object
     * @return 
     */
    public TDFDFitModelTreeWithPopupMenu getFitModelTreeWithPopUp( )
    {
        return m_fitModelTree;
    }
    
    /**
     * Sets the MRLineModelParametersPanel object.
     * @param lineModelParametersPanel 
     */
    public void setMRLineModelParametersPanel( MRLineModelParametersPanel lineModelParametersPanel )
    {
        m_lineModelParametersPanel = lineModelParametersPanel;
    }
    
    /**
     * 
     * Returns the last notified parameters which are currently displayed
     * in the graphical user interface.
     * 
     * @return 
     */
    public MRFittingParameters getLastNotifiedParameters( )
    {
        // First force an update of these values
        m_lineModelParametersPanel.setLastNotifiedValues( );
        // Return them to caller
        return m_lineModelParametersPanel.getLastNotifiedParameters( );
    }
    
    /**
     * Sets the reference to the measured spectral response object of the
     * application.
     * @param measured_response
     */
    public void setMeasuredSpectrum( SingleVoxelMeasuredSpectralData measured_response )
    {
        m_measuredSpectralResponse = measured_response;
    }
    
    /**
     * Sets the reference to the measured spectral response object of the
     * application.
     * @return 
     */
    public SingleVoxelMeasuredSpectralData getMeasuredSpectrum( )
    {
        return m_measuredSpectralResponse;
    }
        
    /**
     * Sets the reference to the measured water reference spectral response object 
     * of the application.
     * @param water_reference_response
     */
    public void setMeasuredWaterReferenceSpectrum( SingleVoxelMeasuredSpectralData water_reference_response )
    {
        m_waterReferenceSpectralResponse = water_reference_response;
        
        // perform quality on spectrum
        if ( m_measuredSpectralResponse != null )
        {
            // eliminate eddy currents by applying Quality algorithm
            m_measuredSpectralResponse.applyQuality( false, m_waterReferenceSpectralResponse );
        }
    }
    
    /**
     * Returns a reference to the TDFDFitModelTreeWithPopupMenu attribute.
     * @return 
     */
    public TDFDFitModelTreeWithPopupMenu getFitModelTree( )
    {
        return m_fitModelTree;
    }
    
    /**
     * Returns the total number of spectral components of the model.
     * @return 
     */
    public int getTotalNumberOfComponents( )
    {
        return ( m_metaboliteList.size( ) + 
                 m_numericPatternList.size( ) + 
                 m_parametricPatternList.size( ) + 
                 m_simpleLineList.size( ) );
    }
    /**
     * Returns the vector of component names in the order of the .par file
     * needed to start tdfdfit.
     * @return 
     */
    public Vector<String> getComponentNameVector( )
    {
        Vector<String> componentNameVector = new Vector<>( );
        Iterator it1 = m_simpleLineList.iterator( );
        while ( it1.hasNext( ) )
        {
            MRLineModelParameters lineModel = (MRLineModelParameters)it1.next( );
            componentNameVector.add( lineModel.getName( ) );
        }    
        
        Iterator it2 =  m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)it2.next( );
            componentNameVector.add( parPat.getName( ) );
        }
        
        Iterator it3 =  m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)it3.next( );
            componentNameVector.add( numPat.getName( ) );
        }
        
        Iterator it4 =  m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRMetaboliteModel metabPat = (MRMetaboliteModel)it4.next( );
            componentNameVector.add( metabPat.getName( ) );
        }
        
        return componentNameVector;
    }
    
    
    /**
     * Adds metabolite pattern to the model
     * TODO: test for uniqueness of name. 
     * @param metaboliteFileRef
     */
    public void addMetaboliteFrom_MPF_File( File[] metaboliteFileRefs )
    {
        for ( int i = 0; i < metaboliteFileRefs.length; i++ )
        {
            m_fitModelTree.addMetabolite_From_MPF_File( metaboliteFileRefs[i].toString() );
            extractModelTreeInformation( true );        
        }
    }
    
    /**
     * Adds parametric pattern to the model
     * TODO: test for uniqueness of name. 
     * @param parPatFileRef
     */
    public void addParametricPatternFrom_PPF_File( File[] parPatFileRefs )
    {
        for ( int i = 0; i < parPatFileRefs.length; i++ )
        {
            m_fitModelTree.addParametricPatternFrom_PPF_File( parPatFileRefs[i].toString( ) );
            extractModelTreeInformation( true );
        }
    }
    
    /**
     * Adds numeric pattern to the model
     * TODO: test for uniqueness of name. 
     * @param numPatFileRef
     */
    public void addNumericPatternFrom_MRUI_Files( File[] numPatFileRefs )
    {
        for ( int i = 0; i < numPatFileRefs.length; i++ )
        {
            m_fitModelTree.addNumericPatternFrom_MRUI_File( numPatFileRefs[i] );
            extractModelTreeInformation( true );
            m_fitModelTree.selectRootNode( );
        }
    }
    
    /**
     * Adds numeric pattern to the model
     * TODO: test for uniqueness of name. 
     * @param binFileRef
     */
    public void addNumericPatternFrom_Binary_Files( File[] binFileRefs )
    {
        for ( int i = 0; i < binFileRefs.length; i++ )
        {
            m_fitModelTree.addNumericPatternFrom_Binary_File( binFileRefs[i] );
            extractModelTreeInformation( true );
            m_fitModelTree.selectRootNode( );
        }
    } 
    
    /**
     * Adds numeric pattern to the model
     * TODO: test for uniqueness of name. 
     * @param numPatFileRef
     */
    public void addNumericPatternFrom_NPF_File( File[] numPatFileRefs )
    {
        for ( int i = 0; i < numPatFileRefs.length; i++ )
        {
            m_fitModelTree.addNumericPatternFrom_NPF_File( numPatFileRefs[i] );
            extractModelTreeInformation( true );
        }
    }
    
    /**
     * Sets a reference to the model tree
     * @param modelTree 
     */
    public void setModelTree( TDFDFitModelTreeWithPopupMenu modelTree ) 
    {
        m_fitModelTree = modelTree;
    }
     
    /**
     * 
     * Update the parameter range parameters.
     * 
     * @param linPars
     * @return 
     */
    private MRFittingParameters updateParameterRanges( MRFittingParameters linPars )
    {
        
        TDFDFittingDefaults defs = TDFDFittingDefaults.getInstance( );
        defs.load( );
        
        String takeAmplitudeFactor = "no";
        
        if ( defs.getProperty( TDFDFittingDefaults.IsAreaMaxDeterminedByFactor_1H ) != null )
        {
            takeAmplitudeFactor = defs.getProperty( TDFDFittingDefaults.IsAreaMaxDeterminedByFactor_1H );
        }
        
        m_fitModelTree.selectRootNode( );
        
        // Fetch all the predefined range values
        DefaultFittingParameterRangeValues defVals = DefaultFittingParameterRangeValues.getInstance( );
        
        float defAreaMinVal = defVals.getAreaMinValDecimalValue( );
        float defAreaMaxVal = defVals.getAreaMaxValDecimalValue( );
        float defAreaFactor = defVals.getAreaMaxFactorDecimalValue( );
        
        if ( takeAmplitudeFactor.compareTo( "yes" ) == 0 )
        {
            defAreaMaxVal = linPars.getArea( ) * defAreaFactor;
        }    
            
        boolean isAreaMaxDeterminedByFactor = defVals.isAreaMaxDeterminedByFactor( );
        
        float defOffsetMinVal = defVals.getOffsetMinValDecimalValue( );
        float defOffsetMaxVal = defVals.getOffsetMaxValDecimalValue( );
        
        float defPhaseMinVal = defVals.getPhaseMinValDecimalValue( );
        float defPhaseMaxVal = defVals.getPhaseMaxValDecimalValue( );
        
        float defLorMinVal = defVals.getLorentzianWidthMinValDecimalValue( );
        float defLorMaxVal = defVals.getLorentzianWidthMaxValDecimalValue( );
        
        float defGaussMinVal = defVals.getGaussianWidthMinValDecimalValue( );
        float defGaussMaxVal = defVals.getGaussianWidthMaxValDecimalValue( );
        
        float curArea = linPars.getArea( );
        
        if ( curArea < 0.0 )
        {
            curArea = Math.abs(curArea);
            linPars.setArea( curArea );
            
            linPars.setPhase( linPars.getPhase( ) + 180f );
            linPars.setMinimumPhase( linPars.getMinimumPhase( ) + 180f );
            linPars.setMaximumPhase( linPars.getMaximumPhase( ) + 180f );
        }
          
        linPars.setMinimumArea( defAreaMinVal );
        if ( isAreaMaxDeterminedByFactor == true )
        {
            linPars.setMaximumArea( curArea * defAreaFactor );
        }
        else
        {
            linPars.setMaximumArea( defAreaMaxVal );
        }

        float curOffset = linPars.getOffset( );
        
        linPars.setMinimumOffset( curOffset - Math.abs( defOffsetMinVal ) );
        linPars.setMaximumOffset( curOffset + Math.abs( defOffsetMaxVal ) );
        
        float curLorWidth = linPars.getLorentzianLineWidth( );
        linPars.setMinimumLorentzianWidth( curLorWidth - Math.abs( defLorMinVal ) );
        linPars.setMaximumLorentzianWidth( curLorWidth + Math.abs( defLorMaxVal ) );

        float curGaussWidth = linPars.getGaussianWidth( );
        
        linPars.setMinimumGaussianWidth( curGaussWidth - Math.abs( defGaussMinVal ) );
        linPars.setMaximumGaussianWidth( curGaussWidth + Math.abs( defGaussMaxVal ) );

        float curPhase = linPars.getPhase( );
        linPars.setMinimumPhase( curPhase - Math.abs( defPhaseMinVal ) );
        linPars.setMaximumPhase( curPhase + Math.abs( defPhaseMaxVal ) );
        return linPars;
    }
    
    /**
     * Updates the valid parameter ranges according to the predefined value
     * ranges.
     */
    public void autoSetParameterRangeValues( )
    {
        
        // Iterate over all simple lines and set the range values to the 
        // predefined default values for the nucleus type.
        Iterator lineIter = m_simpleLineList.iterator( );
        while ( lineIter.hasNext( ) )
        {
            // get reference to the parameters ..
            // have to cast it to a MRFittingParameters object..
            MRFittingParameters simpleLine = (MRFittingParameters)lineIter.next( );  
            simpleLine = updateParameterRanges( simpleLine );  
        }
        
        // Calculate the sum of all ParametricPattern - MRLineModelSpectralResponse
        // objects - if any
        Iterator parPatIter = m_parametricPatternList.iterator( );
        while ( parPatIter.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)parPatIter.next( );
            MRFittingParameters linPars = parPat.getFittingParameters( );
            linPars = updateParameterRanges( linPars ); 
        }
        
        // Calculate the sum of all NumericPattern - MRLineModelSpectralResponse
        // objects
        Iterator mrNumPatIter = m_numericPatternList.iterator( );
        while ( mrNumPatIter.hasNext( ) )
        {
            MRNumericPattern numPat = (MRNumericPattern)mrNumPatIter.next( );
            MRFittingParameters linPars = numPat.getFittingParameters( );
            linPars = updateParameterRanges( linPars ); 
        }
        
        // Calculate the sum of all MRMetaboliteModel - MRLineModelSpectralResponse
        // objects - if any
        Iterator mrMetabIter = m_metaboliteList.iterator( );
        while ( mrMetabIter.hasNext( ) )
        {
            MRMetaboliteModel metabModel = (MRMetaboliteModel)mrMetabIter.next( );
            MRFittingParameters linPars = metabModel.getFittingParameters( );
            linPars = updateParameterRanges( linPars );
        }        
    }
    
    /**
     * This routine inspects for possible incompatibilities with 
     * the fitting algorithm; if these tests pass, the model should be 
     * usable withTDFDFit.
     */
    public boolean isModelConsistent( boolean nonInteractive ) 
    {
        // First create a map of fitting component number to 
        // fitting component name
        boolean isModelConsistent = true;
        createSpectralComponentNumberToNameMap( );  
        int sizeOfMap = m_numberToComponentNameMap.size( );
        
        if ( sizeOfMap == 0 )
        {
            if ( nonInteractive == false )
            {
                isModelConsistent = false;
                JOptionPane.showMessageDialog( null, "The TDFDFit model does not contain components or is not defined!", "Warning", JOptionPane.NO_OPTION );
            }
            return isModelConsistent;
        }
        
        // Verify whether at least one parameter is fitted at all
        if ( isAtLeastOneParameterFitted( ) == false )
        {
            isModelConsistent = false;
            if ( nonInteractive == false )
            {
                JOptionPane.showMessageDialog( null, "The TDFDFit model does not fit one single parameter, therefore cannot be performed!", "Warning", JOptionPane.WARNING_MESSAGE );
            }
            return isModelConsistent;
        }
        
        // verify prior knowledge: if A is fixed to B and B is fixed to C
        // than A should be fixed to C
        if ( verifyOneToOnePriorKnowledge( true ) == true )
        {
            isModelConsistent = false;
        
            if ( nonInteractive == false )
            {
                JOptionPane.showMessageDialog( null, "The TDFDFit model contained A -> B and B -> C prior knowledge!", "Information", JOptionPane.INFORMATION_MESSAGE );
            }
            return isModelConsistent;
        }
        if ( nonInteractive == false )
        {
            JOptionPane.showMessageDialog( null, "<html><font color=\"green\"><b>The model is consistent!</b></html>", "Information", JOptionPane.INFORMATION_MESSAGE );
        }
        return isModelConsistent;
    }
    
    /**
     * Verifies whether at least one parameter is fitted ...
     * @return 
     */
    private boolean isAtLeastOneParameterFitted( ) 
    {
        int degreesOfFreedom = 0;
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            // have to cast it to a MRFittingParameters object..
            MRFittingParameters simpleLine = (MRFittingParameters)it.next( );  
            degreesOfFreedom += simpleLine.getNumberOfParametersToFit( );  
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRParametricPattern parPat = (MRParametricPattern)it2.next( );  
            degreesOfFreedom += parPat.getFittingParameters( ).getNumberOfParametersToFit( );  
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            
            MRNumericPattern numPat = (MRNumericPattern)it3.next( );  
            degreesOfFreedom += numPat.getFittingParameters( ).getNumberOfParametersToFit( );  
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            
            MRMetaboliteModel metabPat = (MRMetaboliteModel)it4.next( );  
            degreesOfFreedom += metabPat.getFittingParameters( ).getNumberOfParametersToFit( );  
        }
        
        return ( degreesOfFreedom > 0 );
    }
    
    
    /**
     * Verify prior knowledge: if A is fixed to B and B is fixed to C
     * than A should be fixed to C etceterera.
     * @param interactive
     * @return 
     */ 
    public boolean verifyOneToOnePriorKnowledge( boolean interactive ) 
    {
        MRFittingParameters referencedObject = null;
        
        boolean returnValue;
        boolean atLeastOnceFalse = false;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            // have to cast it to a MRFittingParameters object..
            MRFittingParameters simpleLine = (MRFittingParameters)it.next( );
            if ( ( simpleLine.getAreaReference( ).compareTo( "Free" ) != 0 ) &&
                 ( simpleLine.getAreaReference( ).compareTo( simpleLine.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeAreaVerification( simpleLine, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( simpleLine.getOffsetReference( ).compareTo( "Free" ) != 0 ) &&
                 ( simpleLine.getOffsetReference( ).compareTo( simpleLine.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeOffsetVerification( simpleLine, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( simpleLine.getPhaseReference( ).compareTo( "Free" ) != 0 ) &&
                 ( simpleLine.getPhaseReference( ).compareTo( simpleLine.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgePhaseVerification( simpleLine, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( simpleLine.getLorentzianLineWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( simpleLine.getLorentzianLineWidthReference( ).compareTo( simpleLine.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeLorentzianVerification( simpleLine, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( simpleLine.getGaussianWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( simpleLine.getGaussianWidthReference( ).compareTo( simpleLine.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeGaussianVerification( simpleLine, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
        }
        
        ///////////////////////////////////////////////////////////////////////
        // Now verify the list of parametric patterns
        ///////////////////////////////////////////////////////////////////////
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            // have to cast it to a MRFittingParameters object..
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( ( fitPars.getAreaReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getAreaReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeAreaVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getOffsetReference( ).compareTo( "Free" ) != 0 ) &&
                  ( fitPars.getOffsetReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeOffsetVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getPhaseReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getPhaseReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgePhaseVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getLorentzianLineWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeLorentzianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getGaussianWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeGaussianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
        }
        
        ///////////////////////////////////////////////////////////////////////
        // Now verify the list of numeric patterns
        ///////////////////////////////////////////////////////////////////////
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            // have to cast it to a MRFittingParameters object..
            MRFittingParameters fitPars = ((MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( ( fitPars.getAreaReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getAreaReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeAreaVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getOffsetReference( ).compareTo( "Free" ) != 0 ) &&
                  ( fitPars.getOffsetReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeOffsetVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getPhaseReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getPhaseReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgePhaseVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getLorentzianLineWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeLorentzianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getGaussianWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeGaussianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
        }
        
        ///////////////////////////////////////////////////////////////////////
        // Now verify the list of numeric patterns
        ///////////////////////////////////////////////////////////////////////
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            // Have to cast it to a MRFittingParameters object..
            MRFittingParameters fitPars = ((MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( ( fitPars.getAreaReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getAreaReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeAreaVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getOffsetReference( ).compareTo( "Free" ) != 0 ) &&
                  ( fitPars.getOffsetReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeOffsetVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getPhaseReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getPhaseReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgePhaseVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getLorentzianLineWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeLorentzianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
            
            if ( ( fitPars.getGaussianWidthReference( ).compareTo( "Free" ) != 0 ) &&
                 ( fitPars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) != 0 ) )
            {
                returnValue = priorKnowledgeGaussianVerification( fitPars, interactive );
                if ( returnValue == false )
                {
                    atLeastOnceFalse = true;
                }
            }
        }
        
        return atLeastOnceFalse;
    }
    
    /**
     * Verify and suggests fixes for mal defined area parameter prior knowledge.
     * @param pars
     * @return 
     */
    private boolean priorKnowledgeAreaVerification( MRFittingParameters pars, boolean interactive )
    {
        MRFittingParameters referencedObject = null;
        boolean isOK = true;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = (MRFittingParameters)it.next( );
            if ( pars.getAreaReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( pars.getAreaReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( pars.getAreaReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( pars.getAreaReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        if ( referencedObject.getAreaReference( ).compareTo( "Free" ) != 0 )
        {
            isOK = false;
            if ( interactive == true )
            {
                JOptionPane.showMessageDialog( null,
                "<html><b>Area parameter prior knowledge defintion error:</b><br /><br /><font color=\"blue\"><b>Area</b></font> parameter of component - <font color=\"blue\"><b>" + pars.getName( ) 
                + "</b></font> references component - <font color=\"blue\"><b>" + referencedObject.getName( ) 
                + "</b></font> that is not fitted itself." 
                + "<br /><br /> <font color=\"red\"> <b>Manually resolve this problem!</b></font></html>",
                "Prior knowledge error.",
                JOptionPane.OK_OPTION );
            }
        }
        
        return isOK;
    }
    
    /**
     * Verify and suggests fixes for mal defined offset parameter prior knowledge.
     * @param pars
     * @return 
     */
    private boolean priorKnowledgeOffsetVerification( MRFittingParameters pars, boolean interactive ) 
    {
        MRFittingParameters referencedObject = null;
        boolean isOK = true;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters simpleLine = (MRFittingParameters)it.next( );
            if ( pars.getOffsetReference( ).compareTo( simpleLine.getName( ) ) == 0 ) 
            {
                referencedObject = simpleLine;  
            }
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( pars.getOffsetReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( pars.getOffsetReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( pars.getOffsetReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        if ( referencedObject.getOffsetReference( ).compareTo( "Free" ) != 0 )
        {
            isOK = false;
            if ( interactive == true )
            {
                JOptionPane.showMessageDialog( null,
                "<html><b>Offset parameter prior knowledge defintion error:</b> <br /> <br /> <font color=\"blue\"><b>Offset</b></font> parameter of component - <font color=\"blue\"><b>" + pars.getName( ) 
                + "</b></font> references component - <font color=\"blue\"><b>" + referencedObject.getName( ) 
                + "</b></font> that is not fitted itself." 
                + "<br /><br /> <font color=\"red\"> <b>Manually resolve this problem!</b></font></html>",
                "Prior knowledge error.",
                JOptionPane.OK_OPTION );
            }
            
        }
        return isOK;
    }
    
     /**
     * Verify and suggests fixes for mal defined phase parameter prior knowledge.
     * @param pars
     * @return 
     */
    private boolean priorKnowledgePhaseVerification( MRFittingParameters pars, 
                                                     boolean interactive ) 
    {
        MRFittingParameters referencedObject = null;
        boolean isOK = true;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = (MRFittingParameters)it.next( );
            if ( pars.getPhaseReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( pars.getPhaseReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( pars.getPhaseReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( pars.getPhaseReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        if ( referencedObject.getPhaseReference( ).compareTo( "Free" ) != 0 )
        {
            isOK = false;
            if ( interactive == true )
            {
                JOptionPane.showMessageDialog( null,
                "<html><b>Phase fit parameter prior knowledge defintion error:</b> <br /> <br /> <font color=\"blue\"><b>Phase</b></font> parameter of line - <font color=\"blue\"><b>" + pars.getName( ) 
                + "</b></font> references line - <font color=\"blue\"><b>" + referencedObject.getName( ) 
                + "</b></font> that is not fitted itself." 
                + "<br /><br /> <font color=\"red\"> <b>Manually resolve this problem!</b></font></html>",
                "Prior knowledge error.",
                JOptionPane.OK_OPTION );
            }
        }
        
        return isOK;
        
    }
    
    /**
     * Verify and suggests fixes for mal defined Lorentzian width parameter 
     * prior knowledge.
     * @param pars
     * @return 
     */
    private boolean priorKnowledgeLorentzianVerification( MRFittingParameters pars, boolean interactive ) 
    {
        MRFittingParameters referencedObject = null;
        boolean isOK = true;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = (MRFittingParameters)it.next( );
            if ( pars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( pars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( pars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( pars.getLorentzianLineWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        if ( referencedObject.getLorentzianLineWidthReference( ).compareTo( "Free" ) != 0 )
        {
            isOK = false;
            if ( interactive == true )
            {
                JOptionPane.showMessageDialog( null,
                "<html><b>Lorentzian width parameter prior knowledge defintion error:</b> <br /> <br /> <font color=\"blue\"><b>Lorentzian width</b></font> parameter of line - <font color=\"blue\"><b>" + pars.getName( ) 
                + "</b></font> references line - <font color=\"blue\"><b>" + referencedObject.getName( ) 
                + "</b></font> that is not fitted itself." 
                + "<br /><br /> <font color=\"red\"> <b>Manually resolve this problem!</b></font></html>",
                "Prior knowledge error.",
                JOptionPane.OK_OPTION );
            }
        }
        return isOK;
    }

    /**
     * Verify and suggests fixes for mal defined Gaussian width parameter 
     * prior knowledge.
     * @param pars
     * @return 
     */
    private boolean priorKnowledgeGaussianVerification( MRFittingParameters pars , boolean interactive )
    {
        MRFittingParameters referencedObject = null;
        boolean isOK = true;
        
        Iterator it = m_simpleLineList.iterator( );
        while ( it.hasNext( ) )
        {
            MRFittingParameters fitPars = (MRFittingParameters)it.next( );
            if ( pars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        while ( it2.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRParametricPattern)it2.next( ) ).getFittingParameters( );
            if ( pars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        while ( it3.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRNumericPattern)it3.next( ) ).getFittingParameters( );
            if ( pars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        while ( it4.hasNext( ) )
        {
            MRFittingParameters fitPars = ( (MRMetaboliteModel)it4.next( ) ).getFittingParameters( );
            if ( pars.getGaussianWidthReference( ).compareTo( fitPars.getName( ) ) == 0 ) 
            {
                referencedObject = fitPars;  
            }
        }
        
        if ( referencedObject.getGaussianWidthReference( ).compareTo( "Free" ) != 0 )
        {
            isOK = false;
            if ( interactive == true )
            {
                JOptionPane.showMessageDialog( null,
                "<html><b>Gaussian width parameterprior knowledge defintion error: </b><br /> <br /> <font color=\"blue\"><b>Gaussian width</b></font> parameter of line - <font color=\"blue\"><b>" + pars.getName( ) 
                + "</b></font> references line - <font color=\"blue\"><b>" + referencedObject.getName( ) 
                + "</b></font> that is not fitted itself." 
                + "<br /><br /> <font color=\"red\"> <b>Manually resolve this problem!</b></font></html>",
                "Prior knowledge error.",
                JOptionPane.OK_OPTION );
            }
        }
        return isOK;
    }
    
    
    /**
     * Returns the current number of spectral components that are fitted in
     * by the TDFDFit model.
     * @return 
     */
    public int getNumberOfSpectralComponentsToFit( )
    {
        int numberOfLines = m_simpleLineList.size( );
        int numberOfParPats = m_parametricPatternList.size( );
        int numberOfNumPats = m_numericPatternList.size( );
        int numberOfMetabs = m_metaboliteList.size( );
        
        return ( numberOfLines + numberOfParPats + numberOfNumPats + numberOfMetabs );
    }
    
    /**
     * Returns the number of simple lines in the model.
     * @return 
     */
    public int getNumberOfSimpleLines( )
    {
        return m_simpleLineList.size( );
    }
    
    /**
     * Returns the number of parametric patterns in the model.
     * @return 
     */
    public int getNumberOfParametricPatterns( )
    {
        return m_parametricPatternList.size( );
    }
    
    /**
     * Returns the number of numeric patterns in the model.
     * @return 
     */
    public int getNumberOfNumericPatterns( )
    {
        return m_numericPatternList.size( );
    }
      
    /**
     * Returns the number of numeric patterns in the model.
     * @return 
     */
    public int getNumberOfMetabolitePatterns( )
    {
        return m_metaboliteList.size( );
    }
    
    /**
     * Returns the list of metabolite names present in the fitting model.
     * @return 
     */
    public LinkedList<String> getSpectralComponentsNames( )
    {
        
        LinkedList<String> currentSpectralComponentNamesList = new LinkedList<String>();
        
        Iterator it = m_simpleLineList.iterator( );
        
        while ( it.hasNext( ) )
        {
            currentSpectralComponentNamesList.add( ((MRLineModelParameters)it.next( ) ).getName( ) );
        }
        
        Iterator it2 = m_parametricPatternList.iterator( );
        
        while ( it2.hasNext( ) )
        {
           currentSpectralComponentNamesList.add( ( (MRParametricPattern)it2.next( ) ).getName( ) ); 
        }
        
        Iterator it3 = m_numericPatternList.iterator( );
        
        while ( it3.hasNext( ) )
        {
           currentSpectralComponentNamesList.add( ( (MRNumericPattern)it3.next( ) ).getName( ) ); 
        }
        
        Iterator it4 = m_metaboliteList.iterator( );
        
        while ( it4.hasNext( ) )
        {
           currentSpectralComponentNamesList.add( ( (MRMetaboliteModel)it4.next( ) ).getName( ) ); 
        }
       return  currentSpectralComponentNamesList;
    }
    
    public void isWaterRemovalModel( boolean thisFitDoesWaterRemoval ) 
    {
        m_thisFitDoesWaterRemoval = thisFitDoesWaterRemoval;
    }
    
    
    // Attributes needed
    private MRLineModelParametersPanel m_lineModelParametersPanel;
    
    // static private attribute is the instance itself (Singleton!)
    private static TDFDFitPrototypeModelSpectrumParameters theInstance = null;
    
    // Listeners to changes in this class: needed to notify registered classes
    // that class have successfully read model objects of different types
    private final LinkedList<ActionListener> m_actionListeners = new LinkedList<>( );
    
    // Action Events raised by this class
    public final static int MODEL_CHANGED           = 0;
    public final static int NODE_SELECTED           = 1;
    public final static int NODE_NAME_CHANGED       = 2;
    public final static int MODEL_ROOT_NAME_CHANGED = 3;
    public final static int PRIOR_KNOWLEDGE_UPDATE  = 4;
    public final static int ROOT_SELECTED           = 5;

    // Contains a list<MRLineModelParameters>
    private LinkedList<MRLineModelParameters> m_simpleLineList = new LinkedList<>( );
    private LinkedList<MRParametricPattern> m_parametricPatternList = new LinkedList<>( ); 
    // Contains a list <MRMetaboliteModel> 
    private LinkedList<MRMetaboliteModel> m_metaboliteList = new LinkedList<>( );
    private LinkedList<MRNumericPattern> m_numericPatternList = new LinkedList<>( );
    
    // Also need a local copy of the model parameters 
    LinkedList<MRLineModelParameters>   m_copyOfSimpleLineList          = new LinkedList<>( );
    LinkedList<MRParametricPattern>     m_copyOfParametricPatternList   = new LinkedList<>( );
    LinkedList<MRNumericPattern>        m_copyOfNumericPatternList      = new LinkedList<>( );
    LinkedList<MRMetaboliteModel>       m_copyOfMetaboliteList          = new LinkedList<>( );
    
    private String m_mrModelSpectrumName = "No model defined";
    private TDFDFitModelTreeWithPopupMenu m_fitModelTree;
    
    // Helper strings
    private String m_currentMetaboliteName;
    private String m_currentParametricPatternName;
    private String m_currentNumericPatternName;
    private String m_currentSimpleLineName;
    
    // Model spectral response : contains the sum response of all components
    private MRLineModelSpectralResponse m_completeModelSpectralResponse;
    private MRLineModelSpectralResponse m_selectedModelSpectralResponse;
    
    // Measured response
    private SingleVoxelMeasuredSpectralData m_measuredSpectralResponse;
    
    // Measured water reference signal
    private SingleVoxelMeasuredSpectralData m_waterReferenceSpectralResponse;
    
    private MRSignalProperties m_MRSignalProperties;
    private LinkedList<String> m_priorKnowledgePossibilities;
    private HashMap<String,Integer> m_numberToComponentNameMap;
    private boolean m_thisFitDoesWaterRemoval = false;
    
    // Attributes for generating random model variations
    MRLineModelSpectralResponse[] m_randomModelResponses;

   
    
    
}
